/*
 * @(#)io.c	1.23 95/05/11 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

#include "oobj.h"
#include "interpreter.h"
#include "javaString.h"
#include "io_md.h"

#include "java_io_FileInputStream.h"
#include "java_io_FileOutputStream.h"
#include "java_io_RandomAccessFile.h"
#include "file_acl.h"
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>

extern long currentClassHasClassLoader();

/**************************************************************
 * Input stream
 */

long
java_io_FileInputStream_open(Hjava_io_FileInputStream *this, HString *name)
{
    char buf[MAXPATHLEN];
    int fd, flags;

    if (name == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return 0;
    }

    javaString2CString(name, buf, sizeof(buf));

    if (! checkForReadAccess(buf,name)) {
	SignalError(0, JAVAPKG "IllegalAccessException", "permission error");
	return 0;
    }

    flags = O_RDONLY;
    if ((fd = sysOpen(buf, flags)) == -1) {
	SignalError(0, JAVAPKG "IOException", buf);
	return 0;
    }
    return fd;
}

long
java_io_FileInputStream_openfd(Hjava_io_FileInputStream *this, long fd)
{
    if (currentClassHasClassLoader()) {
	SignalError(0, JAVAPKG "IllegalAccessException","permission error");
	return;
    }

    if (fd < 0) {
	SignalError(0, JAVAPKG "IOException", "invalid file descriptor");
	return 0;
    }
    return fd;
}

long
java_io_FileInputStream_read(Hjava_io_FileInputStream *this)
{
    Classjava_io_FileInputStream *thisptr = unhand(this);
    unsigned char buf[1];
    int	    cnt;

    if ((cnt = sysRead(thisptr->fd, (char *) buf, 1)) != 1) {
	if (cnt == 0) {
	    return -1;	    /* EOF */
	}
	if (errno != EINTR)
	    SignalError(0, JAVAPKG "IOException", "read error");
    }
    return buf[0];
}

long
java_io_FileInputStream_readBytes(Hjava_io_FileInputStream *this, HArrayOfByte *data, long off, long len)
{
    Classjava_io_FileInputStream *thisptr = unhand(this);
    char *dataptr;
    int datalen, n;

    if (data == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return 0;
    }

    dataptr = unhand(data)->body;
    datalen = obj_length(data);

    if ((off < 0) || (off > datalen)) {
	SignalError(0, JAVAPKG "ArrayIndexOutOfBoundsException", 0);
	return -1;
    }
    if (off + len > datalen) {
	len = datalen - off;
    }
    if (len <= 0) {
	return 0;
    }

    n = sysRead(thisptr->fd, dataptr + off, len);
    if (n == -1) {
	SignalError(0, JAVAPKG "IOException", "read error");
    }
    /* AVH: this is bogus but it stops the pointer from being gc'd */
    if (dataptr == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
    }
    return (n == 0) ? -1 : n;
}

long
java_io_FileInputStream_skip(Hjava_io_FileInputStream *this, long n) 
{
    Classjava_io_FileInputStream *thisptr = unhand(this);
    off_t cur = 0, end = 0;

    if ((cur = lseek(thisptr->fd, 0L, SEEK_CUR)) == -1) {
	SignalError(0, JAVAPKG "IOException", 0);
    } else if ((end = lseek(thisptr->fd, n, SEEK_CUR)) == -1) {
	SignalError(0, JAVAPKG "IOException", 0);
    }
    return end - cur;
}

long
java_io_FileInputStream_available(Hjava_io_FileInputStream *this)
{
    Classjava_io_FileInputStream *thisptr = unhand(this);
    off_t cur = 0, end = 0;

    /* REMIND: this should not throw an exception on sockets, instead
     *         it should return 0
     */
    if ((cur = lseek(thisptr->fd, 0L, SEEK_CUR)) == -1) {
	SignalError(0, JAVAPKG "IOException", 0);
    } else if ((end = lseek(thisptr->fd, 0L, SEEK_END)) == -1) {
	SignalError(0, JAVAPKG "IOException", 0);
    } else if (lseek(thisptr->fd, cur, SEEK_SET) == -1) {
	SignalError(0, JAVAPKG "IOException", 0);
    } 
    return end - cur;
}

void
java_io_FileInputStream_close(Hjava_io_FileInputStream *this) 
{
    Classjava_io_FileInputStream *thisptr = unhand(this);
    sysClose(thisptr->fd);
    thisptr->fd = -1;
}

/**************************************************************
 * Output stream
 */

long
java_io_FileOutputStream_open(Hjava_io_FileOutputStream *this, HString *name)
{
    char buf[MAXPATHLEN];
    int fd, flags;

    if (name == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return 0;
    }

    javaString2CString(name, buf, sizeof(buf));

    if (! checkForWriteAccess(buf,name)) {
	SignalError(0, JAVAPKG "IllegalAccessException", "permission error");
	return 0;
    }

    flags = O_WRONLY|O_CREAT|O_TRUNC;
    if ((fd = sysOpen(buf, flags, 0664)) == -1) {
	SignalError(0, JAVAPKG "IOException", buf);
	return 0;
    }
    return fd;
}

long
java_io_FileOutputStream_openfd(Hjava_io_FileOutputStream *this, long fd)
{
    if (currentClassHasClassLoader()) {
	SignalError(0, JAVAPKG "IllegalAccessException","permission error");
	return;
    }

    if (fd < 0) {
	SignalError(0, JAVAPKG "IOException", "invalid file descriptor");
	return 0;
    }
    return fd;
}

void
java_io_FileOutputStream_write(Hjava_io_FileOutputStream *this, long c) 
{
    Classjava_io_FileOutputStream *thisptr = unhand(this);
    char buf[1];

    buf[0] = c;
    if (sysWrite(thisptr->fd, buf, 1) != 1) {
	SignalError(0, JAVAPKG "IOException", "write error");
    }
}

void
java_io_FileOutputStream_writeBytes(Hjava_io_FileOutputStream *this, HArrayOfByte *data, long off, long len) 
{
    Classjava_io_FileOutputStream *thisptr = unhand(this);
    char *dataptr;
    int datalen, n;

    if (data == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return;
    }
    dataptr = unhand(data)->body;
    datalen = obj_length(data);

    if ((off < 0) || (len < 0) || (off + len > datalen)) {
	SignalError(0, JAVAPKG "ArrayIndexOutOfBoundsException", 0);
	return;
    }

    while (len > 0) {
	n = sysWrite(thisptr->fd, dataptr + off, len);
	if (n == -1) {
	    SignalError(0, JAVAPKG "IOException", "write error");
	    break;
	}
        off += n;
	len -= n;
    }
}

void
java_io_FileOutputStream_close(Hjava_io_FileOutputStream *this) 
{
    Classjava_io_FileOutputStream *thisptr = unhand(this);
    sysClose(thisptr->fd);
    thisptr->fd = -1;
}

/**************************************************************
 * RandomAccessFile
 */

long
java_io_RandomAccessFile_open(Hjava_io_RandomAccessFile *this, HString *name,
			     long writeable)
{
    char buf[MAXPATHLEN];
    int fd;

    if (name == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return 0;
    }

    javaString2CString(name, buf, sizeof(buf));

    if (writeable) {
	if (! checkForWriteAccess(buf,name)) {
	    SignalError(0, JAVAPKG "IllegalAccessException", "permission error");
	    return 0;
	}
	if ((fd = sysOpen(buf, O_RDWR|O_CREAT, 0664)) == -1) {
	    SignalError(0, JAVAPKG "IOException", buf);
	    return 0;
	}
    }
    else {
	if (! checkForReadAccess(buf,name)) {
	    SignalError(0, JAVAPKG "IllegalAccessException", "permission error");
	    return 0;
	}
	if ((fd = sysOpen(buf, O_RDONLY)) == -1) {
	    SignalError(0, JAVAPKG "IOException", buf);
	    return 0;
	}
    }
    return fd;
}

long
java_io_RandomAccessFile_read(Hjava_io_RandomAccessFile *this)
{
    Classjava_io_RandomAccessFile *thisptr = unhand(this);
    unsigned char buf[1];
    int	    cnt;

    if ((cnt = sysRead(thisptr->fd, (char *) buf, 1)) != 1) {
	if (cnt == 0) {
	    return -1;	    /* EOF */
	}
	SignalError(0, JAVAPKG "IOException", "read error");
    }
    return buf[0];
}

long
java_io_RandomAccessFile_readBytes(Hjava_io_RandomAccessFile *this, HArrayOfByte *data, long off, long len)
{
    Classjava_io_RandomAccessFile *thisptr = unhand(this);
    char *dataptr;
    int datalen, n;

    if (data == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return 0;
    }
    dataptr = unhand(data)->body;
    datalen = obj_length(data);

    if ((off < 0) || (off > datalen)) {
	SignalError(0, JAVAPKG "ArrayIndexOutOfBoundsException", 0);
	return -1;
    }
    if (off + len > datalen) {
	len = datalen - off;
    }
    if (len <= 0) {
	return 0;
    }

    n = sysRead(thisptr->fd, dataptr + off, len);
    if (n == -1) {
	SignalError(0, JAVAPKG "IOException", "read error");
    }
    /* AVH: this is bogus but it stops the pointer from being gc'd */
    if (dataptr == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
    }
    return n;
}


void
java_io_RandomAccessFile_write(Hjava_io_RandomAccessFile *this, long c) 
{
    Classjava_io_RandomAccessFile *thisptr = unhand(this);
    char buf[1];

    buf[0] = c;
    if (sysWrite(thisptr->fd, buf, 1) != 1) {
	SignalError(0, JAVAPKG "IOException", "write error");
    }
}

void
java_io_RandomAccessFile_writeBytes(Hjava_io_RandomAccessFile *this, HArrayOfByte *data, long off, long len) 
{
    Classjava_io_RandomAccessFile *thisptr = unhand(this);
    char *dataptr;
    int datalen, n;

    if (data == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return;
    }
    dataptr = unhand(data)->body;
    datalen = obj_length(data);

    if ((off < 0) || (len < 0) || (off + len > datalen)) {
	SignalError(0, JAVAPKG "ArrayIndexOutOfBoundsException", 0);
	return;
    }

    while (len > 0) {
	n = sysWrite(thisptr->fd, dataptr + off, len);
	if (n == -1) {
	    SignalError(0, JAVAPKG "IOException", "write error");
	    break;
	}
        off += n;
	len -= n;
    }
}


long
java_io_RandomAccessFile_getFilePointer(Hjava_io_RandomAccessFile *this)
{
    Classjava_io_RandomAccessFile *thisptr = unhand(this);
    off_t cur = 0;

    if ((cur = lseek(thisptr->fd, 0L, SEEK_CUR)) == -1) {
	SignalError(0, JAVAPKG "IOException", 0);
    }
    return cur;
}


void
java_io_RandomAccessFile_seek(Hjava_io_RandomAccessFile *this, long pos) 
{
    Classjava_io_RandomAccessFile *thisptr = unhand(this);
    off_t cur = 0;

    if ((cur = lseek(thisptr->fd, pos, SEEK_SET)) == -1) {
	SignalError(0, JAVAPKG "IOException", 0);
    }
}

long
java_io_RandomAccessFile_length(Hjava_io_RandomAccessFile *this)
{
    Classjava_io_RandomAccessFile *thisptr = unhand(this);
    off_t cur = 0, end = 0;

    if ((cur = lseek(thisptr->fd, 0L, SEEK_CUR)) == -1) {
	SignalError(0, JAVAPKG "IOException", 0);
    } else if ((end = lseek(thisptr->fd, 0L, SEEK_END)) == -1) {
	SignalError(0, JAVAPKG "IOException", 0);
    } else if (lseek(thisptr->fd, cur, SEEK_SET) == -1) {
	SignalError(0, JAVAPKG "IOException", 0);
    } 
    return end;
}

void
java_io_RandomAccessFile_close(Hjava_io_RandomAccessFile *this) 
{
    Classjava_io_RandomAccessFile *thisptr = unhand(this);
    sysClose(thisptr->fd);
    thisptr->fd = -1;
}





/**************************************************************
 * Data conversions in DataInputStream
 */

float
java_io_DataInputStream_int2float(Hjava_io_FileOutputStream *this, long v)
{
    union {
	int i;
	float f;
    } u;
    u.i = v;
    return u.f;
}


double
java_io_DataInputStream_long2double(Hjava_io_FileOutputStream *this, int64_t v)
{
    union {
	int64_t l;
	double d;
    } u;
    u.l = v;
    return u.d;
}

/**************************************************************
 * Data conversions in DataOutputStream
 */

long
java_io_DataOutputStream_float2int(Hjava_io_FileOutputStream *this, float v)
{
    union {
	int i;
	float f;
    } u;
    u.f = v;
    return u.i;
}


int64_t
java_io_DataOutputStream_double2long(Hjava_io_FileOutputStream *this, double v)
{
    union {
	int64_t l;
	double d;
    } u;
    u.d = v;
    return u.l;
}
