/*
 * @(#)Assembler.java	1.15 95/04/02 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.asm;

import java.tools.java.*;
import java.util.Enumeration;
import java.io.OutputStreamBuffer;
import java.io.DataOutputStream;
import java.io.PrintStream;

/**
 * This class is used to assemble the bytecode instructions for a method.
 */
public final
class Assembler implements Constants {
    static final int NOTREACHED		= 0;
    static final int REACHED		= 1;
    static final int NEEDED		= 2;

    Label first = new Label();
    Instruction last = first;
    int maxdepth;
    int maxvar;

    /**
     * Add an instruction
     */
    public void add(Instruction inst) {
	if (inst != null) {
	    last.next = inst;
	    last = inst;
	}
    }
    public void add(int where, int opc) {
	add(new Instruction(where, opc, null));
    }
    public void add(int where, int opc, Object obj) {
	add(new Instruction(where, opc, obj));
    }
    
    /**
     * Optimize instructions and mark those that can be reached
     */
    void optimize(Environment env, Label lbl) {
	lbl.pc = REACHED;
	
	for (Instruction inst = lbl.next ; inst != null ; inst = inst.next)  {
	    switch (inst.pc) {
	      case NOTREACHED:
		inst.optimize(env);
		inst.pc = REACHED;
		break;
	      case REACHED:
		return;
	      case NEEDED:
		break;
	    }

	    switch (inst.opc) {
	      case opc_label:
	      case opc_dead:
		if (inst.pc == REACHED) {
		    inst.pc = NOTREACHED;
		}
		break;

	      case opc_ifeq:
	      case opc_ifne:
	      case opc_ifgt:
	      case opc_ifge:
	      case opc_iflt:
	      case opc_ifle:
	      case opc_if_icmpeq:
	      case opc_if_icmpne:
	      case opc_if_icmpgt:
	      case opc_if_icmpge:
	      case opc_if_icmplt:
	      case opc_if_icmple:
	      case opc_if_acmpeq:
	      case opc_if_acmpne:
		optimize(env, (Label)inst.value);
		break;
		
	      case opc_goto:
		optimize(env, (Label)inst.value);
		return;

	      case opc_jsr:
		optimize(env, (Label)inst.value);
		break;

	      case opc_ret:
	      case opc_return:
	      case opc_ireturn:
	      case opc_lreturn:
	      case opc_freturn:
	      case opc_dreturn:
	      case opc_areturn:
	      case opc_athrow:
		return;

	      case opc_tableswitch:
	      case opc_lookupswitch: {
		SwitchData sw = (SwitchData)inst.value;
		optimize(env, sw.defaultLabel);
		for (Enumeration e = sw.tab.elements() ; e.hasMoreElements();) {
		    optimize(env, (Label)e.nextElement());
		}
		return;
	      }

	      case opc_try: {
		TryData td = (TryData)inst.value;
		td.getEndLabel().pc = NEEDED;
		for (Enumeration e = td.catches.elements() ; e.hasMoreElements();) {
		    CatchData cd = (CatchData)e.nextElement();
		    optimize(env, cd.getLabel());
		}
		break;
	      }
	    }
	}
    }

    /**
     * Eliminate instructions that are not reached
     */
    boolean eliminate() {
	boolean change = false;
	Instruction prev = first;
	
	for (Instruction inst = first.next ; inst != null ; inst = inst.next) {
	    if (inst.pc != NOTREACHED) {
		prev.next = inst;
		prev = inst;
		inst.pc = NOTREACHED;
	    } else {
		change = true;
	    }
	}
	first.pc = NOTREACHED;
	prev.next = null;
	return change;
    }

    /**
     * Optimize the byte codes
     */
    public void optimize(Environment env) {
	//listing(System.out);
	do {
	    // Figure out which instructions are reached
	    optimize(env, first);

	    // Eliminate instructions that are not reached
	} while (eliminate() && env.optimize());
    }

    /**
     * Collect all constants into the constant table
     */
    public void collect(Environment env, ConstantPool tab) {
	for (Instruction inst = first ; inst != null ; inst = inst.next) {
	    inst.collect(tab);
	}
    }

    /**
     * Determine stack size, count local variables
     */
    void balance(Label lbl, int depth) {
	for (Instruction inst = lbl ; inst != null ; inst = inst.next)  {
	    //System.out.println(inst.toString() + " = " + depth);
	    depth += inst.balance();
	    if (depth < 0) {
	       throw new InternalError("stack under flow: " + inst.toString() + " = " + depth);
	    }
	    if (depth > maxdepth) {
		maxdepth = depth;
	    }
	    switch (inst.opc) {
	      case opc_label:
		lbl = (Label)inst;
		if (inst.pc == REACHED) {
		    if (lbl.depth != depth) {
			throw new InternalError("stack depth error");
		    }
		    return;
		}
		lbl.pc = REACHED;
		lbl.depth = depth;
		break;

	      case opc_ifeq:
	      case opc_ifne:
	      case opc_ifgt:
	      case opc_ifge:
	      case opc_iflt:
	      case opc_ifle:
	      case opc_if_icmpeq:
	      case opc_if_icmpne:
	      case opc_if_icmpgt:
	      case opc_if_icmpge:
	      case opc_if_icmplt:
	      case opc_if_icmple:
	      case opc_if_acmpeq:
	      case opc_if_acmpne:
		balance((Label)inst.value, depth);
		break;

	      case opc_goto:
		balance((Label)inst.value, depth);
		return;

	      case opc_jsr:
		balance((Label)inst.value, depth + 1);
		break;

	      case opc_ret:
	      case opc_return:
	      case opc_ireturn:
	      case opc_lreturn:
	      case opc_freturn:
	      case opc_dreturn:
	      case opc_areturn:
	      case opc_athrow:
		return;

	      case opc_iload:
	      case opc_fload:
	      case opc_aload:
	      case opc_istore:
	      case opc_fstore:
	      case opc_astore: {
		int var = ((Number)inst.value).intValue() + 1;
		if (var > maxvar) {
		    maxvar = var;
		}
		break;
	      }

	      case opc_lload:
	      case opc_dload:
	      case opc_lstore:
	      case opc_dstore: {
		int var = ((Number)inst.value).intValue() + 2;
		if (var > maxvar) {
		    maxvar = var;
		}
		break;
	      }

	      case opc_tableswitch:
	      case opc_lookupswitch: {
		SwitchData sw = (SwitchData)inst.value;
		balance(sw.defaultLabel, depth);
		for (Enumeration e = sw.tab.elements() ; e.hasMoreElements();) {
		    balance((Label)e.nextElement(), depth);
		}
		return;
	      }

	      case opc_try: {
		TryData td = (TryData)inst.value;
		for (Enumeration e = td.catches.elements() ; e.hasMoreElements();) {
		    CatchData cd = (CatchData)e.nextElement();
		    balance(cd.getLabel(), depth + 1);
		}
		break;
	      }
	    }
	}
    }

    /**
     * Generate code
     */
    public void write(Environment env, DataOutputStream out, ConstantPool tab) {
	//listing(System.out);

	// Balance the stack
	try {
	    balance(first, 0);
	} catch (InternalError e) {
	    System.out.println("ERROR: " + e);
	    listing(System.out);
	    throw e;
	}
	
	// Assign PCs
	int pc = 0, nexceptions = 0;
	for (Instruction inst = first ; inst != null ; inst = inst.next) {
	    inst.pc = pc;
	    pc += inst.size(tab);

	    if (inst.opc == opc_try) {
		nexceptions += ((TryData)inst.value).catches.size();
	    }
	}

	// Write header
	out.writeByte(maxdepth);
	out.writeByte(maxvar);
	out.writeShort(pc);

	// Generate code
	for (Instruction inst = first.next ; inst != null ; inst = inst.next) {
	    inst.write(out, tab);
	}

	// write exceptions
	out.writeShort(nexceptions);
	if (nexceptions > 0) {
	    //listing(System.out);
	    writeExceptions(env, out, tab, first, last);
	}

	// write attributes
	if (env.optimize()) {
	    out.writeShort(0);
	} else {
	    OutputStreamBuffer buf = new OutputStreamBuffer(256);
	    writeLineNumbers(env, new DataOutputStream(buf));

	    out.writeShort(1);
	    out.writeShort(tab.index("LineNumberTable"));
	    out.writeInt(buf.size());
	    buf.writeTo(out);
	}
    }

    /**
     * Write the exceptions table
     */
    void writeExceptions(Environment env, DataOutputStream out, ConstantPool tab, Instruction first, Instruction last) {
	for (Instruction inst = first ; inst != last.next ; inst = inst.next) {
	    if (inst.opc == opc_try) {
		TryData td = (TryData)inst.value;
		writeExceptions(env, out, tab, inst.next, td.getEndLabel());
		for (Enumeration e = td.catches.elements() ; e.hasMoreElements();) {
		    CatchData cd = (CatchData)e.nextElement();
		    //System.out.println("EXCEPTION: " + env.getSource() + ", pc=" + inst.pc + ", end=" + td.getEndLabel().pc + ", hdl=" + cd.getLabel().pc + ", tp=" + cd.getType());
		    out.writeShort(inst.pc);
		    out.writeShort(td.getEndLabel().pc);
		    out.writeShort(cd.getLabel().pc);
		    if (cd.getType() != null) {
			out.writeShort(tab.index(cd.getType()));
		    } else {
			out.writeShort(0);
		    }
		}
		inst = td.getEndLabel();
	    }
	}
    }

    /**
     * Write the linenumber table
     */
    public void writeLineNumbers(Environment env, DataOutputStream out) {
	int ln = -1;
	int count = 0;

	for (Instruction inst = first ; inst != null ; inst = inst.next) {
	    int n = (inst.where >> OFFSETBITS);
	    if ((n > 0) && (ln != n)) {
		ln = n;
		count++;
	    }
	}

	ln = -1;
	out.writeShort(count);
	for (Instruction inst = first ; inst != null ; inst = inst.next) {
	    int n = (inst.where >> OFFSETBITS);
	    if ((n > 0) && (ln != n)) {
		ln = n;
		out.writeShort(inst.pc);
		out.writeShort(ln);
		//System.out.println("pc = " + inst.pc + ", ln = " + ln);
	    }
	}
    }

    /**
     * Return true if empty
     */
    public boolean empty() {
	return first == last;
    }

    /**
     * Print the byte codes
     */
    public void listing(PrintStream out) {
	out.println("-- listing --");
	for (Instruction inst = first.next ; inst != null ; inst = inst.next) {
	    out.println(inst.toString());
	}
    }
}

