/*
 * @(#)ClassPath.java	1.9 95/02/06 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.java;

import java.util.Enumeration;
import java.util.Hashtable;
import java.io.File;

/**
 * This class is used to represent the classes in a package.
 * When an instance of this class is created the appropriate
 * directories are scanned and a the .class and .java files
 * are stored in a hashtable so that they can be accessed quickly.
 */
public
class ClassPath {
    /**
     * Hashtables
     */
    Hashtable binaries = new Hashtable();
    Hashtable sources = new Hashtable();

    /**
     * Create a ClassPath, given a path and a package name. Every package directory
     * is scanned for .class and .java files. This information is later used for
     * quick lookups.
     */
    public ClassPath(String path[], Identifier pkg) {
	String pkgname = pkg.toString().replace('.', File.separatorChar);

	for (int i = path.length ; i-- > 0 ;) {
	    File pkgdir = new File(path[i], pkgname);
	    if (pkgdir.isDirectory()) {
		String list[] = pkgdir.list();
		if (list != null) {
		    for (int n = list.length ; n-- > 0 ;) {
			String file = list[n];
			if (file.startsWith(".")) {
			    continue;
			} else if (file.endsWith(".class")) {
			    //System.out.println("FILE=" + file);
			    binaries.put(Identifier.lookup(file.substring(0, file.length() - 6)),
					 new File(pkgdir, file));
			} else if (list[n].endsWith(".java")) {
			    //System.out.println("FILE=" + pkgdir.getPath() + "/" + file);
			    sources.put(Identifier.lookup(file.substring(0, file.length() - 5)),
					new File(pkgdir, file));
			}
		    }
		}
	    }
	}
    }

    /**
     * Check if a class is defined in this package
     */
    public boolean classExists(Identifier className) {
	return (binaries.get(className) != null) || (sources.get(className) != null);
    }

    /**
     * Get the .class file of a class
     */
    public File getBinaryFile(Identifier className) {
	return (File)binaries.get(className);
    }

    /**
     * Get the .java file of a class
     */
    public File getSourceFile(Identifier className) {
	return (File)sources.get(className);
    }
    public File getSourceFile(String fileName) {
	if (fileName.endsWith(".java")) {
	    Identifier className = Identifier.lookup(fileName.substring(0, fileName.length() - 4));
	    return (File)sources.get(className);
	}
	return null;
    }

    /**
     * Decompose a : seperated path into an array of strings
     */
    public static String path2list(String path)[] {
	int i, j, n, len = path.length();

	// Count the number of :'s
	for (i = n = 0 ; i < len ; i++) {
	    if (path.charAt(i) == ':') {
		n++;
	    }
	}

	// allocate the list
	String list[] = new String[n + 1];
	for (i = j = n = 0 ; i < len ; i = ++j) {
            for (; (j < len) && (path.charAt(j) != ':') ; j++);
	    list[n++] = path.substring(i, j);
	}

	return list;
    }

    /**
     * Get all the binary classes names
     */
    public Enumeration getBinaryClassNames() {
	return binaries.keys();
    }

    /**
     * Get all the binary files
     */
    public Enumeration getBinaryFiles() {
	return binaries.elements();
    }

    /**
     * Get all the source files
     */
    public Enumeration getSourceFiles() {
	return sources.elements();
    }
}
