/*
 * @(#)Environment.java	1.18 95/04/05 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.java;

import java.util.Stack;

/**
 * This class defines the environment for a compilation.
 * It is used to load classes, resolve class names and
 * report errors. It is an abstract class, a subclass
 * must define implementations for some of the functions.<p>
 *
 * An environment has a source object associated with it.
 * This is the thing against which errors are reported, it
 * is usually a file name, a field or a class.<p>
 *
 * Environments can be nested to change the source object.<p>
 *
 * @author 	Arthur van Hoff
 * @version 	1.18, 05 Apr 1995
 */

public
class Environment implements Constants {
    /**
     * The actual environment to which everything is forwarded.
     */
    Environment env;
    
    /**
     * The object that is currently being parsed/compiled.
     * It is either a file name (String) or a field (FieldDefinition)
     * or a class (ClassDeclaration or ClassDefinition).
     */
    Object source;

    public Environment(Environment env, Object source) {
	this.env = ((env != null) && (env.env != null)) ? env.env : env;
	this.source = source;
    }
    public Environment() {
	this(null, null);
    }
    
    /**
     * Return a class declaration given a fully qualified class name.
     */
    public ClassDeclaration getClassDeclaration(Identifier nm) {
	return env.getClassDeclaration(nm);
    }
    
    /**
     * Return a class defintion given a fully qualified class name.
     */
    public final ClassDefinition getClassDefinition(Identifier nm) {
	return getClassDeclaration(nm).getClassDefinition(this);
    }

    /**
     * Return a class declaration given a type. Only works for
     * class types.
     */
    public final ClassDeclaration getClassDeclaration(Type t) {
	return getClassDeclaration(t.getClassName());
    }

    /**
     * Return a class definition given a type. Only works for
     * class types.
     */
    public final ClassDefinition getClassDefinition(Type t) {
	return getClassDefinition(t.getClassName());
    }

    /**
     * Check if a class exists (without actually loading it).
     */
    public boolean classExists(Identifier nm) {
	return env.classExists(nm);
    }

    /**
     * Load the definition of a class.
     */
    public void loadDefinition(ClassDeclaration c) {
	env.loadDefinition(c);
    }

    /**
     * Return the source of the environment (ie: the thing being compiled/parsed).
     */
    public final Object getSource() {
	return source;
    }

    /**
     * Resolve a type. Make sure that all the classes referred to by
     * the type have a definition.
     */
    public void resolve(int where, ClassDefinition c, Type t) {
	switch (t.getTypeCode()) {
	  case TC_CLASS: {
	    ClassDefinition def;
	    try {
		def = getClassDefinition(t);
	    } catch (ClassNotFound e) {
		error(where, ERR_CLASS_NOT_FOUND, e.name, "type declaration");
		break;
	    }
	    if (!c.canAccess(this, def.getClassDeclaration())) {
		error(where, ERR_CANT_ACCESS_CLASS, def);
	    }
	    break;
	  }

	  case TC_ARRAY:
	    resolve(where, c, t.getElementType());
	    break;

	  case TC_METHOD:
	    resolve(where, c, t.getReturnType());
	    Type args[] = t.getArgumentTypes();
	    for (int i = args.length ; i-- > 0 ; ) {
		resolve(where, c, args[i]);
	    }
	    break;
	}
    }

    /**
     * Compute the distance between this method type and
     * the given argument list. This method is used to determine
     * which method is the best match given an argument list.
     */
    public int distance(Type t, Type args[]) {
	if (!t.isType(TC_METHOD)) {
	    return Integer.MAX_VALUE;
	}

	Type targs[] = t.getArgumentTypes();
	if (targs.length != args.length) {
	    return Integer.MAX_VALUE;
	}

	int maxdist = 0;
	for (int i = args.length ; i-- > 0 ;) {
	    int dist = distance(args[i], targs[i]);
	    if (maxdist < dist) {
		if (dist == Integer.MAX_VALUE) {
		    return Integer.MAX_VALUE;
		}
		maxdist = dist;
	    }
	}
	return maxdist;
    }

    /**
     * Compute the distance between two types. The distance is 0 is
     * the types are identitical. The distances are computed according
     * to the table in the language spec.
     */
    public int distance(Type from, Type to) {
	if (from == to) {
	    return 0;
	}

	switch (to.getTypeCode()) {
	  case TC_ERROR:
	    return 9;
	    
	  case TC_BOOLEAN:
	    switch (from.getTypeCode()) {
	      case TC_BOOLEAN:	return 0;
	    }
	    break;
	    
	  case TC_BYTE:
	    switch (from.getTypeCode()) {
	      case TC_BYTE:	return 0;
	      case TC_SHORT: 	return 10;
	      case TC_CHAR: 	return 11;
	      case TC_INT: 	return 12;
	      case TC_LONG: 	return 12;
	      case TC_FLOAT:	return 15;
	      case TC_DOUBLE:	return 16;
	    }
	    break;
	    
	  case TC_SHORT:
	    switch (from.getTypeCode()) {
	      case TC_BYTE:	return 1;
	      case TC_SHORT: 	return 0;
	      case TC_CHAR: 	return 10;
	      case TC_INT: 	return 11;
	      case TC_LONG: 	return 11;
	      case TC_FLOAT:	return 14;
	      case TC_DOUBLE:	return 15;
	    }
	    break;
	    
	  case TC_CHAR:
	    switch (from.getTypeCode()) {
	      case TC_BYTE:	return 2;
	      case TC_SHORT: 	return 10;
	      case TC_CHAR: 	return 0;
	      case TC_INT: 	return 11;
	      case TC_LONG: 	return 11;
	      case TC_FLOAT:	return 13;
	      case TC_DOUBLE:	return 14;
	    }
	    break;

	  case TC_INT:
	    switch (from.getTypeCode()) {
	      case TC_BYTE:	return 3;
	      case TC_SHORT: 	return 1;
	      case TC_CHAR: 	return 1;
	      case TC_INT: 	return 0;
	      case TC_LONG: 	return 10;
	      case TC_FLOAT:	return 12;
	      case TC_DOUBLE:	return 13;
	    }
	    break;

	  case TC_LONG:
	    switch (from.getTypeCode()) {
	      case TC_BYTE:	return 4;
	      case TC_SHORT: 	return 2;
	      case TC_CHAR: 	return 2;
	      case TC_INT: 	return 1;
	      case TC_LONG: 	return 0;
	      case TC_FLOAT:	return 11;
	      case TC_DOUBLE:	return 12;
	    }
	    break;

	  case TC_FLOAT:
	    switch (from.getTypeCode()) {
	      case TC_BYTE:	return 6;
	      case TC_SHORT: 	return 4;
	      case TC_CHAR: 	return 4;
	      case TC_INT: 	return 5;
	      case TC_LONG: 	return 6;
	      case TC_FLOAT:	return 0;
	      case TC_DOUBLE:	return 10;
	    }
	    break;

	  case TC_DOUBLE:
	    switch (from.getTypeCode()) {
	      case TC_BYTE:	return 7;
	      case TC_SHORT: 	return 5;
	      case TC_CHAR: 	return 5;
	      case TC_INT: 	return 4;
	      case TC_LONG: 	return 5;
	      case TC_FLOAT:	return 1;
	      case TC_DOUBLE:	return 0;
	    }
	    break;

	  case TC_ARRAY:
	    switch (from.getTypeCode()) {
	      case TC_NULL:
		return 0;

	      case TC_ARRAY:
		Type t1 = from.getElementType();
		Type t2 = to.getElementType();
		while ((t1.getTypeCode() == TC_ARRAY) && (t2.getTypeCode() == TC_ARRAY)) {
		    t1 = t1.getElementType();
		    t2 = t2.getElementType();
		}
		if (t1.inMask(TM_ARRAY|TM_CLASS) && t2.inMask(TM_ARRAY|TM_CLASS)) {
		    return distance(t1, t2);
		}
		return (t1.getTypeCode() == t2.getTypeCode()) ? 0 : Integer.MAX_VALUE;

	      case TC_CLASS:
		return (from == Type.tObject) ? 1 : Integer.MAX_VALUE;
	    }
	    break;

	  case TC_CLASS:
	    switch (from.getTypeCode()) {
	      case TC_NULL:
		return 0;

	      case TC_ARRAY:
		if (to == Type.tObject) {
		    return 1;
		}
		break;
		
	      case TC_CLASS:
		return env.getClassDefinition(to).distance(env, env.getClassDeclaration(from));
	    }
	    break;

	  case TC_METHOD:
	    return Integer.MAX_VALUE;
	}
	
	return (from.getTypeCode() == TC_ERROR) ? 9 : Integer.MAX_VALUE;
    }

    /**
     * Return true if an implicit cast from this type to
     * the given type is allowed.
     */
    public boolean implicitCast(Type from, Type to) {
	switch (from.getTypeCode()) {
	  case TC_BOOLEAN:
	    return to.isType(TC_BOOLEAN);
	    
	  case TC_BYTE:
	  case TC_CHAR:
	  case TC_SHORT:
	  case TC_INT:
	  case TC_LONG:
	  case TC_FLOAT:
	  case TC_DOUBLE:
	    return to.inMask(TM_NUMBER);

	  case TC_NULL:
	    return to.inMask(TM_REFERENCE);

	  case TC_ARRAY:
	    if (to.isType(TC_ARRAY)) {
		Type t1 = from.getElementType();
		Type t2 = to.getElementType();
		while ((t1.getTypeCode() == TC_ARRAY) && (t2.getTypeCode() == TC_ARRAY)) {
		    t1 = t1.getElementType();
		    t2 = t2.getElementType();
		}
		if (t1.inMask(TM_ARRAY|TM_CLASS) && t2.inMask(TM_ARRAY|TM_CLASS)) {
		    return implicitCast(t1, t2);
		}
		return t1 == t2;
	    }
	    if (to == Type.tObject) {
		return true;
	    }
	    break;

	  case TC_CLASS:
	    return (from == to) || (to.isType(TC_CLASS) && env.getClassDefinition(to).implementedBy(env, env.getClassDeclaration(from)));
	}
	
	return false;
    }

    /**
     * Return true if an explicit cast from this type to
     * the given type is allowed.
     */
    public boolean explicitCast(Type from, Type to) {
	if (implicitCast(from, to)) {
	    return true;
	}
	if (from.isType(TC_CLASS) && to.isType(TC_CLASS)) {
	    ClassDefinition fromClass = getClassDefinition(from);
	    ClassDefinition toClass = getClassDefinition(to);
	    return toClass.isInterface() ||
		fromClass.isInterface() ||
		fromClass.superClassOf(env, toClass.getClassDeclaration());
	}
	if ((from == Type.tObject) && to.isType(TC_ARRAY)) {
	    return true;
	}
	if (from.isType(TC_ARRAY) && to.isType(TC_ARRAY)) {
	    Type t1 = from.getElementType();
	    Type t2 = to.getElementType();
	    while ((t1.getTypeCode() == TC_ARRAY) && (t2.getTypeCode() == TC_ARRAY)) {
		t1 = t1.getElementType();
		t2 = t2.getElementType();
	    }
	    if (t1.inMask(TM_ARRAY|TM_CLASS) && t2.inMask(TM_ARRAY|TM_CLASS)) {
		return explicitCast(t1, t2);
	    }
	}
	return false;
    }

    /**
     * Flags.
     */
    public int getFlags() {
        return env.getFlags();
    }

    /**
     * Generate debugging information
     */
    public final boolean debug() {
        return (getFlags() & F_DEBUG) != 0;
    }

    /**
     * Optimize the code
     */
    public final boolean optimize() {
        return (getFlags() & F_OPTIMIZE) != 0;
    }

    /**
     * Verbose
     */
    public final boolean verbose() {
        return (getFlags() & F_VERBOSE) != 0;
    }

    /**
     * Dump debugging stuff
     */
    public final boolean dump() {
        return (getFlags() & F_DUMP) != 0;
    }

    /**
     * Verbose
     */
    public final boolean warnings() {
        return (getFlags() & F_WARNINGS) != 0;
    }

    /**
     * Dependencies
     */
    public final boolean dependencies() {
        return (getFlags() & F_DEPENDENCIES) != 0;
    }

    /**
     * Issue an error.
     *  source	 - the input source, usually a file name string
     *  offset   - the offset in the source of the error
     *  err      - the error number (as defined in this interface)
     *  arg1     - an optional argument to the error (null if not applicable)
     *  arg2     - a second optional argument to the error (null if not applicable)
     *  arg3     - a third optional argument to the error (null if not applicable)
     */
    public void error(Object source, int where, int err, Object arg1, Object arg2, Object arg3) {
	env.error(source, where, err, arg1, arg2, arg3);
    }
    public final void error(int where, int err, Object arg1, Object arg2, Object arg3) {
	error(source, where, err, arg1, arg2, arg3);
    }
    public final void error(int where, int err, Object arg1, Object arg2) {
	error(source, where, err, arg1, arg2, null);
    }
    public final void error(int where, int err, Object arg1) {
	error(source, where, err, arg1, null, null);
    }
    public final void error(int where, int err) {
	error(source, where, err, null, null, null);
    }

    /**
     * Output a string. This can either be an error message or something
     * for debugging. This should be used instead of println.
     */
    public void output(String msg) {
	env.output(msg);
    }
}
