/*
 * @(#)FieldDefinition.java	1.13 95/01/31 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.java;

import java.tools.tree.Node;
import java.tools.tree.Expression;
import java.tools.tree.Context;
import java.tools.asm.Assembler;
import java.io.PrintStream;
import java.util.Vector;

/**
 * This class defines an Java field.
 */
public
class FieldDefinition implements Constants {
    protected int where;
    protected int modifiers;
    protected Type type;
    protected String documentation;
    protected Node value;
    protected ClassDefinition clazz;
    protected Identifier name;
    protected FieldDefinition nextField;
    protected FieldDefinition nextMatch;

    /**
     * Constructor
     */
    public FieldDefinition(int where, ClassDefinition clazz, int modifiers,
			      Type type, Identifier name, Node value) {
	this.where = where;
	this.clazz = clazz;
	this.modifiers = modifiers;
	this.type = type;
	this.name = name;
	this.value = value;
    }

    /**
     * Get the position in the input
     */
    public final int getWhere() {
	return where;
    }

    /**
     * Get the class declaration
     */
    public final ClassDeclaration getClassDeclaration() {
	return clazz.getClassDeclaration();
    }

    /**
     * Get the class definition
     */
    public final ClassDefinition getClassDefinition() {
	return clazz;
    }

    /**
     * Get the field's modifiers
     */
    public final int getModifiers() {
	return modifiers;
    }
    public final void subModifiers(int mod) {
	modifiers &= ~mod;
    }
    public final void addModifiers(int mod) {
	modifiers |= mod;
    }

    /**
     * Get the field's type
     */
    public final Type getType() {
	return type;
    }

    /**
     * Get the field's name
     */
    public final Identifier getName() {
	return name;
    }

    /**
     * Get arguments (a vector of LocalField)
     */
    public Vector getArguments() {
	return isMethod() ? new Vector() : null;
    }

    /**
     * Get the field's final value (may return null)
     */
    public Node getValue(Environment env) {
	return value;
    }
    public final Node getValue() {
	return value;
    }
    public final void setValue(Node value) {
	this.value = value;
    }
    public Object getInitialValue() {
	return null;
    }

    /**
     * Get the next field or the next match
     */
    public final FieldDefinition getNextField() {
	return nextField;
    }
    public final FieldDefinition getNextMatch() {
	return nextMatch;
    }

    /**
     * Get the field's documentation
     */
    public String getDocumentation() {
	return documentation;
    }

    /**
     * Resolve a class name
     */
    public Identifier resolve(Environment env, Identifier nm) {
	throw new InternalError("resolve");
    }

    /**
     * Check the field definition
     */
    public void check(Environment env) {
    }

    /**
     * Generate code
     */
    public void code(Environment env, Assembler asm) {
	throw new InternalError("code");
    }
    public void codeInit(Environment env, Context ctx, Assembler asm) {
	throw new InternalError("codeInit");
    }

    /**
     * Check if a field can access another field (only considers
     * static versus instance access, not the access modifiers).
     */
    public final boolean canAccess(Environment env, FieldDefinition f) {
	return (!isStatic()) || f.isLocal() || f.isStatic();
    }

    /**
     * Check if a field can reach another field (only considers
     * forward references, not the access modifiers).
     */
    public final boolean canReach(Environment env, FieldDefinition f) {
	if ((isVariable() || isInitializer()) && (isStatic() || !f.isStatic()) &&
	    (!f.isLocal()) && getClassDeclaration().equals(f.getClassDeclaration())) {

	    while (((f = f.getNextField()) != null) && (f != this));
	    return f != null;
	}
	return true;
    }

    /**
     * Checks
     */
    public final boolean isPublic() {
	return (modifiers & M_PUBLIC) != 0;
    }
    public final boolean isPrivate() {
	return (modifiers & M_PRIVATE) != 0;
    }
    public final boolean isProtected() {
	return (modifiers & M_PROTECTED) != 0;
    }
    public final boolean isFinal() {
	return (modifiers & M_FINAL) != 0;
    }
    public final boolean isStatic() {
	return (modifiers & M_STATIC) != 0;
    }
    public final boolean isSynchronized() {
	return (modifiers & M_SYNCHRONIZED) != 0;
    }
    public final boolean isAbstract() {
	return (modifiers & M_ABSTRACT) != 0;
    }
    public final boolean isNative() {
	return (modifiers & M_NATIVE) != 0;
    }
    public final boolean isThreadsafe() {
	return (modifiers & M_THREADSAFE) != 0;
    }
    public final boolean isTransient() {
	return (modifiers & M_TRANSIENT) != 0;
    }
    public final boolean isMethod() {
	return type.isType(TC_METHOD);
    }
    public final boolean isVariable() {
	return !type.isType(TC_METHOD);
    }
    public final boolean isInitializer() {
	return getName().equals(idClassInit);
    }
    public final boolean isConstructor() {
	return getName().equals(idInit);
    }
    public boolean isLocal() {
	return false;
    }
    public boolean isInlineable(Environment env, boolean fromFinal) {
	return (isStatic() || isPrivate() || isFinal() || isConstructor() || fromFinal) && 
	    !(isSynchronized() || isNative());
    }

    /**
     * toString
     */
    public String toString() {
	if (isInitializer()) {
	    return "static {}";
	} else if (isConstructor()) {
	    StringBuffer buf = new StringBuffer();
	    buf.append(getClassDefinition().getName());
	    buf.appendChar('(');
	    Type argTypes[] = getType().getArgumentTypes();
	    for (int i = 0 ; i < argTypes.length ; i++) {
		if (i > 0) {
		    buf.appendChar(',');
		}
		buf.append(argTypes[i].toString());
	    }
	    buf.appendChar(')');
	    return buf.toString();
	}
	return type.typeString(getName().toString());
    }

    /**
     * Print for debugging
     */
    public void print(PrintStream out) {
	if (isPublic()) {
	    out.print("public ");
	}
	if (isPrivate()) {
	    out.print("private ");
	}
	if (isProtected()) {
	    out.print("protected ");
	}
	if (isFinal()) {
	    out.print("final ");
	}
	if (isStatic()) {
	    out.print("static ");
	}
	if (isSynchronized()) {
	    out.print("synchronized ");
	}
	if (isAbstract()) {
	    out.print("abstract ");
	}
	if (isNative()) {
	    out.print("native ");
	}
	if (isThreadsafe()) {
	    out.print("threadsafe ");
	}
	if (isTransient()) {
	    out.print("transient ");
	}
	out.println(toString() + ";");
    }
}
