/*
 * @(#)Scanner.java	1.26 95/04/05 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.java;

import java.io.InputStream;
import java.util.Hashtable;

/**
 * A Scanner for Java tokens. Errors are reported 
 * to the environment object.<p>
 *
 * The scanner keeps track of the current token,
 * the value of the current token (if any), and the start
 * position of the current token.<p>
 *
 * The scan() method advances the scanner to the next
 * token in the input.<p>
 *
 * The match() method is used to quickly match opening
 * brackets (ie: '(', '{', or '[') with their closing
 * counter part. This is useful during error recovery.<p>
 *
 * An position consists of: ((linenr << OFFSETBITS) | offset)
 * this means that both the line number and the exact offset into
 * the file are encoded in each postion value.<p>
 *
 * The compiler treats either "\n", "\r" or "\r\n" as the
 * end of a line.<p>
 *
 * @author 	Arthur van Hoff
 * @version 	1.26, 05 Apr 1995
 */

public
class Scanner implements Constants {
    /**
     * The increment for each character.
     */
    protected static final int OFFSETINC = 1;

    /**
     * The increment for each line.
     */
    protected static final int LINEINC = 1 << OFFSETBITS;

    /**
     * End of input
     */
    public static final int EOF	= -1;

    /**
     * Where errors are reported
     */
    protected Environment env;

    /**
     * Input stream
     */
    protected InputStream in;

    /**
     * Current token
     */
    protected int token;

    /**
     * The position of the current token
     */
    protected int pos;

    /**
     * The position of the previous token
     */
    protected int prevPos;

    /**
     * The current position in the input
     */
    private int curPos;

    /**
     * The current character
     */
    private int ch;

    /*
     * Token values.
     */
    protected char charValue;
    protected int intValue;
    protected long longValue;
    protected float floatValue;
    protected double doubleValue;
    protected String stringValue;
    protected Identifier idValue;

    /*
     * A growable character buffer.
     */
    private int count;
    private char buffer[] = new char[32];
    private void putc(int ch) {
	if (count == buffer.length) {
	    char newBuffer[] = new char[buffer.length * 2];
	    System.arraycopy(buffer, 0, newBuffer, 0, buffer.length);
	    buffer = newBuffer;
	}
	buffer[count++] = (char)ch;
    }
    private String bufferString() {
	char buf[] = new char[count];
	System.arraycopy(buffer, 0, buf, 0, count);
	return new String(buf);
    }

    /**
     * Create a scanner to scan an input stream.
     */
    protected Scanner(Environment env, InputStream in) {
	this.env = env;
	this.in = in;

	getc();

	prevPos = curPos = LINEINC; 
	scan();
    }

    /**
     * This method should be called whenever an end-of-line
     * is reached.
     */
    private void newline() {
	curPos += LINEINC;
    }

    /**
     * Get the next character.
     */
    private int getc() {
	curPos += OFFSETINC;
	return ch = in.read();
    }

    /**
     * Define a keyword.
     */
    private static void defineKeyword(int val) {
	Identifier.lookup(opNames[val]).setType(val);
    }

    /**
     * Initialized keyword and token Hashtables
     */
    static {
	// Statement keywords
	defineKeyword(FOR);
	defineKeyword(IF);
	defineKeyword(ELSE);
	defineKeyword(WHILE);
	defineKeyword(DO);
	defineKeyword(SWITCH);
	defineKeyword(CASE);
	defineKeyword(DEFAULT);
	defineKeyword(BREAK);
	defineKeyword(CONTINUE);
	defineKeyword(RETURN);
	defineKeyword(TRY);
	defineKeyword(CATCH);
	defineKeyword(FINALLY);
	defineKeyword(THROW);
	defineKeyword(SYNCHRONIZE);

	// Type defineKeywords
	defineKeyword(BYTE);
	defineKeyword(CHAR);
	defineKeyword(SHORT);
	defineKeyword(INT);
	defineKeyword(LONG);
	defineKeyword(FLOAT);
	defineKeyword(DOUBLE);
	defineKeyword(VOID);
	defineKeyword(BOOLEAN);

	// Expression keywords
	defineKeyword(INSTANCEOF);
	defineKeyword(TRUE);
	defineKeyword(FALSE);
	defineKeyword(NEW);
	defineKeyword(THIS);
	defineKeyword(SUPER);
	defineKeyword(NULL);

	// Declaration keywords
	defineKeyword(IMPORT);
	defineKeyword(CLASS);
	defineKeyword(EXTENDS);
	defineKeyword(IMPLEMENTS);
	defineKeyword(INTERFACE);
	defineKeyword(PACKAGE);

	// Modifier keywords
	defineKeyword(PRIVATE);
	defineKeyword(PUBLIC);
	defineKeyword(PROTECTED);
	defineKeyword(STATIC);
	defineKeyword(TRANSIENT);
	defineKeyword(SYNCHRONIZED);
	defineKeyword(NATIVE);
	defineKeyword(ABSTRACT);
	defineKeyword(THREADSAFE);
	defineKeyword(FINAL);

	// obsolete keywords
	defineKeyword(CONST);
	defineKeyword(GOTO);
	defineKeyword(ASMHACK);
    }

    /**
     * Scan a comment. This method should be
     * called once the initial / and * have
     * been read.
     */
    private void scanComment() {
	token = COMMENT;

	while (true) {
	    switch (ch) {
	      case EOF:
	        env.error(pos, ERR_EOF_IN_COMMENT);
		return;

	      case '*':
		if (getc() == '/')  {
		    getc();
		    return;
		}
		break;

	      case '\r':
		newline();
		if (getc() == '\n') {
		    getc();
		}
		break;

	      case '\n':
		newline();
		// fall through

	      default:
		getc();
		break;
	    }
	}
    }

    /**
     * Scan a doc comment. This method should be called
     * once the initial /, * and * have been read. It gathers
     * the content of the comment (witout leading spaces and '*'s)
     * in the string buffer.
     */
    private void scanDocComment() {
	count = 0;
	token = COMMENT;

	if (ch == '*') {
	    do {
		getc();
	    } while (ch == '*');
	    if (ch == '/') {
		getc();
		stringValue = "";
		return;
	    }
	}
	switch (ch) {
	  case '\r':
	    newline();
	    if (getc() == '\n') {
		getc();
	    }
	    break;

	  case '\n':
	    newline();
	    getc();
	    break;

	  case ' ':
	    getc();
	    break;
	}

	boolean seenstar = false;
	int c = count;
	while (true) {
	    switch (ch) {
	      case EOF:
	        env.error(pos, ERR_EOF_IN_COMMENT);
		stringValue = bufferString();
		return;

	      case '\r':
		putc('\r');
		if (getc() == '\n') {
		    getc();
		}
		newline();
		seenstar = false;
		c = count;
		break;

	      case '\n':
		putc('\n');
		getc();
		newline();
		seenstar = false;
		c = count;
		break;

	      case ' ':
	      case '\t':
		putc(ch);
		getc();
		break;

	      case '*':
		if (seenstar) {
		    if (getc() == '/') {
			getc();
			count = c;
			stringValue = bufferString();
			return;
		    }
		    putc('*');
		} else {
		    seenstar = true;
		    count = c;
		    while (getc() == '*');
		    switch (ch) {
		      case ' ':
			getc();
			break;

		      case '/':
			getc();
			count = c;
			stringValue = bufferString();
			return;
		    }
		}
		break;

	      default:
		if (!seenstar) {
		    seenstar = true;
		}
		putc(ch);
		getc();
		c = count;
		break;
	    }
	}
    }

    /**
     * Scan a real exponent. Should be called when
     * the 'e' or 'E' of a real number is the current
     * character.
     */
    private void scanRealExponent() {
	putc(ch);
	switch (getc()) {
	  case '0': case '1': case '2': case '3': case '4':
	  case '5': case '6': case '7': case '8': case '9':
	  case '+': case '-':
	    while (true) {
		putc(ch);
		switch (getc()) {
		  case '0': case '1': case '2': case '3': case '4':
		  case '5': case '6': case '7': case '8': case '9':
		    break;

		  case 'd': case 'D':
		    getc();
		    try {
			doubleValue = Double.valueOf(bufferString()).doubleValue();
		    } catch (NumberFormatException ee) {
			env.error(pos, ERR_FLOAT_FORMAT);
		    }
		    token = DOUBLEVAL;
		    return;

		  case 'f': case 'F':
		    getc();
		    // fall through

		  default:
		    try {
			floatValue = Float.valueOf(bufferString()).floatValue();
		    } catch (NumberFormatException ee) {
			env.error(pos, ERR_FLOAT_FORMAT);
		    }
		    token = FLOATVAL;
		    return;
		}
	    }
	}
	
	env.error(pos, ERR_FLOAT_FORMAT);
	floatValue = 0;
	token = FLOATVAL;
    }
    
    /**
     * Scan a real fraction. Should be called
     * when the '.' of a real number is the
     * current character.
     */
    private void scanRealFraction() {
	while (true) {
	    putc(ch);
	    switch (getc()) {
	      case '0': case '1': case '2': case '3': case '4':
	      case '5': case '6': case '7': case '8': case '9':
		break;
			
	      case 'e': case 'E':
		scanRealExponent();
		return;

	      case 'd': case 'D':
		getc();
		try {
		    doubleValue = Double.valueOf(bufferString()).doubleValue();
		} catch (NumberFormatException ee) {
		    env.error(pos, ERR_FLOAT_FORMAT);
		}
		token = DOUBLEVAL;
		return;

	      case 'f': case 'F':
		getc();
		// fall through
		
	      default:
		try {
		    floatValue = Float.valueOf(bufferString()).floatValue();
		} catch (NumberFormatException ee) {
		    env.error(pos, ERR_FLOAT_FORMAT);
		}
		token = FLOATVAL;
		return;
	    }
	}
    }

    /**
     * Scan a number. The first digit of the number
     * should be the current character.
     */
    private void scanNumber() {
	long n = ch - '0';

	while (true) {
	    switch (getc()) {
	      case 'e':
	      case 'E':
	      case '.': {
		count = 0;
		do {
		    System.arraycopy(buffer, 0, buffer, 1, count++);
		    buffer[0] = (char)('0' + (n % 10));
		} while ((n /= 10) > 0);
		if (ch == '.') {
		    scanRealFraction();
		} else {
		    scanRealExponent();
		}
		return;
	      }

	      case '0': case '1': case '2': case '3': case '4':
	      case '5': case '6': case '7': case '8': case '9':
		n = (n * 10) + ch - '0';
	        break;

	      case 'l': case 'L':
		getc();
	        longValue = n;
	        token = LONGVAL;
		return;

	      default:
	        if (n == (int)n) {
		    intValue = (int)n;
		    token = INTVAL;
		    return;
		}
	        longValue = n;
	        token = LONGVAL;
		return;
	    }
	}
    }

    /**
     * Scan a hexadecimal number. Should
     * be called when the 'x' of the number
     * is the current character.
     */
    private void scanHexadecimal() {
	long n = 0;

	while (true) {
	    switch (getc()) {
	      case '0': case '1': case '2': case '3': case '4':
	      case '5': case '6': case '7': case '8': case '9':
	        n = (n << 4) + ch - '0';
	        break;

	      case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
	        n = (n << 4) + 10 + ch - 'a';
	        break;

	      case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
	        n = (n << 4) + 10 + ch - 'A';
	        break;

	      case 'l': case 'L':
	        getc();
	        longValue = n;
	        token = LONGVAL;
		return;

	      default:
	        if ((n & ((long)0xFFFFFFFF << 32)) == 0) {
		    intValue = (int)n;
		    token = INTVAL;
		    return;
		}
		longValue = n;
	        token = LONGVAL;
		return;
	    }
	}
    }

    /**
     * Scan an octal number. Should be called
     * when the first digit of the number is the
     * current character.
     */
    private void scanOctal() {
	long n = ch - '0';

	while (true) {
	    switch (getc()) {
	      case '0': case '1': case '2': case '3': case '4':
	      case '5': case '6': case '7': 
	        n = (n << 3) + ch - '0';
	        break;

	      case 'l': case 'L':
	        getc();
	        longValue = n;
	        token = LONGVAL;
		return;

	      default:
	        if ((n & ((long)0xFFFFFFFF << 32)) == 0) {
		    intValue = (int)n;
		    token = INTVAL;
		    return;
	        }
	        longValue = n;
	        token = LONGVAL;
		return;
	    }
	}
    }

    /**
     * Scan an escape character.
     * @return the character or -1 if it escaped an
     * end-of-line or -2 if the escape character was
     * invalid.
     */
    private int scanEscapeChar() {
	int p = curPos;
	
	switch (getc()) {
	  case '0': case '1': case '2': case '3':
	  case '4': case '5': case '6': case '7': {
	    int n = ch - '0';
	    for (int i = 2 ; i > 0 ; i--) {
		switch (getc()) {
		  case '0': case '1': case '2': case '3':
		  case '4': case '5': case '6': case '7':
		    n = (n << 3) + ch - '0';
		    break;

		  default:
		    return n;
		}
	    }
	    getc();
	    return n;
	  }

	  case 'u': case 'x': {
	    int n = 0;
	    for (int i = (ch == 'u') ? 4 : 2 ; i > 0 ; i--) {
		switch (getc()) {
		  case '0': case '1': case '2': case '3': case '4':
		  case '5': case '6': case '7': case '8': case '9':
		    n = (n << 4) + ch - '0';
		    break;

		  case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
		    n = (n << 4) + 10 + ch - 'a';
		    break;

		  case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
		    n = (n << 4) + 10 + ch - 'A';
		    break;

		  default:
		    return n;
		}
	    }
	    getc();
	    return n;
	  }

	  case 'r':  getc(); return '\r';
	  case 'n':  getc(); return '\n';
	  case 'f':  getc(); return '\f';
	  case 'b':  getc(); return '\b';
	  case 't':  getc(); return '\t';
	  case '\\': getc(); return '\\';
	  case '\"': getc(); return '\"';
	  case '\'': getc(); return '\'';

	  case '\r':
	    newline();
	    if (getc() == '\n') {
		getc();
	    }
	    return -1;

	  case '\n':
	    newline();
	    getc();
	    return -1;
	}

	env.error(p, ERR_INVALID_ESCAPE_CHAR);
	getc();
	return -2;
    }

    /**
     * Scan a string. The current character
     * should be the opening " of the string.
     */
    private void scanString() {
	token = STRINGVAL;
	count = 0;
	getc();

	// Scan a String
	while (true) {
	    switch (ch) {
	      case EOF:
		env.error(pos, ERR_EOF_IN_STRING);
		stringValue = bufferString();
		return;

	      case '\r':
		newline();
		if (getc() == '\n') {
		    getc();
		}
		env.error(pos, ERR_NEWLINE_IN_STRING);
		stringValue = bufferString();
		return;

	      case '\n':
		newline();
		getc();
		env.error(pos, ERR_NEWLINE_IN_STRING);
		stringValue = bufferString();
		return;

	      case '"':
		getc();
		stringValue = bufferString();
		return;

	      case '\\': {
		int c = scanEscapeChar();
		if (c >= 0) {
		    putc((char)c);
		}
		break;
	      }	  

	      default:
		putc(ch);
		getc();
		break;
	    }
	}
    }

    /**
     * Scan a character. The current character should be
     * the opening ' of the character constant.
     */
    private void scanCharacter() {
	token = CHARVAL;

	switch (getc()) {
	  case '\\':
	    int c = scanEscapeChar();
	    if (c == -1) {
		env.error(pos, ERR_INVALID_CHAR_CONSTANT);
	    }
	    charValue = (char)c;
	    break;

	  case '\r':
	    newline();
	    charValue = '\r';
	    if (getc() == '\n') {
		getc();
	    }
	    break;

	  case '\n':
	    newline();
	    charValue = '\n';
	    getc();
	    break;

	  default:
	    charValue = (char)ch;
	    getc();
	    break;
	}

	if (ch == '\'') {
	    getc();
	} else {
	    env.error(pos, ERR_INVALID_CHAR_CONSTANT);
	    while (true) {
		switch (ch) {
		  case '\'':
		    getc();
		    return;
		  case ';':
		  case '\n':
		  case '\r':
		  case EOF:
		    return;
		  default:
		    getc();
		}
	    }
	}
    }

    /**
     * Scan an Identifier. The current character should
     * be the first character of the identifier.
     */
    private void scanIdentifier() {
	count = 0;

	while (true) {
	    putc(ch);
	    switch (getc()) {
	      case 'a': case 'b': case 'c': case 'd': case 'e':
	      case 'f': case 'g': case 'h': case 'i': case 'j':
	      case 'k': case 'l': case 'm': case 'n': case 'o':
	      case 'p': case 'q': case 'r': case 's': case 't':
	      case 'u': case 'v': case 'w': case 'x': case 'y':
	      case 'z':
	      case 'A': case 'B': case 'C': case 'D': case 'E':
	      case 'F': case 'G': case 'H': case 'I': case 'J':
	      case 'K': case 'L': case 'M': case 'N': case 'O':
	      case 'P': case 'Q': case 'R': case 'S': case 'T':
	      case 'U': case 'V': case 'W': case 'X': case 'Y':
	      case 'Z':
	      case '0': case '1': case '2': case '3': case '4':
	      case '5': case '6': case '7': case '8': case '9':
	      case '$': case '_':
		break;
		
	      default:
		if (ch < 0x00C0) {
		    idValue = Identifier.lookup(bufferString());
		    token = idValue.getType();
		    return;
		}
	    }
	}
    }

    /**
     * Scan the next token.
     * @return the position of the previous token.
     */
    protected int scan() {
	int retPos = pos;
	prevPos = curPos;

	while (true) {
	    pos = curPos;

	    switch (ch) {
	      case EOF:
		token = EOF;
		return retPos;

		case '\r':
		newline();
		if (getc() == '\n') {
		    getc();
		}
		break;

	      case '\n':
		newline();
		getc();
		break;
		
	      case ' ':
	      case '\t':
	      case '\f':
		getc();
		break;
		
	      case '/':
		switch (getc()) {
		  case '/':
		    // Parse a // comment
		    while ((getc() != EOF) && (ch != '\n') && (ch != '\r'));
                    // if we aborted 'cuz of '\r', check for '\n' following
		    if ((ch == '\r') && (getc() == '\n')) {
			getc();
		    }
		    break;
		    
		  case '*':
		    switch (getc()) {
		      case '*':
			scanDocComment();
			return retPos;

		      default:
			scanComment();
		    }
		    break;
		    
		  case '=':
		    getc();
		    token = ASGDIV;
		    return retPos;
		    
		  default:
		    token = DIV;
		    return retPos;
		}
		break;
		
	      case '"':
		scanString();
		return retPos;
		
	      case '\'':
		scanCharacter();
		return retPos;

	      case '0':
		switch (getc()) {
		  case 'x': case 'X':
		    scanHexadecimal();
		    return retPos;

		  case '.':
		    count = 0;
		    scanRealFraction();
		    return retPos;

		  case 'e': case 'E':
		    scanRealExponent();
		    return retPos;

		  case 'l': case 'L':
		    getc();
		    longValue = 0;
		    token = LONGVAL;
		    return retPos;

		  case 'f': case 'F':
		    getc();
		    floatValue = 0;
		    token = FLOATVAL;
		    return retPos;

		  case 'd': case 'D':
		    getc();
		    doubleValue = 0;
		    token = DOUBLEVAL;
		    return retPos;

		  case '0': case '1': case '2': case '3':
		  case '4': case '5': case '6': case '7':
		    scanOctal();
		    return retPos;

		  default:
		    intValue = 0;
		    token = INTVAL;
		    return retPos;
		}

	      case '1': case '2': case '3': case '4': 
	      case '5': case '6': case '7': case '8': case '9': 
		scanNumber();
		return retPos;

	      case '.':
		switch (getc()) {
		  case '0': case '1': case '2': case '3': case '4':
		  case '5': case '6': case '7': case '8': case '9':
		    count = 0;
		    putc('.');
		    scanRealFraction();
		    return retPos;
		}
		token = FIELD;
		return retPos;
		
	      case '{':
		getc();
		token = LBRACE;
		return retPos;

	      case '}':
		getc();
		token = RBRACE;
		return retPos;

	      case '(':
		getc();
		token = LBRACKET;
		return retPos;

	      case ')':
		getc();
		token = RBRACKET;
		return retPos;

	      case '[':
		getc();
		token = LSQBRACKET;
		return retPos;

	      case ']':
		getc();
		token = RSQBRACKET;
		return retPos;

	      case ',':
		getc();
		token = COMMA;
		return retPos;

	      case ';':
		getc();
		token = SEMICOLON;
		return retPos;

	      case '?':
		getc();
		token = QUESTIONMARK;
		return retPos;

	      case '~':
		getc();
		token = BITNOT;
		return retPos;

	      case ':':
		getc();
		token = COLON;
		return retPos;
		
	      case '-':
		switch (getc()) {
		  case '-':
		    getc();
		    token = DEC;
		    return retPos;

		  case '=':
		    getc();
		    token = ASGSUB;
		    return retPos;
		}
		token = SUB;
		return retPos;
		
	      case '+':
		switch (getc()) {
		  case '+':
		    getc();
		    token = INC;
		    return retPos;

		  case '=':
		    getc();
		    token = ASGADD;
		    return retPos;
		}
		token = ADD;
		return retPos;
		
	      case '<':
		switch (getc()) {
		  case '<':
		    if (getc() == '=') {
			getc();
			token = ASGLSHIFT;
			return retPos;
		    }
		    token = LSHIFT;
		    return retPos;

		  case '=':
		    getc();
		    token = LE;
		    return retPos;
		}
		token = LT;
		return retPos;
		
	      case '>':
		switch (getc()) {
		  case '>':
		    switch (getc()) {
		      case '=':
			getc();
			token = ASGRSHIFT;
			return retPos;
			
		      case '>':
			if (getc() == '=') {
			    getc();
			    token = ASGURSHIFT;
			    return retPos;
			}
			token = URSHIFT;
			return retPos;
		    }
		    token = RSHIFT;
		    return retPos;

		  case '=':
		    getc();
		    token = GE;
		    return retPos;
		}
		token = GT;
		return retPos;
		
	      case '|':
		switch (getc()) {
		  case '|':
		    getc();
		    token = OR;
		    return retPos;
		    
		  case '=':
		    getc();
		    token = ASGBITOR;
		    return retPos;
		}
		token = BITOR;
		return retPos;
		
	      case '&':
		switch (getc()) {
		  case '&':
		    getc();
		    token = AND;
		    return retPos;
		    
		  case '=':
		    getc();
		    token = ASGBITAND;
		    return retPos;
		}
		token = BITAND;
		return retPos;
		
	      case '=':
		if (getc() == '=') {
		    getc();
		    token = EQ;
		    return retPos;
		}
		token = ASSIGN;
		return retPos;
		
	      case '%':
		if (getc() == '=') {
		    getc();
		    token = ASGMOD;
		    return retPos;
		}
		token = MOD;
		return retPos;

	      case '^':
		if (getc() == '=') {
		    getc();
		    token = ASGBITXOR;
		    return retPos;
		}
		token = BITXOR;
		return retPos;

	      case '!':
		if (getc() == '=') {
		    getc();
		    token = NE;
		    return retPos;
		}
		token = NOT;
		return retPos;

	      case '*':
		if (getc() == '=') {
		    getc();
		    token = ASGMUL;
		    return retPos;
		}
		token = MUL;
		return retPos;
		
	      case 'a': case 'b': case 'c': case 'd': case 'e':
	      case 'f': case 'g': case 'h': case 'i': case 'j':
	      case 'k': case 'l': case 'm': case 'n': case 'o':
	      case 'p': case 'q': case 'r': case 's': case 't':
	      case 'u': case 'v': case 'w': case 'x': case 'y':
	      case 'z':
	      case 'A': case 'B': case 'C': case 'D': case 'E':
	      case 'F': case 'G': case 'H': case 'I': case 'J':
	      case 'K': case 'L': case 'M': case 'N': case 'O':
	      case 'P': case 'Q': case 'R': case 'S': case 'T':
	      case 'U': case 'V': case 'W': case 'X': case 'Y':
	      case 'Z':
	      case '$': case '_':
		scanIdentifier();
		return retPos;
		
	      default:
		if (ch >= 0x00C0) {
		    scanIdentifier();
		    return retPos;
		}
		env.error(pos, ERR_FUNNY_CHAR);
		getc();
		break;
 	    }
	}
    }

    /**
     * Scan to a matching '}', ']' or ')'. The current token must be
     * a '{', '[' or '(';
     */
    protected void match(int open, int close) {
	int depth = 1;

	while (true) {
	    scan();
	    if (token == open) {
		depth++;
	    } else if (token == close) {
		if (--depth == 0) {
		    return;
		}
	    } else if (token == EOF) {
		env.error(pos, ERR_UNBALANCED_BRACKET);
		return;
	    }
	}
    }
}
