/*
 * @(#)BatchEnvironment.java	1.29 95/04/06 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.javac;

import java.tools.java.*;

import java.util.Hashtable;
import java.util.Enumeration;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.OutputStream;

/**
 * Main environment of the batch version of the Java compiler,
 * this needs more work.
 */
public 
class BatchEnvironment extends Environment {
    /**
     * The stream where error message are printed.
     */
    OutputStream out;
    
    /**
     * The class path
     */
    String path[];
    
    /**
     * A hashtable of resource contexts.
     */
    Hashtable packages = new Hashtable(31);

    /**
     * The classes, this is only a cache, not a
     * complete list.
     */
    Hashtable classes = new Hashtable(351);

    /**
     * The resource strings
     */
    ResourceStrings strings;

    /**
     * flags
     */
    public int flags;

    /**
     * The number of errors and warnings
     */
    public int nerrors;
    public int nwarnings;

    /**
     * Constructor
     */
    public BatchEnvironment(String path[], ResourceStrings strings) {
	this(System.out, path, strings);
    }

    /**
     * Constructor
     */
    public BatchEnvironment(OutputStream out, String path[], ResourceStrings strings) {
	this.out = out;
	this.path = path;
	this.strings = strings;
    }

    /**
     * Return flags
     */
    public int getFlags() {
        return flags;
    }

    /**
     * Return an enumeration of all the currently defined classes
     */
    public Enumeration getClasses() {
	return classes.elements();
    }

    /**
     * Get a class, given the fully qualified class name
     */
    public ClassDeclaration getClassDeclaration(Identifier nm) {
	ClassDeclaration c = (ClassDeclaration)classes.get(nm);
	if (c == null) {
	    classes.put(nm, c = new ClassDeclaration(nm));
	}
	return c;
    }

    /**
     * Check if a class exists
     */
    public boolean classExists(Identifier nm) {
	return (classes.get(nm) != null) ||
	    getClassPath(nm.getQualifier()).classExists(nm.getName());
    }

    /**
     * Get the package path for a packate
     */
    public ClassPath getClassPath(Identifier pkg) {
	ClassPath p = (ClassPath)packages.get(pkg);
	if (p == null) {
	    packages.put(pkg, p = new ClassPath(path, pkg));
	}
	return p;
    }
    
    /**
     * Parse a source file
     */
    public void parseFile(File file) {
	int tm = System.nowMillis();
	FileInputStream input = new FileInputStream(file);
	BatchParser p = new BatchParser(new Environment(this, file),
					new BufferedInputStream(input));

	try {
	    p.parseFile();
	} catch(Exception e) {
	    e.printStackTrace();
	}
	
	input.close();

	if (verbose()) {
	    tm = System.nowMillis() - tm;
	    output("[parsed " + file.getPath() + " in " + tm + "ms]");
	}
    }

    /**
     * Load a binary file
     */
    BinaryClass loadFile(File file) {
	int tm = System.nowMillis();
	FileInputStream input = new FileInputStream(file);
	BinaryClass c = null;

	try {
	    DataInputStream is = 
		new DataInputStream(new BufferedInputStream(input));
	    c = BinaryClass.load(new Environment(this, file), is, 
				 loadFileFlags());
	} catch (FileFormatException e) {
	    error(0, ERR_CLASS_FORMAT, file.getPath(), e.getMessage());
	    return null;
	} catch (Exception e) {
	    e.printStackTrace();
	}

	input.close();
	if (verbose()) {
	    tm = System.nowMillis() - tm;
	    output("[loaded " + file.getPath() + " in " + tm + "ms]");
	}

	return c;
    }
    
    /**
     * Default flags for loadFile.  Subclasses may override this.
     */
    int loadFileFlags() { 
	return 0; 
    }

    /**
     * Load a binary class
     */
    boolean needsCompilation(Hashtable check, ClassDeclaration c) {
	switch (c.getStatus()) {
	  case CS_UNDEFINED:
	    loadDefinition(c);
	    return needsCompilation(check, c);

	  case CS_UNDECIDED:
	    if (check.get(c) == null) {
		check.put(c, c);

		BinaryClass bin = (BinaryClass)c.getClassDefinition();
		for (Enumeration e = bin.getDependencies() ; e.hasMoreElements() ;) {
		    ClassDeclaration dep = (ClassDeclaration)e.nextElement();
		    if (needsCompilation(check, dep)) {
			// It must be source, dependencies need compilation
			//output(c + " must be source, dependencies need compilation: " + dep);
			c.setDefinition(bin, CS_SOURCE);
			return true;
		    }
		}
	    }
	    return false;

	  case CS_BINARY:
	    return false;
	}
	return true;
    }

    /**
     * Load the definition of a class
     */
    public void loadDefinition(ClassDeclaration c) {
	switch (c.getStatus()) {
	  case CS_UNDEFINED: {
	    Identifier nm = c.getName();
	    ClassPath path = getClassPath(nm.getQualifier());
	    File binfile = path.getBinaryFile(nm.getName());
	    if (binfile == null) {
		// must be source, there is no binary
		//output(c + " must be source, there is no binary");
		c.setDefinition(null, CS_SOURCE);
		return;
	    }

	    File srcfile = path.getSourceFile(nm.getName());
	    if (srcfile == null) {
		BinaryClass bc = loadFile(binfile);
		if ((bc != null) && !bc.getName().equals(nm)) {
		    error(0, WARN_WRONG_CLASS, binfile.getPath(), c, bc);
		    bc = null;
		}
		if (bc == null) {
		    // no source nor binary found
		    //output(c + " is not found, no source nor binary");
		    c.setDefinition(null, CS_NOTFOUND);
		    return;
		}

		// Couldn't find the source, try the one mentioned in the binary
		if (bc.getSource() != null) {
		    srcfile = new File((String)bc.getSource());
		    if (srcfile.exists()) {
			if (srcfile.lastModified() > binfile.lastModified()) {
			    // must be source, it is newer than the binary
			    //output(c + " must be source, source is newer");
			    c.setDefinition(bc, CS_SOURCE);
			    return;
			}
			c.setDefinition(bc, dependencies() ? CS_UNDECIDED : CS_BINARY);
			return;
		    } 
		}

		// It must be binary, there is no source
		//output(c + " must be binary, there is no source");
		c.setDefinition(bc, CS_BINARY);
		return;
	    }
	    if (srcfile.lastModified() > binfile.lastModified()) {
		// must be source, it is newer than the binary
		//output(c + " must be source, source is younger");
		c.setDefinition(null, CS_SOURCE);
		return;
	    }
	    BinaryClass bc = loadFile(binfile);
	    if ((bc != null) && !bc.getName().equals(nm)) {
		error(0, WARN_WRONG_CLASS, binfile.getPath(), c, bc);
		bc = null;
	    }
	    if (bc != null) {
		if (bc.getName().equals(c.getName())) {
		    c.setDefinition(bc, dependencies() ? CS_UNDECIDED : CS_BINARY);
		} else {
		    c.setDefinition(null, CS_NOTFOUND);
		    getClassDeclaration(bc.getName()).setDefinition(bc, dependencies() ? CS_UNDECIDED : CS_BINARY);
		}
	    } else {
		c.setDefinition(null, CS_NOTFOUND);
	    }
	    return;
	  }
	    
	  case CS_UNDECIDED: {
	    Hashtable tab = new Hashtable();
	    if (!needsCompilation(tab, c)) {
		// All undecided classes that this class depends on must be binary
		for (Enumeration e = tab.keys() ; e.hasMoreElements() ; ) {
		    ClassDeclaration dep = (ClassDeclaration)e.nextElement();
		    if (dep.getStatus() == CS_UNDECIDED) {
			// must be binary, dependencies need compilation
			//output(dep + " must be binary, dependencies need no compilation");
			dep.setDefinition(dep.getClassDefinition(), CS_BINARY);
		    }
		}
	    }
	    return;
	  }

	  case CS_SOURCE: {
	    File srcfile;
	    if (c.getClassDefinition() != null) {
		// Use the source file name from the binary class file
		ClassPath path = getClassPath(c.getName().getQualifier());
		srcfile = path.getSourceFile((String)c.getClassDefinition().getSource());
		if (srcfile == null) {
		    srcfile = new File((String)c.getClassDefinition().getSource());
		}
	    } else {
		// Get a source file name from the class path
		Identifier nm = c.getName();
		srcfile = getClassPath(nm.getQualifier()).getSourceFile(nm.getName());
		if (srcfile == null) {
		    // not found, there is no source
		    //output(c + " is not found, no source");
		    c.setDefinition(null, CS_NOTFOUND);
		    return;
		}
	    }
	    parseFile(srcfile);
	    
	    if ((c.getClassDefinition() == null) || (c.getStatus() == CS_SOURCE)) {
		// not found after parsing the file
		c.setDefinition(null, CS_NOTFOUND);
	    } 
	    return;
	  }
	}
    }

    /**
     * Error String
     */
    String errorString(int err, Object arg1, Object arg2, Object arg3) {
	String str = strings.getString(err);
	if (str == null) {
	    return "resource string " + err + " not found";
	}
	
	StringBuffer buf = new StringBuffer();
	for (int i = 0 ; i < str.length() ; i++) {
	    char c = str.charAt(i);
	    if ((c == '%') && (i + 1 < str.length())) {
		switch (str.charAt(++i)) {
		  case 's':
		    buf.append(arg1);
		    arg1 = arg2;
		    arg2 = arg3;
		    break;
		    
		  case '%':
		    buf.appendChar('%');
		    break;

		  default:
		    buf.appendChar('?');
		    break;
		}
	    } else {
		buf.appendChar(c);
	    }
	}
	return buf.toString();
    }

    /**
     * The filename where the last errors have occurred
     */
    String errorFileName;

    /**
     * List of outstanding error messages
     */
    ErrorMessage errors;

    /**
     * Insert an error message in the list of outstanding error messages.
     * The list is sorted on input position.
     */
    void insertError(int where, String message) {
	//output("ERR = " + message);
	ErrorMessage msg = new ErrorMessage(where, message);
	if (errors == null) {
	    errors = msg;
	} else if (errors.where > where) {
	    msg.next = errors;
	    errors = msg;
	} else {
	    ErrorMessage m = errors;
	    for (; (m.next != null) && (m.next.where <= where) ; m = m.next);
	    msg.next = m.next;
	    m.next = msg;
	}
    }

    /**
     * Flush outstanding errors
     */
    public void flushErrors() {
	if (errors == null) {
	    return;
	}

	// Read the file
	FileInputStream in = new FileInputStream(errorFileName);
	byte data[] = new byte[in.available()];
	in.read(data);
	in.close();

	// Report the errors
	for (ErrorMessage msg = errors ; msg != null ; msg = msg.next) {
	    int ln = msg.where >>> OFFSETBITS;
	    int off = msg.where & ((1 << OFFSETBITS) - 1);

	    int i, j;
	    for (i = off ; (i > 0) && (data[i - 1] != '\n') && (data[i - 1] != '\r') ; i--);
	    for (j = off ; (j < data.length) && (data[j] != '\n') && (data[j] != '\r') ; j++);

	    String prefix = errorFileName + ":" + ln + ":";
	    output(prefix + " " + msg.message);
	    output(new String(data, 0, i, j - i));

	    char strdata[] = new char[(off - i) + 1];
	    for (j = i ; j < off ; j++) {
		strdata[j-i] = (data[j] == '\t') ? '\t' : ' ';
	    }
	    strdata[off-i] = '^';
	    output(new String(strdata));
	}
	errors = null;
    }

    /**
     * Report error
     */
    void reportError(Object src, int where, int err, String msg) {
	if (src == null) {
	    if (errorFileName != null) {
		flushErrors();
		errorFileName = null;
	    }
	    output("error: " + msg);
	    nerrors++;

	} else if (src instanceof String) {
	    String fileName = (String)src;

	    if (!fileName.equals(errorFileName)) {
		flushErrors();
		errorFileName = fileName;
	    }
	    if (err < 1000) {
		nerrors++;
	    } else {
		nwarnings++;
	    }
	    insertError(where, msg);

	} else if (src instanceof File) {
	    reportError(((File)src).getPath(), where, err, msg);

	} else if (src instanceof Identifier) {
	    reportError(src.toString(), where, err, msg);

	} else if (src instanceof ClassDeclaration) {
	    reportError(((ClassDeclaration)src).getClassDefinition(this), where, err, msg);

	} else if (src instanceof ClassDefinition) {
	    ClassDefinition c = (ClassDefinition)src;
	    if (err < 1000) {
		c.setError(true);
	    } 
	    reportError(c.getSource(), where, err, msg);

	} else if (src instanceof FieldDefinition) {
	    reportError(((FieldDefinition)src).getClassDeclaration(), where, err, msg);

	} else {
	    output(src + ":error=" + err + ":" + msg);
	}
    }

    /**
     * Issue an error
     */
    public void error(Object source, int where, int err, Object arg1, Object arg2, Object arg3) {
	reportError(source, where, err, errorString(err, arg1, arg2, arg3));
    }

    /**
     * Output a string. This can either be an error message or something
     * for debugging. This should be used instead of println.
     */
    public void output(String msg) {
	int len = msg.length();
	for (int i = 0 ; i < len ; i++) {
	    out.write(msg.charAt(i));
	}
	out.write('\n');
    }
}
