/*
 * @(#)BatchParser.java	1.14 95/01/31 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.javac;

import java.tools.java.*;
import java.tools.tree.*;

import java.util.Vector;
import java.util.Enumeration;
import java.io.InputStream;

/**
 * Batch file parser, this needs more work.
 */

public final
class BatchParser extends Parser {
    /**
     * The current package
     */
    Identifier pkg;

    /**
     * The current imports
     */
    Imports imports;

    /**
     * The classes defined in this file
     */
    Vector classes;
     

    /**
     * The current class
     */
    SourceClass sourceClass;

    /**
     * The toplevel environment
     */
    Environment toplevelEnv;
    
    /**
     * Create a batch file parser
     */
    public BatchParser(Environment env, InputStream in) {
	super(env, in);

	imports = new Imports();
	classes = new Vector();
	toplevelEnv = env;
    }

    /**
     * Resolve a name into a full class name
     */
    protected Identifier resolveClass(Identifier nm) {
	try {
	    return imports.resolve(env, nm);
	} catch (AmbiguousClass e) {
	    env.error(pos, ERR_AMBIG_CLASS, e.name1, e.name2);
	} catch (ClassNotFound e) {
	}
	return (pkg != null) ? Identifier.lookup(pkg, nm) : nm;
    }

    /**
     * Package declaration
     */
    protected void packageDeclaration(int where, Identifier nm) {
	//System.out.println("package " + nm);

	if (pkg == null) {
	    imports.addPackage(pkg = nm);
	} else {
	    env.error(where, ERR_PACKAGE_REPEATED);
	}
    }

    /**
     * Import class
     */
    protected void importClass(int where, Identifier nm) {
	//System.out.println("import class " + nm);
	try {
	    imports.addClass(nm);
	} catch (AmbiguousClass e) {
	    env.error(where, ERR_AMBIG_CLASS, e.name1, e.name2);
	}
    }

    /**
     * Import package
     */
    protected void importPackage(int where, Identifier nm) {
	//println("import package " + nm);
	imports.addPackage(nm);
    }

    /**
     * Define class
     */
    protected void beginClass(int where, String doc, int mod, Identifier nm,
			       Identifier sup, Identifier impl[]) {

	// Find the class
	nm = (pkg != null) ? Identifier.lookup(pkg, nm) : nm;
	ClassDeclaration c = (ClassDeclaration)env.getClassDeclaration(nm);

	// Make sure this is the first definition
	if (c.isDefined()) {
	    env.error(where, ERR_CLASS_MULTIDEF, c.getName(), c.getClassDefinition().getSource());
	}

	// Find the super class
	ClassDeclaration superClass = null;
	if ((sup == null) || ((mod & M_INTERFACE) != 0)) {
	    superClass = env.getClassDeclaration(idJavaLangObject);
	} else if (!c.getName().equals(idJavaLangObject)) {
	    superClass = env.getClassDeclaration(sup);
	}

	// Find interfaces
	ClassDeclaration interfaces[] = new ClassDeclaration[impl.length];
	for (int i = 0 ; i < interfaces.length ; i++) {
	    interfaces[i] = env.getClassDeclaration(impl[i]);
	}

	sourceClass = new SourceClass(toplevelEnv, where, c, doc, mod, superClass, interfaces, imports);
	sourceClass.getClassDeclaration().setDefinition(sourceClass, CS_PARSED);
	env = new Environment(toplevelEnv, sourceClass);

	// Setup interdependencies between the classes in this file,
	// this is needed so that when one is recompiled, all are.
	for (Enumeration e = classes.elements() ; e.hasMoreElements() ;) {
	    ClassDefinition otherClass = (ClassDefinition)e.nextElement();
	    otherClass.addDependency(sourceClass.getClassDeclaration());
	    sourceClass.addDependency(otherClass.getClassDeclaration());
	}
	classes.addElement(sourceClass);
    }

    /**
     * End class
     */
    protected void endClass(int where, Identifier nm) {
	if ((!sourceClass.isInterface()) && (sourceClass.getFirstMatch(idInit) == null)) {
	    defineField(sourceClass.getWhere(), null, M_PUBLIC, Type.tMethod(sourceClass.getType()),
			idInit, new Identifier[0], 
			new CompoundStatement(sourceClass.getWhere(), new Statement[0]));
	}
	sourceClass = null;
	env = toplevelEnv;
    }

    /**
     * Define a method
     */
    protected void defineField(int where, String doc, int mod, Type t, Identifier nm,
			       Identifier args[], Node val) {
	if (sourceClass.isInterface()) {
	    mod |= M_PUBLIC;
	    if (t.isType(TC_METHOD)) {
		mod |= M_ABSTRACT;
	    } else {
		mod |= M_STATIC | M_FINAL;
	    }
	}
	if (nm.equals(idInit)) {
	    if (sourceClass.getType().equals(t.getReturnType())) {
		t = Type.tMethod(Type.tVoid, t.getArgumentTypes());
	    } else {
		env.error(where, ERR_INVALID_METHOD_DECL);
		return;
	    }
	}

	Vector v = null;

	if (args != null) {
	    v = new Vector(args.length);
	    for (int i = 0 ; i < args.length ; i++) {
		v.addElement(args[i]);
	    }
	}

	SourceField f = new SourceField(where, sourceClass, doc, mod, t, nm, v, val);
	sourceClass.addField(env, f);
	if (env.dump()) {
	    f.print(System.out);
	}
    }
}
