/*
 * @(#)SourceClass.java	1.28 95/04/05 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.javac;

import java.tools.java.*;
import java.tools.tree.*;
import java.tools.tree.CompoundStatement;
import java.tools.asm.Assembler;
import java.tools.asm.ConstantPool;
import java.util.Vector;
import java.util.Enumeration;
import java.io.OutputStream;
import java.io.DataOutputStream;
import java.io.OutputStreamBuffer;
import java.io.File;

/**
 * This class represents an Java class as it is read from
 * an Java source file.
 */

public final
class SourceClass extends ClassDefinition {
    /**
     * The classes and packages imported by this class
     */
    Imports imports;

    /**
     * The default constructor
     */
    SourceField defConstructor;

    /**
     * The constant pool
     */
    ConstantPool tab = new ConstantPool();

    /**
     * Constructor
     */
    public SourceClass(Environment env, int where, ClassDeclaration declaration, String documentation,
		       int modifiers, ClassDeclaration superClass, ClassDeclaration interfaces[], 
		       Imports imports) {
	super(env.getSource(), where, declaration, modifiers, superClass, interfaces);
	this.imports = imports;
	this.documentation = documentation;
    }

    /**
     * Return imports
     */
    public final Imports getImports() {
	return imports;
    }

    /**
     * Add a dependency
     */
    public void addDependency(ClassDeclaration c) {
	if (tab != null) {
	    tab.put(c);
	}
    }

    /**
     * Check method override
     * REMIND: rewrite so that it can go into SourceClass, then merge SourceClass and CompilerClass
     */
    void checkOverride(Environment env, ClassDeclaration clazz, FieldDefinition field) {
	ClassDefinition c = clazz.getClassDefinition(env);
	for (FieldDefinition f = c.getFirstMatch(field.getName()) ; f != null ; f = f.getNextMatch()) {
	    if (f.isPrivate() || f.isVariable() || !field.getType().equalArguments(f.getType())) {
		continue;
	    }
	    if (f.isFinal()) {
		env.error(field.getWhere(), ERR_FINAL_METH_OVERRIDE, field, f.getClassDeclaration());
		return;
	    }
	    Type t1 = field.getType().getReturnType();
	    Type t2 = f.getType().getReturnType();
	    if ((!t1.equals(t2)) && ((!t1.isType(TC_CLASS)) || (!t2.isType(TC_CLASS)) ||
		    (!env.getClassDefinition(t2).implementedBy(env, env.getClassDeclaration(t1))))) {
		env.error(field.getWhere(), ERR_REDEF_RETURN_TYPE, field, f);
		return;
	    }
	    if (field.isStatic() != f.isStatic()) {
		env.error(field.getWhere(), ERR_OVERRIDE_STATIC_METH, f, f.getClassDeclaration());
		return;
	    }
	    if (field.isConstructor()) {
		continue;
	    }
	    if (field.isAbstract() && !f.isAbstract()) {
		env.error(field.getWhere(), ERR_ABSTRACT_OVERRIDE, field, f, f.getClassDeclaration());
		return;
	    }
	    if ((!field.isPublic()) && (!f.isPrivate())) {
		if (f.isPublic()) {
		    env.error(field.getWhere(), ERR_OVERRIDE_PUBLIC, f, f.getClassDeclaration());
		    return;
		}
		if (f.isProtected() && !field.isProtected()) {
		    env.error(field.getWhere(), ERR_OVERRIDE_PROTECTED, f, f.getClassDeclaration());
		    return;
		}
		if (field.isPrivate()) {
		    env.error(field.getWhere(), ERR_OVERRIDE_PRIVATE, f, f.getClassDeclaration());
		    return;
		}
	    }
	}

	// check the super class
	if (c.getSuperClass() != null) {
	    checkOverride(env, c.getSuperClass(), field);
	}

	// check the interfaces
	ClassDeclaration intf[] = c.getInterfaces();
	for (int i = 0 ; i < intf.length ; i++) {
	    checkOverride(env, intf[i], field);
	}
    }

    /**
     * Add a field (check it first)
     */
    public void addField(Environment env, FieldDefinition f) {
	if (f.isMethod()) {
	    if (f.isConstructor()) {
		if (f.getClassDefinition().isInterface()) {
		    env.error(f.getWhere(), ERR_INTF_CONSTRUCTOR);
		    return;
		}
		if (f.isNative() || f.isAbstract() || f.isStatic() || f.isSynchronized() || f.isFinal()) {
		    env.error(f.getWhere(), ERR_CONSTR_MODIFIER, f);
		    f.subModifiers(M_NATIVE | M_ABSTRACT | M_STATIC | M_SYNCHRONIZED | M_FINAL);
		}
	    } else if (f.isInitializer()) {
		if (f.getClassDefinition().isInterface()) {
		    env.error(f.getWhere(), ERR_INTF_INITIALIZER);
		    return;
		}
	    }
	    if (f.getClassDefinition().isInterface() && (f.isStatic() || f.isSynchronized() || f.isNative() || f.isFinal())) {
		env.error(f.getWhere(), ERR_INTF_MODIFIER, f);
		f.subModifiers(M_STATIC |  M_SYNCHRONIZED | M_NATIVE | M_FINAL);
	    }
	    if (f.isTransient()) {
		env.error(f.getWhere(), ERR_TRANSIENT_METH, f);
		f.subModifiers(M_TRANSIENT);
	    }
	    if (f.isThreadsafe()) {
		env.error(f.getWhere(), ERR_THREADSAFE_METH, f);
		f.subModifiers(M_THREADSAFE);
	    }
	    if (f.isStatic() && f.isAbstract()) {
		env.error(f.getWhere(), ERR_STATIC_MODIFIER, f);
		f.subModifiers(M_STATIC);
	    }
	    if (f.isAbstract() || f.isNative()) {
		if (f.getValue() != null) {
		    env.error(f.getWhere(), ERR_INVALID_METH_BODY, f);
		    f.setValue(null);
		}
	    } else {
		if (f.getValue() == null) {
		    env.error(f.getWhere(), ERR_NO_METH_BODY, f);
		    f.addModifiers(M_ABSTRACT);
		}
	    }
	} else {
	    if (f.getType().isType(TC_VOID)) {
		env.error(f.getWhere(), ERR_VOID_INST_VAR, f.getName());
		// REMIND: set type to error
		return;
	    }
	    if (f.isSynchronized() || f.isAbstract() || f.isNative()) {
		env.error(f.getWhere(), ERR_VAR_MODIFIER, f);
		f.subModifiers(M_SYNCHRONIZED | M_ABSTRACT | M_NATIVE);
	    }
	    if (f.isTransient() && (f.isFinal() || f.isStatic())) {
		env.error(f.getWhere(), ERR_TRANSIENT_MODIFIER, f);
		f.subModifiers(M_FINAL | M_STATIC);
	    }
	    if (f.isFinal() && (f.getValue() == null)) {
		env.error(f.getWhere(), ERR_INITIALIZER_NEEDED, f);
		f.subModifiers(M_FINAL);
	    }
	}

	if (!f.isInitializer()) {
	    for (FieldDefinition f2 = getFirstMatch(f.getName()) ; f2 != null ; f2 = f2.getNextMatch()) {
		if (f.isMethod()) {
		    if (f.getType().equals(f2.getType())) {
			env.error(f.getWhere(), ERR_METH_MULTIDEF, f);
			return;
		    }
		    if (f.getType().equalArguments(f2.getType())) {
			env.error(f.getWhere(), ERR_METH_REDEF_RETTYPE, f, f2);
			return;
		    }
		} else {
		    env.error(f.getWhere(), ERR_VAR_MULTIDEF, f, f2);
		    return;
		}
	    }
	}

	addField(f);
    }

    /**
     * Check this class.
     */
    public void check(Environment env) {
	Identifier nm = getClassDeclaration().getName();
	if (env.verbose()) {
	    env.output("[checking class " + nm + "]");
	}

	// Make sure it was defined in the right file
	if (isPublic()) {
	    String fname = getName().getName() + ".java";
	    String src = ((File)getSource()).getPath();
	    if (env.warnings() && (!src.equals(fname)) && (!src.endsWith(File.separator + fname))) {
		env.error(getWhere(), WARN_PUBLIC_CLASS_FILE, this, fname);
	    }
	}

	if (isInterface()) {
	    if (isFinal()) {
		env.error(getWhere(), ERR_FINAL_INTF, nm);
	    }
	} else {
	    // Check super class
	    if (getSuperClass() != null) {
		try {
		    ClassDefinition def = getSuperClass().getClassDefinition(env);
		    if (!canAccess(env, getSuperClass())) {
			env.error(getWhere(), ERR_CANT_ACCESS_CLASS, getSuperClass());
			superClass = null;
		    } else if (def.isFinal()) {
			env.error(getWhere(), ERR_SUPER_IS_FINAL, getSuperClass());
			superClass = null;
		    } else if (def.isInterface()) {
			env.error(getWhere(), ERR_SUPER_IS_INTF, getSuperClass());
			superClass = null;
		    } else if (superClassOf(env, getSuperClass())) {
			env.error(getWhere(), ERR_CYCLIC_SUPER);
			superClass = null;
		    }
		} catch (ClassNotFound e) {
		    env.error(getWhere(), ERR_SUPER_NOT_FOUND, e.name, this);
		    superClass = null;
		}
	    }
	}

	// Check interfaces
	for (int i = 0 ; i < interfaces.length ; i++) {
	    ClassDeclaration intf = interfaces[i];
	    try {
		if (!canAccess(env, intf)) {
		    env.error(getWhere(), ERR_CANT_ACCESS_CLASS, intf);
		} else if (!intf.getClassDefinition(env).isInterface()) {
		    env.error(getWhere(), ERR_NOT_INTF, intf);
		} else if (isInterface() && implementedBy(env, intf)) {
		    env.error(getWhere(), ERR_CYCLIC_INTF, intf);
		}
	    } catch (ClassNotFound ee) {
		env.error(getWhere(), ERR_INTF_NOT_FOUND, ee.name, this);
	    }
	}

	// bail out if there were any errors
	if (getError()) {
	    return;
	}

	try {
	    // Check override
	    for (FieldDefinition f = getFirstField() ; f != null ; f = f.getNextField()) {
		if (f.isVariable()) {
		    continue;
		}
		
		// check the super class
		if (getSuperClass() != null) {
		    checkOverride(env, getSuperClass(), f);
		}

		// check the interfaces
		for (int i = 0 ; i < interfaces.length ; i++) {
		    checkOverride(env, interfaces[i], f);
		}
	    }
	} catch (ClassNotFound ee) {
	    env.error(getWhere(), ERR_CLASS_NOT_FOUND, ee.name, this);
	}

	// bail out if there were any errors
	if (getError()) {
	    return;
	}

	// Check field definitions
	for (FieldDefinition f = getFirstField() ; f != null ; f = f.getNextField()) {
	    try {
		f.check(env);
	    } catch (ClassNotFound ee) {
		env.error(f.getWhere(), ERR_CLASS_NOT_FOUND, ee.name, this);
	    }
	}

	// bail out if there were any errors
	if (getError()) {
	    return;
	}

	if (isFinal() && isAbstract(env)) {
	    FieldDefinition f = (FieldDefinition)getAbstractFields(env).nextElement();
	    env.error(getWhere(), ERR_FINAL_ABSTRACT, this, f, f.getClassDeclaration());
	}
    }


    /**
     * A list of active ongoing compilations. This list
     * is used to stop two compilations from saving the
     * same class.
     */
    private static Vector active = new Vector();

    /**
     * Compile this class
     */
    public void compile(Environment env, OutputStream out) {
	synchronize (active) {
	    while (active.contains(getName())) {
		active.wait();
	    }
	    active.addElement(getName());
	}

	try {
	    compileClass(env, out);
	} finally {
	    synchronize (active) {
		active.removeElement(getName());
		active.notifyAll();
	    }
	}
    }

    private void compileClass(Environment env, OutputStream out) {
	env = new Environment(env, this);
	check(env);

	// bail out if there were any errors
	if (getError()) {
	    return;
	}

	Vector variables = new Vector();
	Vector methods = new Vector();
	CompilerField init = new CompilerField(new FieldDefinition(getWhere(), this, M_STATIC, Type.tMethod(Type.tVoid), idClassInit, null), new Assembler());
	Context ctx = new Context(init.field);

	//System.out.println("compile class " + getName());

	// Generate code for all fields
	for (FieldDefinition field = getFirstField() ; field != null ; field = field.getNextField()) {
	    //System.out.println("compile field " + field.getName());

	    try {
		if (field.isMethod()) {
		    if (field.isInitializer()) {
			((SourceField)field).code(env, init.asm);
		    } else {
			CompilerField f = new CompilerField(field, new Assembler());
			((SourceField)field).code(env, f.asm);
			methods.addElement(f);
		    }
		} else {
		    CompilerField f = new CompilerField(field, null);
		    variables.addElement(f);
		    if (field.isStatic()) {
			field.codeInit(env, ctx, init.asm);
		    }
		}
	    } catch (InternalError ee) {
		ee.printStackTrace();
		env.error(field, 0, ERR_GENERIC, field.getClassDeclaration() + ":" + field + "@" + ee.toString(), null, null);
	    }
	}
	if (!init.asm.empty()) {
	    init.asm.add(getWhere(), opc_return);
	    methods.addElement(init);
	}

	// bail out if there were any errors
	if (getError()) {
	    return;
	}

	// Insert constants
	tab.put("Code");
	tab.put("ConstantValue");
	tab.put("LocalVariables");
	tab.put("SourceFile");
	if (!env.optimize()) {
	    tab.put("LineNumberTable");
	}

	String sourceFile = ((File)getSource()).getAbsolutePath();
	tab.put(sourceFile);
	tab.put(getClassDeclaration());
	if (getSuperClass() != null) {
	    tab.put(getSuperClass());
	}
	for (int i = 0 ; i < interfaces.length ; i++) {
	    tab.put(interfaces[i]);
	}

	// Optimize Code and Collect method constants
	for (Enumeration e = methods.elements() ; e.hasMoreElements() ; ) {
	    CompilerField f = (CompilerField)e.nextElement();
	    try {
		f.asm.optimize(env);
		f.asm.collect(env, tab);
		tab.put(f.name);
		tab.put(f.sig);
	    } catch (Exception ee) {
		ee.printStackTrace();
		env.error(f.field, -1, ERR_GENERIC, f.field.getName() + "@" + ee.toString(), null, null);
		f.asm.listing(System.out);
	    }
	}

	// Collect field constants
	for (Enumeration e = variables.elements() ; e.hasMoreElements() ; ) {
	    CompilerField f = (CompilerField)e.nextElement();
	    tab.put(f.name);
	    tab.put(f.sig);

	    Object val = f.field.getInitialValue();
	    if (val != null) {
		tab.put((val instanceof String) ? new StringExpression(f.field.getWhere(), (String)val) : val);
	    }
	}

	// Write header
	DataOutputStream data = new DataOutputStream(out);
	data.writeInt(JAVA_MAGIC);
	data.writeShort(JAVA_MINOR_VERSION);
	data.writeShort(JAVA_VERSION);
	tab.write(env, data);

	// Write class information
	data.writeShort(getModifiers() & MM_CLASS);
	data.writeShort(tab.index(getClassDeclaration()));
	data.writeShort((getSuperClass() != null) ? tab.index(getSuperClass()) : 0);
	data.writeShort(interfaces.length);
	for (int i = 0 ; i < interfaces.length ; i++) {
	    data.writeShort(tab.index(interfaces[i]));
	}

	// write variables
	OutputStreamBuffer buf = new OutputStreamBuffer(256);
	DataOutputStream databuf = new DataOutputStream(buf);

	data.writeShort(variables.size());
	for (Enumeration e = variables.elements() ; e.hasMoreElements() ; ) {
	    CompilerField f = (CompilerField)e.nextElement();
	    Object val = f.field.getInitialValue();

	    data.writeShort(f.field.getModifiers() & MM_FIELD);
	    data.writeShort(tab.index(f.name));
	    data.writeShort(tab.index(f.sig));

	    if (val != null) {
		data.writeShort(1);
		data.writeShort(tab.index("ConstantValue"));
		data.writeInt(2);
		data.writeShort(tab.index((val instanceof String) ? new StringExpression(f.field.getWhere(), (String)val) : val));
	    } else {
		data.writeShort(0);
	    }
	}

	// write methods

	data.writeShort(methods.size());
	for (Enumeration e = methods.elements() ; e.hasMoreElements() ; ) {
	    CompilerField f = (CompilerField)e.nextElement();
	    data.writeShort(f.field.getModifiers() & MM_METHOD);
	    data.writeShort(tab.index(f.name));
	    data.writeShort(tab.index(f.sig));

	    if (!f.asm.empty()) {
		data.writeShort(1);
		f.asm.write(env, databuf, tab);
		data.writeShort(tab.index("Code"));
		data.writeInt(buf.size());
		buf.writeTo(data);
		buf.reset();
	    } else {
		data.writeShort(0);
	    }
	}

	// class attributes
	data.writeShort(1);
	data.writeShort(tab.index("SourceFile"));
	data.writeInt(2);
	data.writeShort(tab.index(sourceFile));

	// Cleanup
	data.flush();
	tab = null;
    }
}
