/*
 * @(#)SourceField.java	1.22 95/02/20 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.javac;

import java.tools.java.*;
import java.tools.tree.*;
import java.tools.asm.*;
import java.util.Vector;
import java.util.Enumeration;
import java.io.PrintStream;

/**
 * A Source Field
 */

public final
class SourceField extends FieldDefinition implements Constants {
    /**
     * The argument names (if it is a method)
     */
    Vector args;

    /**
     * The status of the field
     */
    int status;

    static final int PARSED	= 0;
    static final int CHECKING  	= 1;
    static final int CHECKED	= 2;
    static final int INLINING	= 3;
    static final int INLINED	= 4;
    static final int ERROR	= 5;

    public Vector getArguments() {
	return args;
    }

    /**
     * Constructor
     */
    public SourceField(int where, ClassDefinition clazz, String doc, int modifiers, Type type,
		       Identifier name, Vector argNames, Node value) {
	super(where, clazz, modifiers, type, name, value);
	this.documentation = doc;

	// Create a list of arguments
	if (isMethod()) {
	    args = new Vector();

	    if (isConstructor() || !isStatic()) {
		args.addElement(new LocalField(where, clazz, 0, clazz.getType(), idThis));
	    }

	    Enumeration e = argNames.elements();
	    Type argTypes[] = getType().getArgumentTypes();
	    for (int i = 0 ; i < argTypes.length ; i++) {
		args.addElement(new LocalField(where, clazz, 0, argTypes[i],
					       (Identifier)e.nextElement()));
	    }
	}
    }

    /**
     * Resolve a class name
     */
    public Identifier resolve(Environment env, Identifier nm) {
	return ((SourceClass)getClassDefinition()).imports.resolve(env, nm);
    }

    /**
     * Check this field
     */
    public void check(Environment env) {
	if (status == PARSED) {
	    if (env.dump()) {
		System.out.println("[check field " + getClassDeclaration().getName() + "." + getName() + "]");
		if (getValue() != null) {
		    getValue().print(System.out);
		    System.out.println();
		}
	    }
	    env = new Environment(env, this);
	    env.resolve(where, getClassDefinition(), getType());

	    status = CHECKING;
	    if (getValue() != null) {
		Context ctx = new Context(this);

		if (isMethod()) {
		    Statement s = (Statement)getValue();
		    long vset = 0;

		    for (Enumeration e = args.elements() ; e.hasMoreElements() ; ) {
			LocalField f = (LocalField)e.nextElement();
			vset |= (long)1 << ctx.declare(env, f);
		    }
		    //System.out.println("VSET = " + vset);

		    // undefine this if needed
		    if (isConstructor()) {
			vset &= ~1;
			if ((!s.firstConstructor()) && (getClassDefinition().getSuperClass() != null)) {
			    Statement body[] = {
				new ExpressionStatement(where,
				    new MethodExpression(where, new SuperExpression(where),
					     idInit, new Expression[0])), s
			    };
			    s = new CompoundStatement(where, body);
			    setValue(s);
			}
		    }
		    //System.out.println("VSET = " + vset);
		    s.checkMethod(env, ctx, vset);
		} else {
		    Expression e = (Expression)getValue();
		    e.checkInitializer(env, ctx, isStatic() ? 0 : 1, getType());
		    setValue(e.convert(env, ctx, getType(), e));
		}
		if (env.dump()) {
		    getValue().print(System.out);
		    System.out.println();
		}
	    }
	    status = getClassDefinition().getError() ? ERROR : CHECKED;
	}
    }

    /**
     * Inline the field
     */
    void inline(Environment env) {
	switch (status) {
	  case PARSED:
	    check(env);
	    inline(env);
	    break;

	  case CHECKED:
	    if (env.dump()) {
		System.out.println("[inline field " + getClassDeclaration().getName() + "." + getName() + "]");
	    }
	    status = INLINING;
	    env = new Environment(env, this);

	    if (isMethod()) {
		if ((!isNative()) && (!isAbstract())) {
		    Context ctx = new Context(this);
		    for (Enumeration e = args.elements() ; e.hasMoreElements() ; ) {
			ctx.declare(env, (LocalField)e.nextElement());
		    }
		    setValue(((Statement)getValue()).inline(env, ctx));
		}
	    } else {
		if (getValue() != null)  {
		    Context ctx = new Context(this);
		    setValue(((Expression)getValue()).inlineValue(env, ctx));
		}
	    }
	    if (env.dump()) {
		System.out.println("[inlined field " + getClassDeclaration().getName() + "." + getName() + "]");
		if (getValue() != null) {
		    getValue().print(System.out);
		    System.out.println();
		} else {
		    System.out.println("<empty>");
		}
	    }
	    status = INLINED;
	    break;
	}
    }

    /**
     * Get the value of the field (or null if the value can't be determined)
     */
    public Node getValue(Environment env) {
	if (getValue() == null) {
	    return null;
	}
	switch (status) {
	  case PARSED:
	    check(env);
	    return getValue(env);

	  case CHECKED:
	    inline(env);
	    return getValue(env);

	  case INLINED:
	    return getValue();
	}
	return null;
    }

    public boolean isInlineable(Environment env, boolean fromFinal) {
	if (super.isInlineable(env, fromFinal)) {
	    getValue(env);
	    return (status == INLINED) && !getClassDefinition().getError();
	}
	return false;
    }
	

    /**
     * Get the initial value of the field
     */
    public Object getInitialValue() {
	if (isMethod() || (getValue() == null) || (!isFinal()) || (status != INLINED)) {
	    return null;
	}
	return ((Expression)getValue()).getValue();
    }

    /**
     * Generate code
     */
    public void code(Environment env, Assembler asm) {
	switch (status) {
	  case PARSED:
	    check(env);
	    code(env, asm);
	    return;

	  case CHECKED:
	    inline(env);
	    code(env, asm);
	    return;

	  case INLINED:
	    // Actually generate code
	    if (env.dump()) {
		System.out.println("[code field " + getClassDeclaration().getName() + "." + getName() + "]");
	    }
	    if (isMethod() && (!isNative()) && (!isAbstract())) {
		env = new Environment(env, this);
		Context ctx = new Context(this);
		Statement s = (Statement)getValue();

		for (Enumeration e = args.elements() ; e.hasMoreElements() ; ) {
		    LocalField f = (LocalField)e.nextElement();
		    ctx.declare(env, f);
		    //ctx.declare(env, (LocalField)e.nextElement());
		}

		/*
		if (isConstructor() && ((s == null) || !s.firstConstructor())) {
		    ClassDeclaration c = getClassDefinition().getSuperClass();
		    if (c != null) {
			FieldDefinition field = c.getClassDefinition(env).matchMethod(env, idInit);
			asm.add(getWhere(), opc_aload, new Integer(0));
			asm.add(getWhere(), opc_invokenonvirtual, field);
			asm.add(getWhere(), opc_pop);
		    }

		    // Ouput initialization code
		    for (FieldDefinition f = getClassDefinition().getFirstField() ; f != null ; f = f.getNextField()) {
			if (!f.isStatic()) {
			    f.codeInit(env, ctx, asm);
			}
		    }
		}
		*/
		if (s != null) {
		    s.code(env, ctx, asm);
		}
		if (getType().getReturnType().isType(TC_VOID) && !isInitializer()) {
		    asm.add(getWhere(), opc_return);
		}
	    }
	    return;
	}
    }

    public void codeInit(Environment env, Context ctx, Assembler asm) {
	if (isMethod()) {
	    return;
	}
	switch (status) {
	  case PARSED:
	    check(env);
	    codeInit(env, ctx, asm);
	    return;

	  case CHECKED:
	    inline(env);
	    codeInit(env, ctx, asm);
	    return;

	  case INLINED:
	    // Actually generate code
	    if (env.dump()) {
		System.out.println("[code initializer  " + getClassDeclaration().getName() + "." + getName() + "]");
	    }
	    if (getValue() != null) {
		Expression e = (Expression)getValue();
		if (isStatic()) {
		    if ((getInitialValue() == null) && !e.equals(0)) {
			e.codeValue(env, ctx, asm);
			asm.add(getWhere(), opc_putstatic, this);
		    }
		} else if (!e.equals(0)) {
		    asm.add(getWhere(), opc_aload, new Integer(0));
		    e.codeValue(env, ctx, asm);
		    asm.add(getWhere(), opc_putfield, this);
		}
	    }
	    return;
	}
    }

    /**
     * Print for debugging
     */
    public void print(PrintStream out) {
	super.print(out);
	if (getValue() != null) {
	    getValue().print(out);
	    out.println();
	}
    }
}
