/*
 * @(#)Expression.java	1.19 95/04/02 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.tree;

import java.tools.java.*;
import java.tools.asm.Label;
import java.tools.asm.Assembler;
import java.io.PrintStream;

public
class Expression extends Node {
    Type type;

    /**
     * Constructor
     */
    Expression(int op, int where, Type type) {
	super(op, where);
	this.type = type;
    }

    /**
     * Return the precedence of the operator
     */
    int precedence() {
	return (op < opPrecedence.length) ? opPrecedence[op] : 100;
    }

    /**
     * Order the expression based on precedence
     */
    public Expression order() {
	return this;
    }

    /**
     * Return true if constant
     */
    public boolean isConstant() {
        return false;
    }

    /**
     * Return the constant value
     */
    public Object getValue() {
        return null;
    }

    /**
     * Check if the expression is equal to a value
     */
    public boolean equals(int i) {
	return false;
    }
    public boolean equals(boolean b) {
	return false;
    }
    public boolean equals(Identifier id) {
	return false;
    }

    /**
     * Return true if it is a type
     */
    boolean isType(Environment env, Context ctx) {
	return false;
    }

    /**
     * Convert an expresion to a type
     */
    Type toType(Environment env, Context ctx) {
	env.error(where, ERR_INVALID_TYPE_EXPR);
	return Type.tError;
    }

    /**
     * See if this expression fits in the given type.
     * This is useful because some larger number fit into
     * smaller types.
     */
    public boolean fitsType(Environment env, Type t) {
	return env.distance(type, t) < 10;
    }

    /**
     * Check an expression
     */
    public long checkValue(Environment env, Context ctx, long vset) {
	return vset;
    }
    public long checkInitializer(Environment env, Context ctx, long vset, Type t) {
	return checkValue(env, ctx, vset);
    }
    public long check(Environment env, Context ctx, long vset) {
	env.error(where, WARN_VALUE_NOT_USED);
	return checkValue(env, ctx, vset);
    }

    /**
     * Check a condition
     */
    public void checkCondition(Environment env, Context ctx, ConditionVars cvars) {
	long vs0 = cvars.vsTrue & cvars.vsFalse;
	long vs1 = checkValue(env, ctx, vs0) & ~vs0;
	cvars.vsTrue |= vs1;
	cvars.vsFalse |= vs1;
    }
    public ConditionVars checkCondition(Environment env, Context ctx, long vset) {
	ConditionVars cvars = new ConditionVars();
	cvars.vsTrue = cvars.vsFalse = vset;
	checkCondition(env, ctx, cvars);
	return cvars;
    }

    /**
     * Evaluate
     */
    Expression eval() {
	return this;
    }

    /**
     * Simplify
     */
    Expression simplify() {
	return this;
    }

    /**
     * Inline
     */
    public Expression inline(Environment env, Context ctx) {
	return null;
    }
    public Expression inlineValue(Environment env, Context ctx) {
	return this;
    }
    public Expression inlineLHS(Environment env, Context ctx) {
	return null;
    }

    /**
     * The cost of inlining this expression
     */
    public int costInline(int thresh) {
	return 1;
    }

    /**
     * Code
     */
    void codeBranch(Environment env, Context ctx, Assembler asm, Label lbl, boolean whenTrue) {
	if (type.isType(TC_BOOLEAN)) {
	    codeValue(env, ctx, asm);
	    asm.add(where, whenTrue ? opc_ifne : opc_ifeq, lbl);
	} else {
	    throw new InternalError("codeBranch " + opNames[op]);
	}
    }
    public void codeValue(Environment env, Context ctx, Assembler asm) {
	if (type.isType(TC_BOOLEAN)) {
	    Label l1 = new Label();
	    Label l2 = new Label();

	    codeBranch(env, ctx, asm, l1, true);
	    asm.add(where, opc_ldc2, new Integer(0));
	    asm.add(where, opc_goto, l2);
	    asm.add(l1);
	    asm.add(where, opc_ldc2, new Integer(1));
	    asm.add(l2);
	} else {
	    throw new InternalError("codeValue");
	}
    }
    public void code(Environment env, Context ctx, Assembler asm) {
	codeValue(env, ctx, asm);

	switch (type.getTypeCode()) {
	  case TC_VOID:
	    break;

	  case TC_DOUBLE:
	  case TC_LONG:
	    asm.add(where, opc_pop2);
	    break;
	    
	  default:
	    asm.add(where, opc_pop);
	    break;
	}
    }
    int codeLValue(Environment env, Context ctx, Assembler asm) {
	print(System.out);
	throw new InternalError("invalid lhs");
    }
    void codeLoad(Environment env, Context ctx, Assembler asm) {
	print(System.out);
	throw new InternalError("invalid load");
    }
    void codeStore(Environment env, Context ctx, Assembler asm) {
	print(System.out);
	throw new InternalError("invalid store");
    }
    void codeAppend(Environment env, Context ctx, Assembler asm, ClassDeclaration c) {
	codeValue(env, ctx, asm);
	Type argTypes[] = {type};
	FieldDefinition f = c.getClassDefinition(env).matchMethod(env, idAppend, argTypes);
	asm.add(where, opc_invokevirtual, f);
    }

    /**
     * Code
     */
    void codeDup(Environment env, Context ctx, Assembler asm, int items, int depth) {
	switch (items) {
	  case 0:
	    return;
	    
	  case 1:
	    switch (depth) {
	      case 0:
		asm.add(where, opc_dup);
		return;
	      case 1:
		asm.add(where, opc_dup_x1);
		return;
	      case 2:
		asm.add(where, opc_dup_x2);
		return;
		
	    }
	    break;
	  case 2:
	    switch (depth) {
	      case 0:
		asm.add(where, opc_dup2);
		return;
	      case 1:
		asm.add(where, opc_dup2_x1);
		return;
	      case 2:
		asm.add(where, opc_dup2_x2);
		return;
		
	    }
	    break;
	}
	throw new InternalError("can't dup: " + items + ", " + depth);
    }

    void codeConversion(Environment env, Context ctx, Assembler asm, Type f, Type t) {
	int from = f.getTypeCode();
	int to = t.getTypeCode();

	switch (to) {
	  case TC_BYTE:
	    if (from != TC_BYTE) {
		codeConversion(env, ctx, asm, f, Type.tInt);
		asm.add(where, opc_int2byte);
	    }
	    return;
	  case TC_CHAR:
	    if (from != TC_CHAR) {
		codeConversion(env, ctx, asm, f, Type.tInt);
		asm.add(where, opc_int2char);
	    }
	    return;
	  case TC_SHORT:
	    if (from != TC_SHORT) {
		codeConversion(env, ctx, asm, f, Type.tInt);
		asm.add(where, opc_int2short);
	    }
	    return;
	  case TC_INT:
	    switch (from) {
	      case TC_BYTE:
	      case TC_CHAR:
	      case TC_SHORT:
	      case TC_INT:
		return;
	      case TC_LONG:
		asm.add(where, opc_l2i);
		return;
	      case TC_FLOAT:
		asm.add(where, opc_f2i);
		return;
	      case TC_DOUBLE:
		asm.add(where, opc_d2i);
		return;
	    }
	    break;
	  case TC_LONG:
	    switch (from) {
	      case TC_BYTE:
	      case TC_CHAR:
	      case TC_SHORT:
	      case TC_INT:
		asm.add(where, opc_i2l);
		return;
	      case TC_LONG:
		return;
	      case TC_FLOAT:
		asm.add(where, opc_f2l);
		return;
	      case TC_DOUBLE:
		asm.add(where, opc_d2l);
		return;
	    }
	    break;
	  case TC_FLOAT:
	    switch (from) {
	      case TC_BYTE:
	      case TC_CHAR:
	      case TC_SHORT:
	      case TC_INT:
		asm.add(where, opc_i2f);
		return;
	      case TC_LONG:
		asm.add(where, opc_l2f);
		return;
	      case TC_FLOAT:
		return;
	      case TC_DOUBLE:
		asm.add(where, opc_d2f);
		return;
	    }
	    break;
	  case TC_DOUBLE:
	    switch (from) {
	      case TC_BYTE:
	      case TC_CHAR:
	      case TC_SHORT:
	      case TC_INT:
		asm.add(where, opc_i2d);
		return;
	      case TC_LONG:
		asm.add(where, opc_l2d);
		return;
	      case TC_FLOAT:
		asm.add(where, opc_f2d);
		return;
	      case TC_DOUBLE:
		return;
	    }
	    break;

	  case TC_CLASS:
	    switch (from) {
	      case TC_NULL:
		return;
	      case TC_CLASS:
	      case TC_ARRAY:
		if (!env.implicitCast(f, t)) {
		    asm.add(where, opc_checkcast, env.getClassDeclaration(t));
		}
		return;
	    }
	    
	    break;
	    
	  case TC_ARRAY:
	    switch (from) {
	      case TC_NULL:
		return;
	      case TC_CLASS:
	      case TC_ARRAY:
		if (!env.implicitCast(f, t)) {
		    asm.add(where, opc_checkcast, t);
		}
		return;
	    }
	    break;
	}
	throw new InternalError("codeConversion: " + from + ", " + to);
    }

    /**
     * Check if the first thing is a constructor invokation
     */
    public boolean firstConstructor() {
	return false;
    }

    /**
     * Create a copy of the expression for method inlining
     */
    public Expression copyInline(Context ctx) {
	return (Expression)clone();
    }

    /**
     * Print
     */
    public void print(PrintStream out) {
	out.print(opNames[op]);
    }
}
