/*
 * @(#)FieldExpression.java	1.17 95/04/02 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.tree;

import java.tools.java.*;
import java.tools.asm.*;
import java.io.PrintStream;

public
class FieldExpression extends UnaryExpression {
    Identifier id;
    FieldDefinition field;

    /**
     * constructor
     */
    public FieldExpression(int where, Expression right, Identifier id) {
	super(FIELD, where, Type.tError, right);
	this.id = id;
    }
    public FieldExpression(int where, Context ctx, FieldDefinition field) {
	super(FIELD, where, field.getType(),
	      field.isStatic() ? null : new ThisExpression(where, ctx));
	this.id = field.getName();
	this.field = field;
    }

    /**
     * Convert an '.' expression to a qualified identifier
     */
    Identifier toIdentifier() {
	StringBuffer buf = new StringBuffer();
	FieldExpression e = this;
	buf.appendChar('.');
	buf.append(id);
	while (e.right.op == FIELD) {
	    e = (FieldExpression)e.right;
	    buf.insert(0, e.id);
	    buf.insertChar(0, '.');
	}
	if (e.right.op != IDENT) {
	    return null;
	}
	buf.insert(0, ((IdentifierExpression)e.right).id);
	return Identifier.lookup(buf.toString());
    }

    /**
     * Check if this is a type
     */
    boolean isType(Environment env, Context ctx) {
	Identifier id = toIdentifier();
	return (id != null) && env.classExists(id);
    }

    /**
     * Convert an '.' expression to a type
     */
    Type toType(Environment env, Context ctx) {
	Identifier id = toIdentifier();
	if (id == null) {
	    env.error(where, ERR_INVALID_TYPE_EXPR);
	    return Type.tError;
	}
	return Type.tClass(id);
    }

    /**
     * Check the expression
     */
    public long checkValue(Environment env, Context ctx, long vset) {
	boolean staticRef = false;

	if (right.isType(env, ctx)) {
	   right = new TypeExpression(right.where, right.toType(env, ctx));
	   staticRef = true;
	} else {
	   vset = right.checkValue(env, ctx, vset);
	}

	if (!right.type.isType(TC_CLASS)) {
	    if (right.type.isType(TC_ARRAY) && id.equals(idLength)) {
		type = Type.tInt;
		return vset;
	    }
	    if (!right.type.isType(TC_ERROR)) {
		env.error(where, ERR_INVALID_FIELD_REFERENCE, id, right.type);
	    }
	    return vset;
	}

	try {
	    ClassDefinition clazz = env.getClassDefinition(right.type);
	    field = clazz.getVariable(env, id);
	    if (field == null) {
		if ((field = clazz.findAnyMethod(env, id)) != null) {
		    env.error(where, ERR_INVALID_FIELD, id, field.getClassDeclaration());
		} else {
		    env.error(where, ERR_NO_SUCH_FIELD, id, clazz);
		}
		return vset;
	    }

	    type = field.getType();

	    if (!ctx.field.getClassDefinition().canAccess(env, field)) {
		env.error(where, ERR_NO_FIELD_ACCESS, id, clazz, ctx.field.getClassDeclaration());
		return vset;
	    }
	    if (staticRef && (!field.isStatic()) && (!field.getClassDefinition().implementedBy(env, ctx.field.getClassDeclaration()))) {
		env.error(where, ERR_NO_STATIC_FIELD_ACCESS, id, clazz);
		return vset;
	    }
	    if ((!field.isStatic()) && (right.op == THIS) && ((vset & 1) == 0)) {
		env.error(where, ERR_ACCESS_INST_BEFORE_SUPER, id);
	    }

	    ctx.field.getClassDefinition().addDependency(field.getClassDeclaration());
	} catch (ClassNotFound e) {
	    env.error(where, ERR_CLASS_NOT_FOUND, e.name, ctx.field);

	} catch (AmbiguousField e) {
	    env.error(where, ERR_AMBIG_FIELD, id, e.field1.getClassDeclaration(), e.field2.getClassDeclaration());
	}
	return vset;
    }

    /**
     * Inline
     */
    public Expression inline(Environment env, Context ctx) {
	return (right != null) ? right.inline(env, ctx) : null;
    }
    public Expression inlineValue(Environment env, Context ctx) {
	if (field == null) {
	    if (id.equals(idLength)) {
		return new LengthExpression(where, right).inlineValue(env, ctx);
	    }
	    return this;
	}

	if ((right != null) && (right.op == TYPE)) {
	    right = null;
	}

	if (field.isFinal()) {
	    Expression e = (Expression)field.getValue(env);
	    if ((e != null) && e.isConstant()) {
		return new CommaExpression(where, right, e).inlineValue(env, ctx);
	    }
	}

	if (right != null) {
	    if (field.isStatic()) {
		Expression e = right.inline(env, ctx);
		right = null;
		if (e != null) {
		    return new CommaExpression(where, e, this);
		}
	    } else {
		right = right.inlineValue(env, ctx);
	    }
	}
	return this;
    }
    public Expression inlineLHS(Environment env, Context ctx) {
	if ((right != null) && (right.op == TYPE)) {
	    right = null;
	}
	if (right != null) {
	    if (field.isStatic()) {
		Expression e = right.inline(env, ctx);
		right = null;
		if (e != null) {
		    return new CommaExpression(where, e, this);
		}
	    } else {
		right = right.inlineValue(env, ctx);
	    }
	}
	return this;
    }

    /**
     * The cost of inlining this expression
     */
    public int costInline(int thresh) {
	return 3 + ((right != null) ? right.costInline(thresh) : 0);
    }

    /**
     * Code
     */
    int codeLValue(Environment env, Context ctx, Assembler asm) {
	if (field.isStatic()) {
	    if (right != null) {
		right.code(env, ctx, asm);
		return 1;
	    }
	    return 0;
	}
	right.codeValue(env, ctx, asm);
	return 1;
    }
    void codeLoad(Environment env, Context ctx, Assembler asm) {
	if (field == null) {
	    throw new InternalError("should not be null");
	}
	if (field.isStatic()) {
	    asm.add(where, opc_getstatic, field);
	} else {
	    asm.add(where, opc_getfield, field);
	}
    }
    void codeStore(Environment env, Context ctx, Assembler asm) {
	if (field.isStatic()) {
	    asm.add(where, opc_putstatic, field);
	} else {
	    asm.add(where, opc_putfield, field);
	}
    }
    public void codeValue(Environment env, Context ctx, Assembler asm) {
	codeLValue(env, ctx, asm);
	codeLoad(env, ctx, asm);
    }

    /**
     * Print
     */
    public void print(PrintStream out) {
	out.print("(" + opNames[op] + " ");
	if (right != null) {
	    right.print(out);
	} else {
	    out.print("<empty>");
	}
	out.print(" " + id + ")");
    }
}
