/*
 * @(#)FinallyStatement.java	1.9 95/01/31 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.tree;

import java.tools.java.*;
import java.tools.asm.Assembler;
import java.tools.asm.Label;
import java.tools.asm.TryData;
import java.tools.asm.CatchData;
import java.io.PrintStream;

public
class FinallyStatement extends Statement {
    Statement body;
    Statement finalbody;

    /**
     * Constructor
     */
    public FinallyStatement(int where, Statement body, Statement finalbody) {
	super(FINALLY, where);
	this.body = body;
	this.finalbody = finalbody;
    }

    /**
     * Check statement
     */
    long check(Environment env, Context ctx, long vset) {
	vset = reach(env, vset);
	finalbody.check(env, new CheckContext(ctx, this), vset);

	CheckContext newctx = new CheckContext(ctx, this);
	vset = body.check(env, newctx, vset);

	long deadend = vset & DEAD_END;
	vset &= ~DEAD_END & newctx.vsBreak;
	return finalbody.check(env, new CheckContext(ctx, this), vset) | deadend;
    }

    /**
     * Inline
     */
    public Statement inline(Environment env, Context ctx) {
	if (body != null) {
	    body = body.inline(env, ctx);
	}
	if (finalbody != null) {
	    finalbody = finalbody.inline(env, ctx);
	}
	if (body == null) {
	    return eliminate(env, finalbody);
	}
	if (finalbody == null) {
	    return eliminate(env, body);
	}
	return this;
    }

    /**
     * Create a copy of the statement for method inlining
     */
    public Statement copyInline(Context ctx, boolean valNeeded) {
	throw new InternalError("copyInline");
    }

    /**
     * Code
     */
    public void code(Environment env, Context ctx, Assembler asm) {
	ctx = new Context(ctx, null);
	LocalField f1 = new LocalField(where, ctx.field.getClassDefinition(), 0, Type.tObject, null);
	LocalField f2 = new LocalField(where, ctx.field.getClassDefinition(), 0, Type.tInt, null);
	Integer num1 = new Integer(ctx.declare(env, f1));
	Integer num2 = new Integer(ctx.declare(env, f2));

	Label endLabel = new Label();

	TryData td = new TryData();
	td.add(null);

	// Main body
	CodeContext bodyctx = new CodeContext(ctx, this);
	asm.add(where, opc_try, td);
	body.code(env, bodyctx, asm);
	asm.add(bodyctx.breakLabel);
	asm.add(td.getEndLabel());

	// Cleanup afer body
	asm.add(where, opc_jsr, bodyctx.contLabel);
	asm.add(where, opc_goto, endLabel);

	// Catch code
	CatchData cd = td.getCatch(0);
	asm.add(cd.getLabel());
	asm.add(where, opc_astore, num1);
	asm.add(where, opc_jsr, bodyctx.contLabel);
	asm.add(where, opc_aload, num1);
	asm.add(where, opc_athrow);

	// Final body
	CodeContext finalctx = new CodeContext(ctx, this);
	asm.add(bodyctx.contLabel);
	asm.add(where, opc_astore, num2);
	finalbody.code(env, ctx, asm);
	asm.add(finalctx.breakLabel);
	asm.add(where, opc_ret, num2);

	asm.add(endLabel);
    }

    /**
     * Print
     */
    public void print(PrintStream out, int indent) {
	super.print(out, indent);
	out.print("try ");
	if (body != null) {
	    body.print(out, indent);
	} else {
	    out.print("<empty>");
	}
	out.print(" finally ");
	if (finalbody != null) {
	    finalbody.print(out, indent);
	} else {
	    out.print("<empty>");
	}
    }
}
