/*
 * @(#)IdentifierExpression.java	1.14 95/04/02 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.tree;

import java.tools.java.*;
import java.tools.asm.Assembler;
import java.io.PrintStream;

public
class IdentifierExpression extends Expression {
    Identifier id;
    FieldDefinition field;
    
    /**
     * Constructor
     */
    public IdentifierExpression(int where, Identifier id) {
	super(IDENT, where, Type.tError);
	this.id = id;
    }

    /**
     * Check if the expression is equal to a value
     */
    public boolean equals(Identifier id) {
	return this.id.equals(id);
    }

    /**
     * Assign a value to this identifier
     */
    long assign(Environment env, Context ctx, long vset) {
	if (bind(env, ctx)) {
	    if (field.isLocal()) {
		LocalField local = (LocalField)field;
		vset |= (long)1 << local.number;
		local.writecount++;
	    }
	    if (field.isFinal()) {
		env.error(where, ERR_ASSIGN_TO_FINAL, id);
	    }
	}
	return vset;
    }

    /**
     * Bind to a field
     */
    boolean bind(Environment env, Context ctx) {
	try {
	    field = ctx.getField(env, id);
	    if (field == null) {
		env.error(where, ERR_UNDEF_VAR, id);
		return false;
	    }

	    type = field.getType();

	    // Check access permission
	    if (!ctx.field.getClassDefinition().canAccess(env, field)) {
		env.error(where, ERR_NO_FIELD_ACCESS, id, field.getClassDeclaration(), ctx.field.getClassDeclaration());
		return false;
	    }

	    // Check acces of static variables
	    if (!ctx.field.canAccess(env, field)) {
		env.error(where, ERR_NO_STATIC_FIELD_ACCESS, id, field.getClassDeclaration());
		return false;
	    }

	    // Check forward reference
	    if (!ctx.field.canReach(env, field)) {
		env.error(where, ERR_FORWARD_REF, id, field.getClassDeclaration());
		return false;
	    }
	    return true;
	} catch (AmbiguousField e) {
	    env.error(where, ERR_AMBIG_FIELD, id, e.field1.getClassDeclaration(), e.field2.getClassDeclaration());
	}
	return false;
    }

    /**
     * Check expression
     */
    public long checkValue(Environment env, Context ctx, long vset) {
	if (bind(env, ctx)) {
	    if (field.isLocal()) {
		LocalField local = (LocalField)field;
		if ((vset & ((long)1 << local.number)) == 0) {
		    env.error(where, ERR_VAR_NOT_INITALIZED, id);
		    vset |= (long)1 << local.number;
		}
		local.readcount++;
	    } else if (!field.isStatic()) {
		if ((vset & 1) == 0) {
		    env.error(where, ERR_ACCESS_INST_BEFORE_SUPER, id);
		}
	    }
	    ctx.field.getClassDefinition().addDependency(field.getClassDeclaration());
	}
	return vset;
    }

    /**
     * And identifier can only be a type if it is undefined.
     */
    boolean isType(Environment env, Context ctx) {
	try {
	    return ctx.getField(env, id) == null;
	} catch (ClassNotFound e) {
	} catch (AmbiguousField e) {
	    return true;
	} 
	return false;
    }

    /**
     * Convert an identifier to a type
     */
    Type toType(Environment env, Context ctx) {
	try {
	    return Type.tClass(ctx.field.resolve(env, id));
	} catch (AmbiguousClass ee) {
	    env.error(where, ERR_AMBIG_CLASS, ee.name1, ee.name2);
	} catch (ClassNotFound ee) {
	}
	return Type.tClass(id);
    }

    /**
     * Inline
     */
    public Expression inline(Environment env, Context ctx) {
	return null;
    }
    public Expression inlineValue(Environment env, Context ctx) {
	if (field == null) {
	    return this;
	}
	if (field.isLocal()) {
	    if (field.isFinal()) {
		Expression e = (Expression)field.getValue(env);
		return (e != null) ? e : this;
	    }
	    return this;
	}
	return new FieldExpression(where, ctx, field).inlineValue(env, ctx);
    }
    public Expression inlineLHS(Environment env, Context ctx) {
	if (field.isLocal()) {
	    return this;
	}
	return new FieldExpression(where, ctx, field).inlineValue(env, ctx);
    }

    /**
     * Code
     */
    int codeLValue(Environment env, Context ctx, Assembler asm) {
	return 0;
    }
    void codeLoad(Environment env, Context ctx, Assembler asm) {
	asm.add(where, opc_iload + type.getTypeCodeOffset(),
		new Integer(((LocalField)field).number));
    }
    void codeStore(Environment env, Context ctx, Assembler asm) {
	LocalField local = (LocalField)field;
	asm.add(where, opc_istore + type.getTypeCodeOffset(),
		new Integer(local.number));
    }
    public void codeValue(Environment env, Context ctx, Assembler asm) {
	codeLValue(env, ctx, asm);
	codeLoad(env, ctx, asm);
    }

    /**
     * Print
     */
    public void print(PrintStream out) {
	out.print(id + "#" + ((field != null) ? field.hashCode() : 0));
    }
}
