/*
 * @(#)MethodExpression.java	1.24 95/04/02 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.tree;

import java.tools.java.*;
import java.tools.asm.Assembler;
import java.util.Vector;
import java.io.PrintStream;

public
class MethodExpression extends NaryExpression {
    Identifier id;
    ClassDefinition clazz;
    FieldDefinition field;
    
    /**
     * constructor
     */
    public MethodExpression(int where, Expression right, Identifier id, Expression args[]) {
	super(METHOD, where, Type.tError, right, args);
	this.id = id;
    }
    public MethodExpression(int where, Expression right, FieldDefinition field, Expression args[]) {
	super(METHOD, where, field.getType().getReturnType(), right, args);
	this.id = field.getName();
	this.field = field;
	this.clazz = field.getClassDefinition();
    }

    /**
     * Check expression type
     */
    public long checkValue(Environment env, Context ctx, long vset) {
	ClassDeclaration c = null;
	boolean staticRef = false;
	    
	if (right == null) {
	    staticRef = ctx.field.isStatic();
	    c = ctx.field.getClassDeclaration();
	} else {
	    if (id.equals(idInit)) {
		if (!ctx.field.isConstructor()) {
		    env.error(where, ERR_INVALID_CONSTR_INVOKE);
		    return vset;
		}
		if ((vset & 1) != 0) {
		    env.error(where, ERR_CONSTR_INVOKE_NOT_FIRST);
		    return vset;
		}
		right.checkValue(env, ctx, vset | 1); 
	    } else {
		if (right.isType(env, ctx)) {
		    staticRef = true;
		    right = new TypeExpression(right.where, right.toType(env, ctx));
		} else {
		    vset = right.checkValue(env, ctx, vset);
		}

	    }
	    if (right.type.isType(TC_CLASS)) {
		c = env.getClassDeclaration(right.type);
	    } else if (right.type.isType(TC_ARRAY)) {
		c = env.getClassDeclaration(Type.tObject);
	    } else {
		if (!right.type.isType(TC_ERROR)) {
		    env.error(where, ERR_INVALID_METHOD_INVOKE, right.type);
		}
		return vset;
	    }
	}

	// Compose a list of argument types
	boolean hasErrors = false;
	Type argTypes[] = new Type[args.length];

	for (int i = 0 ; i < args.length ; i++) {
	    vset = args[i].checkValue(env, ctx, vset);
	    argTypes[i] = args[i].type;
	    hasErrors = hasErrors || argTypes[i].isType(TC_ERROR);
	}

	// Check if there are any type errors in the arguments
	if (hasErrors) {
	    return vset;
	}

	try {
	    // Get the method field, given the argument types
	    clazz = c.getClassDefinition(env);
	    field = clazz.matchMethod(env, id, argTypes);
	    if (field == null) {

		if (id.equals(idInit)) {
		    String sig = clazz.getName().getName().toString();
		    sig = Type.tMethod(Type.tError, argTypes).typeString(sig, false, false);
		    env.error(where, ERR_UNMATCHED_CONSTR, sig, c);
		} else {
		    String sig = id.toString();
		    sig = Type.tMethod(Type.tError, argTypes).typeString(sig, false, false);
		    if (clazz.findAnyMethod(env, id) == null) {
			env.error(where, ERR_UNDEF_METH, sig, c);
		    } else {
			env.error(where, ERR_UNMATCHED_METH, sig, c);
		    }
		}
		return vset;
	    }

	    // Check access permissions
	    if (!ctx.field.getClassDefinition().canAccess(env, field)) {
		env.error(where, ERR_NO_METH_ACCESS, field, field.getClassDeclaration(), ctx.field.getClassDeclaration());
		return vset;
	    }
	    
	    // Make sure that static references are allowed
	    if (staticRef && (!field.isStatic()) &&
		(ctx.field.isStatic() || (!field.getClassDefinition().implementedBy(env, ctx.field.getClassDeclaration())))) {
		env.error(where, ERR_NO_STATIC_METH_ACCESS, field, field.getClassDeclaration());
		return vset;
	    }

	    // Make sure that we are not invoking an abstract method
	    if (field.isAbstract() && (right != null) && (right.op == SUPER)) {
		env.error(where, ERR_INVOKE_ABSTRACT, field, field.getClassDeclaration());
		return vset;
	    }

	    // Check for recursive constructor
	    if (field.isConstructor() && ctx.field.equals(field)) {
		env.error(where, WARN_RECURSIVE_CONSTR, field);
	    }

	    ctx.field.getClassDefinition().addDependency(field.getClassDeclaration());
	} catch (ClassNotFound ee) {
	    env.error(where, ERR_CLASS_NOT_FOUND, ee.name, ctx.field);
	    return vset;
	    
	} catch (AmbiguousField ee) {
	    env.error(where, ERR_AMBIG_FIELD, id, ee.field1, ee.field2);
	    return vset;
	}

	// Make sure it is qualified
	if ((right == null) && !field.isStatic()) {
	    right = new ThisExpression(where, ctx);
	}

	// Cast arguments
	argTypes = field.getType().getArgumentTypes();
	for (int i = 0 ; i < args.length ; i++) {
	    args[i] = convert(env, ctx, argTypes[i], args[i]);
	}

	type = field.getType().getReturnType();

	// "this" is defined after the constructor invokation
	if (id.equals(idInit)) {
	    vset |= 1;
	}
	return vset;
    }

    /**
     * Check void expression
     */
    public long check(Environment env, Context ctx, long vset) {
	return checkValue(env, ctx, vset);
    }

    /**
     * Inline
     */
    final int MAXINLINECOST = 30;

    Expression inlineMethod(Environment env, Context ctx, Statement s, boolean valNeeded) {
	if (env.dump()) {
	    System.out.println("INLINE METHOD " + field + " in " + ctx.field);
	}
	Vector v = field.getArguments();
	Statement body[] = new Statement[v.size() + 2];

	int n = 0;
	if (field.isStatic()) {
	    body[0] = new ExpressionStatement(where, right);
	} else {
	    if ((right != null) && (right.op == SUPER)) {
		right = new ThisExpression(right.where, ctx);
	    }
	    body[0] = new VarDeclarationStatement(where, (LocalField)v.elementAt(n++), right);
	}
	for (int i = 0 ; i < args.length ; i++) {
	    body[i + 1] = new VarDeclarationStatement(where, (LocalField)v.elementAt(n++), args[i]);
	}
	//System.out.print("BEFORE:"); s.print(System.out); System.out.println();
	body[body.length - 1] = (s != null) ? s.copyInline(ctx, valNeeded) : null;
	//System.out.print("COPY:"); body[body.length - 1].print(System.out); System.out.println();


	Expression e = new InlineMethodExpression(where, type, field, new CompoundStatement(where, body));
	return valNeeded ? e.inlineValue(env, ctx) : e.inline(env, ctx);
    }
    
    public Expression inline(Environment env, Context ctx) {
	if (right != null) {
	    right = field.isStatic() ? right.inline(env, ctx) : right.inlineValue(env, ctx);
	}
	for (int i = 0 ; i < args.length ; i++) {
	    args[i] = args[i].inlineValue(env, ctx);
	}
	Expression e = this;
	if (env.optimize() && field.isInlineable(env, clazz.isFinal()) &&
	    ((id == null) || !id.equals(idInit)) && 
	    (!ctx.field.isInitializer()) && ctx.field.isMethod() &&
	    (ctx.getInlineFieldContext(field) == null)) {
	    Statement s = (Statement)field.getValue(env);
	    if ((s == null) || (s.costInline(MAXINLINECOST) < MAXINLINECOST))  {
		e = inlineMethod(env, ctx, s, false);
	    }
	}
	if (ctx.field.isConstructor() && field.isConstructor() && (right != null) && (right.op == SUPER) && (id != null)) {
	    id = null;
	    // insert instance initializers
	    for (FieldDefinition f = ctx.field.getClassDefinition().getFirstField() ; f != null ; f = f.getNextField()) {
		if (f.isVariable() && !f.isStatic()) {
		    Expression val = (Expression)f.getValue(env);
		    if ((val != null) && !val.equals(0)) {
			FieldExpression fe = new FieldExpression(f.getWhere(),
						 new ThisExpression(f.getWhere(), ctx), f.getName());
			fe.field = f;
			val = new AssignExpression(f.getWhere(), fe, val.copyInline(ctx));
			val = val.inline(env, ctx);
			if (val != null) {
			    e = new CommaExpression(f.getWhere(), e, val);
			}
		    }
		}
	    }
	}
	return e;
    }
    public Expression inlineValue(Environment env, Context ctx) {
	if (right != null) {
	    right = field.isStatic() ? right.inline(env, ctx) : right.inlineValue(env, ctx);
	}
	for (int i = 0 ; i < args.length ; i++) {
	    args[i] = args[i].inlineValue(env, ctx);
	}
	if (env.optimize() && field.isInlineable(env, clazz.isFinal()) &&
	    (!ctx.field.isInitializer()) && ctx.field.isMethod() &&
	    (ctx.getInlineFieldContext(field) == null)) {
	    Statement s = (Statement)field.getValue(env);
	    if ((s == null) || (s.costInline(MAXINLINECOST) < MAXINLINECOST))  {
		return inlineMethod(env, ctx, s, true);
	    }
	}
	return this;
    }

    /**
     * Code
     */
    public void codeValue(Environment env, Context ctx, Assembler asm) {
	if (field.isStatic()) {
	    if (right != null) {
		right.code(env, ctx, asm);
	    }
	} else if (right == null) {
	    asm.add(where, opc_aload, new Integer(0));
	} else {
	    right.codeValue(env, ctx, asm);
	}

	for (int i = 0 ; i < args.length ; i++) {
	    args[i].codeValue(env, ctx, asm);
	}

	if (field.isStatic()) {
	    asm.add(where, opc_invokestatic, field);
	} else if (field.isConstructor() || field.isPrivate() || ((right != null) && (right.op == SUPER))) {
	    asm.add(where, opc_invokenonvirtual, field);
	} else if (field.getClassDefinition().isInterface()) {
	    asm.add(where, opc_invokeinterface, field);
	} else {
	    asm.add(where, opc_invokevirtual, field);
	}
    }

    /**
     * Check if the first thing is a constructor invokation
     */
    public boolean firstConstructor() {
	return id.equals(idInit);
    }

    /**
     * Print
     */
    public void print(PrintStream out) {
	out.print("(" + opNames[op]);
	if (right != null) {
	    out.print(" ");
	    right.print(out);
	} 
	out.print(" " + ((id == null) ? idInit : id)); 
	for (int i = 0 ; i < args.length ; i++) {
	    out.print(" ");
	    if (args[i] != null) {
		args[i].print(out);
	    } else {
		out.print("<null>");
	    }
	}
	out.print(")");
    }
}
