/*
 * @(#)Statement.java	1.11 95/01/31 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package java.tools.tree;

import java.tools.java.*;
import java.tools.asm.Assembler;
import java.tools.asm.Label;
import java.io.PrintStream;

public
class Statement extends Node {
    public static final long DEAD_END = (long)1 << 63;
    Identifier label;

    /**
     * Constructor
     */
    Statement(int op, int where) {
	super(op, where);
    }

    /**
     * Set the label of a statement
     */
    public void setLabel(Environment env, Expression e) {
	if (e.op == IDENT) {
	    label = ((IdentifierExpression)e).id;
	} else {
	    env.error(e.where, ERR_INVALID_LABEL);
	}
    }

    /**
     * Check a statement
     */
    public long checkMethod(Environment env, Context ctx, long vset) {
	vset = check(env, ctx, vset);

	// Check for return
	if (!ctx.field.getType().getReturnType().isType(TC_VOID)) {
	    if ((vset & DEAD_END) == 0) {
		env.error(ctx.field.getWhere(), ERR_RETURN_REQUIRED_AT_END, ctx.field);
	    }
	}
	return vset;
    }
    long checkDeclaration(Environment env, Context ctx, long vset, int mod, Type t) {
	throw new InternalError("checkDeclaration");
    }
    long check(Environment env, Context ctx, long vset) {
	throw new InternalError("check");
    }

    long reach(Environment env, long vset) {
	if ((vset & DEAD_END) != 0) {
	    if (env.warnings()) {
		env.error(where, WARN_STAT_NOT_REACHED);
	    }
	    vset &= ~DEAD_END;
	}
	return vset;
    }

    /**
     * Inline
     */
    public Statement inline(Environment env, Context ctx) {
	return this;
    }

    /**
     * Eliminate this statement, which is only possible if it has no label.
     */
    public Statement eliminate(Environment env, Statement s) {
	if ((s != null) && (label != null)) {
	    Statement args[] = {s};
	    s = new CompoundStatement(where, args);
	    s.label = label;
	}
	return s;
    }
     

    /**
     * Code
     */
    public void code(Environment env, Context ctx, Assembler asm) {
	throw new InternalError("code");
    }
    void codeFinally(Environment env, Context ctx, Assembler asm, Context stopctx, Type save) {
	Integer num = null;
	
	for (Context c = ctx ; (c != null)  && (c != stopctx) ; c = c.prev) {
	    if ((c.node != null) && ((c.node.op == FINALLY) || (c.node.op == SYNCHRONIZE))) {
		if ((num == null) && (save != null)) {
		    num = new Integer(ctx.declare(env, new LocalField(where, ctx.field.getClassDefinition(), 0, save, null)));
		    asm.add(where, opc_istore + save.getTypeCodeOffset(), num);
		}
		asm.add(where, opc_jsr, ((CodeContext)c).contLabel);
	    }
	}
	if (num != null) {
	    asm.add(where, opc_iload + save.getTypeCodeOffset(), num);
	}
    }

    /**
     * Check if the first thing is a constructor invokation
     */
    public boolean firstConstructor() {
	return false;
    }

    /**
     * Create a copy of the statement for method inlining
     */
    public Statement copyInline(Context ctx, boolean valNeeded) {
	return (Statement)clone();
    }

    /**
     * The cost of inlining this statement
     */
    public int costInline(int thresh) {
	return thresh;
    }

    /**
     * Print
     */
    void printIndent(PrintStream out, int indent) {
	for (int i = 0 ; i < indent ; i++) {
	    out.print("    ");
	}
    }
    public void print(PrintStream out, int indent) {
	if (label != null) {
	    out.print(label + ": ");
	}
    }
    public void print(PrintStream out) {
	print(out, 0);
    }
}
