/*
 * @(#)util.c	1.15 95/03/13 Chris Warth
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*-
	utility routines needed by both the compiler and the interpreter.
 */

#include "oobj.h"
#include "interpreter.h"
#include <stdio.h>
#include <string.h>

#if defined(MACINTOSH)
/* use the nice sysWrite we've set up */
/* REMINDSMZ: done this way since I'm not sure where else util.c is used... */
#define write sysWrite
#endif

unicode *
str2unicode(char *str, unicode *ustr, long len)
{
    unicode *dst = ustr;
    int ll = len;

    memset((char *) dst, 0, len * sizeof(*dst));
    while (*str && --len >= 0)
	*ustr++ = *str++;
    return dst;
}


void
unicode2str(unicode *src, char *dst, long len)
{
    while (--len >= 0)
	*dst++ = *src++;
    *dst = '\0';
}

/*
    print s null terminated C-style character string.
*/
void
prints(char *s)
{
#ifdef PRINTS_STDIO
    int    c;
    while (c = *s++)
	putchar(c);
#else
    if (s)
	write(1, s, strlen(s));
#endif
}

/* print an Java-style unicode string */
void
printus(unicode *str, long len)
{
    char obuf[100];
    register int p;
    if (str == 0 || len <= 0)
	return;
    while (len > 0) {
	unicode c;
	for (p = 0; p < sizeof obuf - 10 && --len >= 0;)
	    if ((c = *str++) >= 256) {
		/* make unicode characters readable */
		sprintf(obuf + p, "\\<%X>", c);
		p += strlen(obuf + p);
	    } else
		obuf[p++] = c;
	if (p)
#ifdef PRINTS_STDIO
	    fwrite(obuf, 1, p, stdout);
#else
	    write(1, obuf, p);
#endif
    }
}

#undef  HTSIZE			/* Avoid conflict on PC */
#define HTSIZE 107		/* must be prime */
#define HTOVERFLOWPOINT ((int)(HTSIZE*0.8))

typedef struct StrIDhash {
    struct StrIDhash *next;
    short used, baseid;
    char *hash[HTSIZE];
    unsigned is_malloced[(HTSIZE + 31) >> 5];
    void **param;
} StrIDhash;

/* Given a string, return a unique 16 bit id number.  If param
 * isn't null, we also set up an array of void * for holding info about each
 * object.  The address of this void * is stored into param
 */
unsigned short
Str2ID(StrIDhash **idhroot, register char *s, void ***param, int CopyNeeded)
{
    /*
     * The database is a hash table.  When the hash table overflows, a new
     * hashtable is created chained onto the previous one.  This is done so
     * that we can use the hashtable slot index as the ID, without worrying
     * about having the IDs change when the hashtable grows.
     */
    StrIDhash *h;
    char *s2;
    long i;
    unsigned raw_hash, hash;
    unsigned secondary_hash;
    for (raw_hash = 0, s2 = s; i = *s2; s2++) 
	raw_hash = raw_hash * 37 + i;
    hash = raw_hash % HTSIZE;
    secondary_hash = (raw_hash & 7) + 1; /* between 1 and 8 */
    if ((h = *idhroot) == 0) {
	h = (StrIDhash *) calloc(1, sizeof(*h));
	*idhroot = h;
	h->baseid = 1;  /* guarantee that no ID is 0 */
    }
    while (1) {
	i = hash;
	while (s2 = h->hash[i]) {
	    if (strcmp(h->hash[i], s) == 0) {
		if (param)
		    *param = &h->param[i];
		return h->baseid + i;
	    }
	    if ((i -= secondary_hash) < 0)
		i += HTSIZE;
	}
	if (h->next == 0)
	    break;
	h = h->next;
    }
    /* Couldn't find the string */
    if (h->used >= HTOVERFLOWPOINT) {
	/* have to create a new hashtable */
	h->next = (StrIDhash *) calloc(1, sizeof(*h));
	h->next->baseid = h->baseid + HTSIZE;
	h = h->next;
	i = hash;
    }
    if (CopyNeeded) {
	h->hash[i] = strdup(s);
	h->is_malloced[i >> 5] |= (1 << (i & 31));
    } else  {
	h->hash[i] = s;
    }
    h->used++;
    if (param) {
	if (h->param == 0)
	    h->param = (void **)calloc(HTSIZE, sizeof(*param));
	*param = &h->param[i];
    }
    return h->baseid + i;
}

void
Str2IDFree(StrIDhash **hash_ptr)
{
    StrIDhash *hash, *next;
    int i;
    hash = *hash_ptr;
    while (hash) {
	next = hash->next;
	for (i = 0; i < HTSIZE; i++) {
	    if (hash->is_malloced[i >> 5] & (1 << (i & 31))) {
		free(hash->hash[i]);
	    }
	}
	if (hash->param)
	    free(hash->param);
	free(hash);
	hash = next;
    }
    *hash_ptr = 0;
}

char *
ID2Str(StrIDhash *h, unsigned short ID, void ***param) {
    while ((long)(ID - h->baseid) >= HTSIZE)
	h = h->next;
    if (param) {
	if (h->param == 0)
	    h->param = (void **)calloc(HTSIZE, sizeof(*param));
	*param = &h->param[ID - h->baseid];
    }
    return h->hash[ID - h->baseid];
}

char       *
addstr(char *s, char *buf,
       char *limit, char term)
{
    char        c;
    while ((c = *s) && c != term && buf < limit) {
	*buf++ = c;
	s++;
    }
    return buf;
}

char       *
unicode2rd(unicode *s, long len)
{       /* unicode string to readable C string */
    static char buf[40];
    char *dp = buf;
    int c;
    if (s == 0)
	return "NULL";
    *dp++ = '"';
    while (--len>=0 && (c = *s++) != 0 && dp < buf + sizeof buf - 10)
	if (040 <= c && c < 0177)
	    *dp++ = c;
	else
	    switch (c) {
	    case '\n':
		*dp++ = '\\';
		*dp++ = 'n';
		break;
	    case '\t':
		*dp++ = '\\';
		*dp++ = 't';
		break;
	    case '\r':
		*dp++ = '\\';
		*dp++ = 'r';
		break;
	    case '\b':
		*dp++ = '\\';
		*dp++ = 'b';
		break;
	    case '\f':
		*dp++ = '\\';
		*dp++ = 'f';
		break;
	    default:
		sprintf(dp, "\\%X", c);
		dp += strlen(dp);
		break;
	    }
    *dp++ = '"';
    if (len >= 0 && c != 0)
	*dp++ = '.', *dp++ = '.', *dp++ = '.';
    *dp++ = 0;
    return buf;
}

char *
int642CString(int64_t number, char *buf, int buflen)
{
    ll2str(number, buf, buf + buflen);
    return buf;
}

void
out_of_memory()
{
    write(2, "**Out of memory, exiting**\n", 27);
    abort();
}

void
panic (const char *format, ...)
{
        va_list ap;
	char buf[256];
        va_start(ap, format);
	printf("\n*** panic: ");
    	vsprintf(buf, format, ap);
	prints(buf);
	abort();
}


char *
classname2string(char *src, char *dst, int size) {
    char *buf = dst;
    for (; (--size > 0) && (*src != '\0') ; src++, dst++) {
	if (*src == '/') {
	    *dst = '.';
	}  else {
	    *dst = *src;
	}
    }
    *dst = '\0';
    return buf;
}
