/*
 * @(#)classloader.c	1.55 95/05/01  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*-
 *      Routines for loading and resolving class definitions.
 *      These routines should not depending up on the interpreter or
 *      garbage collector.
 *
 */



#include <string.h>
#include <sys/types.h>
#include <stddef.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <setjmp.h>

#include "oobj.h"
#include "interpreter.h"
#include "tree.h"
#include "bool.h"
#include "signature.h"

#include "byteorder_md.h"
#include "path_md.h"
#include "installpath.h"

#if defined(MACINTOSH)
/* REMINDSMZ: don't want to break the compiler by always using these,
 * so I'm defining them conditionally for now.
 */

#define read	sysRead
#define close	sysClose
#endif

long SeenFile(char *fn, long state);
int LoadFile(char *fn, char *dir, char *SourceHint);
int DoImport(char *name, char *SourceHint);

void
AddBinClass(ClassClass * cb)
{
    register int i = -1;
    if (cb->loader == 0) {
	for (i = nbinclasses; --i >= 0 && strcmp(classname(binclasses[i]), classname(cb)) != 0;);
    }
    if (i < 0) {
	if (nbinclasses >= sizebinclasses)
	    if (binclasses == 0)
		binclasses = (ClassClass **)
		    malloc(sizeof(ClassClass *) * (sizebinclasses = 50));
	    else
		binclasses = (ClassClass **)
		    realloc(binclasses, sizeof(ClassClass *)
			    * (sizebinclasses = nbinclasses * 2));
	if (binclasses == 0)
	    return;
	i = nbinclasses;
	nbinclasses++;
    }
    binclasses[i] = cb;
}

void
DelBinClass(ClassClass * cb)
{
    register int i;
    for (i = nbinclasses; --i >= 0; )
	if (binclasses[i] == cb) {
	    binclasses[i] = binclasses[--nbinclasses];
	    break;
	}
}

/* LoadFile is shared by the compiler and the interpreter.  It does
   different things for each.

   When the compiler askes to load a file for a particular class,
   in the end all it's looking for is a classinfo structure to be
   created for that class.  It can get that either by loading a
   .class file, and creating the classinfo from that, OR, by
   reading a source file and compiling it.

   When the interpreter askes for a file, in the end it's looking
   for a classblock structure to be created.  The only way it can
   get one of those is by loading a compiled class.

   OpenCode tries to open a .class file first.  -2 means it failed
   because it couldn't open the file, -1 means the requested file
   is being compiled.  Of course, in the case of the interpreter,
   OpenCode never returns -1, only a valid fd or -2.

   If OpenCode returns a valid fd, that means it is a .class file.
   If SkipCheckSource is false (meaning we're checking source
   files), and the source file as specified in the compiled .class
   file is more recent than the .class file, then the loaded class
   is thrown away, and OpenCode is called again, only this time,
   with arguments which will cause it to recompile the source.  In
   the case of java, it forks javac to do this.  When this happens,
   it has to reload the resulting .class file.

   Returns 1 when it succeeds, 0 when it fails. */

int
LoadFile(char *fn, char *dir, char *SourceHint)
{
    struct stat st;
    ClassClass *cb = 0;
    int codefd = -1;
    int i;
    int	retryCount = 0;
    unsigned char *external_class;

    codefd = OpenCode(fn, SourceHint, dir, &st);

again:
    switch (codefd) {
      case -2:		    /* open failed */
	return 0;

      case -1:
	return 1;       /* already compiled (or compiling) source */
    }

    /* if loaded or compiling, return A_OK */
    if (SeenFile(fn, sfs_query) > sfs_query) {
	close(codefd);
	return 1;
    }

    /* Snarf the file into memory. */
    external_class = (unsigned char *)malloc(st.st_size);
    if (external_class == 0)
        goto failed;
    if (read(codefd, external_class, st.st_size) != st.st_size)
        goto failed;
    close(codefd);
    codefd = -1;
    cb = calloc(sizeof(ClassClass), 1);
    if (!createInternalClass(external_class, external_class + st.st_size, cb)) {
	free(external_class);
	goto failed;
    }
    free(external_class);
    /* Valid class - let's put it in the class table. */
    AddBinClass(cb);
    
    /* Now, if we're checking sources, and there's a source file
       name in the .class file we just loaded, check to see if we
       should recompile this class.  If so, call try again. */

    if (!SkipSourceChecks && cb->source_name) {
	struct stat st2;
	if (cb->major_version != JAVA_VERSION) {
	    fprintf(stderr, "Warning: Not using %s because it was\n"
		    "\t compiled with version %d.%d of javac.  "
		    "Current version is %d.%d.\n",
		    classname(cb),
		    cb->major_version, cb->minor_version, 
		    JAVA_VERSION, JAVA_MINOR_VERSION);
	}
	if (stat(classsrcname(cb), &st2) >= 0
	    && (st2.st_mtime > st.st_mtime || cb->major_version != JAVA_VERSION)
	    && ++retryCount < 2) {
	    if (cb->major_version != JAVA_VERSION) {
		fprintf(stderr, "Warning: Attempting to recompile from %s...\n",
			classsrcname(cb));
		unlink(fn);
	    }
	    /* remove the class we just defined - we're going again;  */
	    codefd = OpenCode(fn, classsrcname(cb), dir, &st);
	    DelBinClass(cb);
	    FreeClass(cb);
	    goto again;
	} else if (cb->major_version != JAVA_VERSION) {
	    /* remove the class we just defined - we're going
	       to try again (once) */
	    DelBinClass(cb);
	    goto failed;
	}
    }
    SeenFile(fn, sfs_loaded);
    if (verbose)
	fprintf(stderr, "[Loaded %s]\n", fn);
    ClassDefHook(cb);
    return 1;
failed:
    if (codefd >= 0)
	close(codefd);
    if (cb != 0)
        FreeClass(cb);
    return 0;
}


int 
DoImport(char *name, char *SourceHint)
{
    if (name[0] == SIGNATURE_ARRAY) {
	AddBinClass(createFakeArrayClass(name));
	return 1;
    } else {
	return(import_md(name, SourceHint));
    }
}

long
SeenFile(char *fn, long state) {
    return(SeenFile_md(fn, state));
}

struct CICcontext {
    unsigned char *ptr;
    unsigned char *end_ptr;
    ClassClass *cb;
    jmp_buf jump_buffer;
    char **detail;
};

typedef struct CICcontext CICcontext;

static unsigned long get1byte(CICcontext *);
static unsigned long get2bytes(CICcontext *);
static unsigned long get4bytes(CICcontext *);
static char *getAsciz(CICcontext *);
static char *getAscizFromClass(CICcontext *, int i);
static void getNbytes(CICcontext *, int count, char *buffer);
static void LoadConstantPool(CICcontext *);
static void ReadInCode(CICcontext *, struct methodblock *);
static void ReadLineTable(CICcontext *, struct methodblock *mb);
static void ReadLocalVars(CICcontext *, struct methodblock *mb);
static void ReadInCompiledCode(CICcontext *context, struct methodblock *mb);

static void MungeIncompatibleConstantPool(ClassClass *cb);
static char *MungeClassname(char *classname);
static char *MungeSignature(char *signature);

#undef  ERROR		/* Avoid conflict on PC */
#define ERROR(name) { *(context->detail) = name; \
                        longjmp(context->jump_buffer, 1); }

bool_t 
createInternalClass(unsigned char *ptr, unsigned char *end_ptr, ClassClass *cb)
{
    int i, j;
    union cp_item_type *constant_pool;
    unsigned char *type_table;
    int attribute_count;
    unsigned fields_count, methods_count;
    struct methodblock *mb;
    struct fieldblock *fb;
    unsigned long magic;
    struct CICcontext context_block;
    struct CICcontext *context = &context_block;
    char *detail = 0;

    context->ptr = ptr;
    context->end_ptr = end_ptr;
    context->cb = cb;
    context->detail = &detail;
    if (setjmp(context->jump_buffer)) {
	/* Indicate an error of some sort */
	return FALSE;
    }
    
    if (get4bytes(context) != JAVA_CLASSFILE_MAGIC) 
	ERROR("Bad magic number");

    
    cb->minor_version = get2bytes(context);
    cb->major_version = get2bytes(context);
    if (cb->major_version != JAVA_VERSION) 
	ERROR("Bad major version number");

    LoadConstantPool(context);
    constant_pool = cb->constantpool;
    type_table = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;

    cb->access = get2bytes(context) & ACC_WRITTEN_FLAGS;

    /* Get the name of the class */
    i = get2bytes(context);	/* index in constant pool of class */
    cb->name = getAscizFromClass(context, i);
    constant_pool[i].p = cb;
    CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);

    /* Get the super class name. */
    i = get2bytes(context);	/* index in constant pool of class */
    if (i > 0) 
	cb->super_name = getAscizFromClass(context, i);


    i = cb->implements_count = get2bytes(context);
    if (i > 0) {
	unsigned j;
	cb->implements = calloc(i, sizeof(short));
	for (j = 0; j < i; j++) {
	    cb->implements[j] = get2bytes(context);
	}
    }

    fields_count = cb->fields_count = get2bytes(context);
    if (fields_count > 0) 
        cb->fields = (struct fieldblock *)
	       calloc(cb->fields_count, sizeof(struct fieldblock));
    for (i = fields_count, fb = cb->fields; --i >= 0; fb++) {
	fieldclass(fb) = cb;
	fb->access = get2bytes(context) & ACC_WRITTEN_FLAGS;
	fb->name = getAsciz(context);
	fb->signature = getAsciz(context);
	if (cb->minor_version <= 1)
	    /* backwards compatibility */
	    fb->signature == MungeSignature(fb->signature);
	attribute_count = get2bytes(context);
	for (j = 0; j < (int)attribute_count; j++) {
	    char *name = getAsciz(context);
	    int length = get4bytes(context);
	    if ((fb->access & ACC_STATIC) 
		 && strcmp(name, "ConstantValue") == 0) {
		int index;
		if (length != 2) 
		    ERROR("Wrong size for VALUE attribute");
		fb->access |= ACC_VALKNOWN;
		/* we'll change this below */
		fb->u.offset = get2bytes(context); 
	    } else {
		getNbytes(context, length, NULL);
	    }
	}
	if (fb->access & ACC_STATIC) {
	    int index = fb->u.offset; /* set by "VALUE" attribute */
	    if (fb->signature[0] == SIGNATURE_DOUBLE || 
		   fb->signature[0] == SIGNATURE_LONG) {
		twoword_static_address(fb) = malloc(2 * sizeof(long));
		SET_INT64(twoword_static_address(fb),
		    (fb->access & ACC_VALKNOWN) ? 
			  GET_INT64(&constant_pool[index]) : ll_zero_const);
	    } else if (strcmp(fb->signature, "L" JAVAPKG "String;") == 0) {
		if (fb->access & ACC_VALKNOWN)  {
		    /* In the interpreter, this resovles to an Hjava_lang_String.
		     * For the compiler, this resolves to a struct node * */
		    if (ResolveClassStringConstant(cb, index, 0))
			*(void **)normal_static_address(fb) = 
			                        constant_pool[index].p;
		} else { 
		    *normal_static_address(fb) = 0;
		}
	    } else {
		*normal_static_address(fb) = 
		    (fb->access & ACC_VALKNOWN) ? constant_pool[index].i : 0;
	    }
	}
    }

    if ((cb->methods_count = get2bytes(context)) > 0) 
	cb->methods = (struct methodblock *)
	            calloc(cb->methods_count, sizeof(struct methodblock));
    for (i = cb->methods_count, mb = cb->methods; --i >= 0; mb++) {
	bool_t seenCompiled = FALSE;
	fieldclass(&mb->fb) = cb;
	mb->fb.access = get2bytes(context) & ACC_WRITTEN_FLAGS;
	mb->fb.name = getAsciz(context);
	mb->fb.signature = getAsciz(context);
	if (cb->minor_version <= 1)
	    /* backwards compatibility */
	    mb->fb.signature = MungeSignature(mb->fb.signature);
	mb->args_size = Signature2ArgsSize(mb->fb.signature) 
	                + ((mb->fb.access & ACC_STATIC) ? 0 : 1);
	attribute_count = get2bytes(context);
	for (j = 0; j < attribute_count; j++) {
	    extern char *CompiledCodeAttribute;
	    char *name = getAsciz(context);
	    if (CompiledCodeAttribute != NULL 
		  && !strcmp(name, CompiledCodeAttribute) 
		  && cbLoader(cb) == 0) {
		ReadInCompiledCode(context, mb);
		seenCompiled = TRUE;
	    } else if ((strcmp(name, "Code") == 0) && !seenCompiled) {
	        ReadInCode(context, mb);
	    } else {
		int length = get4bytes(context);
		getNbytes(context, length, NULL);
	    }
	}
    }

    /* See if there are class attributes */
    attribute_count = get2bytes(context); 
    for (j = 0; j < attribute_count; j++) {
	char *name = getAsciz(context);
	int length = get4bytes(context);
	if (strcmp(name, "SourceFile") == 0) {
	    if (length != 2) {
		ERROR("Wrong size for VALUE attribute");
	    }
	    cb->source_name = getAsciz(context);
	} else {
	    getNbytes(context, length, NULL);
	}
    }
    return TRUE;
}

static void LoadConstantPool(CICcontext *context) 
{
    ClassClass *cb = context->cb;
    int nconstants = get2bytes(context);
    unsigned char *initial_ptr = context->ptr;
    cp_item_type *constant_pool = 
	       (cp_item_type *)calloc(nconstants, sizeof(cp_item_type));
    unsigned char *type_table = calloc(nconstants, sizeof(char));
    char *malloced_space;
    int malloc_size;
    int i;
    
    cb->constantpool = constant_pool;
    cb->constantpool_count = nconstants;

    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p = type_table;
    for (malloc_size = 0, i = CONSTANT_POOL_UNUSED_INDEX; i < nconstants; i++) {
	int type = get1byte(context);
	switch (type) {
	  case CONSTANT_Asciz: {
	      int length = get2bytes(context);
	      malloc_size += length + 1;
	      getNbytes(context, length, NULL);
	      break;
	  }
	    
	  case CONSTANT_Class:
	  case CONSTANT_String:
	      getNbytes(context, 2, NULL);
	      break;

	  case CONSTANT_Fieldref:
	  case CONSTANT_Methodref:
	  case CONSTANT_InterfaceMethodref:
	  case CONSTANT_NameAndType:
	  case CONSTANT_Integer:
	  case CONSTANT_Float:
	      getNbytes(context, 4, NULL);
	      break;

	  case CONSTANT_Long:
	  case CONSTANT_Double: 
	      i++;
	      getNbytes(context, 8, NULL);
	      break;

	  default: 
	      ERROR("Illegal constant pool type");
	}
    }

    malloced_space = (char *)malloc(malloc_size);
    
    context->ptr = initial_ptr;

    for (i = CONSTANT_POOL_UNUSED_INDEX; i < nconstants; i++) {
	int type = get1byte(context);
	CONSTANT_POOL_TYPE_TABLE_PUT(type_table, i, type);
	switch (type) {
	  case CONSTANT_Asciz: {
	      int length = get2bytes(context);
	      char *result = malloced_space;
	      malloced_space += (length + 1);
	      getNbytes(context, length, result);
	      result[length] = '\0';
	      constant_pool[i].p = result;
	      CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	      break;
	  }
	
	  case CONSTANT_Class:
	  case CONSTANT_String:
	    constant_pool[i].i = get2bytes(context);
	    break;

	  case CONSTANT_Fieldref:
	  case CONSTANT_Methodref:
	  case CONSTANT_InterfaceMethodref:
	  case CONSTANT_NameAndType:
	    constant_pool[i].i = get4bytes(context);
	    break;
	    
	  case CONSTANT_Integer:
	  case CONSTANT_Float:
	    constant_pool[i].i = get4bytes(context);
	    CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	    break;

	  case CONSTANT_Long:
	  case CONSTANT_Double: {
	      unsigned high = get4bytes(context);
	      unsigned low = get4bytes(context);
		  int64_t value = ll_add(ll_shl(int2ll(high), 32), int2ll(low));
	      SET_INT64(&constant_pool[i], value);
	      CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	      i++;		/* increment i for loop, too */
	      CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	      break;
	  }

	  default:
	      ERROR("Illegal constant pool type");
	}
    }
    if (cb->minor_version <= 1) 
	/* Handle backwards compatility */
	MungeIncompatibleConstantPool(cb);
}



static void ReadInCode(CICcontext *context, struct methodblock *mb)
{
    int attribute_length = get4bytes(context);
    unsigned char *end_ptr = context->ptr + attribute_length;
    int attribute_count;
    int code_length;
    int i;

    mb->maxstack = get1byte(context);
    mb->nlocals = get1byte(context);
    code_length = mb->code_length = get2bytes(context);
    mb->code = malloc(code_length);

    getNbytes(context, code_length, (char *)mb->code);
    if ((mb->exception_table_length = get2bytes(context)) > 0) {
	unsigned exception_table_size = mb->exception_table_length 
	                                  * sizeof(struct CatchFrame);
	mb->exception_table = malloc(exception_table_size);
	for (i = 0; i < (int)mb->exception_table_length; i++) {
	    mb->exception_table[i].start_pc = get2bytes(context);
	    mb->exception_table[i].end_pc = get2bytes(context);		
	    mb->exception_table[i].handler_pc = get2bytes(context); 
	    mb->exception_table[i].catchType = get2bytes(context);
	}
    }
    attribute_count = get2bytes(context);
    for (i = 0; i < attribute_count; i++) {
	char *name = getAsciz(context);
	if (strcmp(name, "LineNumberTable") == 0) {
	    ReadLineTable(context, mb);
	} else if (strcmp(name, "LocalVariableTable") == 0) {
	    ReadLocalVars(context, mb);
	} else {
	    int length = get4bytes(context);
	    getNbytes(context, length, NULL);
	}
    }
    if (context->ptr != end_ptr) 
	ERROR("Code segment was wrong length");
}


static void ReadLineTable(CICcontext *context, struct methodblock *mb)
{
    int attribute_length = get4bytes(context);
    unsigned char *end_ptr = context->ptr  + attribute_length;
    int i;
    if ((mb->line_number_table_length = get2bytes(context)) > 0) {
	struct lineno *ln = 
	    malloc(mb->line_number_table_length * sizeof(struct lineno));
	mb->line_number_table = ln;
	for (i = mb->line_number_table_length; --i >= 0; ln++) {
	    ln->pc = get2bytes(context);
	    ln->line_number = get2bytes(context);
	}
    }
    if (context->ptr != end_ptr)
	ERROR("Line number table was wrong length?");
}	

static void ReadLocalVars(CICcontext *context, struct methodblock *mb)
{
    int attribute_length = get4bytes(context);
    unsigned char *end_ptr = context->ptr  + attribute_length;
    int i;
    if ((mb->localvar_table_length = get2bytes(context)) > 0) {
	struct localvar *lv = 
	        malloc(mb->localvar_table_length * sizeof(struct localvar));
	mb->localvar_table = lv;
	for (i = mb->localvar_table_length; --i >= 0; lv++) {
	    lv->pc0 = get2bytes(context);
	    lv->length = get2bytes(context);
	    lv->nameoff = get2bytes(context);
	    lv->sigoff = get2bytes(context);
	    lv->slot = get2bytes(context);
	}
    }
    if (context->ptr != end_ptr)
	ERROR("Local variables table was wrong length?");
}	

/* Read in Compiled code.  The code has a very strict format:
 *    4 bytes:  length of code
 *    < code >
 *    4 bytes:  length of patches
 *    < patches>
 */
static void 
ReadInCompiledCode(CICcontext *context, struct methodblock *mb) {
    int attribute_length = get4bytes(context);
    unsigned char *end_ptr = context->ptr + attribute_length;
    int code_length = get4bytes(context);
    char *code = malloc(code_length); 
    int patch_length;
    unsigned char *patch;
    mb->CompiledCode = code;
    mb->code = 0;
    getNbytes(context, code_length, code);
    mb->maxstack = mb->nlocals = 0;
    mb->code_length = patch_length = get4bytes(context);
    mb->code = patch = malloc(patch_length);
    getNbytes(context, patch_length, (char *)patch);
    if (context->ptr != end_ptr) 
	ERROR("Compiled Code segment was wrong length");
    mb->fb.access |= ACC_MACHINE_COMPILED;
}

unsigned Signature2ArgsSize(char *method_signature)
{
    char *p;
    int args_size = 0;
    for (p = method_signature; *p != SIGNATURE_ENDFUNC; p++) {
	switch (*p) {
	  case SIGNATURE_BOOLEAN:
	  case SIGNATURE_BYTE:
	  case SIGNATURE_CHAR:
	  case SIGNATURE_SHORT:
	  case SIGNATURE_INT:
	  case SIGNATURE_FLOAT:
	    args_size += 1;
	    break;
	  case SIGNATURE_CLASS:
	    args_size += 1;
	    while (*p != SIGNATURE_ENDCLASS) p++;
	    break;
	  case SIGNATURE_ARRAY:
	    args_size += 1;
	    while ((*p == SIGNATURE_ARRAY)) p++;
	    /* If an array of classes, skip over class name, too. */
	    if (*p == SIGNATURE_CLASS) { 
		while (*p != SIGNATURE_ENDCLASS) 
		  p++;
	    } 
	    break;
	  case SIGNATURE_DOUBLE:
	  case SIGNATURE_LONG:
	    args_size += 2;
	    break;
	  case SIGNATURE_FUNC:	/* ignore initial (, if given */
	    break;
	  default:
	    /* Indicate an error. */
	    return 0;
	}
    }
    return args_size;
}


ClassClass*
createFakeArrayClass(char *name) 
{
    char *name_p;
    ClassClass *cb = calloc(sizeof(ClassClass), 1);
    int depth, base_type;
    char buffer[256];
    cp_item_type *constant_pool = calloc(CONSTANT_POOL_ARRAY_LENGTH, 
					 sizeof(cp_item_type));
    unsigned char *type_table = calloc(CONSTANT_POOL_ARRAY_LENGTH,sizeof(char));
    
    assert(name[0] == SIGNATURE_ARRAY);
    for (name_p = name, depth = 0; *name_p == SIGNATURE_ARRAY; 
	 name_p++, depth++);
    switch(*name_p) {
        case SIGNATURE_INT:    base_type = T_INT; break;
	case SIGNATURE_LONG:   base_type = T_LONG; break;
	case SIGNATURE_FLOAT:  base_type = T_FLOAT; break;
	case SIGNATURE_DOUBLE: base_type = T_DOUBLE; break;
	case SIGNATURE_BOOLEAN:base_type = T_BOOLEAN; break; 
	case SIGNATURE_BYTE:   base_type = T_BYTE; break;
	case SIGNATURE_CHAR:   base_type = T_CHAR; break;
	case SIGNATURE_SHORT:  base_type = T_SHORT; break;
	case SIGNATURE_CLASS: {
	    char *bp;
	    for (bp = buffer, name_p++; /* skip over SIGNATURE_ARRAY */
		 *name_p != SIGNATURE_ENDCLASS; *bp++ = *name_p++);
	    *bp = '\0';
	    base_type = T_CLASS;
	}
    }
    cb->major_version = JAVA_VERSION;
    cb->minor_version = JAVA_MINOR_VERSION;
    cb->access = ACC_FINAL | ACC_ABSTRACT | ACC_PUBLIC;
    cb->name = strdup(name);
    cb->super_name = JAVAPKG "Object";
    cb->constantpool = constant_pool;
    cb->constantpool_count = CONSTANT_POOL_ARRAY_LENGTH;

    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p = type_table;
    constant_pool[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i = depth;     
    constant_pool[CONSTANT_POOL_ARRAY_TYPE_INDEX].i = base_type;
    type_table[CONSTANT_POOL_ARRAY_DEPTH_INDEX] = 
	type_table[CONSTANT_POOL_ARRAY_TYPE_INDEX] = 
	    CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED;
    
    if (base_type == T_CLASS) {
	constant_pool[CONSTANT_POOL_ARRAY_CLASS_INDEX].i = 
	      CONSTANT_POOL_ARRAY_CLASSNAME_INDEX;
	constant_pool[CONSTANT_POOL_ARRAY_CLASSNAME_INDEX].cp = strdup(buffer);
	type_table[CONSTANT_POOL_ARRAY_CLASS_INDEX] = CONSTANT_Class;
	type_table[CONSTANT_POOL_ARRAY_CLASSNAME_INDEX] = 
	      CONSTANT_Asciz | CONSTANT_POOL_ENTRY_RESOLVED;
    } else {
	type_table[CONSTANT_POOL_ARRAY_CLASS_INDEX] = 
	    type_table[CONSTANT_POOL_ARRAY_CLASSNAME_INDEX] = 
		CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED;
    }
    return cb;
}


void FreeClass(ClassClass *cb) 
{
    int i;
    unsigned char *type_table;
    struct methodblock *mb;
    struct fieldblock *fb;

    free(cb->implements);
    for (i = cb->fields_count, fb = cb->fields; --i >= 0; fb++) {
	if (fb->access & ACC_STATIC) {
	    if (fb->signature[0] == SIGNATURE_DOUBLE || fb->signature[0] == SIGNATURE_LONG) {
		free(twoword_static_address(fb));
	    }
	}
    }
    free(cb->fields);
    for (i = cb->methods_count, mb = cb->methods; --i >= 0; mb++) {
	if (mb->code_length) /* not external */
	    free(mb->code);
	free(mb->exception_table);
	free(mb->localvar_table);
	free(mb->line_number_table);
    }
    free(cb->methods);
    if (cb->constantpool) {
	type_table = cb->constantpool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;
	for (i = CONSTANT_POOL_UNUSED_INDEX; 
	         i < (int)cb->constantpool_count; i++) {
	    /* The first item in the constant pool of type CONSTANT_Asciz
	     * has the malloc of all the strings. */
	    if ((type_table[i] & CONSTANT_POOL_ENTRY_TYPEMASK) == 
		    CONSTANT_Asciz) {
		free(cb->constantpool[i].p);
		break;
	    }
	}
	free(cb->constantpool);
    }
    free(cb);
}
	
static unsigned long get1byte(CICcontext *context)
{
    unsigned char *ptr = context->ptr;
    if (context->end_ptr - ptr < 1) {
	ERROR("Truncated class file");
    } else {
	unsigned char *ptr = context->ptr;
	unsigned long value = ptr[0];
	(context->ptr) += 1;
	return value;
    }
}

static unsigned long get2bytes(CICcontext *context)
{
    unsigned char *ptr = context->ptr;
    if (context->end_ptr - ptr < 2) {
	ERROR("Truncated class file");
    } else {
	unsigned long value = (ptr[0] << 8) + ptr[1];
	(context->ptr) += 2;
	return value;
    }
}


static unsigned long get4bytes(CICcontext *context)
{
    unsigned char *ptr = context->ptr;
    if (context->end_ptr - ptr < 4) {
	ERROR("Truncated class file");
    } else {
	unsigned long value = (ptr[0] << 24) + (ptr[1] << 16) + 
	                             (ptr[2] << 8) + ptr[3];
	(context->ptr) += 4;
	return value;
    }
}

static void getNbytes(CICcontext *context, int count, char *buffer) 
{
    unsigned char *ptr = context->ptr;
    if (context->end_ptr - ptr < count)
	ERROR("Truncated class file");
    if (buffer != NULL) 
	memcpy(buffer, ptr, count);
    (context->ptr) += count;
}


static char *getAsciz(CICcontext *context) 
{
    ClassClass *cb = context->cb;
    union cp_item_type *constant_pool = cb->constantpool;
    int nconstants = cb->constantpool_count;
    unsigned char *type_table = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;

    int value = get2bytes(context);
    if ((value == 0) || (value >= nconstants) || 
	   type_table[value] != (CONSTANT_Asciz | CONSTANT_POOL_ENTRY_RESOLVED))
	ERROR("Illegal constant pool index");
    return constant_pool[value].cp;
}

static char *getAscizFromClass(CICcontext *context, int value) 
{
    ClassClass *cb = context->cb;
    union cp_item_type *constant_pool = cb->constantpool;
    int nconstants = cb->constantpool_count;
    unsigned char *type_table = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;
    if ((value <= 0) || (value >= nconstants) 
	|| (type_table[value] != CONSTANT_Class))
	ERROR("Illegal constant pool index");
    value = constant_pool[value].i;
    if ((value <= 0) || (value >= nconstants) ||
	 (type_table[value] != (CONSTANT_Asciz | CONSTANT_POOL_ENTRY_RESOLVED)))
	ERROR("Illegal constant pool index");
    return constant_pool[value].cp;
}


/* Gag.  Find all strings in the constant pool that refer to classes or
 * to signatures.  Rewrite them, if necessary, so that oak/ is now java/
 */
static void MungeIncompatibleConstantPool(ClassClass *cb) {
    cp_item_type *constant_pool = cb->constantpool;
    int nconstants = cb->constantpool_count;
    char *type_table = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;
    int i;

/* temporary CONSTANT_ tags unused anywhere else */

#define CONSTANT_Class_Asciz 50
#define CONSTANT_Signature_Asciz 51

    for (i = CONSTANT_POOL_UNUSED_INDEX; i < nconstants; i++) {
	switch (type_table[i]) {
	    case CONSTANT_Class:
		/* Indicate this Asciz is a class name */
	        type_table[constant_pool[i].i] = CONSTANT_Class_Asciz;
		break;
	    case CONSTANT_NameAndType:
		/* Indicate this Asciz is a type signature */
	        type_table[constant_pool[i].i & 0xFFFF] = 
		      CONSTANT_Signature_Asciz;
		break;
	}
    }
    /* Fix up all the class names and type signatures */
    for (i = CONSTANT_POOL_UNUSED_INDEX; i < nconstants; i++) {
	switch (type_table[i]) {
	   case CONSTANT_Class_Asciz:
	        constant_pool[i].cp = MungeClassname(constant_pool[i].cp);
		type_table[i] = CONSTANT_Asciz | CONSTANT_POOL_ENTRY_RESOLVED;
		break;
	    case CONSTANT_Signature_Asciz:
	        constant_pool[i].cp = MungeSignature(constant_pool[i].cp);
		type_table[i] = CONSTANT_Asciz | CONSTANT_POOL_ENTRY_RESOLVED;
		break;
	}
    }
}

/* If a classname starts with "oak/", change it to "java/"
 */

struct translation_struct {
    char *from;			/* old string */
    char *to;			/* new string */
    int from_length, to_length;	/* lengths */
    int delta;			/* to_length - from_length */
};

#define TRANS(from, to) \
     { from, to, sizeof(from)-1, sizeof(to)-1, (int)(sizeof(to)-sizeof(from)) }
#define EOTRANS { NULL, NULL, 0, 0, 0 }

struct translation_struct translation_table[] = 
    { TRANS("oak/",       "java/"), 
      TRANS("webrunner/", "browser/"),
      EOTRANS
    };

static char *MungeClassname(char *classname) {
    struct translation_struct *tp;
    char *cp = classname;
    while (*cp == SIGNATURE_ARRAY) cp++;
    for (tp = translation_table; tp->from != NULL; tp++) {
	if (strncmp(cp, tp->from, tp->from_length) == 0) {
	    /* tp->from is a prefix of cp.  Create a new string */
	    char *result = (char *)malloc(strlen(classname) + tp->delta + 1);
	    sprintf(result, "%.*s%s%s", 
		    cp - classname, classname, tp->to, cp + tp->from_length);
	    return result;
	} 
    }
    return classname;
}

/* If a field or method signature has an imbedded classname that starts with
 * "oak/", change it to start with "java/"
 */

static char *MungeSignature(char *signature) {
    struct translation_struct *tp;
    if (signature[0] == SIGNATURE_FUNC) {
	/* Method signature.  Copy everything, but if we find a class name
	 * that starts with oak/, change it
	 */
	bool_t change = FALSE;
	char buffer[500], *bp = buffer, *sp = signature;
	for (;;) {
	    switch (*bp++ = *sp++) {
	        case '\0':
		    return change ? strdup(buffer) : signature;
	        case SIGNATURE_CLASS:
		    for (tp = translation_table; tp->from; tp++) {
			if (strncmp(sp, tp->from, tp->from_length) == 0) {
			    strcpy(bp, tp->to);
			    sp += tp->from_length;
			    bp += tp->to_length;
			    change = TRUE;
			    break;
			}
		    }
		    /* skip over the rest of the class name */
		    while ((*bp++ = *sp++) != SIGNATURE_ENDCLASS);
		    break;
	    }
	}
    } else {
	/* A field signature.  We just need to look for Loak/" and change it */
	char *cp = signature;
	while (*cp == SIGNATURE_ARRAY) cp++;
	if (*cp++ != SIGNATURE_CLASS) 
	    return signature;
	for (tp = translation_table; tp->from != NULL; tp++) {
	    if (strncmp(cp, tp->from, tp->from_length) == 0) {
		/* tp->from is a prefix of cp.  Create a new string */
		char *result = 
		    (char *)malloc(strlen(signature) + tp->delta + 1);
		sprintf(result, "%.*s%s%s", 
			cp - signature, signature, 
			tp->to, cp + tp->from_length);
		return result;
	    } 
	}
	return signature;
    }
}

/* In certain cases, we only need a class's name.  We don't want to resolve
 * the class reference if it isn't, already. 
 */

char *GetClassConstantClassName(cp_item_type *constant_pool, int index)
{
    unsigned char *type_table = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;
    switch(type_table[index]) {
        case CONSTANT_Class | CONSTANT_POOL_ENTRY_RESOLVED: {
	    ClassClass *cb = constant_pool[index].p;
	    return classname(cb);
	}

        case CONSTANT_Class: {
	    int name_index = constant_pool[index].i;
	    return constant_pool[name_index].cp;
	}
	    
	default:
	    return (char *)0;
    }
}
