/*
 * @(#)classresolver.c	1.35 95/04/05  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*-
 *      Routines for loading and resolving class definitions.
 *      These routines should not depending up on the interpreter or
 *      garbage collector.
 *
 */

#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <timeval.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>

#include <oobj.h>
#include <interpreter.h>
#include <signature.h>
#include <tree.h>
#include <monitor.h>
#include <bool.h>
#include <path.h>
#include <sys_api.h>

ClassClass *classclass;
ClassClass *classobject;

char *RuntimeInitClass(struct execenv *, ClassClass *);

static ClassClass *Locked_FindClassFromClass(struct execenv *ee, char *name, bool_t resolve, ClassClass *from);
static char *Locked_InitializeClass(ClassClass * cb, char **detail);
static char *Locked_ResolveClass(ClassClass * cb, char **detail);

extern void lock_classes(), unlock_classes();
static sys_mon_t _class_lock;
static int class_lock_depth = 0;


static struct fieldblock **
addslots(struct fieldblock ** fb, ClassClass * clb)
{
    long n = clb->fields_count;
    struct fieldblock *sfb = cbFields(clb);
    if (cbSuperclass(clb))
	fb = addslots(fb, unhand(cbSuperclass(clb)));
    while (--n >= 0) {
	*fb++ = sfb;
	sfb++;
    }
    return fb;
}

void
makeslottable(ClassClass * clb)
{
    ClassClass *sclb;
    int         nslots = 0;
    if (cbSlotTable(clb))
	return;
    sclb = clb;
    while (sclb) {
	long        n = sclb->fields_count;
	struct fieldblock *fb = cbFields(sclb);
	while (--n >= 0) {
	    nslots++;
	    fb++;
	}
	if (cbSuperclass(sclb) == 0)
	    break;
	sclb = unhand(cbSuperclass(sclb));
    }
    cbSlotTableSize(clb) = nslots;
    if (nslots == 0)
	nslots++;
    cbSlotTable(clb) = (struct fieldblock **)
	malloc(nslots * sizeof(struct fieldblock *));
    if (cbSlotTable(clb) == 0)
	out_of_memory();
    addslots(cbSlotTable(clb), clb);
}

static char *
copyclassname(char *src, char *dst)
{
    assert(*src == SIGNATURE_CLASS);
    src++;
    while (*src && *src != SIGNATURE_ENDCLASS)
	*dst++ = *src++;
    *dst = 0;
    return src;
}

static void
ResolveFields(ClassClass *cb, unsigned slot)
{
    struct fieldblock *fb;
    int thishash = 0;
    char *s;
    int c;
    int size;

    fb = cbFields(cb);
    for (size = 0; size < (int) cb->fields_count; size++, fb++) {
	char *signature = fieldsig(fb);
	int size = (((signature[0] == SIGNATURE_LONG || 
		      signature[0] == SIGNATURE_DOUBLE)) ? 2 : 1);
	if (fb->access & ACC_STATIC) {
	    /* Do nothing.  Handled when the class is loaded. */
	} else {
	    if (slot > 0xff * sizeof (OBJECT)) {
#ifndef TRIMMED
		fprintf(stderr, "%s.%s: Number of instance variables "
		    "> 256 for this class (internal error)\n",
		    classname(cb), fieldname(fb));
#endif
		abort();
	    }
	    fb->u.offset = slot;
	    slot += size * sizeof(OBJECT);
	}
	if ((fb->access & (ACC_STATIC | ACC_TRANSIENT)) == 0) {
	    for (s = fieldname(fb); c = *s++;)
		thishash = thishash * 7 + c;
	    for (s = fieldsig(fb); c = *s++;)
		thishash = thishash * 7 + c;
	}
    }
    cbInstanceSize(cb) = slot;
    cbThisHash(cb) = thishash;
    if (cbSuperclass(cb))
	cbTotalHash(cb) = thishash - cbTotalHash(unhand(cbSuperclass(cb)));
    else
        cbTotalHash(cb) = thishash;
    if (cbTotalHash(cb) < N_TYPECODES)
	cbTotalHash(cb) += N_TYPECODES;
}


char *
ResolveMethods(ClassClass *cb)
{
    struct methodblock *mb;
    int size;
    ClassClass *super = 0;
    unsigned mslot;

    if (cb->super_name != 0 
	&& (super = FindClassFromClass(0, classsupername(cb), TRUE, cb))) {
	mslot = cbMethodTableSize(super);
    } else {
	mslot = 1;
    }
    mb = cbMethods(cb);
    for (size = 0; size < (int) cb->methods_count; size++, mb++) {
	struct methodblock *mbb = 0, *mbp;
	ClassClass *superp;
	int cnt;
	if ((superp = super))
	    while (1) {
		for (mbp = cbMethods(superp), cnt = superp->methods_count;
			--cnt >= 0; mbp++)
		    if ((strcmp(fieldname(&mbp->fb), fieldname(&mb->fb)) == 0)
			&& (strcmp(fieldsig(&mbp->fb), fieldsig(&mb->fb)) == 0))
		        mbb = mbp;
		if (cbSuperclass(superp) == 0)
		    break;
		superp = unhand(cbSuperclass(superp));
	    }
	if (mbb)
	    mb->fb.u.offset = mbb->fb.u.offset;
	else {
	    if (mslot > 0xff) {
#ifndef TRIMMED
		fprintf(stderr, "%s.%s: Number of methods "
		    "> 256 for this class (internal error)\n",
		    classname(cb), fieldname(&mb->fb));
#endif
		abort();
	    }
	    mb->fb.u.offset = mslot;
	    mslot++;
	}
    }
    cbMethodTableSize(cb) = mslot;
    /*
	This should be the only place that method tables are
	allocated.  We allocate more than we need here and mask the
	resulting pointer because the methodtable pointer field in
	object handles is overloaded and sometimes hold array types
	and array lengths.  We must ensure that method table pointers
	are allocated on an appropriate boundary so we don't lose any
	address bits when we mask off the type portion of the pointer.
    */
    cbMethodTable(cb) = (struct methodtable *) 
       malloc(sizeof(struct methodtable)
	          + (cbMethodTableSize(cb)-1)*sizeof(struct methodblock *)
						+ FLAG_MASK);
    if (cbMethodTable(cb) == 0) {
	CCSet(cb, Resolved);
	return "OutOfMemoryException";
    }
    cbMethodTable(cb) = (struct methodtable *)((((int) cbMethodTable(cb)) + FLAG_MASK) & LENGTH_MASK);
    cbMethodTable(cb)->classdescriptor = cb;
    memset((char *) cbMethodTable(cb)->methods, 0,
	   cbMethodTableSize(cb) * sizeof(struct methodblock *));
    if (super) {
	memcpy((char *) cbMethodTable(cb)->methods,
	       (char *) cbMethodTable(super)->methods,
	       cbMethodTableSize(super)* sizeof(struct methodblock *));
    }
    return 0;
}



char *
InitializeClass(ClassClass * cb, char **detail) 
{ 
    char *result;
    monitorEnter(obj_monitor(cb));
    result = Locked_InitializeClass(cb, detail);
    monitorExit(obj_monitor(cb));
    return result;
}


char *
ResolveClass(ClassClass *cb, char **detail) 
{
    char *result;
    monitorEnter(obj_monitor(cb));
    result = Locked_ResolveClass(cb, detail);
    monitorExit(obj_monitor(cb));
    return result;
}



static char *
Locked_InitializeClass(ClassClass * cb, char **detail)
{
    ClassClass *super = 0;
    char *ret = 0;

    if (CCIs(cb, Initialized))
	return 0;

#ifndef TRIMMED
    if (verbose)
	fprintf(stderr, "[Initializing %s]\n", classname(cb));
#endif

    if (cb->fields_count > 2000) {
	return "FileFormatException";
    }
    if (strcmp(classname(cb), CLS_RESLV_INIT_CLASS) == 0)
	classclass = cb;
    if (strcmp(classname(cb), CLS_RESLV_INIT_OBJECT) == 0)
	classobject = cb;

    cbClassnameArray(cb) = MakeString(classname(cb), strlen(classname(cb)));

    if (cbHandle(cb) == 0) {
	cbHandle(cb) = 
	    (HClass *) AllocHandle(mkatype(T_BYTE, 1), (ClassObject *) cb);
	if (cbHandle(cb) == 0)
	    return "OutOfMemoryException";
    }
    if (strcmp(classname(cb), CLS_RESLV_INIT_REF) == 0)
	CCSet(cb, SoftRef);
    if (cbSuperclass(cb) == 0) {
	if (cb->super_name) {
	    if ((super = FindClassFromClass(0, classsupername(cb), FALSE, cb))) {
		assert(CCIs(super, Initialized));
		cbSuperclass(cb) = cbHandle(super);
		if (CCIs(super, SoftRef))
		    CCSet(cb, SoftRef);
	    } else {
		ret = "NoClassDefFoundException";
		*detail = classsupername(cb);
		cbSuperclass(cb) = 0;
	    } 
	} else if (cb == classobject) {
	    cbSuperclass(cb) = 0;	    
	} else {
	    *detail = classname(cb);
	    return "FileFormatException";
	}
    }
    CCSet(cb, Initialized);

    /* Make sure we know what classclass is, and that it's method table
     * is filled in.
     */
    if (classclass == 0) {
	classclass = FindClassFromClass(0, CLS_RESLV_INIT_CLASS, TRUE, cb);
	if (classclass == 0) {
#ifndef TRIMMED
	    fprintf(stderr, "class Class missing -- internal error\n");
#endif
	    exit(1);
	}
    }
    
    /* The following may not do anything useful if classclass hasn't been
     * completely resolved.  But we clean up in ResolveClass 
     */

    cbHandle(cb)->methods = cbMethodTable(classclass);
    return ret;
}

static char *
Locked_ResolveClass(ClassClass * cb, char **detail)
{
    ClassClass *super = 0;
    unsigned slot = 0;
    char *ret = 0;

    if (CCIs(cb, Error)) {
	return "NoClassDefFoundException";
    }
   
    assert(CCIs(cb, Initialized));

    if (CCIs(cb, Resolved)) {
	return 0;
    }

    cbInstanceSize(cb) = -1;

    if (cbSuperclass(cb)) {
	/* If this object has a superclass. . . */
	super = unhand(cbSuperclass(cb));
	if (!CCIs(super, Resolved)) {
	    if (ret = ResolveClass(super, detail)) {
		CCSet(cb, Error);
		return ret;
	    }
	}
	assert(CCIs(super, Resolved));
	slot = cbInstanceSize(super);
	if ((int) slot < 0) {
#ifndef TRIMMED
	    fprintf(stderr, "%s: Can't resolve class because of interdependant static initializers\n", classname(cb));
#endif
	    CCSet(cb, Error);
	    return "NoClassDefFoundException";
	}
    }
    
#ifndef TRIMMED
    if (verbose)
        fprintf(stderr, "[Resolving %s]\n", classname(cb));
#endif
    CCSet(cb, Resolved);
    ResolveFields(cb, slot);
    ResolveMethods(cb);

    /*	Perform runtime-specific initialization on the class.  The
	definition of this routine if different depending on if we
	are compiling the interpreter or not.
    */
    if ((ret = RuntimeInitClass(0, cb))) {
	*detail = cb->name;
        return ret;
    }

    /* We need this for bootstrapping.  We can't set the Handle's class block
     * pointer in Object or Class until Class has been resolved.
     */
    if (cb == classclass) {
	int         j;
	ClassClass **pcb;
	for (j = nbinclasses, pcb = binclasses; --j >= 0; pcb++) {
	    cbHandle(*pcb)->methods = cbMethodTable(cb);
	}
    }
    return 0;
}


/* Find an already loaded class with the given name.  If no such class exists
 * then return 0.
 *
 * We attempt to resolve it, also, if the "resolve" argument is true.
 * Otherwise, we only perform a minimal initialization on it, such that
 * it points to its superclass, which points to its superclass. . . .
 */

#ifdef DEBUG
void
PrintLoadedClasses()
{
    int         j;
    ClassClass **pcb, *cb;
    pcb = binclasses;
    for (j = nbinclasses; --j >= 0; pcb++) {
	cb = *pcb;
	printf("class=%s, 0x%x\n", classname(cb), cb->loader);
    }
}
#endif

static ClassClass *
FindLoadedClass(char *name, bool_t resolve, HObject *loader)
{
    int         j;
    ClassClass **pcb, *cb;

    assert(class_lock_depth > 0);

    pcb = binclasses;
    for (j = nbinclasses; --j >= 0; pcb++) {
	cb = *pcb;
	if ((cb->loader == loader) && (strcmp(name, classname(cb)) == 0)) {
	    char *err, *detail;
	    if (!CCIs(cb, Initialized)) {
		if ((err = InitializeClass(cb, &detail))) {
		    SignalError(0, err, detail);
		    return 0;
		}
	    }
	    if (resolve) {
		unlock_classes();
		err = ResolveClass(cb, &detail);
		lock_classes();
		if (err) {
		    SignalError(0, err, detail);
		    return 0;
		}
	    }
	    /*
	    if (resolve && !CCIs(cb, Resolved)) {
		if ((err = ResolveClass(cb, &detail))) {
		    SignalError(0, err, detail);
		    return 0;
		}
	    }
	    */
	    return cb;
	}
    }
    return 0;
}

/* Find the class with the given name.  
 * If "resolve" is true, then we completely resolve the class.  Otherwise,
 * we only make sure that it points to its superclass, which points to its
 * superclass, . . . .
 */
ClassClass *(*class_loader_func)(HObject *loader, char *name, long resolve) = 0;

ClassClass *
FindClass(struct execenv *ee, char *name, bool_t resolve)
{
    return FindClassFromClass(ee, name, resolve,
    	    (ee && ee->current_frame && ee->current_frame->current_method) ?
  	          fieldclass(&ee->current_frame->current_method->fb) : 0);
}

static ClassClass *
FindClassLoaderArrayClass(struct execenv *ee, char *name, 
			  bool_t resolve, ClassClass *from);


void
lock_classes()
{
    static bool_t initialized = FALSE;
    if (!initialized) {
	monitorRegister(&_class_lock, "Class lock");
	initialized = TRUE;
    }
    sysMonitorEnter(&_class_lock);
    class_lock_depth++;
}

void
unlock_classes()
{
    class_lock_depth--;
    sysMonitorExit(&_class_lock);
}

static ClassClass *
Locked_FindClassFromClass(struct execenv *ee, char *name, bool_t resolve, ClassClass *from)
{
    ClassClass *cb = 0;

    assert(class_lock_depth > 0);

    if (from && from->loader && class_loader_func) {
	if (name[0] == SIGNATURE_ARRAY) 
	    cb = FindClassLoaderArrayClass(ee, name, resolve, from);
	else {
	    /* unlock the classes while the class is located by
 	     * the class loader. The class loader should do this
	     * in a synchronized method
	     */
	    unlock_classes();
	    cb = (*class_loader_func)(from->loader, name, resolve);
	    lock_classes();
	}
    }
    if (cb == 0) {
	cb = FindLoadedClass(name, resolve, 0);
	if (cb == 0) {
	    DoImport(name, 0);
	    cb = FindLoadedClass(name, resolve, 0);
	}
    }
    return cb;
}

ClassClass *
FindClassFromClass(struct execenv *ee, char *name, bool_t resolve, ClassClass *from) {
    ClassClass *cb;
    lock_classes();
    cb = Locked_FindClassFromClass(ee, name, resolve, from);
    unlock_classes();
    return cb;
}

static ClassClass *
FindClassLoaderArrayClass(struct execenv *ee, char *name, 
			  bool_t resolve, ClassClass *from)
{
    ClassClass *cb;
    char *ptr;
    HObject *loader = from->loader;
    if ((cb = FindLoadedClass(name, resolve, loader)) != NULL)
	return cb;
    for (ptr = name; *ptr == SIGNATURE_ARRAY; ptr++);
    if (*ptr != SIGNATURE_CLASS) {
	return NULL;
    } else {
	ClassClass *inner_cb;
	char buffer[256];
	strcpy(buffer, ptr + 1);
	buffer[strlen(buffer) - 1] = '\0';
	inner_cb = FindClassFromClass(ee, buffer, FALSE, from);
	if (inner_cb == NULL || inner_cb->loader != loader) 
	    return NULL;
	cb = (ClassClass *)createFakeArrayClass(name);
	cb->loader = loader;
	AddBinClass(cb);
	/* this initializes */
	return FindLoadedClass(name, resolve, loader) ;
    }
}	
