/*
 * @(#)gc.c	1.53 95/04/18
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * The Java garbage collector
 *
 * Implementation notes:
 *
 * 1. The locking mechanisms need to be cleaned up, and used in malloc,
 * range alloc, etc.
 *
 * 2. We really don't need to allocate mark bits for soft references.
 * It might be better to keep a small list of the soft references we
 * find while doing the mark phase.  This would save 24KB of mark bits,
 * and shouldn't cost too much in the way of cpu.
 */

#include <typedefs.h>
#include <profile.h>
#include <tree.h>
#include <oobj.h>
#include <interpreter.h>
#include <code.h>
#include <monitor.h>
#include <threads.h>
#include <signature.h>
#include <monitor_cache.h>
#include <finalize.h>

#include <sys_api.h>

#include <java_lang_Thread.h>

extern void out_of_memory();

void gc_locked(int, unsigned int);	/* Call when holding heap lock */

/* Define this if you want lots of verbose info at runtime */
#define TRACEGC 1
int tracegc = 0;

/*
 * Lock against heap modification.
 */
sys_mon_t _heap_lock;
#define HEAP_LOCK_INIT()    monitorRegister(&_heap_lock, "Heap lock")
#define HEAP_LOCK()	    sysMonitorEnter(&_heap_lock)
#define HEAP_UNLOCK()	    sysMonitorExit(&_heap_lock)


/*
 * Define this if you want the mark phase to detect pointers into the
 * interior of objects.
 */
/* #define CHECK_INTERIOR_POINTERS */

#define ALLOCGRAIN      8
#define BITSPERCHAR     8

static unsigned char *pool, *poollimit;
static unsigned char *allochand;
static unsigned int *markbits;
#define BITSPERMARK (BITSPERCHAR * sizeof(*markbits))
static long FreeMemoryCtr;
static long TotalMemoryCtr;
static long marksize;
static long FreeMemoryLowWaterMark;

#define hpool (poollimit + sizeof(hdr)) /* the Handle pool immediatly follows
					 * the memory pool */
#define hpmin hpool
static unsigned char *hpoollimit, *hpoolhand;

static HObject *realObjAlloc(struct methodtable *, long );
int compactHeap();
void gc(int,  unsigned int);

/*
 * Memory block header (bottom three bits are flags):
 *
 * -------------------------------------------------------------
 * | <--- length --->| unused | <- obj swapped -> | <- free -> |
 * -------------------------------------------------------------
 * 31                3        2                   1            0
 */
typedef long hdr;

#define obj_geth(p) (*((hdr *)(p)))
#define obj_seth(p, h) (*((hdr *)(p)) = (h))
#define h_len(h) ((h) & ~(ALLOCGRAIN-1))
#define h_free(h) ((h) & 1)
#define h_bumplen(h, l) ((h) += (l))

#define obj_len(p) (obj_geth(p)&~(ALLOCGRAIN-1))
#define obj_setlf(p, l, f) (obj_geth(p) = (l)|(f))
#define obj_bumplen(p, l) (obj_geth(p) += (l))
#define obj_free(p) (obj_geth(p)&1)
#define obj_setfree(p) (obj_geth(p) |= 1)
#define obj_clearfree(p) (obj_geth(p) &= ~1)

#define HardMark 3
#define SoftMark 1
#define ValidPtr(p) ((((int)(p)) & (ALLOCGRAIN-1)) == 0 && \
		(unsigned char *)(p)>=pmin && \
		(unsigned char *)(p)< pmax)
#define ValidHandle(p) (((int) (p) & (sizeof(Handle)-1)) == 0 && \
		(unsigned char *)(p)>=pmax && \
		(unsigned char *)(p)<= hpmax)
#define ValidHorP(p) (((int) (p) & (ALLOCGRAIN-1)) == 0 && \
		(unsigned char *)(p)>=pmin && \
		(unsigned char *)(p) <= hpmax)


#define MARKINDEX(p)    (((unsigned char *)(p) - pmin) >> 7)
#define BITOFFSET(p)    ((((unsigned char *)(p) - pmin) >> 2) & 0x1e)
#define _MarkPtr(p, v)  (markbits[MARKINDEX(p)] |= (v) << BITOFFSET(p))
#define _ClearMarkPtr(p, v) (markbits[MARKINDEX(p)] &= ~((v) << BITOFFSET(p)))
#define _IsMarked(p)    ((markbits[MARKINDEX(p)] >> BITOFFSET(p)) &3)
#define MarkPtr(p, v) _MarkPtr(((unsigned int) (p) & ~(ALLOCGRAIN - 1)), v)
#define ClearMarkPtr(p, v) _ClearMarkPtr(((unsigned int)(p)&~(ALLOCGRAIN-1)),v)
#define IsMarked(p) _IsMarked((unsigned int) (p) & ~(ALLOCGRAIN - 1))

#define SetLimits() register unsigned char *const pmin = pool, \
    *const pmax = poollimit, *const hpmax = hpoollimit-sizeof(Handle)

#define SOFTREFBAGSIZE 200  /* max number of soft refs to kill in one cycle */

struct arrayinfo arrayinfo[] = {
    { 0,             0, 		 "N/A",          0},
    { T_ARRAY,      SIGNATURE_ARRAY,	"[]",           sizeof(OBJECT)},
    { T_CLASS,      SIGNATURE_CLASS,	"class[]",      sizeof(OBJECT)},
    { T_BOOLEAN,    SIGNATURE_BOOLEAN,	"bool[]",       sizeof(char)},
    { T_CHAR,       SIGNATURE_CHAR,	"char[]",       sizeof(unicode)},
    { T_FLOAT,      SIGNATURE_FLOAT,	"float[]",      sizeof(float)},
    { T_DOUBLE,     SIGNATURE_DOUBLE,	"double[]",     sizeof(double)},
    { T_BYTE,       SIGNATURE_BYTE,	"byte[]",       sizeof(char)},
    { T_SHORT,      SIGNATURE_SHORT,	"short[]",      sizeof(short)},
    { T_INT,        SIGNATURE_INT,	"int[]",        sizeof(int)},
    { T_LONG,       SIGNATURE_LONG,	"long[]",       sizeof(int64_t)},
#ifdef NO_LONGER_SUPPORTED
    { T_UBYTE,      SIGNATURE_BYTE,	"ubyte[]",      sizeof(char)},
    { T_USHORT,     SIGNATURE_SHORT,	"ushort[]",     sizeof(unsigned short)},
    { T_UINT,       SIGNATURE_INT,	"uint[]",       sizeof(int)},
    { T_ULONG,      SIGNATURE_LONG,	"ulong[]",      sizeof(int64_t)},
#endif
};

/*
 * Rough counter of heap modification events: a lack of known permu-
 * tations of the heap is used to suppress async GC.  While this isn't
 * a perfect indicator of good times to GC, it is unlikely that there 
 * will be many good times to GC when we aren't turning over memory.
 * Current incrementers of heap_memory_changes include realObjectAlloc()
 * and execute_finalizer().
 */
int heap_memory_changes = 0;

int
TotalMemory()
{
    return TotalMemoryCtr;
}

int
FreeMemory()
{
    return FreeMemoryCtr;
}

HObject *
AllocHandle(struct methodtable *mptr, ClassObject *p)
{
    register unsigned char *hp = hpoolhand;
    register unsigned char *limit = hpoollimit - sizeof(Handle);
    if (p == 0) {
	return 0;
    }
    while (hp <= limit) {
	if (((Handle *) hp)->obj == 0) {
	    goto GotHandle;
	} else {
	    hp += sizeof(Handle);
	}
    }
    limit = hpoolhand;
    hp = hpool;
#ifdef TRACEGC
    if (tracegc) {
	prints("[handle pool wraps]\n");
    }
#endif /* TRACEGC */
    while (hp < limit) {
	if (((Handle *) hp)->obj == 0) {
	    goto GotHandle;
	} else {
	    hp += sizeof(Handle);
	}
    }
#ifdef DEBUG
    if (verbosegc) {
	prints("<GC: out of handle space>\n");
    }
#endif /* DEBUG */
    return 0;
GotHandle:
    ((Handle *) hp)->methods = mptr;
    ((Handle *) hp)->obj = (ClassObject *) p;
    hpoolhand = hp + sizeof(Handle);
    return (HObject *) hp;
}

/*
 * Search the handle table for instances of class Ref, and clear the
 * object pointer in the lowest priority group.  If async_call is true,
 * we are being called asynchronously, and should watch for pending
 * interrupts.
 *
 * Return the number of refs zeroed, and set the int argument pointer
 * to the number of refs found in the system.
 */
static int
clearRefPointers(int async_call, int free_space_goal, int *totalrefs)
{
    int nsoftrefs = 0;
    struct {
	Handle *ref;
	long priority;
    } bag[SOFTREFBAGSIZE];
    Handle *hp;
    int ninbag = 0;
    SetLimits();

    for (hp = (Handle *) hpool; hp <= (Handle *) hpmax; hp++) {
	if (ValidPtr(hp->obj)
		&& obj_flags(hp) == T_NORMAL_OBJECT
		&& CCIs(obj_classblock(hp), SoftRef)
		&& ValidHandle(obj_getslot(hp, 0))
		&& IsMarked(obj_getslot(hp, 0)) == SoftMark) {
	    int prio = obj_getslot(hp, 1);
	    if (ninbag >= SOFTREFBAGSIZE
		    && prio < bag[SOFTREFBAGSIZE - 1].priority) {
		ninbag--;
	    }
	    if (ninbag < SOFTREFBAGSIZE) {
		int i = ninbag;
		while (i > 0 && prio < bag[i - 1].priority) {
		    bag[i] = bag[i - 1];
		    i--;
		}
		bag[i].ref = hp;
		bag[i].priority = prio;
		ninbag++;
	    }
	    nsoftrefs++;
	}
    }

    if (async_call && INTERRUPTS_PENDING()) {
	return 0;
    }

    if (ninbag) {
	int i;
	int freemem;

	/* Don't free more than half the Refs when there are more than 10 */
	if ((i = (nsoftrefs + 1) >> 1) <= ninbag && ninbag > 10) {
	    ninbag = i;
	}

	freemem = FreeMemoryCtr;
	for (i = ninbag; --i >= 0;) {
	    freemem -= obj_len((char *)bag[i].ref - sizeof(hdr));
#ifdef TRACEGC
	    if (tracegc > 1) {
		fprintf(stderr, "clear soft ref: 0x%x: %s\n",
			(unsigned int) bag[i].ref, Object2CString(bag[i].ref));
	    }
#endif /* TRACEGC */
	    obj_setslot(bag[i].ref, 0, 0);
	    if (freemem >= free_space_goal) {
		break;
	    }
	}
    }

    *totalrefs = nsoftrefs;
    return ninbag;
}


/*
 * Calculate the size in bytes of an array of type t and length l.
 */
int
sizearray(int t, int l)
{
    int size = 0;

    switch(t){
    case T_CLASS:
    case T_ARRAY:
	size = sizeof(OBJECT);
	break;
    default:
	size = T_ELEMENT_SIZE(t);
	break;
    }
    size *= l;
    return size;
}

/*
 * Allocate an array of type 't' with initial size 'l'.  This routine
 * takes care of setting the method table pointer correctly.  It also
 * takes into consideration any special sizeing requirements, like 
 * those needed for arrays of classes.
 */
HObject *
ArrayAlloc(int t, int l)
{
    assert(t >= T_ARRAY && t < T_MAXNUMERIC);
/*
 *  Uncomment me to find all the places where the code creates zero length
 *  arrays.
 *
 *  if (l == 0) {
 *	extern void DumpThreads();
 *	printf("zero length array created %s[%d]\n", arrayinfo[t].name, l);
 *	DumpThreads();
 *  }
 */
    if (l > TotalMemoryCtr) {
	/* This is guaranteed to fail.  It may also cause us arithmetic
	 * overflow in calculating sizearray().  So catch it early.
	 */
	return 0;
    }
    return realObjAlloc((struct methodtable *) mkatype(t, l),
	   sizearray(t, l) + ((t == T_CLASS || t == T_ARRAY) 
			          ? sizeof(OBJECT) : 0));
}

/*
 * Allocate an object.  This routine is in flux.  For now, the second 
 * parameter should always be zero and the first must alwaysd point to
 * a valid classblock.  This routine should not be used to allocate
 * arrays; use ArrayAlloc for that.
 */
HObject *
ObjAlloc(ClassClass *cb, long n0)
{
    HObject *handle;

#ifdef DEBUG
    if (n0 != 0 || cb == 0) {
	abort();
    }
#endif /* DEBUG */

    n0 = cbInstanceSize(cb);
    handle = realObjAlloc(cbMethodTable(cb), n0);

    /* 
     * If the class is a normal class and has a finalization method, flag
     * the instance and prepend it to the head of the HasFinalizerQ.  Note
     * that handle can be nil if we are out of memory!
     */
    if (handle && obj_flags(handle) == T_NORMAL_OBJECT && cb->finalizer) {
	finalizer_t *final;

	HASFINALQ_LOCK();
        final = (finalizer_t *) malloc(sizeof(finalizer_t));
        final->handle = handle;
        final->next = HasFinalizerQ;
        HasFinalizerQ = final;
	HASFINALQ_UNLOCK();
    }

    return handle;
}

static HObject *
realObjAlloc(struct methodtable *mptr, long n0)
{
    register unsigned char *p;
    register unsigned char *limit = poollimit;
    int JustGCed = 0;
    register long n;

    n = (n0 + sizeof(hdr) + (ALLOCGRAIN - 1)) & ~(ALLOCGRAIN - 1);
    HEAP_LOCK();
    p = allochand;
    while (1) {
	while (p < limit) {     /* search for a free block */
	    register unsigned char *next;
	    register hdr h = obj_geth(p);
	    register hdr h2;
	    assert(h_len(h) >= sizeof(hdr));
	    next = p + h_len(h);
	    assert(next <= poollimit + sizeof(ClassObject));
	    if (h_free(h)) {
		while (h_free(h2 = obj_geth(next))) {
		    h_bumplen(h, h_len(h2));
		    next = p + h_len(h);
		    obj_seth(p, h);
		}
		h2 = h_len(h) - n;
		if (h2 >= 0) {  /* bingo! */
		    if (h2 > 0) {
			obj_setlf(p + n, h2, 1);
#ifdef DEBUG
			((long *) (p + n))[1] = 0x55555555;
#endif /* DEBUG */
		    }
		    obj_setlf(p, n, 0);
#ifdef DEBUG
		    ((long *) p)[1] = 0x55555555;
#endif /* DEBUG */
		    allochand = p + n;
		    assert(((int) p + sizeof(hdr) & (ALLOCGRAIN - 1)) == 0);
		    FreeMemoryCtr -= n;
		    goto GotIt;
		}
	    }
	    p = next;
	}
	if (limit == poollimit && allochand > pool) {
	    /* not found, search other half */
	    p = pool;
	    limit = allochand;
	    if (allochand == poollimit) {
		allochand = p;
	    }
	    continue;
	}
	/* no space */
ForceGC:
	p = pool;
	allochand = p;
	limit = poollimit;
	{                       /* try running the garbage collector */
	    int freespace = FreeMemoryCtr;
	    int nsoftrefs = 0;
	    struct {
		Handle *ref;
		long priority;
	    } bag[SOFTREFBAGSIZE];
	    Handle *hp;
	    int ninbag = 0;
	    SetLimits();
	    if (!JustGCed) {
#ifdef DEBUG
		if (verbosegc) {
	            prints("<GC: out of heap space>\n");
	    	}
#endif /* DEBUG */
		gc(0, n);
	    }
	    JustGCed = 1;
	    if (FreeMemoryCtr > freespace) {
		continue;       /* Made some progress, try again */
	    }

	    /*
	     * If there are objects on the FinalizeMeQ they are known to
	     * be garbage that can't be freed until they are finalized.
	     * Synchronously finalize them, then garbage collect again.
	     */
	    if (FinalizeMeQ) {
		if (verbosegc) {
		    fprintf(stderr,
			"<GC: synchronously running pending finalizers>\n");
		}
		runFinalization();
		gc(0, n);
		if (FreeMemoryCtr > freespace) {
		    continue;
		}
	    }

	    /*
	     * Finally, if there are any Refs defined, free some of them
	     * and try garbage collecting again.
	     */
	    if (ninbag = clearRefPointers(0, TotalMemoryCtr, &nsoftrefs)) {
		freespace = FreeMemoryCtr;
		p = pool;
		if (verbosegc) {
		    prints("<GC: tried to free ref pointers>\n");
		}
		gc(0, n);
		if (verbosegc) {
		    fprintf(stderr, "<GC: Zeroed %d of %d soft refs, which freed %d bytes>\n",
			ninbag, nsoftrefs, FreeMemoryCtr - freespace);
		}
		continue;
	    }
	}
	HEAP_UNLOCK();
#ifdef DEBUG
	if (verbosegc) {
	    prints("<GC: totally out of heap space>\n");
	}
#endif /* DEBUG */
	return 0;
    }
GotIt:                          /* p points to new object */
    {
	HObject *ret = AllocHandle(mptr, (ClassObject *) (p + sizeof(hdr)));
	if (ret == 0) {
	    obj_setfree(p);
	    FreeMemoryCtr += obj_len(p);
	    goto ForceGC;
	}

	heap_memory_changes++;

	HEAP_UNLOCK();
	    
	return ret;
    }
}

void
markChildren(register Handle * h, register Handle * limit)
{
    register ClassClass *cb;
    register Handle *sub, *needsmark;
    register ClassObject *p;
    SetLimits();
ScanNext:
    needsmark = 0;
    p = unhand(h);
#ifdef TRACEGC
    if (tracegc > 1) {
	fprintf(stderr, "<%X/%X>\n", h, p);
	fprintf(stderr, " <  %s\n", Object2CString(h));
    }
#endif /* TRACEGC */
    switch (obj_flags(h)) {
	case T_NORMAL_OBJECT:
	    cb = obj_classblock(h);
	    do {
		register long n = cb->fields_count;
		register struct fieldblock *fb = cbFields(cb);
		while (--n >= 0) {
		    if ((fieldIsArray(fb) || fieldIsClass(fb))
			&& !(fb->access & ACC_STATIC)) {
			sub = *(Handle **) ((char *) p + fb->u.offset);
			if (ValidHandle(sub) && !IsMarked(sub)) {
			    MarkPtr(sub, CCIs(cb, SoftRef)
				    && fb->u.offset == 0
				    ? SoftMark : HardMark);
#ifdef TRACEGC
			    if (tracegc > 1)  {
				fprintf(stderr, " mo %s\n", Object2CString(sub));
			    }
#endif /* TRACEGC */
			    if (sub < limit) {
				if (needsmark) {
				    markChildren(needsmark, limit);
				}
				needsmark = sub;
			    }
			}
		    }
		    fb++;
		}
		if (cbSuperclass(cb) == 0) {
		    break;
		}
		cb = unhand(cbSuperclass(cb));
	    } while (cb);
	    break;

	case T_ARRAY:
	case T_CLASS: {     /* an array of classes */
	    register long n = obj_length(h);
	    while (--n >= 0) {
		sub = ((ArrayOfObject *) p)->body[n];
		if (ValidHandle(sub) && !IsMarked(sub)) {
		    MarkPtr(sub, HardMark);
#ifdef TRACEGC
		    if (tracegc > 1) {
			fprintf(stderr, " ma %s\n", Object2CString(sub));
		    }
#endif /* TRACEGC */
		    if (sub < limit) {
			if (needsmark) {
			    markChildren(needsmark, limit);
			}
			needsmark = sub;
		    }
		}
	    }
	}
    }
    if (h = needsmark) {
	goto ScanNext;          /* tail recursion */
    }
#ifdef TRACEGC
    if (tracegc > 1) {
	fprintf(stderr, " >  %s\n", Object2CString(h));
    }
#endif /* TRACEGC */
}


static int
scanThread(sys_thread_t *t, void *arg)
{
    extern Handle *getThreadNext();
    Handle *me = (Handle *) arg;
    Handle *p = (Handle *) t->cookie;
    unsigned char **ssc, **limit;
    stackp_t base;
    SetLimits();

#ifdef DEBUG
    if (!ValidHandle(p)) {
	abort();
    }
#endif /* DEBUG */
#ifdef TRACEGC
    if (tracegc > 1 && !IsMarked(p)) {
	if (((Handle *)p)->obj && ((Handle *)p)->methods) {
	    fprintf(stderr, " ms %s\n", Object2CString((Handle *)p));
	} else {
	    fprintf(stderr, " ms %8X\n", p);
	}
    }
#endif /* TRACEGC */
    /* Mark thread object */
    MarkPtr(p, HardMark);

    base = threadStackBase((TID)p);
    ssc  = threadStackPointer((TID)p);
    assert(base != 0);

    if (p == me) {
	ssc = (unsigned char **) &me;
    }
    if (ssc == 0) {
	/*
	 * if the stack does not have a top_of_stack pointer then it
	 * hasn't run yet and we don't need to scan its stack.
	 */
	return SYS_OK;
    }
    limit = (unsigned char **) base;

#ifdef TRACEGC
    if (tracegc)  {
	fprintf(stderr, "%8X [%10X,%10X] %d\n", p, ssc, limit, limit - ssc);
    }
#endif /* TRACEGC */
    assert(ssc != limit);

    while (ssc < limit) {
	register unsigned char *ptr = *ssc;
	if (ValidHorP(ptr)) {
#ifdef TRACEGC
	    if (tracegc > 1 && !IsMarked(ptr) && ValidHandle(ptr)) {
		if (((Handle *)ptr)->obj && ((Handle *)ptr)->methods) {
		    fprintf(stderr, " ms %s\n", Object2CString((Handle *) ptr));
		} else {
		    fprintf(stderr, " ms %8X\n", ptr);
		}
	    }
#endif /* TRACEGC */
	    MarkPtr(ptr, HardMark);
	}
	ssc++;
    }

    {
	ExecEnv *ee = (ExecEnv *)THREAD(p)->eetop;
	JavaFrame *frame;
	register unsigned char *ptr;
	if ((ee != 0) && ((frame = ee->current_frame) != 0)) {
	    stack_item *top_top_stack = frame->current_method 
		  ? &frame->ostack[frame->current_method->maxstack] 
		  : frame->optop;
	    limit = (unsigned char **)top_top_stack;
	    for(;;) {
#ifdef TRACEGC
		if (tracegc) {
		    fprintf(stderr, " scanning frame %8X\n", frame);
		}
#endif /* TRACEGC */

		for ( ssc= (unsigned char **)(frame->ostack); 
		       ssc < limit; ssc++) {
		    ptr = *ssc;
		    if (ValidHorP(ptr)) {
#ifdef TRACEGC
			if (tracegc > 1 && !IsMarked(ptr) && ValidHandle(ptr)) {
			    fprintf(stderr, " ms %8X\n", ptr);
			}
#endif /* TRACEGC */
			MarkPtr(ptr, HardMark);
		    }
		}
		ssc = (unsigned char **)(frame->vars);
		if (ssc) {
		    limit = (unsigned char **)frame;
		    for (; ssc < limit; ssc ++) {
			ptr = *ssc;
			if (ValidHorP(ptr)) {
#ifdef TRACEGC
			    if (tracegc > 1 && !IsMarked(ptr) && ValidHandle(ptr)) {
			    	fprintf(stderr, " ms %8X\n", ptr);
			    }
#endif /* TRACEGC */
			    MarkPtr(ptr, HardMark);
			}
		    }
		}
		frame = frame->prev;
		if (!frame) break;
		limit = (unsigned char **)(frame->optop);

	    }
	} else {
#ifdef TRACEGC 
	    if (tracegc) {
		fprintf(stderr, "ee=%x\n", ee);
		if (ee) {
		    fprintf(stderr, "frame=%x\n", ee->current_frame);
		}
	    }
#endif /* TRACEGC */
	}

    }

    return SYS_OK;
}

static void
scanThreads()
{
    Handle *self;
    int err;

    self = (Handle *) sysThreadSelf()->cookie;
    err = sysThreadEnumerateOver(scanThread, (void *) self);
    /* REMIND - should do something with the return value. */
}


/* Look through the constant pool of each of the classes to see if it
 * contains any objects (like Strings) that need to be marked.
 */
 
static void
scanClasses()
{
    int i, j;
    ClassClass **pcb =  binclasses;
    struct fieldblock *fb;
    SetLimits();

    for (j = nbinclasses; --j >= 0; pcb++) {
	ClassClass *cb = *pcb;
	if (CCIs(cb, Resolved)) {
	    union cp_item_type *constant_pool = cbConstantPool(cb);
	    union cp_item_type *cpp = constant_pool+ CONSTANT_POOL_UNUSED_INDEX;
	    union cp_item_type *end_cpp = &constant_pool[cb->constantpool_count];
	    for ( ; cpp < end_cpp; cpp++) {
		void *ptr = (*cpp).p;
		if (ValidHorP(ptr)) {
#ifdef TRACEGC
		    if (tracegc > 1 && !IsMarked(ptr) && ValidHandle(ptr)) {
		        if (((Handle *)ptr)->obj && ((Handle *)ptr)->methods) {
			    fprintf(stderr, " ms %s\n", 
				    Object2CString((Handle *) ptr));
			} else {
			    fprintf(stderr, " ms %8X\n", ptr);
			}
		    }
#endif /* TRACEGC */
		    MarkPtr(ptr, HardMark);
		}
	    }
	}
	/* Scan class definitions looking for statics */
	for (i = cb->fields_count, fb = cbFields(cb); --i >= 0; fb++) {
	    if ((fieldIsArray(fb) || fieldIsClass(fb)) && (fb->access & ACC_STATIC)) {
		Handle *sub = *(Handle **)normal_static_address(fb);
		if (ValidHandle(sub) && !IsMarked(sub)) {
		    MarkPtr(sub, HardMark);
#ifdef TRACEGC
		    if (tracegc > 1) {
			fprintf(stderr, " mc %s\n", Object2CString(sub));
		    }
#endif /* TRACEGC */
		}
	    }
	}
	/* don't have to mark cb->superclass, because it's caught by
	   the HandleToSelf for that class */
	if (ValidHandle(cbHandle(cb))) {
	    MarkPtr((Handle *) cbHandle(cb), HardMark);
	}
	if (ValidHandle(cbClassnameArray(cb))) {
	    MarkPtr((Handle *) cbClassnameArray(cb), HardMark);
	}
	if (ValidHandle(cbPrototype(cb))) {
	    MarkPtr((Handle *) cbPrototype(cb), HardMark);
	}
	if (ValidHandle(cbLoader(cb))) {
	    MarkPtr((Handle *) cbLoader(cb), HardMark);
	}
    }
}

/*
 * Scan the heap and chase pointers.  Scanning from pointers into the
 * middle of objects is normally disabled, apparently because it was
 * slow.  This should be revisited, especially if it reduces pinning.
 */
static void
scanHeap()
{
    SetLimits();
    unsigned char *o, *oo, *next;
    int i, size;

#ifdef TRACEGC
    if (tracegc) {
        fprintf(stderr, "\nRecursive mark scan:\n");
    }
#endif /* TRACEGC */

#ifdef CHECK_INTERIOR_POINTERS
    /*
     * Search the entire memory pool looking for references into the
     * the middle of objects.  If any are found, mark the object they
     * are part of.  Pointers into the middle of objects can happen
     * frequently, especially in the case of arrays.
     *
     * This piece of code is tricky because it needs to be fast.
     * It is basically walking through most of the mark bits looking
     * for set bits.  It does it in 32-bit increments for speed.
     */
    o = oo = pmin;
    next = o + obj_len(o);
    size = ((int) (poollimit - pmin) / (ALLOCGRAIN * BITSPERMARK) + 1) * 2;
    for (i=0; i < size; i++) {
        assert(((unsigned int)(oo - pmin) >> 7) == i);
        if (markbits[i]) {
	    unsigned int bits = markbits[i];
	    unsigned int shifts = ALLOCGRAIN * BITSPERMARK / 2;
	    while (bits && oo < pmax) {
		if (bits & 0x3) {
		    /* Mark the object, not the header */
		    MarkPtr((o + sizeof(hdr)), HardMark);
		}
		bits >>= 2;
		shifts -= ALLOCGRAIN;
		oo += ALLOCGRAIN;
		if (oo == next) {
		    o = next;
		    next = o + obj_len(o);
		}
	    }
	    oo += shifts;
	} else {
	    oo += ALLOCGRAIN * BITSPERMARK / 2;
	}

	if (oo >= pmax) {
	    break;
	}

	while (oo >= next) {
	    o = next;
	    next = o + obj_len(o);
	}
    }
#endif /* CHECK_INTERIOR_POINTERS */
}

/*
 * Walk the entire handle pool, and do a recursive search for
 * references.
 */
static void
scanHandles()
{
    SetLimits();
    Handle *hp;

    hp = (Handle *) hpool;
    while ((unsigned char *) hp <= hpmax) {
        if (hp->obj) {
	    if (ValidPtr(hp->obj) && IsMarked(hp->obj)) {
		MarkPtr((Handle *) hp, HardMark);
	    }
	    if (IsMarked(hp)) {
		markChildren(hp, hp + 1);
	    }
        }
        hp++;
    }
}

/*
 * Take care of finalization
 *
 * At this point we have marked all reachable objects.  Anything
 * not marked and having a finalization method will be queued for
 * finalization.  However, it is not the case that everything
 * else that's unmarked can just be freed.  An object to be
 * finalized expects everything it references to be retained by
 * GC, whether or not those things are themselves finalized.  So
 * we make a pass over the HasFinalizerQ that moves freed objects
 * needing finalization to the FinalizeMeQ.  We then make a pass
 * over the FinalizeMeQ recursively marking everything on it.  It
 * is important that there is one pass over each queue: doing it
 * in a single pass over the HasFinalizerQ causes different
 * behavior as finalizable objects referenced by earlier finali-
 * zable objects will be marked and not be themselves finalized.
 * We could potentially allow interruption between the two passes.
 *
 * After this pass we are in a consistent state (only objects that
 * are unreachable have been put on the FinalizeMeQ), so we can
 * allow async GC to be interrupted.  Note that this means that
 * an object may be finalized earlier than its non-finalized peers
 * are freed.  Note that the FinalizeMeQ scan also marks objects that
 * were on the FinalizeMeQ when we started, which we needed to
 * do anyhow.  Removing an object from the HasFinalizerQ ensures
 * that we will never try to finalize it again.
 */
static void
prepareFinalization()
{    
    if (HasFinalizerQ) {
        SetLimits();
        Handle *hp;
        finalizer_t *prev = (finalizer_t *) &HasFinalizerQ;
        finalizer_t *final;

        /* Pass over HasFinalizerQ: */
        while (final = prev->next) {
	    hp = final->handle;
	    if (hp->obj && !IsMarked(hp) && ValidPtr(hp->obj)) {
		/* Important: Put in FinalizeMeQ in reverse order */
		prev->next = final->next;
		final->next = FinalizeMeQ;
		FinalizeMeQ = final;
	    } else { 	/* still in use, so leave alone */
		prev = final;
	    }
	}

	/* Trace the FinalizeMeQ: */
	final = FinalizeMeQ;
	while (final) {
	    hp = final->handle;
	    /* Mark the handle itself */
	    MarkPtr((Handle *) hp, HardMark);
	    /* Recursively mark all objects referencable via the handle */
	    markChildren(hp, hp + 1);
	    final = final->next;
	}

	/* Trace the BeingFinalized object (see comments in finalize.c): */
	final = BeingFinalized;
	if (final) {
	    hp = final->handle;
	    MarkPtr((Handle *) hp, HardMark);
	    markChildren(hp, hp + 1);
	}
    }
}

/*
 * Scan for remaining unmarked objects and free them.  All objects
 * that are to be finalized have been marked.
 */ 
static long nfreed;		/* For -verbosegc */
static long bytesfreed;		/* For -verbosegc */

static int
freeSweep(unsigned int free_space_goal)
{
    SetLimits();
    Handle *hp;
    int do_compact_heap = 1;

#ifdef TRACEGC
    if (tracegc) {
        fprintf(stderr, "\nfree scan:\n");
    }
#endif /* TRACEGC */
    hp = (Handle *) hpool;
    while ((unsigned char *) hp <= hpmax) {
        if (hp->obj && !IsMarked(hp) && ValidPtr(hp->obj)) {
#ifdef TRACEGC
	    if (tracegc) {
		fprintf(stderr, " fr %s (%8X/%8X)\n",
			Object2CString(hp), hp, hp->obj);
	    }
#endif /* TRACEGC */
#ifdef DEBUG
	    memset((char *) hp->obj, 0x55,
		obj_len((char *) hp->obj - sizeof(hdr)) - sizeof(hdr));
#endif /* DEBUG */
	    FreeMemoryCtr += obj_len((char *) hp->obj - sizeof(hdr));
	    if (verbosegc) {
	        nfreed++;
	        bytesfreed += obj_len((char *) hp->obj - sizeof(hdr));
	    }
	    if (obj_len((char *)hp->obj - sizeof(hdr)) >= free_space_goal)
	        do_compact_heap = 0;
	    obj_setfree((char *) hp->obj - sizeof(hdr));
	    hp->obj = 0;
#ifdef DEBUG
	    hp->methods = 0;    /* paranoia */
#endif /* DEBUG */
	}
	hp++;
    }

    /* Reset allocation hands, so that freed memory gets used */
    allochand = pmin;
    hpoolhand = hpool;

    return do_compact_heap;
}


/*
 * An explicit call to GC, not implicit as from a memory allocator.
 * This guy should not have the heap lock, so needs to grab it before
 * collecting.
 */
void
gc(int async_call, unsigned int free_space_goal)
{
    HEAP_LOCK();
    HASFINALQ_LOCK();
    FINALMEQ_LOCK();

    if (sysThreadSingle() == SYS_OK) {
	gc_locked(async_call, free_space_goal);
    }
    sysThreadMulti();

    /* Notify the finalizer thread if objects pending finalization */
    if (FinalizeMeQ) {
	FINALMEQ_NOTIFY();
    }

    heap_memory_changes = 0;

    FINALMEQ_UNLOCK();
    HASFINALQ_UNLOCK();
    HEAP_UNLOCK();
}

/*
 * Garbage collect memory.  If 'async_call' is true, this function is being
 * called from the asych GC thread, and should immediately exit whenever
 * the INTERRUPTS_PENDING() macro returns true.  The 'free_space_goal' is
 * the amount of memory in a single block which the caller is interested
 * in having free.  If such a block does not exist, compact the heap.
 *
 * gc0_locked() should only be called when holding the heap lock and the
 * finalization queue lock.  gc() acquires the heap lock, then calls
 * gc_locked(), which calls gc0_locked().
 */

void
gc0_locked(int async_call, unsigned int free_space_goal)
{
    SetLimits();
    /* SPARC only: "ta ST_FLUSH_WINDOWS" done by "gc" stub in machgc.s */
    int do_compact_heap;
    long start_time = now();

    if (verbosegc) {
	nfreed = 0;
	bytesfreed = 0;
    }

    memset((char *) markbits, 0, marksize);
    if (async_call && INTERRUPTS_PENDING()) {
	goto unlock;
    }

    MARK(GCTHREAD);
    scanThreads();
    if (async_call && INTERRUPTS_PENDING()) {
	goto unlock;
    }

    MARK(GCCLASS);
    scanClasses();
    if (async_call && INTERRUPTS_PENDING()) {
	goto unlock;
    }

    MARK(GCHEAP);
    scanHeap();
    if (async_call && INTERRUPTS_PENDING()) {
	goto unlock;
    }

    MARK(GCHANDLES);
    scanHandles();
    if (async_call && INTERRUPTS_PENDING()) {
	goto unlock;
    }

    MARK(GCFINAL);
    prepareFinalization();
    if (async_call && INTERRUPTS_PENDING()) {
       goto unlock;
    }

    MARK(GCSWEEP);
    do_compact_heap = freeSweep(free_space_goal);
    if (async_call && INTERRUPTS_PENDING()) {
	goto unlock;
    }

    if (do_compact_heap) {
	while (compactHeap(async_call)) ;
    }

    if (verbosegc) {
	long total_time = now() - start_time;
	fprintf(stderr, "<GC%s: %d objects, %d bytes, %d msec, %d%% free>\n",
	    async_call ? "(async)" : "",
	    nfreed, bytesfreed, total_time,
	    100 * FreeMemoryCtr / TotalMemoryCtr);
    }

  unlock:
    if (java_monitor && !async_call) {
	ExecEnv *ee = EE();
	if (ee && ee->current_frame && ee->current_frame->current_method) {
	    java_mon(ee->current_frame->current_method,
		    (struct methodblock *)-1, now() - start_time);
	}
    }
}


/*
 * Concept: Assuming optimal packing, there are only two places that
 * holes, or bubbles in the heap can occur.  These are immediately before
 * each pinned range of memory, and at the tail of the heap (the last
 * range is always free).
 *
 * This algorithm walks the heap from beginning to end, performing
 * several operations:
 *  1. Coalesce adjacent free blocks.
 *  2. Record bubbles in the free_blocks[] array.  This is an array
 *     containing the biggest free blocks found.
 *  3. Copy unpinned, allocated blocks down into the lowest free
 *     block available.
 *
 * The goal of this scheme is to leave as many large free blocks available
 * as is possible.
 */
static int
compactHeap(int async_call)
{
    SetLimits();
    register Handle *hp;
    unsigned char *p, *next, *prev;
#define NUMFREEBLOCKS 16
    unsigned char *free_blocks[NUMFREEBLOCKS];
    unsigned int plen;
    int blocks_moved = 0;

    p = pmin;
    prev = 0;
    memset((char *)free_blocks, 0, sizeof(free_blocks));

    /*
     * Put a pointer to each object's handle in the first word of the object,
     * saving the first word of the object in the object pointer of the
     * handle. This way we can compact the heap looking only at the heap,
     * then go back and fix up the handles.  (we need to scan all objects in
     * heap order, not handle order to effectively compact everything).
     */
#define obj_swapped(p)      (obj_geth(p) & 2)
#define obj_setswapped(p)   (obj_geth(p) |= 2)
#define obj_clearswapped(p) (obj_geth(p) &= ~2)

#define handle_swapped(hp)      (IsMarked(hp) & 0x1)
#define handle_setswapped(hp)   (MarkPtr(hp, 0x1))
#define handle_clearswapped(hp) (ClearMarkPtr(hp, 0x1))

    for (hp = (Handle *) hpool; hp <= (Handle *) hpmax; hp++) {
	if (ValidPtr(hp->obj)) {
	    OBJECT T = *hp->obj;
	    unsigned char *p = (unsigned char *)(hp->obj) - sizeof(hdr);
	    *hp->obj = (OBJECT) hp;
	    hp->obj = (OBJECT *) T;
	    if (obj_swapped(p)) {
		handle_setswapped(hp);
	    } else {
		obj_setswapped(p);
		handle_clearswapped(hp);
	    }
	}
    }

    /*
     * This is the reverse of the above swap.  It is called once
     * on each allocated block and fixes the handle/pointer swap.
     */
#define REVERSE_SWAP(p) { \
    Handle *hp, *nexthp; \
    assert(obj_swapped(p)); \
    obj_clearswapped(p); \
    hp = *(Handle **) (p + sizeof(hdr)); \
    while (handle_swapped(hp)) { \
	nexthp = (Handle *) (hp->obj); \
	hp->obj = (OBJECT *) (p + sizeof(hdr)); \
	hp = nexthp; \
    } \
    *(OBJECT *) (p + sizeof(hdr)) = (OBJECT) hp->obj; \
    hp->obj = (OBJECT *) (p + sizeof(hdr)); \
    handle_setswapped(hp); \
}

    /*
     * The main loop - look at every object in the pool
     */
    while (p < pmax) {
	plen = obj_len(p);
	next = p + plen;

	if (async_call && INTERRUPTS_PENDING()) {
	    blocks_moved = 0;
	    /*
	     * Cycle through the remaining blocks as quickly as possible,
	     * reversing all of the handle pointers
	     */
	    while (p < pmax) {
		if (!obj_free(p)) {
		    REVERSE_SWAP(p);
		}
		p += obj_len(p);
	    }
	    break;
	}

	if (obj_free(p)) {
	    unsigned int i, smallest = 0, len = (unsigned) ~0;

	    /* 1. Coalesce adjacent free blocks */

	    while (next && obj_free(next)) {
		plen += obj_len(next);
		next += obj_len(next);
		blocks_moved++;
	    }
	    obj_setlf(p, plen, 1);

	    /* 2. Record the fact that p is a free block */

	    /*
	     * Search for the smallest sized free range, and if it
	     * is smaller than the size of p, replace it with p.
	     * There may be empty slots in the buffer. Use them
	     * if possible.
	     */
	    for (i = 0; i < NUMFREEBLOCKS; i++) {
		assert(free_blocks[i] == 0 || obj_free(free_blocks[i]));
		if (free_blocks[i] == 0 || free_blocks[i] == p) {
		    len = 0;
		    smallest = i;
		    break;
		} else if (obj_len(free_blocks[i]) < len) {
		    smallest = i;
		    len = obj_len(free_blocks[i]);
		}
	    }

	    if (len < obj_len(p)) {
		free_blocks[smallest] = p;
	    }

	    prev = p;

	} else if (!IsMarked(p + sizeof(hdr))) {
	    int i, lowest, prev_block = -1;
	    unsigned int len, lowest_len;
	    unsigned char *lowest_addr;

	    /* 3. Copy p down into the lowest available free block */

	    /*
	     * Find lowest block that is big enough to hold the block.
	     * Ignore 'prev' as we will handle it with a simpler case.
	     */
	    lowest = -1;
	    lowest_len = 0;
	    lowest_addr = pmax;
	    for (i = 0; i < NUMFREEBLOCKS; i++) {
		if (!free_blocks[i]) {
		    continue;
		}
		if (free_blocks[i] == prev) {
		    prev_block = i;
		}
		len = obj_len(free_blocks[i]);
		if (len > plen && lowest_addr > free_blocks[i]) {
		    lowest = i;
		    lowest_addr = free_blocks[i];
		    lowest_len = len;
		}
	    }

	    /* did we find a slot? */
	    if (lowest >= 0 && lowest != prev_block) {
		unsigned char *oldp = p;
		/*
		 * If a slot is found, move the block, and mark the old
		 * slot as free.  If the free space was bigger than we needed,
		 * break off a new chunk.
		 */
		assert(obj_free(free_blocks[lowest]));
		memmove(free_blocks[lowest], p, plen);
		blocks_moved++;
		obj_setfree(p);
#ifdef DEBUG
		((long *) p)[1] = 0x55555555;
#endif /* DEBUG */
		p = free_blocks[lowest];

		/* it was bigger than we needed, break off a piece. */
		if (lowest_len > plen) {
		    unsigned char *fp = free_blocks[lowest];
		    fp += plen;
		    obj_setlf(fp, (lowest_len - plen), 1);
#ifdef DEBUG
		    ((long *) fp)[1] = 0x55555555;
#endif /* DEBUG */
		    if (prev <= free_blocks[lowest]) {
			free_blocks[lowest] = 0;
		    } else {
			free_blocks[lowest] = fp;
		    }
		} else {
		    free_blocks[lowest] = 0;
		}

		if (obj_free(prev)) {
		    next = prev;
		    prev = 0;
		} else {
		    next = oldp;
		}
	    } else {
		/*
		 * no free slot found or it is the previous slot.  See if prev
		 * is free, and if so, just slide this range back.
		 */
		if (prev && obj_free(prev)) {
		    memmove(prev, p, plen);
		    blocks_moved++;
		    next = prev + plen;
		    obj_setlf(next, (p - prev), 1);
#ifdef DEBUG
		    ((long *) next)[1] = 0x55555555;
#endif /* DEBUG */
		    p = prev;
		    if (prev_block >= 0) {
			free_blocks[prev_block] = 0;
		    }

		} else {
		    prev = p;
		}
	    }
	} else {
	    prev = p;
	}

	if (next > p && !obj_free(p)) {
	    REVERSE_SWAP(p);
	}
	p = next;
    }

#ifdef TRACEGC
    if (tracegc && blocks_moved) {
	fprintf(stderr, "<GC: compacted %d objects>\n", blocks_moved);
    }
#endif /* TRACEGC */

    return async_call ? blocks_moved : 0;
}

/*
 * Initialize the memory allocator.  The argument 'n' is the initial
 * memory pool size.
 */
void
InitializeAlloc(register long n)
{
    assert(pool == 0);
    pool = (unsigned char *) sysAllocateMem(n);
    if (pool == 0) {
	out_of_memory();
    }
    marksize = ((n / (ALLOCGRAIN * BITSPERMARK) + 1) * 2) * sizeof(*markbits);
    markbits = (unsigned int *) malloc(marksize);
    if (markbits == 0) {
	out_of_memory();
    }
    poollimit = pool + n - sizeof(hdr);
    while (((int) pool + sizeof(hdr)) & (ALLOCGRAIN - 1)) {
	pool += sizeof(hdr);
    }

    /* leave 80% of the pool for objects, the rest for handles */
    hpoollimit = poollimit + sizeof(hdr);
    poollimit = pool + ((int) ((poollimit - pool) * 0.80) & ~(ALLOCGRAIN - 1));
    hpoolhand = hpool;

    /* Make sure that all the handles are zero.  We recognize a free handle by
     * the fact that its ->obj field is 0.     */
    memset(hpoolhand, 0, hpoollimit - hpoolhand);

    obj_setlf(pool, poollimit - pool, 1);
#ifdef DEBUG
    ((long *) pool)[1] = 0x55555555;
#endif /* DEBUG */
    obj_setlf(poollimit, 0, 0);
    FreeMemoryCtr = poollimit - pool;
    TotalMemoryCtr = FreeMemoryCtr;
    FreeMemoryLowWaterMark = (long) (0.2 * FreeMemoryCtr) & ~(ALLOCGRAIN - 1);
    allochand = pool;

    HEAP_LOCK_INIT();
}

/*
 * Called whenever the system thinks it is idle, and might want to
 * do a GC.
 */
void
asyncGC()
{
    int freemem = FreeMemoryCtr;
    int totalrefs, freedrefs;

    HEAP_LOCK();
    HASFINALQ_LOCK();
    FINALMEQ_LOCK();

    if (sysThreadSingle() != SYS_OK) {
	goto unlock;
    }
    gc_locked(1, ~0);
    if (INTERRUPTS_PENDING()) {
	goto unlock;
    }

    /* Don't want to zero this unless gc actually completed... */
    heap_memory_changes = 0;

    /*
     * If we didn't free anything, and we hit the low water mark,
     * try clearing some soft refs.
     */
    if ((freemem == FreeMemoryCtr) && (freemem < FreeMemoryLowWaterMark)) {
	/* If we can't clear any, just return */
	if (!(freedrefs = clearRefPointers(1, FreeMemoryLowWaterMark,
					     &totalrefs))) {
	    goto unlock;
	}

	if (verbosegc) {
	    fprintf(stderr, "<GC: Asynchronously zeroed %d of %d soft refs\n",
		freedrefs, totalrefs);
	}

	if (INTERRUPTS_PENDING()) {
	    goto unlock;
	}

	gc_locked(1, ~0);
    }

unlock:
    sysThreadMulti();

    /* Notify the finalizer thread if objects pending finalization */
    if (FinalizeMeQ) {
	FINALMEQ_NOTIFY();
    }

    FINALMEQ_UNLOCK();
    HASFINALQ_UNLOCK();
    HEAP_UNLOCK();

    return;
}


/*
 * Support for DumpMonitors(): we want to print whose monitor this is
 * if it is associated with an object.  To check whether that's so we
 * need to see whether the handle is in the handle pool, and the limits
 * of the handle pool aren't visible out of gc.c.
 */ 
void
monitorDumpHelper(monitor_t *mid, void *name)
{
    TID t;
    unsigned int key = mid->key;
    SetLimits();

    if (mid && (mid->flags & MON_IN_USE) != 0) {
	if (name == 0) {
	    if (ValidHandle(key)) {
		name = Object2CString((Handle *) key);
	    } else {
		name = "unknown key";
	    }
	}
	fprintf(stderr, "    %s (key=0x%x): ", name, key);
	sysMonitorDumpInfo(&mid->mid);
    }
    return;
}


/*
 * Profiling support
 */
#define MAX_CLASS_HASH 1023
#define ARRAYTYPES 	(sizeof(arrayinfo) / sizeof(arrayinfo[0]))

void
profHandles(FILE *fp) {
    SetLimits();
    register Handle *hp = (Handle *) hpmin;
    ClassClass *cb;
    struct hash_entry {
	ClassClass *cb;
	int count;
	int arraycount;
	int arraylength;
    } *tab, *p;
    int arraycount[ARRAYTYPES];
    int arraylength[ARRAYTYPES];
    int handles_count = (Handle *)hpmax - (Handle *)hpmin;
    int handles_used = 0;
    int i;

    /* array type mapping */
    int tmap[64];
    for (i = 0 ; i < ARRAYTYPES ; i++) {
	tmap[arrayinfo[i].index] = i;
    }

    tab = (struct hash_entry *)malloc(sizeof(struct hash_entry) * MAX_CLASS_HASH);
    memset((char *)tab, 0, sizeof(struct hash_entry) * MAX_CLASS_HASH);
    memset((char *)arraycount, 0, sizeof(arraycount));
    memset((char *)arraylength, 0, sizeof(arraylength));

    for (; hp <= (Handle *) hpmax ; hp++) {
	if ((hp->obj == 0) || (obj_free((unsigned char *)unhand(hp)))) {
	    continue;
	}
	switch (obj_flags(hp)) {
	  case T_NORMAL_OBJECT:
	    cb = obj_classblock(hp);
	    p = &tab[(((long)cb) >> 2) % MAX_CLASS_HASH];
	    while ((p->cb != 0) && (p->cb != cb)) {
		if (p-- == tab) {
		    p = &tab[MAX_CLASS_HASH - 1];
		}
	    }
	    if (p->cb == 0) {
		p->cb = cb;
	    }
	    p->count++;
	    break;

	  case T_CLASS:
	    cb = ((ClassClass **)unhand((HArrayOfObject*)hp)->body)[obj_length(hp)];
	    p = &tab[(((long)cb) >> 2) % MAX_CLASS_HASH];
	    while ((p->cb != 0) && (p->cb != cb)) {
		if (p-- == tab) {
		    p = &tab[MAX_CLASS_HASH - 1];
		}
	    }
	    if (p->cb == 0) {
		p->cb = cb;
	    }
	    p->arraycount++;
	    p->arraylength += obj_length(hp);
	    break;

	  default:
	    if (obj_flags(hp) < T_BOOLEAN) {
		continue;
	    }
	    arraycount[tmap[obj_flags(hp)]]++;
	    arraylength[tmap[obj_flags(hp)]] += obj_length(hp);
	    break;
	}
	handles_used++;
    }

    fprintf(fp, "# handles-used, handles-free heap-used heap-free\n");
    fprintf(fp, "%d %d %d %d\n", handles_used, handles_count - handles_used, 
	    TotalMemoryCtr - FreeMemoryCtr, FreeMemoryCtr);

    fprintf(fp, "# type count bytes\n");
    for (i = 0 ; i < ARRAYTYPES ; i++) {
	if (arraycount[i]) {
	    fprintf(fp, "[%c %d %d\n", arrayinfo[i].sig, arraycount[i], arraylength[i] * arrayinfo[i].factor);
	}
    }
    p = &tab[MAX_CLASS_HASH];
    while (p-- != tab) {
	if (p->cb != 0) {
	    if (p->count > 0) {
		fprintf(fp, "L%s; %d %d\n", 
			classname(p->cb), p->count, 
			p->count * cbInstanceSize(p->cb));
	    }
	    if (p->arraycount > 0) {
		fprintf(fp, "[L%s; %d %d\n", classname(p->cb), p->arraycount, p->arraylength * sizeof(OBJECT));
	    }
	}
    }
    free(tab);
}

void
prof_heap(FILE *fp)
{

    gc(0, ~0);
    profHandles(fp);
}


/*
 * Debugging facilities
 */

#ifdef DEBUG

int GCValidHandle(Handle *h) {
    SetLimits();

    return ValidHandle(h) && h->obj != 0;
}

void
printPool(char *s)
{
    SetLimits();
    register unsigned char *p = pmin;
    register col = 0;
    char buf[100];
    if (s) {
	printf("\014%s\n", s);
    }
    while (p < pmax) {
	if (++col > 7) {
	    putchar('\n'), col = 1;
	}
	if (obj_len(p) < 4 || obj_len(p) > (poollimit-pool)) {
	    printf("Bogus length");
	    break;
	}
	printf("%d%c%c%c\t", obj_len(p),
	       p == allochand ? '*' : ' ',
	       obj_free(p) ? 'F' : ' ',
	       IsMarked(p + sizeof(hdr)) ? 'P' : ' ');
	p += obj_len(p);
    }
    putchar('\n');
}

void
printHandles(char *s)
{
    SetLimits();
    Handle *hp;
    printf("\n\n%s\n", s);
    for (hp = (Handle *) hpool; hp <= (Handle *) hpmax; hp++) {
	if (hp->obj != 0) {
	    printf("%X %X%c\n", hp, hp->obj,
		 ValidPtr(hp->obj) ? (IsMarked(hp->obj) ? 'P' : ' ') : '*');
	}
    }
    printf("\n");
}

void
validatePool(long quick)
{
    SetLimits();
    register unsigned char *p = pmin;
    register Handle *hp;
    int err = 0;
    char buf[100];

    while (p < pmax) {
	assert(obj_len(p) >= 4);
	assert(obj_len(p) < (poollimit-pool));
	if (obj_free(p)) { 
	    assert(((long *) p)[1] == 0x55555555);
	} else if (!quick) {
	    for (hp = (Handle *) hpool; hp <= (Handle *) hpmax; hp++) {
		if (hp->obj == (ClassObject *) (p + sizeof(hdr))) {
		    goto HasHandle;
		}
	    }
	    fprintf(stderr, "@%X -- alloced object with no handle\n", p), err++;
	}
HasHandle:
	p += obj_len(p);
    }
    for (hp = (Handle *) hpool; hp <= (Handle *) hpmax; hp++) {
	if (hp->obj)
	    if (ValidPtr(hp->obj) && obj_free((char *) hp->obj - sizeof(hdr))) {
		fprintf(stderr, "@%X/%X -- Handle points to free block\n", hp, hp->obj), err++;
	    } else {
		switch (obj_flags(hp)) {
		case T_NORMAL_OBJECT:
		    if (obj_methodtable(hp) == 0)
			fprintf(stderr, "1 @%X/%X -- Invalid method table\n", hp, hp->methods), err++;
		    else if (obj_methodtable(hp) != cbMethodTable(obj_classblock(hp)))
			fprintf(stderr, "2 @%X/%X -- Invalid method table\n", hp, hp->methods), err++;
		   break;

		default:
		   if (obj_flags(hp) < T_ARRAY )
		       fprintf(stderr, "3 @%X/%X -- Invalid method table\n", hp, hp->methods), err++;
		}
	    }
    }
    fprintf(stderr, err == 0 ? "OK memory pool\n" : "***%d errors in the memory pool***\n", err);
}
#else
#define validatepool(a) 0
#endif /* DEBUG */


#ifdef NOTCURRENTLYUSED

void
objFree(HObject * o)
{
    register unsigned char *p;

    HEAP_LOCK();
#ifdef DEBUG
    if (o == 0) {
	printf("freeing null object\n");
	abort();
    }
#endif /* DEBUG */

    p = ((unsigned char *) unhand(o)) - sizeof(hdr);

#ifdef DEBUG
    if (p > poollimit || p < pool) {
	printf("freeing object not in the pool\n");
	abort();
    }
#endif /* DEBUG */

    if (!obj_free(p)) {
	FreeMemoryCtr += obj_len(p);
	obj_setfree(p);
	unhand(o) = 0;
#ifdef DEBUG
	((long *) p)[1] = 0x55555555;
#endif /* DEBUG */
    }
    HEAP_UNLOCK();
}

#endif /* NOTCURRENTLYUSED */

