/*
 * @(#)interpreter.c	1.102 95/05/01
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Interpret an Java object file
 */

#include <sys/types.h>
#include <string.h>
#include <stdarg.h>
#include <bool.h>
#include <memory.h>

#include <opcodes.h>
#include <oobj.h>
#include <interpreter.h>
#include <signature.h>
#include <code.h>
#include <tree.h>
#include <threads.h>
#include <monitor.h>
#include <exceptions.h>
#include <javaString.h>
#include <path.h>
#include <byteorder.h>

#include <sys_api.h>

#include <java_lang_Thread.h>

#ifdef ADE
#include <fp_old_DbgBreakpointQueue.h>
#endif

#ifdef TRACING
int trace;
int tracem;
#  define TRACE(a) (trace ? printf("%6X %8X", (int)threadSelf(), (int)pc), \
	printf a, fflush(stdout), 0 : 0)
#  define TRACEIF(o, a) (opcode == o ? TRACE(a) : 0)
#else
#  define TRACE(a)
#  define TRACEIF(a, b)
#endif

#ifdef STATISTICS
int instruction_count[256];
#endif

extern struct StrIDhash *interfaceHash;
ExecEnv *DefaultExecEnv;
struct execenv *ee;

/*
 * Lock around rewriting method code for quick_ instruction optimization
 */
static sys_mon_t _code_lock;
#define CODE_LOCK_INIT()    monitorRegister(&_code_lock, "Code rewrite lock")
#define CODE_LOCK()	    sysMonitorEnter(&_code_lock)
#define CODE_UNLOCK()	    sysMonitorExit(&_code_lock)

/*
 * Java stack management
 */

/*
 * Lock around code freeing deleted Java stack segments
 */
static sys_mon_t _ostack_lock;
#define OSTACK_LOCK_INIT()  monitorRegister(&_ostack_lock, "Java stack lock")
#define OSTACK_LOCK()	    sysMonitorEnter(&_ostack_lock)
#define OSTACK_UNLOCK()	    sysMonitorExit(&_ostack_lock)

void FreeJavaStackMemory();

JavaStack *CreateNewJavaStack(ExecEnv *ee, JavaStack *previous_stack)
{
    JavaStack *stack;

    FreeJavaStackMemory();	/* Free any stack segments for dead threads */

    stack = malloc(sizeof (*stack));
    stack->execenv = ee;
    stack->prev = previous_stack;
    stack->next = 0;
    if (previous_stack) {
	assert(previous_stack->next == 0);
        previous_stack->next = stack;
	stack->stack_so_far = previous_stack->stack_so_far 
	          + JAVASTACK_CHUNK_SIZE * sizeof(stack->data[0]);
    } else {
	stack->stack_so_far = JAVASTACK_CHUNK_SIZE * sizeof(stack->data[0]);
    }
    stack->end_data = &stack->data[JAVASTACK_CHUNK_SIZE];
    return stack;
}

void InitializeExecEnv(ExecEnv *ee, Handle *thread) {
    if (DefaultExecEnv == 0 && thread == 0)
	DefaultExecEnv = ee;
    ee->current_frame = 0;
    ee->thread = thread;
    ee->initial_stack = CreateNewJavaStack(ee, 0);
    if (thread) { /* we don't have thread at initialization */
        unhand((Hjava_lang_Thread *)thread)->eetop = (long)ee;
    }
    exceptionClear(ee);
}

#ifdef DEBUG
int java_stack_chunks_freed = 0;
#endif DEBUG

/*
 * The list on which we put freed thread stacks pending reclamation.
 */
typedef struct free_javastack_t {
    struct free_javastack_t *next;
} free_javastack_t;

static free_javastack_t *javastackFreeList = (free_javastack_t *) 0;

void DeleteExecEnv(ExecEnv *ee, Handle *thread) {
    JavaStack *this_stack;

    if (thread) { 
        unhand((Hjava_lang_Thread *)thread)->eetop = (long)0;
    }
    for (this_stack = ee->initial_stack; this_stack != 0; ) {
        JavaStack *next_stack = this_stack->next;
     
        ((free_javastack_t *) this_stack)->next = javastackFreeList;
        javastackFreeList = (free_javastack_t *) this_stack;
#ifdef DEBUG
        java_stack_chunks_freed++;
#endif DEBUG
        this_stack = next_stack;
    }
    ee->initial_stack = 0;
}

/*
 * Free Java stack segments for dead threads.  We do this whenever the
 * idle thread runs, and whenever we grow an Java stack.
 */
void FreeJavaStackMemory() {
    free_javastack_t* mem;
    free_javastack_t* next;

    if (javastackFreeList != NULL) {
	OSTACK_LOCK();
	mem = javastackFreeList;
	javastackFreeList = 0;
	OSTACK_UNLOCK();
	while (mem) {
	    next = mem->next;
	    free(mem);
	    mem = next;
	}
    }
}

bool_t ExecuteJava(unsigned char  *, ExecEnv *ee);


/* Call an Java method from C.  The java method may have any number of arguments
 * and may return a single 4 byte value which may be an integer, a pointer to
 * an object handle, whatever.
 */
long
execute_java_dynamic_method(ExecEnv *ee, HObject *obj, 
			   char *method_name, char *signature, ...)
{
    va_list args;
    long result;
    va_start(args, signature);
        result = do_execute_java_method_vararg(ee, obj, method_name, signature, 
					      0, FALSE, args, 0, FALSE);
    va_end(args);
    return result;
}


/* Call a static Java method from C.  The java method may have any number of
 * arguments and may return a single 4-byte value.`
 */
long
execute_java_static_method(ExecEnv *ee, ClassClass *cb, 
			  char *method_name, char *signature, ...)
{
    va_list args;
    long result;
    va_start(args, signature);
        result = do_execute_java_method_vararg(ee, cb, method_name, signature,
					      0, TRUE, args, 0, FALSE);
    va_end(args);
    return result;
}

/* Create a new Java argument from C.  For example:
 *      obj = execute_constructor(0, "<init>(II)Lfp/ade/gui/MouseEvent;", x, y);
 */

HObject *execute_java_constructor(struct execenv *ee,
				 char *classname,
				 ClassClass *cb,
				 char *signature, ...)
{
#ifdef TRACING
    int pc = 0;			/* for TRACE */
#endif

    HObject *obj;
    va_list args;
    char real_signature[256];

    /* Check that the method name is <init> */
    /* Look for the SIGNATURE_ENDFUNC in the method signature */
    sprintf(real_signature, "%sV", signature);

    /* Find the class, and allocate the object. */
    if (cb == (ClassClass *)0 && (cb = FindClass(ee, classname, TRUE)) == 0)
	return 0;
    if (ee == 0)
        ee = EE();
    if ((obj = newobject(cb, 0, ee)) == 0)
        return 0;

    TRACE(("execute_java_constructor new %s => %s\n", 
	   classname(cb), Object2CString(obj)));

    /* Call the constructor */
    va_start(args, signature);
       do_execute_java_method_vararg(ee, obj, "<init>", real_signature,
				    0, FALSE, args, 0, FALSE);
    va_end(args);
    return obj;
}

long
do_execute_java_method(ExecEnv *ee, void *obj, 
		      char *method_name, char *signature, 
		      struct methodblock *mb,
		      bool_t isStaticCall, ...) {
    va_list args;
    long result;
    va_start(args, isStaticCall);
        result = do_execute_java_method_vararg(ee, obj, 
					  method_name, signature, 
					  mb, isStaticCall, args, 0, FALSE);
    va_end(args);
    return result;
}


long
do_execute_java_method_vararg(ExecEnv *ee, void *obj, 
			      char *method_name, char *method_signature, 
			      struct methodblock *mb,
			      bool_t isStaticCall, va_list args, 
			      long *otherBits, bool_t shortFloats) {

    JavaFrame *current_frame, *previous_frame;
    JavaStack *current_stack;
    int retval;
    char *p;

    if (ee == 0)
        ee = EE();

    if (mb) {
        method_name = fieldname(&mb->fb);
	method_signature = fieldsig(&mb->fb);
    } else {
	Str2ID(&interfaceHash, method_signature, 0, TRUE);
	Str2ID(&interfaceHash, method_name, 0, TRUE);
    }
    
    previous_frame = ee->current_frame;
    if (previous_frame == 0) {	
	/* bottommost frame on this Exec Env. */
	current_stack = ee->initial_stack;
	current_frame = (JavaFrame *)(current_stack->data); /* no vars */
    } else {
	int args_size = mb ? mb->args_size 
	                   /* We don't need the +1 if a static call, but it's
			    * okay to make args_size too big.  */
	                   : (Signature2ArgsSize(method_signature) + 1);
	current_stack = previous_frame->javastack; /* assume same stack */
	if (previous_frame->current_method) {
	    int size = previous_frame->current_method->maxstack;
	    current_frame = (JavaFrame *)(&previous_frame->ostack[size]);
	} else {
	    /* The only frames that don't have a mb are pseudo frames like 
	     * this one and they don't really touch their stack. */
	    current_frame = (JavaFrame *)(previous_frame->optop + 3);
	}
	if (current_frame->ostack + args_size > current_stack->end_data) {
	    /* Ooops.  The current stack isn't big enough.  */
	    current_stack = current_stack->next 
	       ? current_stack->next : CreateNewJavaStack(ee, current_stack);
	    current_frame = (JavaFrame *)(current_stack->data); /* no vars */	
	} 
    }
    ee->current_frame = current_frame;

    current_frame->prev = previous_frame;
    current_frame->javastack = current_stack;
    current_frame->optop = current_frame->ostack;
    current_frame->vars = 0;	/* better not reference any! */
    current_frame->monitor = 0;	/* not monitoring anything */
    current_frame->current_method = 0; /* not part of a method. */
    
    /* Push the target object, if not a static call */
    if (!isStaticCall) 
        (current_frame->optop++)->p = obj;

    for (p = method_signature + 1; *p != SIGNATURE_ENDFUNC; p++) {
	switch (*p) {
	  case SIGNATURE_BOOLEAN:
	  case SIGNATURE_BYTE:
	  case SIGNATURE_CHAR:
	  case SIGNATURE_INT:
	    (current_frame->optop++)->i = va_arg(args, long);
	    break;
	  case SIGNATURE_FLOAT:
	      if (shortFloats) 
		  (current_frame->optop++)->i = va_arg(args, long);
	      else 
		  (current_frame->optop++)->f = va_arg(args, double);
	      break;
	  case SIGNATURE_CLASS:
	    (current_frame->optop++)->h = va_arg(args, HObject *);
	    while (*p != SIGNATURE_ENDCLASS) p++;
	    break;
	  case SIGNATURE_ARRAY:
	    while ((*p == SIGNATURE_ARRAY)) p++;
	    if (*p == SIGNATURE_CLASS) 
	        { while (*p != SIGNATURE_ENDCLASS) p++; } 
	    (current_frame->optop++)->p = va_arg(args, void *);
	    break;
	  case SIGNATURE_LONG:
	    SET_INT64(current_frame->optop, va_arg(args, int64_t));
	    current_frame->optop += 2;
	    break;
	  case SIGNATURE_DOUBLE:
	    SET_DOUBLE(current_frame->optop, va_arg(args, double));
	    current_frame->optop += 2;
	    break;
	  default:
	    fprintf(stderr, "Invalid method signature '%s'\n", method_name);
	    return 0;
	}
    }
    {
	ClassClass *cb = (isStaticCall || mb != 0) ? obj 
	                       : obj_array_classblock((Handle *)obj);
	unsigned char pc[6];
	cp_item_type  constant_pool[10];
	unsigned char cpt[10];
	bool_t isInitializer = (method_name[0] == '<'  &&  /* fast test */
				 !isStaticCall && 
				 !strcmp(method_name, "<init>"));
	
	constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p = cpt;
	cpt[0] = CONSTANT_POOL_ENTRY_RESOLVED;

	if (isStaticCall || mb) {
	    pc[0] = isStaticCall ? (mb ? opc_invokestatic_quick 
				       : opc_invokestatic) 
	                         : opc_invokenonvirtual_quick;
	    pc[1] = 0; pc[2] = 1;	/* constant pool entry #1 */
	    pc[3] = opc_return;
	} else {
	    int nargs = (current_frame->optop - current_frame->ostack);
	    pc[0] = isInitializer ? opc_invokenonvirtual 
		                  : opc_invokeinterface;
	    pc[1] = 0; pc[2] = 1;	/* constant pool entry #1 */
	    pc[3] = nargs;
	    pc[4] = 0;			/* padding needed for guess */
	    pc[5] = opc_return;
	}

	if (mb) {
	    constant_pool[1].p = mb;
	    cpt[1] = CONSTANT_POOL_ENTRY_RESOLVED | CONSTANT_Methodref;
	} else {
	    constant_pool[1].i = (2 << 16) + 3; /* Class, NameAndType */
	    cpt[1] = (isStaticCall || isInitializer)
		        ? CONSTANT_Methodref : CONSTANT_InterfaceMethodref; 
	    constant_pool[2].p = cb; /* class */
	    cpt[2] = CONSTANT_POOL_ENTRY_RESOLVED | CONSTANT_Class;
	    constant_pool[3].i = (4 << 16) + 5;	/* Name, Type */
	    cpt[3] = CONSTANT_NameAndType;
	    constant_pool[4].cp = method_name;
	    cpt[4] = CONSTANT_POOL_ENTRY_RESOLVED | CONSTANT_Asciz;
	    constant_pool[5].cp = method_signature;
	    cpt[5] = CONSTANT_POOL_ENTRY_RESOLVED | CONSTANT_Asciz;	    
	}

	current_frame->constant_pool = constant_pool;

	/* Run the byte codes outside the lock and catch any exceptions. */
	ee->exceptionKind = EXCKIND_NONE;
	if (otherBits != 0)
	    *otherBits = 0;
	if (!ExecuteJava(pc, ee) || (p[1] == SIGNATURE_VOID)) {
	    retval = 0; 
	} else { 
	    retval = current_frame->optop[-1].i;
	    if ((p[1] == SIGNATURE_DOUBLE || p[1] == SIGNATURE_LONG) 
		  && (otherBits != 0))
		*otherBits = current_frame->optop[-2].i;
	}
	/* Our caller can look at ee->exceptionKind and ee->exception. */
	ee->current_frame = previous_frame;
	return retval;
    }
}

#define STACK_REDZONE 4096

char       *
Object2CString(Handle *oh)
{                               /* ARB to C string */
    ClassObject *o;
    ClassClass *cb;
    static char buf[100];
    int maxlen = sizeof(buf) - 10;

    if (oh == 0)
	return "NULL";
    o = unhand(oh);
    if (-1000 < (long) o && (long) o < 1000)
	sprintf(buf, "BOGUS-PTR[%d]", (int)o);
    else {
	switch (obj_flags(oh)) {
	case T_NORMAL_OBJECT:
	    {
		char cbuf[128];
		cb = obj_classblock(oh);
		sprintf(buf, "%s@%X%c%X", classname2string(classname(cb), cbuf, sizeof(cbuf)),
			(int)oh, DIR_SEPARATOR, (int)o);
	    }
	    break;
	case T_CHAR: {
	    HArrayOfChar *str = (HArrayOfChar *) oh;
	    int len = obj_length(str);
	    unicode *s = unhand(str)->body;
	    char *p = buf;
	    
	    len = (len > maxlen ? maxlen : len);

	    *p++ = '"';
	    while (--len >= 0)
		*p++ = *s++;
	    *p++ = '"';
	    *p++ = '\0';
	    break;
	    }
	case T_CLASS:
#define BODYOF(h)   unhand(h)->body
	    {
		char cbuf[128];

		cb = ((ClassClass **) BODYOF((HArrayOfObject*)oh))[obj_length(oh)];
		sprintf(buf, "%s[%d]",
			cb ? classname2string(classname(cb), cbuf, sizeof(cbuf)) : "*MissingClassName*",
			obj_length(oh));
	    }
	    break;
	default:
	    sprintf(buf, "%s[%d]", 
		    arrayinfo[obj_flags(oh)].name, obj_length(oh));
	}
    }
    return buf;
}

#ifdef ADE

static Hfp_old_DbgBreakpointQueue *the_bkptQ = 0;

static unsigned char 
get_breakpoint_opcode(unsigned char *pc, ExecEnv *ee)
{
    HThread *hp = threadSelf();

    int real_opcode;

    /* If this is the first time we're called, perform initialization. */
    if (the_bkptQ == 0) {
	/* Find the class */
	ClassClass *c = FindClass(ee, INTRP_BRKPT_STRING, TRUE);
	/* Find the static variable that contains the queue */
	Hfp_old_DbgBreakpointQueue **bkptQ =
	      (Hfp_old_DbgBreakpointQueue **)getclassvariable(c, "the_bkptQ");
	if (bkptQ == 0 || *bkptQ == 0) {
	    printf("?breakpoint queue is null?\n");
	    SignalError(ee, JAVAPKG "InternalException", 
			"debugger not initialized");
	    return 0;
	} else {
	    the_bkptQ = *bkptQ;
	}
    }

    /* printf("BREAKPOINT at pc=%x\n", ee->lastpc); */

    /* transfer control to the debugger thread by signalling the monitor. */
    monitorEnter(obj_monitor(the_bkptQ));

    /* communicate the information about this breakpoint */
    unhand(the_bkptQ)->thread = hp;
    unhand(the_bkptQ)->pc = (int)pc;
    unhand(the_bkptQ)->opcode = -1;

    /* this will yield to the debugger thread since it is running at */
    /* MaximumPriority */
    monitorNotifyAll(obj_monitor(the_bkptQ));
    monitorExit(obj_monitor(the_bkptQ));

    real_opcode = unhand(the_bkptQ)->opcode;
    unhand(the_bkptQ)->opcode = -1;
    return real_opcode;
}

static void
set_breakpoint_opcode(unsigned char *pc, unsigned char opcode)
{
    HThread *hp = threadSelf();

    int real_opcode;

    /* transfer control to the debugger thread by signalling the monitor. */
    monitorEnter(obj_monitor(the_bkptQ));

    /* communicate the information about this breakpoint */
    unhand(the_bkptQ)->thread = hp;
    unhand(the_bkptQ)->pc = (int)pc;
    unhand(the_bkptQ)->opcode = opcode;

    /* Yield to the debugger thread, running at Maximum priority. */
    monitorNotifyAll(obj_monitor(the_bkptQ));
    monitorExit(obj_monitor(the_bkptQ));
}

void 
dbg_breakpoint(HThread *hp, int kind, ExecEnv *ee) { }

#endif

static bool_t is_subclass_of(ClassClass *, ClassClass *, ExecEnv *);
static bool_t array_is_instance_of_array_type(Handle*, ClassClass *, ExecEnv *);


bool_t
is_instance_of(Handle * h, ClassClass *dcb, ExecEnv *ee)
{
    if (h == 0) {		/* casting 0 always succeeds */
	return TRUE;
    } else if (obj_flags(h) == T_NORMAL_OBJECT) {
	/* h is an object.  Get its class and to the work */
	return is_subclass_of(obj_classblock(h), dcb, ee);
    } else {
	/* h is an array.  dcb must either be classobject, or an array class */
	return dcb->name[0] == SIGNATURE_ARRAY
	           ? array_is_instance_of_array_type(h, dcb, ee)
		   : (dcb == classobject);
    }
}

static bool_t
is_subclass_of(ClassClass *cb, ClassClass *dcb, ExecEnv *ee)
{
    if (dcb->access & ACC_INTERFACE) {
	for (; ; cb = unhand(cbSuperclass(cb))) {
	    if (ImplementsInterface(cb, dcb, ee)) 
		return TRUE;
	    if (cbSuperclass(cb) == 0)
		return FALSE;
	}
    } else {
	for (; ; cb = unhand(cbSuperclass(cb))) {
	    if (cb == dcb)
		return TRUE;
	    if (cbSuperclass(cb) == 0)
		return FALSE;
	}
    }
}

/* Return TRUE if cb implements the interface icb */

bool_t 
ImplementsInterface(ClassClass *cb, ClassClass *icb, ExecEnv *ee)
{
    /* Let's avoid resolving the class unless we really have to.  So first,
     * First, do a string comparison of icb against all the interfaces thant
     * cb implements.
     */
    char *icb_name = icb->name;
    cp_item_type *constant_pool = cb->constantpool;
    int i;
    for (i = 0; i < (int)(cb->implements_count); i++) {
	int interface_index = cb->implements[i];
	char *interface_name = 
	        GetClassConstantClassName(constant_pool, interface_index);
	if (strcmp(icb_name, interface_name) == 0)
	    return TRUE;
    }
    /* See if any of the interfaces that we implement possibly implement icb */
    for (i = 0; i < (int)(cb->implements_count); i++) {
	int interface_index = cb->implements[i];
	ClassClass *sub_cb;
	if (!ResolveClassConstantFromClass(cb, interface_index, ee, 
					   1 << CONSTANT_Class))
	    return FALSE;
	sub_cb = (ClassClass *)(constant_pool[interface_index].p);
	if (ImplementsInterface(sub_cb, icb, ee))
	    return TRUE;
    }
    return FALSE;
}

bool_t array_is_instance_of_array_type(Handle * h, ClassClass *cb, ExecEnv *ee)
{
    int class_depth = cb->constantpool[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i;
    int class_type = cb->constantpool[CONSTANT_POOL_ARRAY_TYPE_INDEX].i;
    ClassClass *class_class = 
	             cb->constantpool[CONSTANT_POOL_ARRAY_CLASS_INDEX].p;
    int item_depth, item_type;
    ClassClass *item_class;
    if (obj_flags(h) == T_CLASS) {
	/* We have an array of classes.  Find out the types of the elements. */
	ArrayOfObject *array = (ArrayOfObject *)unhand(h);
	int length = obj_length(h);
	ClassClass *array_cb = (ClassClass *) array->body[length];
	if (array_cb->name[0] == SIGNATURE_ARRAY) {
	    /* an Array of arrays */
	    item_depth = 
		  array_cb->constantpool[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i + 1;
	    item_type = 
		  array_cb->constantpool[CONSTANT_POOL_ARRAY_TYPE_INDEX].i;
	    item_class = 
		  array_cb->constantpool[CONSTANT_POOL_ARRAY_CLASS_INDEX].p;
	} else {
	    /* An array of clases */
	    item_depth = 1;
	    item_type = T_CLASS;
	    item_class = array_cb;
	}
    } else {
	item_depth = 1;
	item_type = obj_flags(h);
	item_class = 0;		/* immaterial */
    }
    if ((class_depth == item_depth) && (class_type == item_type)) {
	if ( (class_type != T_CLASS) || 
	     (class_class == item_class) || 
	     is_subclass_of(item_class, class_class, ee))
	    return TRUE;
    }
    return FALSE;
}
    
		
#ifndef ADE
#    define WITH_OPCODE_CHANGE 
#    define END_WITH_OPCODE_CHANGE
#else
#    define WITH_OPCODE_CHANGE { \
               bool_t debugging = (*pc == opc_breakpoint); 
#    define END_WITH_OPCODE_CHANGE \
               if (debugging && (*pc != opc_breakpoint)) { \
                   opcode = *pc; \
                   *pc = opc_breakpoint;  \
		   set_breakpoint_opcode(pc, opcode); \
                   goto again; } }
#endif                         


/* An exception has been thrown.  Find the pc at which this exception
 * is handled.  Alternatively, rethrow the error if no appropriate pc
 * is found.
 */

static unsigned char *
ProcedureFindThrowTag(JavaFrame *frame, Handle *object, unsigned char *pc)
{
    if (frame->current_method == 0) {
	return 0;
    } else {
	/* Search current class for catch frame, based on pc value,
	   matching specified type. */
	struct methodblock *mb = frame->current_method;
	struct CatchFrame *cf = mb->exception_table;
	int cnt = mb->exception_table_length;
	int pcTarget = pc - mb->code; /* offset from start of code */
	for (; --cnt >= 0; cf += 1) {
	    if (cf->start_pc <= pcTarget && pcTarget < cf->end_pc) {
		/* Is the class name a superclass of object's class. */
		if (cf->catchType == 0) {
		    /* Special hack put in for finally.  Always succeed. */
		    goto found;
		} else {
		    /* The usual case */
		    ClassClass *cb = obj_array_classblock(object);
		    char *catchname = 
			 GetClassConstantClassName(frame->constant_pool, 
						   cf->catchType);
		    for (;; cb = unhand(cbSuperclass(cb))) {
			if (!strcmp(classname(cb), catchname)) {
			    goto found;
			}
			if (cbSuperclass(cb) == 0) {
			    /* Not found.  Try next catch frame. */
			    break;
			}
		    }
		}
	    }
	}
	/* not found in this execution environment */
	return 0;
      found:
	return mb->code + cf->handler_pc;
    }
}
    
/* Called by opc_multianewarray.  Create a filled in multidimensional array
 *    dimensions:  The number of dimensions to fill in
 *    array_cb:    The type of the array
 *    sizes:       An array of the dimension sizes
 *
 * Return 0 if we've run out of space.  Otherwise, return the allocated array.
 */

HObject *
MultiArrayAlloc(int dimensions, ClassClass *array_cb, stack_item *sizes)
{
    /* Get the base type and depth of the array class we're constructing */
    cp_item_type *constant_pool = cbConstantPool(array_cb);
    unsigned char *type_table = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;
    int depth = constant_pool[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i;
    int base_type = constant_pool[CONSTANT_POOL_ARRAY_TYPE_INDEX].i;
    ClassClass *sub_array_cb[256];
    char *class_name = classname(array_cb);
    int i;
    static HObject *MultiArrayAlloc2(int, ClassClass **, stack_item *, int);

    assert(class_name[0] == SIGNATURE_ARRAY);
    assert(dimensions <= depth);

    /* Fill in sub_array_cb with the array types of each of the lower.
     * dimensions.  Put a 0 in the slot if we've reached a non-class base type
     */
    for (i = 1; i <= dimensions; i++) {
	if (i < depth) 
	    sub_array_cb[i - 1] = 
		FindClassFromClass(ee, class_name + i, FALSE, array_cb);
	else if (i == depth && base_type == T_CLASS) 
	    sub_array_cb[i - 1] = 
		constant_pool[CONSTANT_POOL_ARRAY_CLASS_INDEX].p;
	else 
	    sub_array_cb[i - 1] = 0;
    }
    /* Call a recursive function that does all the work */
    return MultiArrayAlloc2(dimensions, sub_array_cb, sizes, base_type);
}

/* This function does all the actual work for MultiArrayAlloc(), after that
 * function has done all the checking and initial work. 
 */
static HObject *
MultiArrayAlloc2(int dimensions, ClassClass **types, stack_item *sizes, 
		 int base_type)
{
    int size = sizes[0].i;	/* size of this dimension */
    HObject *ret;
    int i;

    if (types[0] != 0) {
	ret = ArrayAlloc(T_CLASS, size); /* generate the array */
	if (ret == 0) 
	    return 0;
	/* fill in the type  */
	unhand(((HArrayOfObject *)ret))->body[size] = (HObject*)types[0];
	if (dimensions > 1) {
	    /* fill in the subarrays */
	    for (i = 0; i < size; i++) {
		HObject *sub = MultiArrayAlloc2(dimensions - 1, types + 1, 
						sizes + 1, base_type);
		if (sub == 0)	/* out of memory? */
		    return 0;
		unhand(((HArrayOfObject *)ret))->body[i] = sub;
	    } 
	} else {
	    /* this is the last dimension.  Just zero the array */
	    memset((char *) unhand(ret), 0, sizearray(T_CLASS, size));
	}
    } else {
	assert(dimensions == 1);
	ret = ArrayAlloc(base_type, size);
	if (ret == 0)		/* out of memory? */
	    return 0;
	memset((char *) unhand(ret), 0, sizearray(base_type, size));
    }
    return ret;
}



/*
    This routine is used to get the current time in milliseconds for
    the profiling routines.
*/
long now()
{
    return(sysGetMilliTicks());
}


/* Execute bytecodes at "pc" in the specified execution environment.
   Returns TRUE if completed OK, FALSE if finished because of some
   uncaught exception. */

bool_t ExecuteJava(unsigned char *initial_pc, ExecEnv *ee) 
{
/* This is ugly.  Some machines (and compilers) produce better code from 
 * positive offsets than from negative offsets.  So we cache ee->optop in
 * a local variable, but have it point OPTOP_OFFSET words further down the 
 * stack.  That way, stack references will be a nonnegative index off offset
 */

/* avh: using a static buffer to print longs is unsafe */
static char longstringbuf[30];

#define OPTOP_OFFSET 4

#define S_INFO(offset)    (optop[offset + OPTOP_OFFSET])
#define S_INT(offset)     (S_INFO(offset).i)
#define S_FLOAT(offset)   (S_INFO(offset).f)
#define S_HANDLE(offset)  (S_INFO(offset).h)
#define S_OBJECT(offset)  (S_INFO(offset).o)
#define S_POINTER(offset) (S_INFO(offset).p)
#define S_OSTRING(offset) (Object2CString(S_HANDLE(offset)))
#define S_LONGSTRING(offset) \
        (int642CString(S_LONG(offset), longstringbuf, sizeof(longstringbuf)))

#define S_DOUBLE(offset)  GET_DOUBLE(&(S_INFO(offset)))
#define S_LONG(offset)    GET_INT64(&(S_INFO(offset)))
#define SET_S_DOUBLE(offset, value) SET_DOUBLE(&(S_INFO(offset)), (value))
#define SET_S_LONG(offset, value)   SET_INT64(&(S_INFO(offset)), (value))

#define REAL_OPTOP()      (&S_INFO(0))

#define DECACHE_OPTOP()   (frame->optop = REAL_OPTOP())
#define CACHE_OPTOP()     (optop = frame->optop - OPTOP_OFFSET)

#define SIZE_AND_STACK(size, stack) pc += size; optop += stack; continue
#define SIZE_AND_STACK_BUT(size, stack) pc += size; optop += stack;
#define pc2signedshort(pc)  ((((signed char *)(pc))[1] << 8) | pc[2])

/* I can't use the do { } while(0) trick because the compiler complains about
 * the test not bein reachable.  So I'm forced to use an uglier trick.
 */
#undef  ERROR			/* Avoid conflict on PC */
#define ERROR(name, data) \
        if (1) { frame->lastpc = pc;  \
	         SignalError(ee, JAVAPKG name, data); \
                 goto handle_exception; } else 

    register unsigned char  *pc = initial_pc;
    register JavaFrame       *frame = ee->current_frame;
    JavaFrame                *initial_frame = frame;
    register stack_item     *optop;
    register stack_item     *vars = frame->vars;

    unsigned char            opcode;
    cp_item_type            *constant_pool = frame->constant_pool;
    struct methodblock      *mb;

    Handle                  *o;	/* object whose method we're calling  */
    int	    args_size;
    sys_thread_t    *stack_var;
  
    CACHE_OPTOP();
    if (threadCheckStack(STACK_REDZONE, stack_var)) {
	/* got a stack overflow: have to be careful to not do
	   anything that writes to the stack.
	   "StackOverflowException" is a special kind of
	   exception, just because we can't call SignalError */
	frame->lastpc = pc;
	ee->exceptionKind = EXCKIND_STKOVRFLW;
	return FALSE;
    }

    if (trace) {
	char                where_am_i[100];
	pc2string(pc, frame->current_method, 
		  where_am_i, where_am_i + sizeof where_am_i);
	fprintf(stdout, "Entering %s\n", where_am_i);
	fflush(stdout);
    }

    while (1) {
	assert(REAL_OPTOP() >= frame->ostack);
	assert(!frame->current_method || 
	       REAL_OPTOP() <= frame->ostack + frame->current_method->maxstack);
#ifdef DBINFO
	frame->lastpc = pc;
#endif
	opcode = *pc;
again:
#ifdef STATISTICS
	instruction_count[opcode]++;
#endif
	switch (opcode) { 

	      /* NOPs.  Don't do anything. */
	case opc_nop:
	    TRACE(("\tnop\n"));
	    SIZE_AND_STACK(1, 0);

	case opc_aload:
	case opc_iload:
        case opc_fload:
	    S_INFO(0) = vars[pc[1]];
	    TRACEIF(opc_iload, ("\tiload r%d => %d\n", pc[1], S_INT(0)));
	    TRACEIF(opc_aload, ("\taload r%d => %s\n", pc[1], S_OSTRING(0)));
	    TRACEIF(opc_fload, ("\tfload r%d => %g\n", pc[1], S_FLOAT(0)));
	    SIZE_AND_STACK(2, 1);
	      
	case opc_lload:
	case opc_dload: {
	    int reg = pc[1];
	    S_INFO(0) = vars[reg];
	    S_INFO(1) = vars[reg + 1];
	    TRACEIF(opc_lload, ("\tlload r%d => %s\n", reg, S_LONGSTRING(0)));
	    TRACEIF(opc_dload, ("\tdload r%d => %g\n", reg, S_DOUBLE(0)));
	    SIZE_AND_STACK(2, 2);
	}

#define OPC_DO_LOAD_n(num)                                              \
	case opc_iload_##num:                                           \
	case opc_aload_##num:                                           \
	case opc_fload_##num:                                           \
	    S_INFO(0) = vars[num];                                      \
	    TRACEIF(opc_iload_##num,                                    \
		    ("\t%s => %d\n", opnames[opcode], S_INT(0)));       \
	    TRACEIF(opc_aload_##num,                                    \
		    ("\t%s => %s\n", opnames[opcode], S_OSTRING(0)));   \
	    TRACEIF(opc_fload_##num,                                    \
		    ("\t%s => %g\n", opnames[opcode], S_FLOAT(0)));     \
	    SIZE_AND_STACK(1, 1);					\
        case opc_lload_##num:                                         	\
	case opc_dload_##num:	                                        \
	    S_INFO(0) = vars[num];                                      \
	    S_INFO(1) = vars[num + 1];                                  \
	    TRACEIF(opc_lload_##num,                                    \
		    ("\t%s => %s\n", opnames[opcode], S_LONGSTRING(0))); \
	    TRACEIF(opc_dload_##num,                                    \
		    ("\t%s => %g\n", opnames[opcode], S_DOUBLE(0)));    \
	    SIZE_AND_STACK(1, 2);


	OPC_DO_LOAD_n(0)
	OPC_DO_LOAD_n(1)
	OPC_DO_LOAD_n(2)
	OPC_DO_LOAD_n(3)

	      /* store a local variable onto the stack. */
	case opc_istore:
	case opc_astore:
	case opc_fstore:
	    vars[pc[1]] = S_INFO(-1);
	    TRACEIF(opc_istore,("\tistore %d => r%d\n", S_INT(-1), pc[1]));
	    TRACEIF(opc_astore,("\tastore %s => r%d\n", S_OSTRING(-1),pc[1]));
	    TRACEIF(opc_fstore,("\tfstore %g => r%d\n", S_FLOAT(-1), pc[1]));
	    SIZE_AND_STACK(2, -1);

	case opc_lstore:
	case opc_dstore: {
	    int reg = pc[1];
	    vars[reg] = S_INFO(-2);
	    vars[reg + 1] = S_INFO(-1);
	    TRACEIF(opc_lstore, ("\tlstore %s => r%d\n", S_LONGSTRING(-2),reg));
	    TRACEIF(opc_dstore, ("\tdstore %g => r%d\n", S_DOUBLE(-2), reg));
	    SIZE_AND_STACK(2, -2);
	}

#define OPC_DO_STORE_n(num)                                             \
	case opc_istore_##num:				        	\
	case opc_astore_##num: 						\
	case opc_fstore_##num: 						\
	    vars[num] = S_INFO(-1);  					\
	    TRACEIF(opc_istore_##num, ("\t%s %d => r%d\n", 		\
		       opnames[opcode], S_INT(-1), num));  		\
	    TRACEIF(opc_astore_##num, ("\t%s %d => r%d\n", 		\
			      opnames[opcode], S_OSTRING(-1), num)); 	\
	    TRACEIF(opc_fstore_##num, ("\t%s %g => r%d\n", 		\
			      opnames[opcode], S_FLOAT(-1), num)); 	\
	    SIZE_AND_STACK(1, -1);                                      \
	case opc_lstore_##num: 						\
	case opc_dstore_##num: 						\
	    vars[num] = S_INFO(-2); 					\
	    vars[num + 1] = S_INFO(-1); 				\
	    TRACEIF(opc_lstore_##num, ("\t%s %s => r%d\n", 		\
			     opnames[opcode], S_LONGSTRING(-2), num)); 	\
	    TRACEIF(opc_dstore_##num, ("\t%s %g => r%d\n", 		\
			     opnames[opcode], S_DOUBLE(-2), num)); 	\
	    SIZE_AND_STACK(1, -2);

	OPC_DO_STORE_n(0)
	OPC_DO_STORE_n(1)
	OPC_DO_STORE_n(2)
	OPC_DO_STORE_n(3)


	    /* Return a value from a function */
	case opc_areturn:
	case opc_ireturn:
	case opc_freturn:
	    frame->prev->optop[0] = S_INFO(-1);
	    frame->prev->optop++;
	    TRACEIF(opc_ireturn, ("\tireturn %d\n", S_INT(-1)));
	    TRACEIF(opc_areturn, ("\tareturn %s\n", S_OSTRING(-1)));
	    TRACEIF(opc_freturn, ("\tfreturn %g\n", S_FLOAT(-1)));
	    goto finish_return;
	      
	case opc_lreturn:
	case opc_dreturn:
	    frame->prev->optop[0] = S_INFO(-2);
	    frame->prev->optop[1] = S_INFO(-1);
	    frame->prev->optop += 2;
	    TRACEIF(opc_lreturn, ("\tlreturn %s\n", S_LONGSTRING(-2)));
	    TRACEIF(opc_dreturn, ("\tdreturn %g\n", S_DOUBLE(-2)));
	    goto finish_return;

	    /* Return from a function with no value */
	case opc_return:
	    TRACE(("\treturn\n"));
	    goto finish_return;

	case opc_i2f:		/* Convert top of stack int to float */
	    S_FLOAT(-1) = S_INT(-1);
	    TRACE(("\ti2f => %g\n", S_FLOAT(-1)));
	    SIZE_AND_STACK(1, 0);

	case opc_i2l:		/* Convert top of stack int to long */
	    SET_S_LONG(-1, int2ll(S_INT(-1)));
	    TRACE(("\ti2l => %s\n", S_LONGSTRING(-1)));
	    SIZE_AND_STACK(1, 1);

	case opc_i2d:
	    SET_S_DOUBLE(-1, S_INT(-1));
	    TRACE(("\ti2d => %g\n", S_DOUBLE(-1)));
	    SIZE_AND_STACK(1, 1);
	      
	case opc_f2i:		/* Convert top of stack float to int */
	    S_INT(-1) = S_FLOAT(-1);
	    TRACE(("\tf2i => %d\n", S_INT(-1)));
	    SIZE_AND_STACK(1, 0);

	case opc_f2l:		/* convert top of stack float to long */
	    SET_S_LONG(-1, float2ll(S_FLOAT(-1)));
	    TRACE(("\tf2l => %s\n", S_LONGSTRING(-1)));
	    SIZE_AND_STACK(1, 1);
	      
	case opc_f2d:		/* convert top of stack float to double */
	    SET_S_DOUBLE(-1, S_FLOAT(-1));
	    TRACE(("\tf2d => %g\n", S_DOUBLE(-1)));
	    SIZE_AND_STACK(1, 1);

	case opc_l2i:		/* convert top of stack long to int */
	    S_INT(-2) = ll2int(S_LONG(-2));
	    TRACE(("\tl2i => %d\n", S_INT(-2)));
	    SIZE_AND_STACK(1, -1);
	      
	case opc_l2f:		/* convert top of stack long to float */
	    S_FLOAT(-2) = ll2float(S_LONG(-2));
	    TRACE(("\tl2f => %g\n", S_FLOAT(-2)));
	    SIZE_AND_STACK(1, -1);

	case opc_l2d:		/* convert top of stack long to double */
	    SET_S_DOUBLE(-2, ll2double(S_LONG(-2)));
	    TRACE(("\tl2d => %g\n", S_DOUBLE(-2)));
	    SIZE_AND_STACK(1, 0);

	case opc_d2i:
	    S_INT(-2) = S_DOUBLE(-2);
	    TRACE(("\td2i => %d\n", S_INT(-2)));
	    SIZE_AND_STACK(1, -1);
	      
	case opc_d2f:
	    S_FLOAT(-2) = S_DOUBLE(-2);
	    TRACE(("\td2f => %g\n", S_FLOAT(-2)));
	    SIZE_AND_STACK(1, -1);

	case opc_d2l:
	    SET_S_LONG(-2, double2ll(S_DOUBLE(-2)));
	    TRACE(("\td2l => %s\n", S_LONGSTRING(-2)));
	    SIZE_AND_STACK(1, 0);

	case opc_int2byte:
	    S_INT(-1) = (signed char) S_INT(-1);
	    TRACE(("\tint2byte => %d\n", S_INT(-1)));
	    SIZE_AND_STACK(1, 0);
	      
	case opc_int2char:
	    S_INT(-1) &= 0xFFFF;
	    TRACE(("\tint2char => %d\n", S_INT(-1)));
	    SIZE_AND_STACK(1, 0);

	case opc_int2short:
	    S_INT(-1) = (signed short) S_INT(-1);
	    TRACE(("\tint2short => %d\n", S_INT(-1)));
	    SIZE_AND_STACK(1, 0);

        case opc_fcmpl:
        case opc_fcmpg: {
	    float l = S_FLOAT(-2);
	    float r = S_FLOAT(-1);
	    int value =   (l < r) ? -1 : (l > r) ? 1 : (l == r ) ?  0 
	                /* Beyond here, must be NaN. */
	                : (opcode == opc_fcmpl) ? -1 
			: (opcode == opc_fcmpg) ? 1 
			: 0;
	    S_INT(-2) = value;
	    TRACE(("\t%s => %d\n", opnames[opcode], S_INT(-2))); 
	    SIZE_AND_STACK(1, -1);
	}

	case opc_lcmp: {
	    int64_t l = S_LONG(-4);
	    int64_t r = S_LONG(-2);
	    S_INT(-4) = (ll_lt(l, r)) ? -1 : (ll_gt(l, r)) ? 1 : 0;
	    TRACE(("\tlcmp => %d\n", S_INT(-4))); 
	    SIZE_AND_STACK(1, -3);
	}

	case opc_dcmpl: 
	case opc_dcmpg: {
	    double l = S_DOUBLE(-4);
	    double r = S_DOUBLE(-2);
	    int value =   (l < r) ? -1 : (l > r) ? 1 : (l == r ) ?  0 
	               /* Beyond here, must be NaN. */
			: (opcode == opc_dcmpl) ? -1 
			: (opcode == opc_dcmpg) ? 1 
			: 0;
	    S_INT(-4) = value;
	    TRACE(("\t%s => %d\n", opnames[opcode], S_INT(-4))); 
	    SIZE_AND_STACK(1, -3);
	}
	      
	    /* Push a 1-byte signed integer value onto the stack. */
	case opc_bipush:	
	    S_INT(0) = (signed char)(pc[1]);
	    TRACE(("\tbipush %d\n", S_INT(0)));
	    SIZE_AND_STACK(2, 1);

	    /* Push a 2-byte signed integer constant onto the stack. */
	case opc_sipush:	
	    S_INT(0) = pc2signedshort(pc);
	    TRACE(("\tsipush %d\n", S_INT(0)));
	    SIZE_AND_STACK(3, 1);

	case opc_ldc1:
	    WITH_OPCODE_CHANGE
	        ResolveClassConstant(constant_pool, pc[1], ee, 
				     (1 << CONSTANT_Integer) | 
				     (1 << CONSTANT_Float) |
				     (1 << CONSTANT_String));
	        if (!exceptionOccurred(ee))
		    pc[0] = opc_ldc1_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_ldc1_quick:
	    S_POINTER(0) = constant_pool[pc[1]].p;
	    TRACE(("\tldc1 #%d => 0x%x\n", pc[1], S_INT(0)));
	    SIZE_AND_STACK(2, 1);

	case opc_ldc2:
	    WITH_OPCODE_CHANGE
	        ResolveClassConstant(constant_pool, GET_INDEX(pc + 1), ee, 
				     (1 << CONSTANT_Integer) | 
				     (1 << CONSTANT_Float) |
				     (1 << CONSTANT_String));
	        if (!exceptionOccurred(ee))
		    pc[0] = opc_ldc2_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_ldc2_quick:
	    S_POINTER(0) = constant_pool[GET_INDEX(pc + 1)].p;
	    TRACE(("\tldc2 #%d => 0x%x\n", GET_INDEX(pc + 1), S_INT(0)));
	    SIZE_AND_STACK(3, 1);

	case opc_ldc2w:	/* load a two word constant */
	    WITH_OPCODE_CHANGE
	        ResolveClassConstant(constant_pool, GET_INDEX(pc + 1), ee,
				 (1 << CONSTANT_Double) | (1 << CONSTANT_Long));
	        if (!exceptionOccurred(ee))
		    pc[0] = opc_ldc2w_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_ldc2w_quick:
	    SET_S_LONG(0, GET_INT64(&constant_pool[GET_INDEX(pc + 1)]));
	    TRACE(("\tldc2 #%d => long=%s double=%g\n", GET_INDEX(pc + 1), 
		   S_LONGSTRING(0), S_DOUBLE(0)));
	    SIZE_AND_STACK(3, 2);

#define CONSTANT_OP(opcode, this_type, value) 				\
	case opcode: 							\
	    S_INFO(0).this_type = value; 				\
	    TRACE(("\t%s\n", opnames[opcode])); 			\
	    SIZE_AND_STACK(1, 1);
	  
	    /* Push miscellaneous constants onto the stack. */
	CONSTANT_OP(opc_aconst_null, p,        0)
	CONSTANT_OP(opc_iconst_m1,   i,       -1)
	CONSTANT_OP(opc_iconst_0,    i,        0)
	CONSTANT_OP(opc_iconst_1,    i,        1)
	CONSTANT_OP(opc_iconst_2,    i,        2)
	CONSTANT_OP(opc_iconst_3,    i,        3)
	CONSTANT_OP(opc_iconst_4,    i,        4)
	CONSTANT_OP(opc_iconst_5,    i,        5)
	CONSTANT_OP(opc_fconst_0,    f,      0.0)
	CONSTANT_OP(opc_fconst_1,    f,      1.0)
	CONSTANT_OP(opc_fconst_2,    f,      2.0)

	case opc_dconst_0:
	    SET_S_DOUBLE(0, 0.0);
	    TRACE(("\t%s\n", opnames[opcode]));
	    SIZE_AND_STACK(1, 2);

	case opc_dconst_1:
	    SET_S_DOUBLE(0, 1.0);
	    TRACE(("\t%s\n", opnames[opcode]));
	    SIZE_AND_STACK(1, 2);

	case opc_lconst_0:
	    SET_S_LONG(0, ll_zero_const);
	    TRACE(("\t%s\n", opnames[opcode]));
	    SIZE_AND_STACK(1, 2);

	case opc_lconst_1:
   	    SET_S_LONG(0, ll_one_const);
	    TRACE(("\t%s\n", opnames[opcode]));
	    SIZE_AND_STACK(1, 2);

#define BINARY_INT_OP(name, op, test)                               \
	case opc_i##name:                                          	\
	    if (test && S_INT(-1) == 0)                              	\
               ERROR("ArithmeticException", "/ by zero");               \
	    S_INT(-2) op##= S_INT(-1);                                	\
	    TRACE(("\t%s => %d\n", opnames[opcode], S_INT(-2)));      	\
	    SIZE_AND_STACK(1, -1);                                    	\
	case opc_l##name:                                             	\
	    if (test && ll_eqz(S_LONG(-2)))                             \
               ERROR("ArithmeticException", "/ by zero");               \
	    SET_S_LONG(-4, ll_##name(S_LONG(-4), S_LONG(-2)));          \
	    TRACE(("\t%s => %s\n", opnames[opcode], S_LONGSTRING(-4))); \
	    SIZE_AND_STACK(1, -2);

	      /* Perform various binary integer operations */
	BINARY_INT_OP(add, +, 0) 
	BINARY_INT_OP(sub, -, 0)
	BINARY_INT_OP(mul, *, 0)
	BINARY_INT_OP(and, &, 0)
	BINARY_INT_OP(or,  |, 0)
	BINARY_INT_OP(xor, ^, 0)
	BINARY_INT_OP(div, /, 1)
	BINARY_INT_OP(mod, %, 1)

#define BINARY_SHIFT_OP(name, op)                               	\
	case opc_i##name:                                          	\
	    S_INT(-2) = ((int)(S_INT(-2))) op (S_INT(-1) & 0x1F);       \
	    TRACE(("\t%s => %d\n", opnames[opcode], S_INT(-2)));      	\
	    SIZE_AND_STACK(1, -1);                                    	\
	case opc_l##name:                                             	\
	    SET_S_LONG(-3, ll_##name(S_LONG(-3), (S_INT(-1) & 0x3F)));	\
	    TRACE(("\t%s => %s\n", opnames[opcode], S_LONGSTRING(-3))); \
	    SIZE_AND_STACK(1, -1);                                      \

	BINARY_SHIFT_OP(shl, <<)
	BINARY_SHIFT_OP(shr, >>)

	case opc_iushr:
	    S_INT(-2) = ((unsigned int)(S_INT(-2))) >> (S_INT(-1) & 0x1F);
	    TRACE(("\t%s => %d\n", opnames[opcode], S_INT(-2)));
	    SIZE_AND_STACK(1, -1);
	case opc_lushr:
	    SET_S_LONG(-3, ll_ushr(S_LONG(-3), (S_INT(-1) & 0x3F)));
	    TRACE(("\t%s => %s\n", opnames[opcode], S_LONGSTRING(-3)));
	    SIZE_AND_STACK(1, -1); 

	      /* Unary negation */
	case opc_ineg:
	    S_INT(-1) = -S_INT(-1);
	    TRACE(("\tineg => %d\n", S_INT(-1)));
	    SIZE_AND_STACK(1, 0);

	case opc_lneg:
            SET_S_LONG(-2, ll_neg(S_LONG(-2)));
	    TRACE(("\tlneg => %d\n", S_LONG(-2)));
	    SIZE_AND_STACK(1, 0);
	    
	    /* Add a small constant to a local variable */
	case opc_iinc:
	    vars[pc[1]].i += ((signed char *) pc)[2];
	    TRACE(("\tiinc r%d+%d => %d\n", 
		   pc[1], ((signed char *) pc)[2], vars[pc[1]].i));
	    SIZE_AND_STACK(3, 0);

#define BINARY_FLOAT_OP(name, op) 					\
	case opc_f##name: 						\
	    S_FLOAT(-2) op##= S_FLOAT(-1); 				\
	    TRACE(("\t%s => %g\n", opnames[opcode], S_FLOAT(-2))); 	\
	    SIZE_AND_STACK(1, -1); 					\
	case opc_d##name: 						\
	    SET_S_DOUBLE(-4, S_DOUBLE(-4) op S_DOUBLE(-2));   	        \
	    TRACE(("\t%s => %g\n", opnames[opcode], S_DOUBLE(-4)));     \
	    SIZE_AND_STACK(1, -2);

	BINARY_FLOAT_OP(add, +)
	BINARY_FLOAT_OP(sub, -)
	BINARY_FLOAT_OP(mul, *)
	BINARY_FLOAT_OP(div, /)

	case opc_fmod: {
	    float left = S_FLOAT(-2);
	    float right = S_FLOAT(-1);
	    S_FLOAT(-2) = left - ((int)(left/right)) * right;
	    TRACE(("\t%s => %g\n", opnames[opcode], S_FLOAT(-2))); 
	    SIZE_AND_STACK(1, -1);
	}
	      
	case opc_dmod: {
	    double left = S_DOUBLE(-4);
	    double right = S_DOUBLE(-2);
	    SET_S_DOUBLE(-4, left - ((int)(left/right)) * right);
	    TRACE(("\t%s => %g\n", opnames[opcode], S_DOUBLE(-4)));
	    SIZE_AND_STACK(1, -2);
	}

	case opc_fneg:
	    S_FLOAT(-1) = -S_FLOAT(-1);
	    TRACE(("\tfneg => %g\n", S_FLOAT(-1)));
	    SIZE_AND_STACK(1, 0);

	case opc_dneg:
	    SET_S_DOUBLE(-2, -S_DOUBLE(-2));
	    TRACE(("\tdneg => %g\n", S_DOUBLE(-2)));
	    SIZE_AND_STACK(1, 0);

	    /* Unconditional goto.  Maybe save return address in register. */
	case opc_jsr:
	    S_INFO(0).addr = pc + 3;
	    optop++;
	    /* Fall through to opc_goto */

	case opc_goto:
	    TRACE(("\t%s %8X\n", opnames[opcode],
		   pc + (short) ((pc[1] << 8) | pc[2])));
	    SIZE_AND_STACK(pc2signedshort(pc), 0);

	    /* Return from subroutine.  June to address in register. */
	case opc_ret:
	    TRACE(("\tret %d (%8X)\n", pc[1], vars[pc[1]].addr));
	    pc = vars[pc[1]].addr;
	    SIZE_AND_STACK(0, 0);
	      
#define COMPARISON_OP(name, comparison)                                   \
	case opc_if_icmp##name: {					  \
            int skip = (S_INT(-2) comparison S_INT(-1))                   \
	           ? pc2signedshort(pc) : 3;                              \
            TRACE(("\t%s goto %8X (%staken)\n",                           \
		   opnames[opcode],                                       \
		   pc + pc2signedshort(pc),                               \
		   (S_INT(-2) comparison S_INT(-1)) ? "" : "not "));	  \
            SIZE_AND_STACK(skip, -2); 					  \
        }								  \
	case opc_if##name: {					     	  \
            int skip = (S_INT(-1) comparison 0) ? pc2signedshort(pc) : 3; \
            TRACE(("\t%s goto %8X (%staken)\n",                           \
		   opnames[opcode],                                       \
		   pc + pc2signedshort(pc),                               \
		   (S_INT(-1) comparison 0) ? "" : "not "));	          \
            SIZE_AND_STACK(skip, -1); 					  \
        }                                                                 \
            
#define COMPARISON_OP2(name, comparison)                                  \
        COMPARISON_OP(name, comparison)                                   \
	case opc_if_acmp##name: {					  \
            int skip = (S_POINTER(-2) comparison S_POINTER(-1))           \
	           ? pc2signedshort(pc) : 3;                              \
            TRACE(("\t%s goto %8X (%staken)\n",                           \
		   opnames[opcode],                                       \
		   pc + pc2signedshort(pc),                               \
		   (S_POINTER(-2) comparison S_POINTER(-1)) ? "" : "not ")); \
            SIZE_AND_STACK(skip, -2); 					  \
        }								  \


	COMPARISON_OP2(eq, ==)	/* also generate acmp_eq and acomp_ne */
	COMPARISON_OP2(ne, !=)
	COMPARISON_OP(lt, <)
	COMPARISON_OP(gt, >)
	COMPARISON_OP(le, <=)
	COMPARISON_OP(ge, >=)


	    /* Discard the top item on the stack */
	case opc_pop:
	    TRACE(("\tpop\n"));
	    SIZE_AND_STACK(1, -1);

	    /* Discard the top item on the stack */
	case opc_pop2:
	    TRACE(("\tpop2\n"));
	    SIZE_AND_STACK(1, -2);

	    /* Duplicate the top item on the stack */
	case opc_dup:
	    S_INFO(0) = S_INFO(-1);
	    TRACE(("\tdup\n"));
	    SIZE_AND_STACK(1, 1);

	case opc_dup2:		/* Duplicate the top two items on the stack */
	    S_INFO(0) = S_INFO(-2);
	    S_INFO(1) = S_INFO(-1);
	    TRACE(("\tdup2\n"));
	    SIZE_AND_STACK(1, 2);

	case opc_dup_x1:	/* insert S_INFO(-1) on stack past 1 */
	    S_INFO(0) = S_INFO(-1);
	    S_INFO(-1) = S_INFO(-2);
	    S_INFO(-2) = S_INFO(0);
	    TRACE(("\tdup_x1\n"));
	    SIZE_AND_STACK(1, 1);
	      
	case opc_dup_x2:	/* insert S_INFO(-1) on stack past 2*/
	    S_INFO(0) = S_INFO(-1);
	    S_INFO(-1) = S_INFO(-2);
	    S_INFO(-2) = S_INFO(-3);
	    S_INFO(-3) = S_INFO(0);
	    TRACE(("\tdup_x2\n"));
	    SIZE_AND_STACK(1, 1);

	case opc_dup2_x1:	/* insert S_INFO(-1,-2) on stack past 1 */
	    S_INFO(1) = S_INFO(-1);
	    S_INFO(0) = S_INFO(-2);
	    S_INFO(-1) = S_INFO(-3);
	    S_INFO(-2) = S_INFO(1);
	    S_INFO(-3) = S_INFO(0);
	    TRACE(("\tdup2_x1\n"));
	    SIZE_AND_STACK(1, 2);
	      
	case opc_dup2_x2:	/* insert S_INFO(-1,2) on stack past 2 */
	    S_INFO(1) = S_INFO(-1);
	    S_INFO(0) = S_INFO(-2);
	    S_INFO(-1) = S_INFO(-3);
	    S_INFO(-2) = S_INFO(-4);
	    S_INFO(-3) = S_INFO(1);
	    S_INFO(-4) = S_INFO(0);
	    TRACE(("\tdup2_x2\n"));
	    SIZE_AND_STACK(1, 2);
	      
	case opc_swap: {        /* swap top two elements on the stack */
	    stack_item j = S_INFO(-1);
	    S_INFO(-1) = S_INFO(-2);
	    S_INFO(-2) = j;
	    SIZE_AND_STACK(1, 0);
	}

	    /* Find the length of an array */
	case opc_arraylength: {
	    HArrayOfChar *arr = (HArrayOfChar *)S_HANDLE(-1);
	    if (arr == 0) 
		ERROR("NullPointerException", 0);
	    S_INT(-1) = obj_length(arr);
	    TRACE(("\tarraylength %s => %d\n", 
		   Object2CString((Handle*)arr), S_INT(-1)));
	    SIZE_AND_STACK(1, 0);
	}

	    /* Goto one of a sequence of targets. */
	case opc_tableswitch: {
	    long *lpc = (long *) UCALIGN(pc + 1);
	    int key = S_INT(-1);
	    int low = ntohl(lpc[1]);
	    int high = ntohl(lpc[2]);
	    TRACE(("\ttableswitch %d [%d-%d]\n", key, low, high));
	    key -= low;
	    if ((unsigned) key > high - low)
		pc += ntohl(lpc[0]);
	    else 
		pc += ntohl(lpc[key + 3]);
	    SIZE_AND_STACK(0, -1);
	}
	      
	    /* Goto one of a sequence of targets */
	case opc_lookupswitch: {
	    long *lpc = (long *) UCALIGN(pc + 1);
	    int key = S_INT(-1);
	    int off = ntohl(lpc[0]);
	    int npairs = ntohl(lpc[1]);
	    TRACE(("\tlookupswitch %d\n", key));
	    while (--npairs >= 0) {
		lpc += 2;
		if (key == ntohl(lpc[0])) {
		    pc += ntohl(lpc[1]);
		    goto done;
		}
	    }
	    pc += off;
	  done:
	    SIZE_AND_STACK(0, -1);
	}


	    /* Throw to a given value */
	case opc_athrow: {
	    Handle *o = S_HANDLE(-1);
	    TRACE(("\tathrow => %s\n", Object2CString(o)));
	    if (o == 0) {
	        ERROR("NullPointerException", 0);
	    } else {
		exceptionThrow(ee, o);
		goto handle_exception;
	    }

	}
	    
	case opc_getfield:
	case opc_putfield: 
	    DECACHE_OPTOP();
	    WITH_OPCODE_CHANGE {
		int field_index = GET_INDEX(pc+1);
		int offset;
		struct fieldblock *fb;
		char *signature;

		ResolveClassConstant(constant_pool, field_index, ee, 
				     1 << CONSTANT_Fieldref);
		fb = constant_pool[field_index].p;

		if (!exceptionOccurred(ee)) {
		    signature = fieldsig(fb);
		    TRACE(("\t%s %s.%s\n", opnames[opcode], 
			   classname(fieldclass(fb)), fieldname(fb)));
		    if (fb->access & ACC_STATIC) {
			char buf[256];
			classname2string(classname(fieldclass(fb)), 
					 buf, sizeof(buf));
			    sprintf(buf + strlen(buf), 
				": field %s used to be dynamic",
				fb->name);
			ERROR("IncompatibleClassChangeException", buf);
		    } else {
			int delta = (signature[0] == SIGNATURE_LONG 
				       || signature[0] == SIGNATURE_DOUBLE)
			           ? (opc_getfield2_quick - opc_getfield)
			           : (opc_getfield_quick - opc_getfield);
			offset = fb->u.offset / sizeof (OBJECT);
			assert(offset <= 0xFF);
			CODE_LOCK();
			  pc[0] = opcode + delta;
			  pc[1] = offset;
			CODE_UNLOCK();
		    }
		}
	    } END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_getstatic:
	case opc_putstatic:
	    DECACHE_OPTOP();
	    WITH_OPCODE_CHANGE {
		int field_index = GET_INDEX(pc+1);
		struct fieldblock *fb;

		ResolveClassConstant(constant_pool, field_index, ee, 
				     1 << CONSTANT_Fieldref);
		fb = constant_pool[field_index].p;
		if (!exceptionOccurred(ee)) {
		    TRACE(("\t%s %s.%s\n", opnames[opcode], 
			   classname(fieldclass(fb)), fieldname(fb)));
		    if ((fb->access & ACC_STATIC) == 0) {
			char buf[256];
			classname2string(classname(fieldclass(fb)), 
					 buf, sizeof(buf));
			sprintf(buf + strlen(buf), 
				": field %s used to be static",
				fb->name);
			ERROR("IncompatibleClassChangeException", buf);
		    } else {
			char *signature = fieldsig(fb);
			int delta = (signature[0] == SIGNATURE_LONG 
				       || signature[0] == SIGNATURE_DOUBLE)
			           ? (opc_getstatic2_quick - opc_getstatic)
			           : (opc_getstatic_quick - opc_getstatic);
			pc[0] = opcode + delta;
		    }
		}
	    } END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_invokevirtual:
	case opc_invokenonvirtual:
	case opc_invokestatic:
	    WITH_OPCODE_CHANGE {
		int method_index = GET_INDEX(pc+1);
		struct methodblock *mb;
		
		ResolveClassConstant(constant_pool, method_index, ee, 
				     1 << CONSTANT_Methodref);

		if (!exceptionOccurred(ee)) {
		    mb = constant_pool[method_index].p;
		    TRACE(("\t%s %s.%s%s\n", opnames[opcode], 
			   classname(fieldclass(&mb->fb)), 
			   fieldname(&mb->fb), fieldsig(&mb->fb)));
		    assert(mb->fb.u.offset <= 0xFF);
		    if (opcode == opc_invokestatic) {
			if ((mb->fb.access & ACC_STATIC) == 0) {
			    char buf[256];
			    classname2string(classname(fieldclass(&mb->fb)), 
					     buf, sizeof(buf));
				sprintf(buf + strlen(buf), 
					": method %s%s used to be static",
					mb->fb.name, mb->fb.signature);
			    ERROR("IncompatibleClassChangeException", buf);
			} 
		    } else {
			if (mb->fb.access & ACC_STATIC) {
			    char buf[256];
			    classname2string(classname(fieldclass(&mb->fb)), 
					     buf, sizeof(buf));
				sprintf(buf + strlen(buf), 
					": method %s%s used to be dynamic",
					mb->fb.name, mb->fb.signature);
			    ERROR("IncompatibleClassChangeException", buf);
			} 
		    }
		    if (opcode == opc_invokevirtual) {
			CODE_LOCK();
			  pc[0] = fieldclass(&mb->fb) == classobject 
			      ? opc_invokevirtualobject_quick
			      : opc_invokevirtual_quick;
			  pc[1] = mb->fb.u.offset; 
			  pc[2] = mb->args_size; 
			CODE_UNLOCK();
		    } else if (opcode == opc_invokenonvirtual) {
			pc[0] = opc_invokenonvirtual_quick;
		    } else {
			pc[0] = opc_invokestatic_quick;
		    }
		}
	    } END_WITH_OPCODE_CHANGE
	    goto check_for_exception;
	  
	case opc_invokeinterface:
	    WITH_OPCODE_CHANGE {
		unsigned InterfaceMethodref_index = GET_INDEX(pc + 1);
		ResolveClassConstant(constant_pool, InterfaceMethodref_index,
				     ee, 1 << CONSTANT_InterfaceMethodref);
		if (!exceptionOccurred(ee)) {
		    unsigned InterfaceMethodref = 
		         constant_pool[InterfaceMethodref_index].i;
		    unsigned NameAndType_index = InterfaceMethodref & 0xFFFF;
		    unsigned ID = constant_pool[NameAndType_index].i;
		    TRACE(("\t%s %s%s\n", opnames[opcode], 
			   ID2Str(interfaceHash, ID >> 16, 0),
			   ID2Str(interfaceHash, ID & 0xFFFF, 0)));
		    CODE_LOCK();
		      pc[1] = NameAndType_index >> 8;
		      pc[2] = NameAndType_index & 0xFF;
		      pc[4] = 0;	/* no guess, yet */
		      pc[0] = opc_invokeinterface_quick;
		    CODE_UNLOCK();
		}
	    } END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_putfield_quick: {
	    Handle *o = S_HANDLE(-2);
	    if (o == 0) 
	        ERROR("NullPointerException", 0);
	    TRACE(("\t%s %s.(#%d) <= %X\n", opnames[opcode], 
		   classname(obj_classblock(o)), pc[1], S_POINTER(-1)));
	    obj_setslot(o, pc[1], (long)S_POINTER(-1));
	    SIZE_AND_STACK(3, -2);
	}

	case opc_getfield_quick: {
	    Handle *o = S_HANDLE(-1);
	    if (o == 0)
	        ERROR("NullPointerException", 0);
	    S_POINTER(-1) = (void *)obj_getslot(o, pc[1]);
	    TRACE(("\t%s %s.(#%d) <= %X\n", opnames[opcode], 
		   classname(obj_classblock(o)), pc[1], S_POINTER(-1)));
	    SIZE_AND_STACK(3, 0);
	}
	    
	case opc_putfield2_quick: {
	    Handle *o = S_HANDLE(-3);
	    if (o == 0) 
	        ERROR("NullPointerException", 0);
	    TRACE(("\t%s %s.(#%d) <= long=%s double=%g\n", opnames[opcode], 
		   classname(obj_classblock(o)), pc[1], 
		   S_LONGSTRING(-2), S_DOUBLE(-2)));
	    obj_setslot(o, pc[1], (long)S_POINTER(-2));
	    obj_setslot(o, pc[1] + 1, (long)S_POINTER(-1));
	    SIZE_AND_STACK(3, -3);
	}

	case opc_getfield2_quick: {
	    Handle *o = S_HANDLE(-1);
	    if (o == 0) 
		ERROR("NullPointerException", 0);
	    S_POINTER(-1) = (void *)obj_getslot(o, pc[1]);
	    S_POINTER(0) = (void *)obj_getslot(o, pc[1] + 1);
	    TRACE(("\t%s %s.(#%d) <= long=%s double=%g\n", opnames[opcode], 
		   classname(obj_classblock(o)), pc[1], 
		   S_LONGSTRING(-1), S_DOUBLE(-1)));
	    SIZE_AND_STACK(3, 1);
	}


	case opc_putstatic_quick: {   
	    struct fieldblock *fb = constant_pool[GET_INDEX(pc+1)].p;
	    OBJECT *d = (OBJECT *)normal_static_address(fb);
	    *d = S_OBJECT(-1);
	    if (trace) {
		char isig = fieldsig(fb)[0];
		if (isig == SIGNATURE_FLOAT) {
		    TRACE(("\t%s %s.%s <= %g\n", opnames[opcode], 
			   classname(fieldclass(fb)), fieldname(fb), 
			   S_FLOAT(-1)));
		} else if (isig == SIGNATURE_CLASS || isig == SIGNATURE_ARRAY) {
		    TRACE(("\t%s %s.%s <= %s\n", opnames[opcode], 
			   classname(fieldclass(fb)), fieldname(fb), 
			   S_OSTRING(-1)));
		} else {
		    TRACE(("\t%s %s.%s <= %d\n", opnames[opcode], 
			   classname(fieldclass(fb)), fieldname(fb), 
			   S_INT(-1)));
		}
	    }
	    SIZE_AND_STACK(3, -1);
	}

	case opc_getstatic_quick: {
	    struct fieldblock *fb = constant_pool[GET_INDEX(pc+1)].p;
	    OBJECT *d = (OBJECT *) normal_static_address(fb);
	    S_OBJECT(0) = *d;
	    if (trace) {
		char isig = fieldsig(fb)[0];
		if (isig == SIGNATURE_FLOAT) {
		    TRACE(("\t%s %s.%s <= %g\n", opnames[opcode], 
			   classname(fieldclass(fb)), fieldname(fb), 
			   S_FLOAT(0)));
		} else if (isig == SIGNATURE_CLASS || isig == SIGNATURE_ARRAY) {
		    TRACE(("\t%s %s.%s <= %s\n", opnames[opcode], 
			   classname(fieldclass(fb)), fieldname(fb), 
			   S_OSTRING(0)));
		} else {
		    TRACE(("\t%s %s.%s <= %d\n", opnames[opcode], 
			   classname(fieldclass(fb)), fieldname(fb), 
			   S_INT(0)));
		} 
	    }
	    SIZE_AND_STACK(3, 1);
	}

	case opc_putstatic2_quick: {   
	    struct fieldblock *fb = constant_pool[GET_INDEX(pc+1)].p;
	    stack_item *location = (stack_item *)twoword_static_address(fb);
	    location[0] = S_INFO(-2);
	    location[1] = S_INFO(-1);
	    if (trace) {
		char *signature = fieldsig(fb);
		if (signature[0] == SIGNATURE_LONG) {
		    TRACE(("\t%s %s.%s <= %s\n", opnames[opcode], 
			   classname(fieldclass(fb)), fieldname(fb), 
			   S_LONGSTRING(-2)));
		} else {
		    TRACE(("\t%s %s.%s <= %g\n", opnames[opcode], 
			   classname(fieldclass(fb)), fieldname(fb), 
			   S_DOUBLE(-2)));
		}
	    }
	    SIZE_AND_STACK(3, -2);
	}

	case opc_getstatic2_quick: {
	    struct fieldblock *fb = constant_pool[GET_INDEX(pc+1)].p;
	    stack_item *location = (stack_item *)twoword_static_address(fb);
	    S_INFO(0) = location[0];
	    S_INFO(1) = location[1];
	    if (trace) {
		char *signature = fieldsig(fb);
		if (signature[0] == SIGNATURE_LONG) {
		    TRACE(("\t%s %s.%s <= %s\n", opnames[opcode], 
			   classname(fieldclass(fb)), fieldname(fb), 
			   S_LONGSTRING(0)));
		} else {
		    TRACE(("\t%s %s.%s <= %g\n", opnames[opcode], 
			   classname(fieldclass(fb)), fieldname(fb), 
			   S_DOUBLE(0)));
		}
	    }
	    SIZE_AND_STACK(3, 2);
	}

	case opc_invokenonvirtual_quick:
	    mb = constant_pool[GET_INDEX(pc + 1)].p;
	    args_size = mb->args_size;
	    optop -= args_size;
	    frame->returnpc = pc + 3;
	    o = S_HANDLE(0);
	    if (o == 0) 
	        ERROR("NullPointerException", 0);
	    goto callmethod;

	case opc_invokestatic_quick:
	    mb = constant_pool[GET_INDEX(pc + 1)].p;
	    args_size = mb->args_size;
	    optop -= args_size;
	    frame->returnpc = pc + 3;
	    /* o should be the handle to the method's class */
	    o = (Handle *)(cbHandle(fieldclass(&mb->fb)));
	    goto callmethod;

	case opc_invokeinterface_quick: {
	    unsigned NameAndType_index = GET_INDEX(pc + 1);
	    unsigned ID = constant_pool[NameAndType_index].i;
	    int mslot = pc[4];	/* guess the slot of the method name */
	    int nslots;
	    struct methodblock **mbt;
	    ClassClass *cb;
	    
	    args_size = pc[3];
	    optop -= args_size;
	    frame->returnpc = pc + 5;
	    o = S_HANDLE(0);
	    if (o == 0)
	        ERROR("NullPointerException", 0);
	    mb = mt_slot(obj_array_methodtable(o), pc[1]);
	    cb = o->methods->classdescriptor;
	    nslots = cbMethodTableSize(cb);
	    mbt = o->methods->methods;
	    if (mslot >= nslots || (mb = mbt[mslot]) == 0 || mb->fb.ID != ID) {
		/* The guess didn't pay off, search */
		int n = cbMethodTableSize(cb);
		int slot = 0;
		    
		for (; --n > 0;) {
		    mb = mbt[++slot];
		    if (mb->fb.ID == ID) {
			goto HaveIt;
		    }
		}
		{
		    char buf[256];
		    classname2string(classname(cb), buf, sizeof(buf));
		    sprintf(buf + strlen(buf), 
			    ": dynamic method %s%s not found",
			    ID2Str(interfaceHash, ID >> 16, 0),
			    ID2Str(interfaceHash, ID & 0xFFFF, 0));
		    ERROR("IncompatibleClassChangeException", buf);
		}
	    HaveIt:
		if (mb->fb.access & ACC_STATIC) {
		    char buf[256];
		    classname2string(classname(cb), buf, sizeof(buf));
		    sprintf(buf + strlen(buf), 
			    ": method %s%s used to be dynamic",
			    mb->fb.name, mb->fb.signature);
		    ERROR("IncompatibleClassChangeException", buf);
		}
	        if (exceptionOccurred(ee)) {
		    frame->lastpc = pc;  
		    goto handle_exception;
		}
		/* After searching, set the guess for the next time around */
		pc[4] = slot;
	    }
	    goto callmethod;
	}

	case opc_invokevirtualobject_quick:
	    frame->returnpc = pc + 3;
	    args_size = pc[2];
	    optop -= args_size;
	    o = S_HANDLE(0);
	    if (o == 0) 
	        ERROR("NullPointerException", 0);
	    mb = mt_slot(obj_array_methodtable(o), pc[1]);
	    goto callmethod;

	case opc_invokevirtual_quick:
	    frame->returnpc = pc + 3;
	    args_size = pc[2];
	    optop -= args_size;
	    o = S_HANDLE(0);
	    if (o == 0) 
	        ERROR("NullPointerException", 0);
	    mb = mt_slot(obj_methodtable(o), pc[1]);
		
callmethod:
	    assert(o != 0);
	    TRACE(("\t%s %s.%s%s (%d)\n", opnames[opcode], 
		   classname(fieldclass(&mb->fb)), fieldname(&mb->fb), 
		   fieldsig(&mb->fb), args_size));

#ifndef DBINFO		/* don't do it twice if DBINFO is defined */
	    frame->lastpc = pc;
#endif

	    DECACHE_OPTOP();
	    if (!mb->invoker(o, mb, args_size, ee)) {
		/* FALSE indicates an exception might have occurred */
		goto check_for_exception;
	    } else {
		frame = ee->current_frame;
		vars = frame->vars;
		pc = frame->returnpc;
		constant_pool = frame->constant_pool;
		CACHE_OPTOP();
		SIZE_AND_STACK(0, 0);
	    }

	case opc_verifystack:
#ifdef DEBUG
	    if (REAL_OPTOP() != frame->ostack) {
		char        where[100];
		pc2string(pc, frame->current_method, 
			  where, where + sizeof where);
		fprintf(stderr, 
			"Stack has height %d at %s.  Should be 0\n", 
			REAL_OPTOP() - frame->ostack, where);
	    }
#endif
	    frame->optop = frame->ostack;
	    CACHE_OPTOP();
	    TRACE(("\tverifystack\n"));
	    SIZE_AND_STACK(1, 0);


	case opc_instanceof:
	    WITH_OPCODE_CHANGE
	        ResolveClassConstant(constant_pool, GET_INDEX(pc + 1), ee, 
				     1 << CONSTANT_Class);
	        if (!exceptionOccurred(ee))
		    pc[0] = opc_instanceof_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_instanceof_quick: {
	    ClassClass *cb = constant_pool[GET_INDEX(pc + 1)].p;
	    S_INT(-1) = is_instance_of(S_HANDLE(-1), cb, ee);
	    TRACE(("\tinstanceof %s => %d\n", classname(cb), S_INT(-1)));
	    SIZE_AND_STACK_BUT(3, 0); /* may error on implementations */
	    goto check_for_exception;
	}
	
	case opc_checkcast:
	    WITH_OPCODE_CHANGE
	        ResolveClassConstant(constant_pool, GET_INDEX(pc + 1), ee, 
				     1 << CONSTANT_Class);
	        if (!exceptionOccurred(ee))
	            pc[0] = opc_checkcast_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_checkcast_quick: {
	    Handle *h = S_HANDLE(-1);
	    ClassClass *cb = constant_pool[GET_INDEX(pc + 1)].p;
	    if (!is_instance_of(h, cb, ee)) {
	        /* Checkcast fails  */
		char buf[128];
		ERROR("ClassCastException", 
		      classname2string(classname(obj_array_classblock(h)), 
				       buf, sizeof(buf)));
	    }
	    TRACE(("\tcheckcast %s\n", classname(cb)));
	    SIZE_AND_STACK(3, 0);
	}

	case opc_new:
	    WITH_OPCODE_CHANGE
	        if (ResolveClassConstant(constant_pool, GET_INDEX(pc + 1), ee, 
 					 1 << CONSTANT_Class)) {
		    ClassClass *current_class = frame->current_method ? 
			        fieldclass(&frame->current_method->fb) : NULL;
		    ClassClass *cb = constant_pool[GET_INDEX(pc + 1)].p;
		    if (cbAccess(cb) & (ACC_INTERFACE | ACC_ABSTRACT))
			ERROR("IncompatibleTypeException", cb->name);
		    if (!VerifyClassAccess(current_class, cb, FALSE))
			ERROR("IllegalAccessException", cb->name);
		}
	        if (!exceptionOccurred(ee))
	            pc[0] = opc_new_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_new_quick: {
	    ClassClass *cb = constant_pool[GET_INDEX(pc + 1)].p;
	    S_HANDLE(0) = newobject(cb, pc, ee);
	    TRACE(("\tnew %s => %s\n", classname(cb), S_OSTRING(0)));
	    SIZE_AND_STACK_BUT(3, 1);
	    goto check_for_exception;
	}

	case opc_anewarray:
	    WITH_OPCODE_CHANGE
	        ResolveClassConstant(constant_pool, GET_INDEX(pc + 1), ee, 
				    1 << CONSTANT_Class);
	        if (!exceptionOccurred(ee))
		    pc[0] = opc_anewarray_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_anewarray_quick: {
	    int size = S_INT(-1);
	    HArrayOfObject *ret;
	    if (size < 0) 
		ERROR("NegativeArraySizeException", 0);
	    ret = (HArrayOfObject *)ArrayAlloc(T_CLASS, size);
	    if (ret == 0) 
	        ERROR("OutOfMemoryException", 0);
	    else {
		ClassClass *array_cb = constant_pool[GET_INDEX(pc + 1)].p;
		memset((char *) unhand(ret), 0, sizearray(T_CLASS, size));
		unhand(ret)->body[size] = (HObject*)array_cb;
		TRACE(("\tanewarray_quick %s[%d] => %s\n", 
		       array_cb->name, size, Object2CString((Handle*)ret)));
		S_HANDLE(-1) = (Handle *)ret;
		SIZE_AND_STACK(3, 0);
	    }
	}

	case opc_multianewarray:
	    WITH_OPCODE_CHANGE
	       ResolveClassConstant(constant_pool, GET_INDEX(pc + 1), ee, 
				    1 << CONSTANT_Class);
	       if (!exceptionOccurred(ee))
		   pc[0] = opc_multianewarray_quick;
	    END_WITH_OPCODE_CHANGE
	    goto check_for_exception;

	case opc_multianewarray_quick: {
	    ClassClass *array_cb = constant_pool[GET_INDEX(pc + 1)].p;
	    int dimensions = pc[3];
	    HObject *ret;
	    int i;

	    ResolveClassConstantFromClass(array_cb, 
					  CONSTANT_POOL_ARRAY_CLASS_INDEX,
					  ee, 1 << CONSTANT_Class);

	    optop -= dimensions;
	    for (i = 1; i <= dimensions; i++) 
		if (S_INT(i) < 0)
		    ERROR("NegativeArraySizeException", 0);
	    ret = MultiArrayAlloc(dimensions, array_cb, &S_INFO(0));
	    if (ret == 0)
		ERROR("OutOfMemoryException", 0);
	    S_HANDLE(0) = ret;
	    SIZE_AND_STACK(4, 1);
	}

	case opc_newarray: {
	    int size = S_INT(-1);
	    int type = pc[1];
	    int array_size = sizearray(pc[1], size) + 
		               (type == T_ARRAY ? sizeof(OBJECT) : 0);
	    Handle *ret;
	    if (size < 0) 
	        ERROR("NegativeArraySizeException", 0);
	    ret = ArrayAlloc(type, size);
	    if (ret == 0) 
	        ERROR("OutOfMemoryException", 0);
	    memset((char *) unhand(ret), 0, array_size);
	    TRACE(("\tnewarray %d %d => %s\n", 
		   size, sizearray(type, size), Object2CString(ret)));
	    S_HANDLE(-1) = ret;
	    SIZE_AND_STACK(2, 0);
	}

	case opc_newfromname: {
	    ClassClass *current_class = frame->current_method ? 
		        fieldclass(&frame->current_method->fb) : NULL;
	    char tmp[200];
	    HString *ho = (HString *)(S_HANDLE(-1));
	    ClassClass *cb;
	    HObject *ret;
	    HString *o;
	    int len;
	    char *ch;
	    DECACHE_OPTOP(); /* Just in case */
	    if (ho == 0) 
	        ERROR("NullPointerException", 0);
	    javaString2CString(ho, tmp, sizeof (tmp));
	    for (ch = tmp; *ch; ch++)
		if (*ch == '.')
		    *ch = '/';
	    if ((cb = FindClass(ee, tmp, TRUE)) == 0) 
	        ERROR("NoClassDefFoundException", tmp);
	    if (cbAccess(cb) & (ACC_INTERFACE | ACC_ABSTRACT))
	        ERROR("IncompatibleTypeException", tmp);
	    if (!VerifyClassAccess(current_class, cb, FALSE))
		ERROR("IllegalAccessException", tmp);
	    if ((ret = newobject(cb, pc, ee)) == 0) 
	        ERROR("OutOfMemoryException", 0);
	    S_HANDLE(-1) = ret;
	    TRACE(("\tnewfromname %s => %s\n", 
		   classname(cb), Object2CString(ret)));

	    /* execute the constructor the hard way */
	    for (len = cb->methodtable_size; --len>=0; )
		if ((mb = cb->methodtable->methods[len]) != 0
		    && strcmp(mb->fb.name,"<init>") == 0
		    && strncmp(mb->fb.signature, "()", 2) == 0)
		break;
	    if (len < 0)
		ERROR("NoSuchMethodException", 0);
	    do_execute_java_method_vararg(ee, ret, 0, 0, mb, FALSE, 0, 0, FALSE);
	    
	    SIZE_AND_STACK(1, 0);
	}
	      
	case opc_iaload:
	case opc_laload:
	case opc_faload:
	case opc_daload:
	case opc_aaload:
	case opc_baload:
	case opc_caload:
	case opc_saload: {
	    int index = S_INT(-1);
	    int stack_use = -1;
	    HArrayOfChar *arrh = (HArrayOfChar *)(S_HANDLE(-2));
	    ArrayOfChar *arr;
	    
	    if (arrh == 0) 
	        ERROR("NullPointerException", 0);
	    arr = unhand(arrh);
	    if (index < 0 || index >= obj_length(arrh)) 
	        ERROR("ArrayIndexOutOfBoundsException", 0);
	    switch(opcode) {
	      case opc_iaload: 
		S_INT(-2) = ((ArrayOfInt *)arr)->body[index];
		TRACE(("\tiaload ?[%d] => %d\n", index, S_INT(-2)));
		break;
	      case opc_laload: 
		SET_S_LONG(-2,  ((ArrayOfLong *)arr)->body[index]);
		TRACE(("\tlaload ?[%d] => %s\n", index, S_LONGSTRING(-2)));
		stack_use = 0;
		break;
	      case opc_faload:
		S_FLOAT(-2) = ((ArrayOfFloat *)arr)->body[index];
		TRACE(("\tfaload ?[%d] => %g\n", index, S_FLOAT(-2)));
		break;
	      case opc_daload:
		SET_S_DOUBLE(-2, ((ArrayOfDouble *)arr)->body[index]);
		TRACE(("\tdaload ?[%d] => %g\n", index, S_DOUBLE(-2)));
		stack_use = 0;
		break;
	      case opc_aaload:
		S_HANDLE(-2) = ((ArrayOfObject *)arr)->body[index];
		TRACE(("\taaload ?[%d] => %s\n", index, S_OSTRING(-2)));
		break;
	      case opc_baload:		/* signed bytes */
		S_INT(-2) = ((ArrayOfByte *)arr)->body[index];
		TRACE(("\tbaload ?[%d] => %d\n", index, S_INT(-2)));
		break;
	      case opc_caload:		/* chars */
		S_INT(-2) = ((ArrayOfChar *)arr)->body[index];
		TRACE(("\tcaload ?[%d] => %d\n", index, S_INT(-2)));
		break;
	      case opc_saload:		/* unsigned shorts */
		S_INT(-2) = ((ArrayOfShort *)arr)->body[index];
		TRACE(("\tsaload ?[%d] => %d\n", index, S_INT(-2)));
		break;
	    }
	    SIZE_AND_STACK(1, stack_use);
	}

	case opc_iastore:
	case opc_fastore:
	case opc_aastore:
	case opc_bastore:
	case opc_castore:
	case opc_sastore: {
	    HArrayOfChar *arrh = (HArrayOfChar *)S_HANDLE(-3);
	    int index = S_INT(-2);

	    ArrayOfChar *arr;
	    if (arrh == 0) 
		ERROR("NullPointerException", 0);
	    arr = unhand(arrh);
	    if (index < 0 || obj_length(arrh) <= index) 
	        ERROR("ArrayIndexOutOfBoundsException", 0);
	    switch(opcode) {
	      case opc_iastore:
		((ArrayOfInt *)arr)->body[index] = S_INT(-1);
		TRACE(("\tiastore ?[%d] <= %d\n", index, S_INT(-1)));
		break;
	      case opc_fastore:
		((ArrayOfFloat *)arr)->body[index] = S_FLOAT(-1);
		TRACE(("\tiastore ?[%d] <= %d\n", index, S_FLOAT(-1)));
		break;
	      case opc_aastore: {
		  Handle *value = S_HANDLE(-1);
		  ArrayOfObject *arrc = (ArrayOfObject *)arr;
		  int alen = obj_length(arrh);
		  ClassClass *array_cb = (ClassClass *)(arrc->body[alen]);
		  if (array_cb && !is_instance_of(value, array_cb, ee)) {
		      /* we're assigning an object, not 0, and the array is
		       * is an array of objects, not an array of arrays or
		       * and the types don't match   
		       */
		      ERROR("IncompatibleTypeException", 0);
		  }
		  /* don't dereference optop twice */
		  arrc->body[index] = value;
		  TRACE(("\taastore ?[%d] <= %d\n", index, 
			 Object2CString(value)));
		  break;
	      }
	      case opc_bastore:
		((ArrayOfByte *)arr)->body[index] = S_INT(-1);
		TRACE(("\tbastore ?[%d] <= %d\n", index, S_INT(-1)));
		break;
	      case opc_castore: 
		((ArrayOfChar *)arr)->body[index] = S_INT(-1);
		TRACE(("\tcastore ?[%d] <= %d\n", index, (unicode) S_INT(-1)));
		break;
	      case opc_sastore: 
		((ArrayOfShort *)arr)->body[index] = S_INT(-1);
		TRACE(("\tsastore ?[%d] <= %d\n", index, (short) S_INT(-1)));
		break;
	    }
	    SIZE_AND_STACK(1, -3);
	}

	case opc_lastore:
	case opc_dastore: {
	    HArrayOfChar *arrh = (HArrayOfChar *)S_HANDLE(-4);
	    int index = S_INT(-3);

	    ArrayOfChar *arr;
	    if (arrh == 0) 
		ERROR("NullPointerException", 0);
	    arr = unhand(arrh);
	    if (index < 0 || obj_length(arrh) <= index) 
		ERROR("ArrayIndexOutOfBoundsException", 0);
	    if (opcode == opc_lastore) {
		((ArrayOfLong *)arr)->body[index] = S_LONG(-2);
		TRACE(("\tlastore ?[%d] <= %s\n", index, S_LONGSTRING(-2)));
	    } else {
		((ArrayOfDouble *)arr)->body[index] = S_DOUBLE(-2);
		TRACE(("\tdastore ?[%d] <= %g\n", index, S_DOUBLE(-2)));
	    }
	    SIZE_AND_STACK(1, -4);
	}


	case opc_monitorenter:
	    monitorEnter(obj_monitor(S_HANDLE(-1)));
	    SIZE_AND_STACK(1, -1);

	case opc_monitorexit:
	    monitorExit(obj_monitor(S_HANDLE(-1)));
	    SIZE_AND_STACK(1, -1);

#ifdef ADE
	case opc_breakpoint:
	    DECACHE_OPTOP();
	    opcode = get_breakpoint_opcode(pc, ee);
	    goto again;
#endif

	default:
	    ERROR("InternalException", "undefined opcode");
	    
	} /* end of switch */

      check_for_exception:
	if (!exceptionOccurred(ee))
	   continue;
	/* Fall through */

      handle_exception: {
	  /* We've got an exception. */
	  Handle *object = ee->exception.exc;
	  unsigned char *new_pc = ProcedureFindThrowTag(frame, object, pc);
	  if (new_pc != 0) {
	      /* Found the exception handler. */
	      exceptionClear(ee);
	      vars = frame->vars;
	      constant_pool = frame->constant_pool;
	      frame->optop = frame->ostack; /* Reset to empty */
	      CACHE_OPTOP();
	      /* Push exception object onto the stack. */
	      S_HANDLE(0) = object;
	      optop++;
	      pc = new_pc;
	      if (trace) {
		  char where_am_i[100];
		  pc2string(pc, frame->current_method, 
			    where_am_i, where_am_i + sizeof where_am_i);
		  fprintf(stdout, "Catch at %s\n", where_am_i);
		  fflush(stdout);
	      }
	      continue;		/* continue with the outer loop */
	  }
      }
	/* Fall through */
      finish_return:
	/* Exit from a monitor, if necessary. */
	if (frame->monitor)
	    monitorExit(obj_monitor(frame->monitor));
	if (java_monitor) {
	    /* Indicate how much time we've used. */
	    struct methodblock *mb = frame->current_method;
	    if (mb != 0) {
	    	int time = now() - frame->mon_starttime;
		java_mon(frame->prev->current_method, mb, time);
            }
        }
	if (trace) {
	    char where_am_i[100];
	    pc2string(pc, frame->current_method, 
		      where_am_i, where_am_i + sizeof where_am_i);
	    fprintf(stdout, "Leaving %s\n", where_am_i);
	    fflush(stdout);
	}
	if (frame == initial_frame) {
	    return !exceptionOccurred(ee);
	} else {
	    ee->current_frame = frame = frame->prev;
	    if (exceptionOccurred(ee)) {
		/* Go back, and look for an exception handler. */
		pc = frame->lastpc; /*  */
	        goto handle_exception;
	    }
	    pc = frame->returnpc;
	    vars = frame->vars;
	    constant_pool = frame->constant_pool;
	    CACHE_OPTOP();
	    continue;
	}
    } /* end of infinite loop */
}


/*
 * Note: Call InitializeInterpreter() BEFORE InitializeClassThread() to
 * ensure that the Java stack lock is initialized before the idle thread
 * has a chance to run!
 */
void
InitializeInterpreter()
{
    int i;

    CODE_LOCK_INIT();
    OSTACK_LOCK_INIT();
    InstallCompiledClasses();
#ifdef STATISTICS
    memset(instruction_count, 0, sizeof(instruction_count));
#endif
}


#ifdef TRACING
#include <opcodes.init>
#else
#ifdef STATISTICS
#include <opcodes.init>
#endif
#endif

#ifdef DEBUG
void PrintObject(Handle *oh) {
    if (oh == 0)
        printf("NULL\n");
    else {
	ClassObject *o = unhand(oh);
	if (-1000 < (long) o && (long) o < 1000)
	    printf("BOGUS-PTR[%d]\n", (int)o);
	else if (obj_flags(oh) != T_NORMAL_OBJECT) 
	    printf("%s\n", Object2CString(oh));
	else {
	    ExecEnv *current_ee = threadSelf() 
	                            ? (ExecEnv *)(unhand(threadSelf())->eetop)
				    : 0;
	    ExecEnv lee;
	    ExecEnv *ee = current_ee ? current_ee 
	                         : (InitializeExecEnv(&lee, 0), &lee);
	    ClassClass *system_class = FindClassFromClass(&lee, "java/lang/System", TRUE, NULL);
	    if (system_class == 0) {
		fprintf(stderr, "Unable to find System class\n");
	    } else {
		Handle *out = *(Handle **)getclassvariable(system_class, "out");
		int oldtrace = trace;   /* we don't want to see this */
		trace = 0;
		execute_java_dynamic_method(ee, out, 
					   "println", "(Ljava/lang/Object;)V",
					   oh);
		trace = oldtrace;
	    }
	    if (current_ee == 0)
	        DeleteExecEnv(&lee, 0);
	}
    }
}    
#endif


#ifdef STATISTICS
void
DumpStatistics ()
{

    int i;
    for (i = 0; i < 256; i++) {
	if (instruction_count[i]) {
	    printf("%30.-30s %6d\n", opnames[i], instruction_count[i]);
	}
    }
}
#endif

