/*
 * @(#)monitor_cache.c	1.18 95/02/16  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/* Code to manage the monitor cache.
 *
 * The monitor cache is a hash table which contains monitors.  The
 * hash function hashs an arbitrary object address to a single monitor
 * structure.  The hash table is not growable, and if it becomes full,
 * threads which try to add entries will be blocked until there is a
 * free slot.
 *
 * JP: Rewritten to be a bucket hash table, in hopes of
 *     simplifying and speeding up the code.  No more mod's in the
 *     hash function, or arbitraryily long loops looking up
 *     monitors.  I hope monitor lookup is no more than a couple
 *     dozen instructions.
 */

#include <threads.h>
#include <monitor.h>
#include <monitor_cache.h>

#include <sys_api.h>

/* The monitor hash table: */

monitor_t  *monitorHashTable[MONITOR_HT_SIZE] = {0};
monitor_t  *monitorFreeList = 0;

/* This is a simple last used monitor cache, based on the
   assumption that a single monitor is used repeatedly before
   another monitor is used. */
static MID  lastMonitorUsed = 0;

#ifdef DEBUG
int mon_lookups = 0;
int mon_quick_find = 0;
int mon_hash_find = 0;
int mon_create = 0;
#endif

monitor_t *moncache_low;	/* Base of the monitor cache */
monitor_t *moncache_high;	/* Top of monitor cache */

sys_mon_t _moncache_lock;

/* Initialize the monitor cache
 * 
 * Called during bootstrapping, so should be no chance of interruption.
 *
 * We need a sys_-level monitor initialization step.  We don't need
 * a sys_-level monitor destruction, as we never really destroy a
 * cache monitor -- they are just left in the table to be used again.
 */

bool_t  monitorsInitialized = FALSE;

void
monitorCacheInit()
{
    int count = MONITOR_HT_SIZE * 3;
    int nbytes = count * sizeof(monitor_t);
    monitor_t *mon = (monitor_t *) malloc(nbytes);

    moncache_low = mon;
    moncache_high = mon + count;

    CACHE_LOCK_INIT();
    memset((char *) mon, 0, nbytes);
    monitorFreeList = mon;
    while (--count > 0) {
	monitor_t *next = mon + 1;
	monitorInit(mon);
	mon->next = next;
	mon = next;
    }
    monitorInit(mon);
    /* monitorFreeList is zero terminated by the memset()... */
    monitorsInitialized = TRUE;
}

/*
 * Look up a monitor in the hash table.  Return MID_NULL if it is not there.
 * Assumes we have control of the monitor cache mutex.
 */
monitor_t *
lookupMonitor(unsigned int key)
{
    monitor_t *mid, *ret;
    int	hash;

    /* assert(that we have cache mutex...) */
    CACHE_LOCK();
    
#ifdef DEBUG
    mon_lookups += 1;
#endif

    if ((mid = lastMonitorUsed) != 0 && mid->key == key) {
	assert((mid->flags & MON_IN_USE) != 0);
#ifdef DEBUG
	mon_quick_find += 1;
#endif
	ret = mid;
	goto unlock;
    }

    hash = HASH(key);
    mid = monitorHashTable[hash];
    while (mid != 0) {
	/* all monitors in hash table are in use */
	assert((mid->flags & MON_IN_USE) != 0);
	if (mid->key == key) {
#ifdef DEBUG
	    mon_hash_find += 1;
#endif
	    ret = (lastMonitorUsed = mid);
	    goto unlock;
	}
	mid = mid->next;
    }
    ret = 0;
unlock:
    CACHE_UNLOCK();
    return ret;
}


/*
 * Create a monitor for the specified key.  If it already exists,
 * just return it.
 */
monitor_t *
createMonitor(unsigned int key)
{
    int	hash;
    monitor_t **hashp, *mid;
    monitor_t *ret;	/* the return value. */

    CACHE_LOCK();

#ifdef DEBUG
    mon_lookups += 1;
#endif

    if ((mid = lastMonitorUsed) != 0 && mid->key == key) {
	assert((mid->flags & MON_IN_USE) != 0);
#ifdef DEBUG
	mon_quick_find += 1;
#endif
	ret = mid;
	goto unlock;
    }

    /* first look for the monitor, in case it's in use */
    hash = HASH(key);
    mid = monitorHashTable[hash];
    while (mid != 0) {
	/* all monitors in hash table are in use */
	assert((mid->flags & MON_IN_USE) != 0);
	if (mid->key == key) {
#ifdef DEBUG
	    mon_hash_find += 1;
#endif
	    ret = (lastMonitorUsed = mid);
	    goto unlock;
	}
	mid = mid->next;
    }

    /* If we get here, the monitor is currently not being used.  Now
       we grab one from the free list, and add it to the hash table.
       If the free list is empty, we can either add some more to the
       free list, or we can wait until some old ones are freed.  There
       is the possibility that will never happen, I think.  Perhaps
       waiting with a timeout is in order? */

    while (monitorFreeList == 0) {
#ifdef DEBUG
	fprintf(stderr, "Monitor cache is full!\n");
#endif
	CACHE_WAIT();
    }

#ifdef DEBUG
    mon_create += 1;
#endif
    mid = monitorFreeList;
    monitorFreeList = mid->next;
    assert((mid->flags & MON_IN_USE) == 0);

    mid->key = key;
    mid->flags = MON_IN_USE;
    hashp = &monitorHashTable[hash];
    mid->next = *hashp;
    *hashp = mid;
    ret = mid;

unlock:
    ret->uninit_count++;
    CACHE_UNLOCK();
    return ret;
}

/*
 * Destroy the specified monitor.  Note that it is not actually reclaimed,
 * just taken out of the hash table of the monitor cache.
 */
void
monitorDestroy(monitor_t *mid, unsigned int key)
{
    int block = 0;
    int	hash;
    MID	*hp;
    MID	next;

    assert(CACHE_LOCKED());

    if (mid->uninit_count != 0) {
	return;
    }

    assert(mid == lookupMonitor(key));
    assert(mid != 0);

    mid->flags = 0;
    mid->key = 0;
    hash = HASH(key);
    for (hp = &monitorHashTable[hash]; *hp != 0; hp = &((*hp)->next)) {
	if (*hp == mid) {
	    *(hp) = mid->next;
	    break;
	}
    }

    if (mid == lastMonitorUsed) {
	lastMonitorUsed = 0;
    }
    mid->next = monitorFreeList;
    monitorFreeList = mid;

    /*
     * Here we tell anyone who is waiting for a free monitor that a
     * new one is available.
     */
    CACHE_NOTIFY();

    return;
}

void
monitorEnumerate(void (*fcn)(monitor_t *, void *), void *cookie)
{
    monitor_t **hashtable = monitorHashTable;
    monitor_t *mid, *next;
    int	i;

    CACHE_LOCK();
    for (i = MONITOR_HT_SIZE; --i >= 0; ) {
	mid = hashtable[i];
	while (mid != 0) {
	    next = mid->next;
	    fcn(mid, cookie);
	    mid = next;
	}
    }
    CACHE_UNLOCK();
}
