/*
 * @(#)threadruntime.c	1.38 95/05/14  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*-
 *      Stuff for dealing with threads.
 */

#include <stdio.h>
#include <signal.h>

#include <tree.h>
#include <monitor.h>
#include <signature.h>
#include <threads.h>
#include <exceptions.h>
#include <log.h>
#include <counter.h>

#include "java_lang_Thread.h"

/*
 * Pointer to the Thread class block.
 */
ClassClass *Thread_classblock;

int ProcStackSize = PROCSTACKSIZE;
int JavaStackSize = JAVASTACKSIZE;

long *
getclassvariable(struct ClassClass *cb, char *fname)
{
    int    n = cb->fields_count;
    struct fieldblock *fb;

    for (fb = cbFields(cb); --n >= 0; fb++)
    if ((fb->access & ACC_STATIC) && strcmp(fname, fieldname(fb)) == 0) {
	char *signature = fieldsig(fb);
	if (signature[0] == SIGNATURE_LONG ||
	    signature[0] == SIGNATURE_DOUBLE)
	    return (long *)twoword_static_address(fb);
	else 
	    return (long *)normal_static_address(fb);
    }
    return (long *)0;
}

HThread *
InitializeClassThread(ExecEnv *ee, char **errmsg)
{
    HThread *main;
    ClassClass *cb;
    long *addr;
    struct Classjava_lang_Thread *tid;
    extern void InitializeGCThread();
    extern int noasyncgc;
    extern void ResolveInit();

    ResolveInit();
    cb = FindClassFromClass(ee, JAVAPKG "Thread", TRUE, NULL);
    if (!cb) {
	*errmsg = "cannot find class " JAVAPKG "Thread";
	return 0;
    }

    Thread_classblock = cb;

    /*
     * We want to bootstrap the currently executing context into the
     * threads package.  Instantiate Thread, and hand to the thread's
     * package.
     */
    main = (HThread *) ObjAlloc(cb, 0);
    if (main == 0) {
	out_of_memory();
    }

    tid = (struct Classjava_lang_Thread *) unhand(main);
    memcpy((char *) tid, (char *) unhand(cbPrototype(cb)), cbInstanceSize(cb));

    /* Need to set unhand(main)->eetop before threadBootstrap() */
    unhand(main)->eetop = (long)ee;
    ee->thread = (Handle *)main;
    threadBootstrap(main, &ee);

    if (noasyncgc == 0) {
	InitializeGCThread();
    }
    *errmsg = 0;
    return main;
}


void
TraceMethodCalls(int on)
{
#ifdef TRACING
    tracem = on;
#endif
}


void
TraceInstructions(int on)
{
#ifdef TRACING
    trace = on;
#endif
}


/**
 ** These are stubs that are required by other parts of Java
 **/

void
setThreadName(HThread *ht, HArrayOfChar *newName)
{
    struct Classjava_lang_Thread *t;
    t = (struct Classjava_lang_Thread *) unhand(ht);
    t->name = newName;
}

HArrayOfChar *
getThreadName()
{
    struct Classjava_lang_Thread *t;
    t = (struct Classjava_lang_Thread *) unhand(threadSelf());
    return (HArrayOfChar *)t->name;
}


HThread *
getThreadNext(HThread *p)
{
    struct Classjava_lang_Thread *tid;

    tid = (struct Classjava_lang_Thread *) unhand(p);
    return (tid->threadQ);
}

long
setThreadEETop(HThread *p, long val)
{
    long r;
    struct Classjava_lang_Thread *tid;

    tid = (struct Classjava_lang_Thread *) unhand(p);
    r = tid->eetop;
    tid->eetop = val;
    return r;
}


/*
 * Idle thread
 */

static counter_t idle_count;

static void
idle_thread(TID tid)
{
    int sb;
    extern void FreeJavaStackMemory();

    threadInit(tid, (stackp_t) &sb);
    counterInit(&idle_count, 0);
    while (1) {
	Log1(2, "Idle loop is running, %d active threads\n", 
	     ActiveThreadCount);
	FreeJavaStackMemory();	/* Free Java stacks from dead threads */
	counterIncrement(&idle_count, 1);
	pause();	/* Suspend idle thread */
    }
}

/*
 * Create the idle thread.
 */
void
InitializeIdle()
{
    TID tid;
    char *name = "Idle thread";

    tid = (TID) ObjAlloc(Thread_classblock, 0);
    memcpy((char *) unhand(tid), 
	   (char *) unhand(cbPrototype(Thread_classblock)), 
	   cbInstanceSize(Thread_classblock));

    if (threadCreate(tid, THR_SYSTEM, 10*1024,
		     (void * (*)(void *)) idle_thread) != 0) {
	SignalError(0, JAVAPKG "OutOfMemoryException", 0);
    } else {
	THREAD(tid)->name = MakeString(name, strlen(name));
	threadSetPriority(tid, MinimumPriority - 1);
	threadResume(tid);
    }
}


/*
 * Asynchronous GC thread
 *
 * Asynchronous GC is decided upon as follows: the idle thread is running
 * at a priority lower than anything else, including the async GC thread.
 * If the async GC thread wakes up, it remembers the current idle count
 * and then goes to sleep for a second.  (The idle count essentially counts
 * the number of times the idle thread has run.)  If when the GC thread
 * wakes up again the idle count is higher than it had been, it means that
 * the idle thread ran while the GC thread slept.  That's a reasonably good
 * indication that nothing was going on during that time, and thus that
 * nothing will be going on in the near future, so the garbage collector is
 * run.  If the idle thread has not been run while the GC thread, it goes
 * back to sleep again.
 *
 * Note that there is nothing to keep the system from garbage collecting
 * even though nothing has happened since it last GC'ed, and this will in
 * fact happen when the system is truly idle.  This isn't really desirable
 * as it takes CPU cycles and, if one of these useless GCs is interrupted,
 * may mean worse system responsiveness.  We could avoid this by, say, having
 * a flag that is set when GC completes and is cleared when the system does
 * something significant, like runs some Java code.
 */
void
gc_loop(TID tid)
{
    extern int heap_memory_changes;
    counter_t *idle_counter = &idle_count;
    int previous_idle_value;

    threadInit(tid, (stackp_t) &idle_counter);
    while (1) {
	previous_idle_value = counterValue(idle_counter);
	threadSleep(1000);	/* 1000 milliseconds */

	/* run GC on gc'able heap */
	if (counterValue(idle_counter) > (previous_idle_value) &&
	    heap_memory_changes) {
	    asyncGC();
	}
    }
}

void
InitializeGCThread()
{
    TID tid;
    char *name = "Async Garbage Collector";

    tid = (TID) ObjAlloc(Thread_classblock, 0);
    memcpy((char *) unhand(tid), 
	   (char *) unhand(cbPrototype(Thread_classblock)), 
	   cbInstanceSize(Thread_classblock));

    /* 32768 is a guess at an adequate stack size for GC... */
    if (threadCreate(tid, THR_SYSTEM, 32768,
		     (void * (*)(void *)) gc_loop) != 0) {
	out_of_memory();		/* fatal error */
    }
    THREAD(tid)->name = (HArrayOfChar *)MakeString(name, strlen(name));
    threadSetPriority(tid, MinimumPriority);
    threadResume(tid);
}
