/*
 * @(#)threads.c	1.29 95/03/15  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

#include <sys/types.h>
#include <memory.h>
#include <string.h>

#include <exceptions.h>
#include <threads.h>
#include <debug.h>
#include <log.h>
#include <timeval.h>
#include <monitor.h>
#include <monitor_cache.h>
#include <sys_api.h>

#include <java_lang_Thread.h>

#ifdef LOGGING
int logging_level;	/* this has to go somewhere! */
#endif

/*
 * A count of all active threads and of those all user threads.  These
 * counts are maintained by code below the HPI.
 */
int ActiveThreadCount = 0;		/* All threads */
int UserThreadCount   = 0;		/* User threads */

/*
 * The following are defined as macros in threads.h:
 *
 * void threadInit(TID tid, stackp_t sb)
 * int threadKill(TID tid)
 * stackp_t threadStackBase(TID tid)
 * void * threadStackPointer(TID tid)
 * void threadYield()
 * int threadSetPriority(TID tid, int priority)
 * int threadGetPriority(TID tid, int *priority)
 * int threadResume(TID tid)
 * int threadSuspend(TID tid)
 * void threadExit() 
 */

/*
 * Initialize the threads package.  The caller will return, executing
 * in the context of the first thread.
 */
void
threadBootstrap(TID tid, stackp_t sb)
{
    sys_thread_t *t;

    assert(SYSTHREAD(tid) == 0);

    /* Bootstrap the current process into the main, primordial thread */

    sysThreadBootstrap(&t);

    THREAD(tid)->PrivateInfo = (long) t;

    /* setup backpointer to java thread object */
    t->cookie = (void *) tid;

    THREAD(tid)->priority = NormalPriority;
    threadSetPriority(tid, NormalPriority);

    /* Create the idle thread */
    InitializeIdle();

    /* Create the finalization thread */
    InitializeFinalizer();

    /* Preallocate the internal exception objects */
    exceptionInit();

    threadInit(tid, sb);
}

int
threadCreate(TID tid, unsigned int flags, size_t stack_size, void *(*func)(void *))
{
    int err;
    sys_thread_t *t;

    assert(SYSTHREAD(tid) == 0);

    err = sysThreadCreate(stack_size, flags,
			(void * (*)(void *)) func, &t, (void *) tid);
    
    if (err == SYS_ERR) {
        return 1;
    }

    /* Set up backpointer to java thread object */
    t->cookie = (void *) tid;

    THREAD(tid)->PrivateInfo = (long) t;

    return 0;
}

TID
threadSelf()				/* Common: should be a macro? */
{
    sys_thread_t *t = sysThreadSelf();
    if (t) {
	return (TID) t->cookie;
    } else {
	return 0;
    }
}

/*
 * This structure is used during the enumeration of threads.  It
 * holds the state of the enumeration while walking through the
 * threads list.
 */
typedef struct {
    TID *array;
    int index;
    int maxindex;
} tid_array_t;

static int
threadEnumeratorHelper(sys_thread_t *t, void *arg)
{
    tid_array_t *tarr = (tid_array_t *) arg;
    TID tid;

    tid = (TID) t->cookie;
    if (tarr->array != NULL) {
	if (tarr->index >= tarr->maxindex) {
	    return SYS_ERR;
	}
	tarr->array[tarr->index] = tid;
    }
    tarr->index++;
    return SYS_OK;
}

/*
 * Enumerate all threads in the system
 */
int
threadEnumerate(TID *tarr, int count)
{
    tid_array_t helper_data;
    int i, err;

    helper_data.array = tarr;
    helper_data.maxindex = count;
    helper_data.index = 0;

    err = sysThreadEnumerateOver(threadEnumeratorHelper, (void *) &helper_data);
    for (i = helper_data.index; i < count; i++) {
	tarr[i] = (TID) 0;
    }

    return helper_data.index;
}

/*
 * Count all threads in the system.
 */
int
threadCount()
{
    tid_array_t helper_data;

    helper_data.array = 0;
    helper_data.maxindex = 0;
    helper_data.index = 0;

    sysThreadEnumerateOver(threadEnumeratorHelper, (void *) &helper_data);
    return helper_data.index;
}

void
threadSleep(int millis)
{
    /* Need a unique key per thread */
    int sleep_key = (int) SYSTHREAD(threadSelf());

    if (millis == 0) {
	threadYield();
    } else {
	Log2(4, "threadSleep: TID %s millis %d\n",
	     threadName(threadSelf()), millis);
	monitorEnter((uint_t) sleep_key);
	monitorWait((uint_t) sleep_key, millis);
	monitorExit((uint_t) sleep_key);
    }
}

int
threadPostException(TID tid, struct Hjava_lang_Object *exc) 
{ 
    struct execenv *ee;  

    ee = (ExecEnv *) THREAD(tid)->eetop;
    if (ee != (ExecEnv *) 0) {   
        /* For now we throw this random exception */ 
	struct ClassClass *cb;

	cb = FindClass(ee, JAVAPKG "Exception", TRUE);
	
	if (cb != 0 && is_instance_of(exc, cb, ee)) {
	    _fillInStackTrace(unhand(exc), ee);
	}
        exceptionThrow(ee, exc);
	return TRUE;
    }
    return FALSE;
} 
