/*
 * @(#)verify_data.c	1.11 95/01/31 Arthur van Hoff
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

#include <stdio.h>
#include <stdarg.h>
#include <oobj.h>
#include <tree.h>
#include <signature.h>
#include <verify.h>

/*
 * This file contains some really paranoid code for verifying the
 * consistency of class data. The function verify_data checks for 
 * the following:
 *
 * - data size, magic number, version
 * - the range of all offsets 
 * - the allocation of all tables, code, and strings
 * - legality of all field and method type signatures 
 * - correctness of all fields in fieldblocks and methodblock
 * - redefinition of fields and methods
 * - legality of field and method modifiers
 * - consistency of the constant table 
 * - consistency of the exception table
 * - and more..
 *
 * It does not check the byte codes, this is done in
 * a seperate pass (see verify_code). It also does not
 * check anything to do with external type information
 * such as the superclass, overriding of methods etc,
 * this should be done in a seperate pass (to be defined)...
 *
 * This code does not deal with byte order, I need to talk to
 * Eugene and Frank about that...
 */

/*
 * get data
 */

long get_ubyte(uchar *ptr)
{
    return ptr[0];
}

long get_ushort(uchar *ptr)
{
    return (ptr[0] << 8) | ptr[1];
}

long get_uint(uchar *ptr)
{
    return (ptr[0] << 24) | (ptr[1] << 16) | (ptr[2] << 8) | ptr[3];
}

long get_byte(uchar *ptr)
{
    return ((char *)ptr)[0];
}

long get_short(uchar *ptr)
{
    return (((char *)ptr)[0] << 8) | ptr[1];
}

long get_int(uchar *ptr)
{
    return (((char *)ptr)[0] << 24) | (ptr[1] << 16) | (ptr[2] << 8) | ptr[3];
}

long read_ubyte(uchar **pptr)
{
    uchar *ptr = *pptr;
    *pptr = ptr + 1;
    return get_ubyte(ptr);
}

long read_ushort(uchar **pptr)
{
    uchar *ptr = *pptr;
    *pptr = ptr + 2;
    return get_ushort(ptr);
}

long read_uint(uchar **pptr)
{
    uchar *ptr = *pptr;
    *pptr = ptr + 4;
    return get_uint(ptr);
}

long read_byte(uchar **pptr)
{
    uchar *ptr = *pptr;
    *pptr = ptr + 1;
    return get_byte(ptr);
}

long read_short(uchar **pptr)
{
    uchar *ptr = *pptr;
    *pptr = ptr + 2;
    return get_short(ptr);
}

long read_int(uchar **pptr)
{
    uchar *ptr = *pptr;
    *pptr = ptr + 4;
    return get_int(ptr);
}

/*
 * Error message
 */
int verify_verbose = 0;
int verify_errors = 0;
char verify_error_message[128];

void verify_error(char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    vsprintf(verify_error_message, fmt, args);

    if (verify_verbose) {
	fprintf(stderr, "verify: %s\n", verify_error_message);
    }
    verify_errors++;
}

/*
 * Verify type signatures
 */
int verify_field_signature(char *sig, int len) 
{
    if (len <= 0) {
	return 0;
    }
    
    switch (sig[0]) {
      case SIGNATURE_CLASS: {
	int i = 1;
	while ((sig[i] != SIGNATURE_ENDCLASS) && (i < len)) i++;
	return (i == len - 1) && (sig[i] == SIGNATURE_ENDCLASS);
      }

      case SIGNATURE_ARRAY:
	return verify_field_signature(sig + 1, len - 1);

      case SIGNATURE_BOOLEAN:
      case SIGNATURE_BYTE:
      case SIGNATURE_USHORT:
      case SIGNATURE_CHAR:
      case SIGNATURE_INT:
      case SIGNATURE_LONG:
      case SIGNATURE_FLOAT:
      case SIGNATURE_DOUBLE:
	return len == 1;
    }
    return 0;
}

int verify_method_signature(char *sig, int len) 
{
    int i;

    if ((len <= 0) || (sig[0] != SIGNATURE_FUNC)) {
	return 0;
    }

    for (i = 1 ; (i < len) && (sig[i] != SIGNATURE_ENDFUNC) ;) {
	switch (sig[i++]) {
	  case SIGNATURE_CLASS:
	    while ((sig[i] != SIGNATURE_ENDCLASS) && (i < len)) i++;
	    if ((i >= len) || (sig[i++] != SIGNATURE_ENDCLASS)) {
		return 0;
	    }
	    break;

	  case SIGNATURE_ARRAY:
	  case SIGNATURE_BOOLEAN:
	  case SIGNATURE_BYTE:
	  case SIGNATURE_USHORT:
	  case SIGNATURE_CHAR:
	  case SIGNATURE_INT:
	  case SIGNATURE_LONG:
	  case SIGNATURE_FLOAT:
	  case SIGNATURE_DOUBLE:
	    break;

	  default:
	    return 0;
	}
    }
    if (++i >= len) {
	return 0;
    }
    return (sig[i] == SIGNATURE_VOID) ? (i == len - 1) : verify_field_signature(sig + i , len - i);
}

/**
 * Verify constant pool entries
 */

int verify_Asciz(int nconstants, uchar **cpool, uint n)
{
    if ((n == 0) || (n >= nconstants)) {
	verify_error("invalid constant reference: ref=%d", n);
	return 0;
    }
    if ((n != 1) && (get_ubyte(cpool[n]) != CONSTANT_Asciz)) {
	verify_error("invalid Asciz constant reference: ref=%d", n);
	return 0;
    }
    return 1;
}

int verify_Class(int nconstants, uchar **cpool, uint n)
{
    if ((n == 0) || (n >= nconstants)) {
	verify_error("invalid constant reference: ref=%d", n);
	return 0;
    }
    if ((n != 1) && (get_ubyte(cpool[n]) != CONSTANT_Class)) {
	verify_error("invalid Class constant reference: ref=%d", n);
	return 0;
    }
    return 1;
}

int verify_NameAndType(int nconstants, uchar **cpool, uint n)
{
    if ((n == 0) || (n >= nconstants)) {
	verify_error("invalid constant reference: ref=%d", n);
	return 0;
    }
    if ((n <= 1) && (get_ubyte(cpool[n]) != CONSTANT_NameAndType)) {
	verify_error("invalid NameAndType constant reference: ref=%d", n);
	return 0;
    }
    return 1;
}

/*
 * Verify class data
 */
int verify_data(uchar *data, int length) 
{
    uchar *ptr = data;
    uchar **cpool;
    int n, i, j;
    int nconst, nintf, nfields, nmeth, natt, attlen, attname;
    int flags, name, sig;

    /* verify the size */
    if ((length < MINCLASSSIZE) || (length >= MAXCLASSSIZE)) {
	verify_error("bad class data length: len=%d", length);
	goto error;
    }

    /* verify the magic number */
    n = read_uint(&ptr);
    if (n != JAVA_CLASSFILE_MAGIC) {
	verify_error("bad magic number: expected=0x%X, magic=0x%X", JAVA_CLASSFILE_MAGIC, n);
	goto error;
    }

    /* verify the version number */
    n = read_uint(&ptr);
    if (n != JAVA_VERSION) {
	verify_error("bad version number: expected=%d, version=%d", JAVA_VERSION, n);
	goto error;
    }

    /* allocate constant pool */
    nconst = read_ushort(&ptr);
    if ((nconst < MINCONSTANTS) || (length >= MAXCONSTANTS)) {
	verify_error("bad constant pool length: len=%d", nconst);
	goto error;
    }
    cpool = (uchar **)malloc(sizeof(uchar *) * nconst);
    if (cpool == NULL) {
	verify_error("can't allocate constant pool: len=%d", nconst);
	goto error;
    }

    /* initialize constant pool */
    cpool[0] = 0;
    cpool[1] = 0;
    for (i = 2 ; i < nconst ; i++) {
	cpool[i] = ptr;
	n = read_ubyte(&ptr);
	switch (n) {
	  case CONSTANT_Asciz:
	    n = read_ushort(&ptr);
	    ptr += n * 1;
	    break;

	  case CONSTANT_Unicode:
	    n = read_ushort(&ptr);
	    ptr += n * 2;
	    break;

	  case CONSTANT_Integer:
	  case CONSTANT_Float:
	  case CONSTANT_Fieldref:
	  case CONSTANT_Methodref:
	  case CONSTANT_InterfaceMethodref:
	  case CONSTANT_NameAndType:
	    ptr += 4;
	    break;

	  case CONSTANT_Long:
	  case CONSTANT_Double:
	    ptr += 8;
	    if (++i >= nconst) {
		verify_error("last constant extends beyond end of constant pool");
		goto error;
	    }
	    cpool[i] = (uchar *)-1;
	    break;

	  case CONSTANT_Class:
	  case CONSTANT_String:
	    ptr += 2;
	    break;

	  default:
	    verify_error("invalid constant in constant pool: index=%d, id=%d", i, n);
	    goto error;
	}
    }

    /* verify constant pool */
    for (i = 2 ; i < nconst ; i++) {
	switch (get_ubyte(cpool[i])) {
	  case CONSTANT_Class:
	    n = get_ushort(cpool[i] + 1);
	    if (!verify_Asciz(nconst, cpool, n)) {
		verify_error("invalid data reference in Class constant: index=%d, ref=%d", i, n);
		goto error;
	    }
	    break;

	  case CONSTANT_String:
	    n = get_ushort(cpool[i] + 1);
	    if (!verify_Asciz(nconst, cpool, n)) {
		verify_error("invalid data reference in String constant: index=%d, ref=%d", i, n);
		goto error;
	    }
	    break;

	  case CONSTANT_Fieldref:
	    n = get_ushort(cpool[i] + 1);
	    if (!verify_Class(nconst, cpool, n)) {
		verify_error("invalid Class reference in Fieldref constant: index=%d, ref=%d", i, n);
		goto error;
	    }
	    n = get_ushort(cpool[i] + 3);
	    if (!verify_NameAndType(nconst, cpool, n)) {
		verify_error("invalid NameAndType reference in Fieldref constant: index=%d, ref=%d", i, n);
		goto error;
	    }
	    break;

	  case CONSTANT_Methodref:
	    n = get_ushort(cpool[i] + 1);
	    if (!verify_Class(nconst, cpool, n)) {
		verify_error("invalid Class reference in Methodref constant: index=%d, ref=%d", i, n);
		goto error;
	    }
	    n = get_ushort(cpool[i] + 3);
	    if (!verify_NameAndType(nconst, cpool, n)) {
		verify_error("invalid NameAndType reference in Methodref constant: index=%d, ref=%d", i, n);
		goto error;
	    }
	    break;

	  case CONSTANT_InterfaceMethodref:
	    n = get_ushort(cpool[i] + 1);
	    if ((n >= nconst) || (get_ubyte(cpool[n]) != CONSTANT_Class)) {
		verify_error("invalid Class reference in InterfaceMethodref constant: index=%d, ref=%d", i, n);
		goto error;
	    }
	    n = get_ushort(cpool[i] + 3);
	    if ((n >= nconst) || (get_ubyte(cpool[n]) != CONSTANT_NameAndType)) {
		verify_error("invalid NameAndType reference in InterfaceMethodref constant: index=%d, ref=%d", i, n);
		goto error;
	    }
	    break;

	  case CONSTANT_NameAndType:
	    n = get_ushort(cpool[i] + 1);
	    if (!verify_Asciz(nconst, cpool, n)) {
		verify_error("invalid name reference in NameAndType constant: index=%d, ref=%d", i, n);
		goto error;
	    }
	    n = get_ushort(cpool[i] + 3);
	    if (!verify_Asciz(nconst, cpool, n)) {
		verify_error("invalid signature reference in NameAndType constant: index=%d, ref=%d", i, n);
		goto error;
	    }
	    break;

	  case CONSTANT_Long:
	  case CONSTANT_Double:
	    i++;
	    break;
	}
    }

    /* verify signatures in constant pool */
    for (i = 2 ; i < nconst ; i++) {
	switch (get_ubyte(cpool[i])) {
	  case CONSTANT_Fieldref:
	    n = get_ushort(cpool[i] + 3);
	    n = get_ushort(cpool[n] + 3);
	    if (!verify_field_signature((char *)cpool[n] + 3, get_ushort(cpool[n] + 1))) {
		verify_error("invalid field signature in constant pool: index=%d, ref=%d", i, n);
		goto error;
	    }
	    break;

	  case CONSTANT_Methodref:
	    n = get_ushort(cpool[i] + 3);
	    n = get_ushort(cpool[n] + 3);
	    if (!verify_method_signature((char *)cpool[n] + 3, get_ushort(cpool[n] + 1))) {
		verify_error("invalid method signature in constant pool: index=%d, ref=%d", i, n);
		goto error;
	    }
	    break;

	  case CONSTANT_InterfaceMethodref:
	    n = get_ushort(cpool[i] + 3);
	    n = get_ushort(cpool[n] + 3);
	    if (!verify_method_signature((char *)cpool[n] + 3, get_ushort(cpool[n] + 1))) {
		verify_error("invalid interface method signature in constant pool: index=%d, ref=%d", i, n);
		goto error;
	    }
	    break;
	    
	  case CONSTANT_Long:
	  case CONSTANT_Double:
	    i++;
	    break;
	}
    }

    /* verify flags */
    n = read_ushort(&ptr);

    /* verify class name */
    n = read_ushort(&ptr);
    if (!verify_Asciz(nconst, cpool, n)) {
	verify_error("invalid class name constant: ref=%d", n);
	goto error;
    }

    /* verify super class name */
    n = read_ushort(&ptr);
    if ((n != 0) && (!verify_Asciz(nconst, cpool, n))) {
	verify_error("invalid super class name constant: ref=%d", n);
	goto error;
    }

    /* verify interface names */
    nintf = read_ushort(&ptr);
    if ((nintf < MININTERFACES) && (nintf >= MAXINTERFACES)) {
	verify_error("invalid number of interfaces: n=%d", nintf);
	goto error;
    }
    for (i = 0 ; i < nintf ; i++) {
	n = read_ushort(&ptr);
	if (!verify_Asciz(nconst, cpool, n)) {
	    verify_error("invalid interface name constant: index=%d, ref=%d", i, n);
	    goto error;
	}
    }

    /* verify fields */
    nfields = read_ushort(&ptr);
    if ((nfields < MINFIELDS) && (nfields >= MAXFIELDS)) {
	verify_error("invalid number of fields: n=%d", nfields);
	goto error;
    }
    for (i = 0 ; i < nfields ; i++) {
	/* verify flags */
	flags = read_ushort(&ptr);

	/* verify name */
	name = read_ushort(&ptr);
	if (!verify_Asciz(nconst, cpool, name)) {
	    verify_error("invalid field name constant: index=%d, ref=%d", i, name);
	    goto error;
	}

	/* verify signature */
	sig = read_ushort(&ptr);
	if (!verify_Asciz(nconst, cpool, sig)) {
	    verify_error("invalid field signature constant: index=%d, ref=%d", i, sig);
	    goto error;
	}
	if (!verify_field_signature((char *)cpool[sig] + 3, get_ushort(cpool[sig] + 1))) {
	    verify_error("invalid field signature: index=%d, ref=%d", i, sig);
	    goto error;
	}

	/* verify field attributes */
	natt = read_ushort(&ptr);
	if ((natt < MINATTR) || (natt > MAXATTR)) {
	    verify_error("invalid number of field attributes: n=%d", natt);
	    goto error;
	}
	for (j = 0 ; j < natt ; j++) {
	    attname = read_ushort(&ptr);
	    if (!verify_Asciz(nconst, cpool, attname)) {
		verify_error("invalid field attribute name constant: index=%d,%d, ref=%d", i, j, attname);
		goto error;
	    }
	    attlen = read_uint(&ptr);
	    if ((attlen < MINATTRLEN) || (attlen > MAXATTRLEN)) {
		verify_error("invalid field attribute length: index=%d,%d, len=%d", i, j, attlen);
		goto error;
	    }
	    ptr += attlen;
	}
    }

    /* verify methods */
    nmeth = read_ushort(&ptr);
    if ((nmeth < MINMETHODS) && (nmeth >= MAXMETHODS)) {
	verify_error("invalid number of methods: n=%d", nmeth);
	goto error;
    }
    for (i = 0 ; i < nmeth ; i++) {
	/* verify flags */
	flags = read_ushort(&ptr);

	/* verify name */
	name = read_ushort(&ptr);
	if (!verify_Asciz(nconst, cpool, name)) {
	    verify_error("invalid method name constant: index=%d, ref=%d", i, name);
	    goto error;
	}

	/* verify signature */
	sig = read_ushort(&ptr);
	if (!verify_Asciz(nconst, cpool, sig)) {
	    verify_error("invalid method signature constant: index=%d, ref=%d", i, sig);
	    goto error;
	}
	if (!verify_method_signature((char *)cpool[sig] + 3, get_ushort(cpool[sig] + 1))) {
	    verify_error("invalid method signature: index=%d, ref=%d", i, sig);
	    goto error;
	}

	/* verify method attributes */
	natt = read_ushort(&ptr);
	if ((natt < MINATTR) || (natt > MAXATTR)) {
	    verify_error("invalid number of method attributes: n=%d", natt);
	    goto error;
	}
	for (j = 0 ; j < natt ; j++) {
	    attname = read_ushort(&ptr);
	    if (!verify_Asciz(nconst, cpool, attname)) {
		verify_error("invalid method attribute name constant: index=%d,%d, ref=%d", i, j, attname);
		goto error;
	    }
	    attlen = read_uint(&ptr);
	    if ((attlen < MINATTRLEN) || (attlen > MAXATTRLEN)) {
		verify_error("invalid method attribute length: index=%d,%d, len=%d", i, j, attlen);
		goto error;
	    }
	    if (strncmp("Code", (char *)cpool[attname] + 3, get_ushort(cpool[attname] + 1)) == 0) {
		if (!verify_code(nconst, cpool, name, sig, ptr, attlen)) {
		    verify_error("invalid code attribute: index=%d,%d", i, j);
		    goto error;
		}
	    }
	    ptr += attlen;
	}
    }

    /* verify attributes */
    natt = read_ushort(&ptr);
    if ((natt < MINATTR) || (natt > MAXATTR)) {
	verify_error("invalid number of class attributes: n=%d", natt);
	goto error;
    }
    for (j = 0 ; j < natt ; j++) {
	attname = read_ushort(&ptr);
	if (!verify_Asciz(nconst, cpool, attname)) {
	    verify_error("invalid class attribute name constant: index=%d,%d, ref=%d", i, j, attname);
	    goto error;
	}
	attlen = read_uint(&ptr);
	if ((attlen < MINATTRLEN) || (attlen > MAXATTRLEN)) {
	    verify_error("invalid class attribute length: index=%d,%d, len=%d", i, j, attlen);
	    goto error;
	}
	ptr += attlen;
    }
    if (ptr - data != length) {
	verify_error("trailing data: remain=%d", length - (ptr - data));
	goto error;
    }

    free(cpool);
    return 1;

    /* clean up after an error */
  error:
    if (cpool != NULL) {
	free(cpool);
    }
    return 0;
}
