/*
 * @(#)color.c	1.16 95/05/14 Arthur van Hoff, Patrick Naughton
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*-
 *	Image dithering and rendering code for X11.
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include "color.h"
#include "debug.h"

#define CMAP_BITS		3
#define CMAP_COLS		(1 << CMAP_BITS)
#define CMAP_BIAS		(1 << (7 - CMAP_BITS))

/* efficiently map an RGB value to a pixel value */
#define RGBMAP(r, g, b)		\
	(rgbcube[(int)((r) + CMAP_BIAS) >> (int)(8 - CMAP_BITS)]\
	 	[(int)((g) + CMAP_BIAS) >> (int)(8 - CMAP_BITS)]\
	 	[(int)((b) + CMAP_BIAS) >> (int)(8 - CMAP_BITS)])

#define COMPOSE24(I, r, g, b)		\
	((((b) & 0xff) << ((I)->bOff)) |	\
	 (((g) & 0xff) << ((I)->gOff)) |	\
	 (((r) & 0xff) << ((I)->rOff)))

#define RGBTOGRAY(r, g, b) ((int) (.299 * r + .587 * g + .114 * b))

enum {
    FREE_COLOR		= 0,
    LIKELY_COLOR	= 1,
    ALLOCATED_COLOR	= 2
};

typedef struct ColorEntry {
	unsigned char r, g, b;
	unsigned char flags;
} ColorEntry;

static int awt_num_colors;
static ColorEntry colors[256];

/* the color cube */
static unsigned char rgbcube[CMAP_COLS + 1][CMAP_COLS + 1][CMAP_COLS + 1];

/* the StaticGray ramp */
static int grays[256];

/* the function pointers for doing image manipulation */
awtImageData *awtImage;

int awt_color_match(int, int, int);
int awt_color_match24(int, int, int);
int awt_color_matchGS(int, int, int);
void awt_dither_image2(u_char *, u_char *, u_char *,
		       u_char *, int, u_long *, int,
		       int, int, int);
void awt_dither_image24(u_char *, u_char *, u_char *,
			u_char *, int, u_long *, int,
			int, int, int);
void awt_dither_imageGS(u_char *, u_char *, u_char *,
			u_char *, int, u_long *, int,
			int, int, int);
void awt_dither_scale_image(u_char *, u_char *, u_char *,
			    u_char *, int, int, int,
			    u_long *, u_char *, int,
			    int, int, int);
int awt_retrieve_colors(u_char *, u_char *, u_char *);

awtImageData awtImage8 = {
    8,
    8,
    0, 0, 0,
    { 0 },
    { 8, 8, 8 },
    awt_color_match,
    awt_dither_image2,
    awt_dither_scale_image,
    awt_retrieve_colors
};

awtImageData awtImage24 = {
    24,
    32,
    0, 8, 16,
    { 0 },
    { 8, 8, 8 },
    awt_color_match24,
    awt_dither_image24,
    awt_dither_scale_image,
    awt_retrieve_colors
};

awtImageData awtImageGS = {
    8,
    8,
    -1, -1, -1,
    { 0 },
    { 8, 8, 8 },
    awt_color_matchGS,
    awt_dither_imageGS,
    awt_dither_scale_image,
    awt_retrieve_colors
};

#ifdef DEBUG
static int debug_colormap = 0;
#endif

/*
 * Find the best color.
 */
int
awt_color_match24(int r, int g, int b)
{
    return COMPOSE24(awtImage, r, g, b);
}

int
awt_color_matchGS(int r, int g, int b)
{
    return grays[RGBTOGRAY(r, g, b)];
}

int
awt_color_match(int r, int g, int b)
{
    int mindist = 256 * 256 * 256, besti = 0;
    int i, t, d;
    ColorEntry *p = colors;

    if (r > 255) {
	r = 255;
    }
    if (g > 255) {
	g = 255;
    }
    if (b > 255) {
	b = 255;
    }

    for (i = 0 ; i < awt_num_colors ; i++, p++)
	if (p->flags == ALLOCATED_COLOR) {
	    t = p->r - r;
	    d = t * t;
	    if (d >= mindist)
		continue;
	    t = p->g - g;
	    d += t * t;
	    if (d >= mindist)
		continue;
	    t = p->b - b;
	    d += t * t;
	    if (d >= mindist)
		continue;
	    if (d == 0)
		return i;
	    if (d < mindist) {
		besti = i;
		mindist = d;
	    }
	}
    return besti;
}

/*
 * Allocate a color in the X color map and return the index.
 */
static int
alloc_col(Display *dpy, Colormap cm, int r, int g, int b)
{
    XColor col;

    if (r > 255) {
	r = 255;
    }
    if (g > 255) {
	g = 255;
    }
    if (b > 255) {
	b = 255;
    }

    col.flags = DoRed | DoGreen | DoBlue;
    col.red   = (r << 8) | r;
    col.green = (g << 8) | g;
    col.blue  = (b << 8) | b;
    if (XAllocColor(dpy, cm, &col)) {
#ifdef DEBUG
	if (debug_colormap)
	    printf("allocated %d (%d,%d, %d)\n", col.pixel, r, g, b);
#endif
	colors[col.pixel].flags = ALLOCATED_COLOR;
	colors[col.pixel].r = col.red   >> 8;
	colors[col.pixel].g = col.green >> 8;
	colors[col.pixel].b = col.blue  >> 8;
	return col.pixel;
    }

    return awt_color_match(r, g, b);
}

/*
 * called from X11Server_create() in xlib.c 
 */
int
awt_allocate_colors(Display *dpy)
{
    unsigned long freecolors[256], plane_masks[1];
    XColor cols[256], col;
    Colormap cm;
    int i, ri, gi, bi, nfree, depth, screen;
    XPixmapFormatValues *pPFV;
    int numpfv;
    XVisualInfo *pVI, viTmp;
    int numvi;

    /*
     * Get the colormap and make sure we have the right visual
     */
    screen = DefaultScreen(dpy);
    cm = DefaultColormap(dpy, screen);
    depth = DefaultDepth(dpy, screen);

    if (depth == 24) {
	awtImage = &awtImage24;
    } else {
	awtImage = &awtImage8;
	awtImage->Depth = depth;
    }

    viTmp.visualid = XVisualIDFromVisual(DefaultVisual(dpy, screen));
    viTmp.depth = depth;
    pVI = XGetVisualInfo(dpy, VisualIDMask | VisualDepthMask, &viTmp, &numvi);
    if (pVI) {
	if (pVI->class == StaticGray || pVI->class == GrayScale) {
	    awtImage = &awtImageGS;
	    awtImage->Depth = depth;
	}
	awtImage->wsVisualInfo = *pVI;
	if (depth == 24) {
	    awtImage->rOff = 0;
	    for (i = pVI->red_mask; (i & 1) == 0; i >>= 1) {
		awtImage->rOff++;
	    }
	    awtImage->gOff = 0;
	    for (i = pVI->green_mask; (i & 1) == 0; i >>= 1) {
		awtImage->gOff++;
	    }
	    awtImage->bOff = 0;
	    for (i = pVI->blue_mask; (i & 1) == 0; i >>= 1) {
		awtImage->bOff++;
	    }
	}
	XFree(pVI);
    }

    pPFV = XListPixmapFormats(dpy, &numpfv);
    if (pPFV) {
	for (i = 0; i < numpfv; i++) {
	    if (pPFV[i].depth == depth) {
		awtImage->wsImageFormat = pPFV[i];
		break;
	    }
	}
	XFree(pPFV);
    }

    awt_num_colors = XDisplayCells(dpy, screen);

    if (depth == 24) {
	return 1;
    }

    if (awt_num_colors > 256) {
	return 0;
    }

    /*
     * Initialize colors array
     */
    for (i = 0; i < awt_num_colors; i++) {
	cols[i].pixel = i;
    }

    XQueryColors(dpy, cm, cols, awt_num_colors);
    for (i = 0; i < awt_num_colors; i++) {
	colors[i].r = cols[i].red >> 8;
	colors[i].g = cols[i].green >> 8;
	colors[i].b = cols[i].blue >> 8;
	colors[i].flags = LIKELY_COLOR;
    }

    /*
     * Determine which colors in the colormap can be allocated and mark
     * them in the colors array
     */
    nfree = 0;
    for (i = 128; i > 0; i >>= 1) {
	if (XAllocColorCells(dpy, cm, False, plane_masks, 0,
			     freecolors + nfree, i)) {
	    nfree += i;
	}
    }

    for (i = 0; i < nfree; i++) {
	colors[freecolors[i]].flags = FREE_COLOR;
    }

#ifdef DEBUG
    if (debug_colormap) {
	printf("%d free.\n", nfree);
    }
#endif

    XFreeColors(dpy, cm, freecolors, nfree, 0);

    /*
     * Allocate the colors that are already allocated by other
     * applications
     */
    for (i = 0; i < awt_num_colors; i++) {
	if (colors[i].flags == LIKELY_COLOR) {
	    colors[i].flags = FREE_COLOR;
	    alloc_col(dpy, cm, colors[i].r, colors[i].g, colors[i].b);
	}
    }
#ifdef DEBUG
    if (debug_colormap) {
	printf("got the already allocated ones\n");
    }
#endif

    /*
     * Allocate more colors, filling the color space evenly.
     */

    alloc_col(dpy, cm, 255, 255, 255);
    alloc_col(dpy, cm, 255, 0, 0);

    if (awtImage == &awtImageGS) {
	int g;

	for (i = 128; i > 0; i >>= 1) {
	    for (g = i; g < 256; g += (i * 2)) {
		alloc_col(dpy, cm, g, g, g);
	    }
	}

	for (g = 0; g < 256; g++) {
	    ColorEntry *p;
	    int mindist, besti;
	    int d;

	    p = colors;
	    mindist = 256;
	    besti = 0;
	    for (i = 0 ; i < awt_num_colors ; i++, p++) {
		if (p->flags == ALLOCATED_COLOR) {
		    d = p->g - g;
		    if (d < 0) d = -d;
		    if (d < mindist) {
			besti = i;
			if (d == 0) {
			    break;
			}
			mindist = d;
		    }
		}
	    }

	    grays[g] = besti;
	}

#ifdef DEBUG
	if (debug_colormap) {
	    fprintf(stderr, "GrayScale initialized\n");
	    fprintf(stderr, "color table:\n");
	    for (i = 0; i < awt_num_colors; i++) {
		fprintf(stderr, "%3d: %3d %3d %3d\n",
			i, colors[i].r, colors[i].g, colors[i].b);
	    }
	    fprintf(stderr, "gray table:\n");
	    for (i = 0; i < 256; i++) {
		fprintf(stderr, "%3d ", grays[i]);
		if ((i & 7) == 7)
		    fprintf(stderr, "\n");
	    }
	}
#endif
	return 1;
    }

    alloc_col(dpy, cm, 0, 255, 0);
    alloc_col(dpy, cm, 0, 0, 255);
    alloc_col(dpy, cm, 255, 255, 0);
    alloc_col(dpy, cm, 255, 0, 255);
    alloc_col(dpy, cm, 0, 255, 255);
    alloc_col(dpy, cm, 235, 235, 235);
    alloc_col(dpy, cm, 224, 224, 224);
    alloc_col(dpy, cm, 214, 214, 214);
    alloc_col(dpy, cm, 192, 192, 192);
    alloc_col(dpy, cm, 162, 162, 162);
    alloc_col(dpy, cm, 128, 128, 128);
    alloc_col(dpy, cm, 105, 105, 105);
    alloc_col(dpy, cm, 64, 64, 64);
    alloc_col(dpy, cm, 32, 32, 32);
    alloc_col(dpy, cm, 255, 128, 128);
    alloc_col(dpy, cm, 128, 255, 128);
    alloc_col(dpy, cm, 128, 128, 255);
    alloc_col(dpy, cm, 255, 255, 128);
    alloc_col(dpy, cm, 255, 128, 255);
    alloc_col(dpy, cm, 128, 255, 255);

#ifdef DEBUG
    if (debug_colormap)
	printf("got critical ones\n");
#endif

    /* allocate some colors */
    for (bi = 0 ; bi <= CMAP_COLS ; bi += 4) {
	int b = bi << (8 - CMAP_BITS);
	for (gi = 0 ; gi <= CMAP_COLS ; gi += 2) {
	    int g = gi << (8 - CMAP_BITS);
	    for (ri = 0; ri <= CMAP_COLS ; ri += 1) {
		int r = ri << (8 - CMAP_BITS);
		if (rgbcube[ri][gi][bi] == 0) {
		    rgbcube[ri][gi][bi] = alloc_col(dpy, cm, r, g, b);
		}
	    }
	}
    }
    /* match remainder */
    for (bi = 0 ; bi <= CMAP_COLS ; bi += 1) {
	int b = bi << (8 - CMAP_BITS);
	for (gi = 0 ; gi <= CMAP_COLS ; gi += 1) {
	    int g = gi << (8 - CMAP_BITS);
	    for (ri = 0 ; ri <= CMAP_COLS ; ri += 1) {
		int r = ri << (8 - CMAP_BITS);
		if (rgbcube[ri][gi][bi] == 0) {
		    rgbcube[ri][gi][bi] = awt_color_match(r, g, b);
		}
	    }
	}
    }

#ifdef DEBUG
    if (debug_colormap) {
        printf("got cube\n");
    }
#endif

#ifdef DEBUG
    if (debug_colormap) {
	int alloc_count = 0;
	int reuse_count = 0;
	int free_count = 0;
	int saw[256];
	for (i = 0; i < awt_num_colors; i++) {
	    switch (colors[i].flags) {
	      case ALLOCATED_COLOR:
		alloc_count++;
		break;
	      case LIKELY_COLOR:
		reuse_count++;
		break;
	      case FREE_COLOR:
		free_count++;
		break;
	    }
	}
	printf("%d total, %d allocated, %d reused, %d still free.\n",
	       awt_num_colors, alloc_count, reuse_count, free_count);

	for (i = 0; i < awt_num_colors; i++) {
	    saw[i] = 0;
	}
	for (bi = 0; bi <= CMAP_COLS; bi++) {
	    for (gi = 0; gi <= CMAP_COLS; gi++) {
		for (ri = 0; ri <= CMAP_COLS; ri++) {
		    saw[rgbcube[ri][gi][bi]]++;
		}
	    }
	}
	for (i = 0; i < awt_num_colors; i++) {
	    printf("%d %d\n", i, saw[i]);
	}
    }
#endif
}

/* Dithering algorithms
 */

void
awt_dither_image0(u_char *rm, u_char *gm, u_char *bm,
		  u_char *p1, int off1,
		  u_long *pa, int off2, int w, int h, int trans_index)
{
    int x, y;
    u_char *p2 = (u_char *) pa;
    for (y = 0 ; y < h ; y++, p1 += off1, p2 += off2) {
	for (x = 0 ; x < w ; x++, p1++, p2++) {
	    *p2 = awt_color_match(rm[*p1], gm[*p1], bm[*p1]);
	}
    }
}

void
awt_dither_image1(u_char *rm, u_char *gm, u_char *bm,
		  u_char *p1, int off1,
		  u_long *pa, int off2, int w, int h, int trans_index)
{
    int x, y;
    u_char *p2 = (u_char *) pa;
    for (y = 0 ; y < h ; y++, p1 += off1, p2 += off2) {
	for (x = 0 ; x < w ; x++, p1++, p2++) {
	    *p2 = RGBMAP(rm[*p1], gm[*p1], bm[*p1]);
	}
    }
}

void
awt_dither_image24(u_char *rm, u_char *gm, u_char *bm,
		   u_char *p1, int off1,
		   u_long *p2, int off2, int w, int h, int trans_index)
{
    int x, y;
    for (y = 0 ; y < h ; y++, p1 += off1, p2 += off2) {
	for (x = 0 ; x < w ; x++, p1++, p2++) {
	    *p2 = COMPOSE24(awtImage, rm[*p1], gm[*p1], bm[*p1]);
	}
    }
}

void
awt_convert_imageGS(u_char *rm, u_char *gm, u_char *bm,
		    u_char *p1, int off1,
		    u_long *pa, int off2, int w, int h, int trans_index)
{
    int x, y;
    u_char *p2 = (u_char *) pa;

    for (y = 0 ; y < h ; y++, p1 += off1, p2 += off2) {
	for (x = 0 ; x < w ; x++, p1++, p2++) {
	    *p2 = grays[RGBTOGRAY(rm[*p1], gm[*p1], bm[*p1])];
	}
    }
}

void
awt_dither_imageGS(u_char *rm, u_char *gm, u_char *bm,
		   u_char *p1, int off1,
		   u_long *pa, int off2, int w, int h, int trans_index)
{
    int *err, *ep;
    ColorEntry *cp;
    int x, y, eg, e1, e2, e3;
    u_char *p2 = (u_char *) pa;

    /* Check for 0 dimensions */
    if ((w <= 0) || (h <= 0))
	return;

    /* Allocate the error vector */
    err = (int *)malloc(sizeof(int) * (w + 2));
    if (err == 0) {
	awt_convert_imageGS(rm, gm, bm, p1, off1, pa, off2, w, h, trans_index);
	return;
    }

    memset(err, 0, sizeof(int) * (w + 2));

    for (y = h ; y-- != 0 ; p1 += off1, p2 += off2) {
	eg = 0;
	ep = err;

	for (x = w ; x-- != 0 ; ep++) {
	    /* get the next pixel */
	    u_char pix = *p1++;

	    /* decompose a pixel and add the error */
	    int g = RGBTOGRAY(rm[pix], gm[pix], bm[pix]) + ep[1];

	    /* bounds checking */
	    if (g < 0) e2 = 0;
	    else if (g > 255) e2 = 255;
	    else e2 = g;

	    /* Store the closest color in the destination pixel */
	    cp = colors + (*p2++ = grays[e2]);

	    /* Set the error from the previous lap */
	    ep[1] = eg;

	    /* compute the errors */
	    eg = e2 - cp->g;

	    /* distribute the gray error */
	    e3 = (eg << 1);
	    e1 = e3 + eg;
	    e2 = e3 + e1;
	    e3 += e2;

	    ep[0] += e1 >>= 4;
	    ep[1] += e2 >>= 4;
	    ep[2] += e3 >>= 4;
	    eg -= e1 + e2 + e3;
	}
    }

    free(err);
}

void
awt_dither_image2(u_char *rm, u_char *gm, u_char *bm,
		  u_char *p1, int off1,
		  u_long *pa, int off2, int w, int h, int trans_index)
{
    struct col {
	int r, g, b;
    } *err, *ep;
    ColorEntry *cp;
    int x, y, er, eg, eb, e1, e2, e3;
    u_char *p2 = (u_char *) pa;

    /* Check for 0 dimensions */
    if ((w <= 0) || (h <= 0))
	return;

    /* Allocate the error vector */
    err = (struct col *)malloc(sizeof(struct col) * (w + 2));
    if (err == 0) {
	awt_dither_image1(rm, gm, bm, p1, off1, pa, off2, w, h, trans_index);
	return;
    }

    memset(err, 0, sizeof(struct col) * (w + 2));

    for (y = h ; y-- != 0 ; p1 += off1, p2 += off2) {
	er = eg = eb = 0;
	ep = err;

	for (x = w ; x-- != 0 ; ep++) {
	    /* get the next pixel */
	    u_char pix = *p1++;

	    /* decompose a pixel and add the error */
	    int r = (int)rm[pix] + ep[1].r;
	    int g = (int)gm[pix] + ep[1].g;
	    int b = (int)bm[pix] + ep[1].b;

	    /* bounds checking */
	    if (r < 0) e1 = 0;
	    else if (r > 255) e1 = 255;
	    else e1 = r;

	    if (g < 0) e2 = 0;
	    else if (g > 255) e2 = 255;
	    else e2 = g;
	    
	    if (b < 0) e3 = 0;
	    else if (b > 255) e3 = 255;
	    else e3 = b;

	    /* Store the closest color in the destination pixel */
	    cp = colors + (*p2++ = RGBMAP(e1, e2, e3));

	    /* Set the error from the previous lap */
	    ep[1].r = er;
	    ep[1].g = eg;
	    ep[1].b = eb;

	    /* compute the errors */
	    er = e1 - cp->r;
	    eg = e2 - cp->g;
	    eb = e3 - cp->b;

	    /* distribute the red error */
	    e3 = (er << 1);
	    e1 = e3 + er;
	    e2 = e3 + e1;
	    e3 += e2;

	    ep[0].r += e1 >>= 4;
	    ep[1].r += e2 >>= 4;
	    ep[2].r += e3 >>= 4;
	    er -= e1 + e2 + e3;

	    /* distribute the green error */
	    e3 = (eg << 1);
	    e1 = e3 + eg;
	    e2 = e3 + e1;
	    e3 += e2;

	    ep[0].g += e1 >>= 4;
	    ep[1].g += e2 >>= 4;
	    ep[2].g += e3 >>= 4;
	    eg -= e1 + e2 + e3;

	    /* distribute the blue error */
	    e3 = (eb << 1);
	    e1 = e3 + eb;
	    e2 = e3 + e1;
	    e3 += e2;

	    ep[0].b += e1 >>= 4;
	    ep[1].b += e2 >>= 4;
	    ep[2].b += e3 >>= 4;
	    eb -= e1 + e2 + e3;
	}
    }

    free(err);
}

#ifdef SLOWFLOAT
typedef int	Accumulator;
#define DeclareScale	int scalefactor = 1
#define FindScale()							\
    for (mult = dstWtX * dstWtY; mult >= (1 << 23); mult >>= 1)		\
	scalefactor <<= 1
#define ScaleMul(x, y)		((x) * (y) / scalefactor)
#else
typedef float	Accumulator;
#define DeclareScale
#define FindScale()
#define ScaleMul(x, y)		(((float) (x)) * ((float) (y)))
#endif

static int
lcm(int m, int n)
{
    int product = m * n;
    int r;

    if (m < n) {
	r = m;
	m = n;
	n = r;
    }

    do {
	r = m % n;
	m = n;
	n = r;
    } while (n);

    return (product / m);
}

void
awt_dither_scale_image(u_char *rm, u_char *gm, u_char *bm,
		       u_char *p1, int off1, int sw, int sh,
		       u_long *pa, u_char *t2, int off2, int dw, int dh,
		       int trans_index)
{
    struct col {
	Accumulator ar, ag, ab, at;
	int er, eg, eb;
    } *err, *ep;
    ColorEntry *cp;
    int er, eg, eb, e1, e2, e3;
    int x, y;
    int srcWtX, srcWtY, dstWtX, dstWtY;
    int amtx, amty;
    Accumulator mult;
    int sxrem, syrem;
    int dxrem, dyrem;
    u_char *p, pix;
    u_char *p2;
    u_long *p3;
    DeclareScale;

    /* Check for 0 dimensions */
    if ((dw <= 0) || (dh <= 0))
	return;
    if ((sw <= 0) || (sh <= 0))
	/* Error! */
	return;
    if ((sw == dw) && (sh == dh)) {
	/* Don't have to scale it */
	awtImage->Ditherfrom8(rm, gm, bm, p1, off1, pa, off2, sw, sh,
			      trans_index);
	/* REMIND: Need to search the source pixels looking for
	 * transparent indices.
	 */
	memset(t2, 0, dw * dh);
	return;
    }

    err = (struct col *)malloc(sizeof(struct col) * (dw + 2));

    if (err == 0) {
	/* REMIND: Should write a memory efficient scaling algorithm. */
	if (sw < dw) amtx = sw; else amtx = dw;
	if (sh < dh) amty = sh; else amty = dh;
	assert((awtImage->awtBitsPerPixel & 7) == 0);
	awtImage->Ditherfrom8(rm, gm, bm,
			      p1, off1 + sw - amtx,
			      pa, off2 + (dw-amtx)*awtImage->awtBitsPerPixel/8,
			      amtx, amty,
			      trans_index);
	/* REMIND: Should search the source pixels looking for
	 * transparent indices.
	 */
	memset(t2, 0, dw * dh);
	return;
    }

    memset(err, 0, sizeof(struct col) * (dw + 2));
    if (awtImage->Depth == 24) {
	p3 = (u_long *) pa;
    } else {
	p2 = (u_char *) pa;
    }

    amtx = lcm(sw, dw);
    amty = lcm(sh, dh);
    dstWtX = amtx / dw;		/* Destination Weight in X direction */
    dstWtY = amty / dh;		/* Destination Weight in Y direction */
    srcWtX = amtx / sw;		/* Source Weight in X direction */
    srcWtY = amty / sh;		/* Source Weight in Y direction */

    FindScale();

    syrem = srcWtY;
    for (y = dh; --y >= 0; ) {
	/* First accumulate the destination pixels for this line */
	/* Clear the accumulation variables. */
	for (x = dw, ep = &err[1]; --x >= 0; ep++) {
	    ep->ar = 0;
	    ep->ag = 0;
	    ep->ab = 0;
	    ep->at = 0;
	}
	dyrem = dstWtY;
	while (dyrem) {
	    if (syrem < dyrem)
		amty = syrem;
	    else
		amty = dyrem;
	    x = sw;
	    p = p1;
	    pix = *p++;
	    sxrem = srcWtX;
	    ep = &err[1];
	    dxrem = dstWtX;
	    while (x) {
		if (sxrem < dxrem)
		    amtx = sxrem;
		else
		    amtx = dxrem;
		mult = ScaleMul(amtx, amty);
		ep->ar += mult * (Accumulator)rm[pix];
		ep->ag += mult * (Accumulator)gm[pix];
		ep->ab += mult * (Accumulator)bm[pix];
		if (pix == trans_index)
		    ep->at += mult;
		if ((sxrem -= amtx) == 0) {
		    x--;
		    pix = *p++;
		    sxrem = srcWtX;
		}
		if ((dxrem -= amtx) == 0) {
		    ep++;
		    dxrem = dstWtX;
		}
	    }
	    if ((syrem -= amty) == 0) {
		syrem = srcWtY;
		p1 += off1 + sw;
	    }
	    dyrem -= amty;
	}

	/* Now dither the destination line into the destination image */
	mult = ScaleMul(dstWtX, dstWtY);
	ep = err;

	if (awtImage->Depth == 24) {
	    for (x = dw ; x-- != 0 ; ep++) {
		/* decompose a pixel */
		int r = ((int) (ep[1].ar / mult));
		int g = ((int) (ep[1].ag / mult));
		int b = ((int) (ep[1].ab / mult));

		/* bounds checking */
		if (r < 0) e1 = 0;
		else if (r > 255) e1 = 255;
		else e1 = r;

		if (g < 0) e2 = 0;
		else if (g > 255) e2 = 255;
		else e2 = g;

		if (b < 0) e3 = 0;
		else if (b > 255) e3 = 255;
		else e3 = b;

		*p3++ = COMPOSE24(awtImage, e1, e2, e3);

		/* New transparent pixels require >= 75% transparency. */
		*t2++ = ((int) ((ep[1].at * 4) / (mult * 3)));
	    }
	    p3 += off2;
	    continue;
	}

	er = eg = eb = 0;

	if (awtImage->rOff == -1) {
	    /* GrayScale or StaticGray visual */
	    for (x = dw ; x-- != 0 ; ep++) {
		/* decompose a pixel */
		int r = ((int) (ep[1].ar / mult));
		int g = ((int) (ep[1].ag / mult));
		int b = ((int) (ep[1].ab / mult));

		/* composite RGB bounds checking */
		if (r < 0) e1 = 0;
		else if (r > 255) e1 = 255;
		else e1 = r;

		if (g < 0) e2 = 0;
		else if (g > 255) e2 = 255;
		else e2 = g;

		if (b < 0) e3 = 0;
		else if (b > 255) e3 = 255;
		else e3 = b;

		g = RGBTOGRAY(e1, e2, e3) + ep[1].eg;

		/* resulting gray bounds checking */
		if (g < 0) e2 = 0;
		else if (g > 255) e2 = 255;
		else e2 = g;

		/* Store the closest color in the destination pixel */
		cp = colors + (*p2++ = grays[e2]);

		/* Set the error from the previous lap */
		ep[1].eg = eg;

		/* compute the errors */
		eg = e2 - cp->g;

		/* distribute the gray error */
		e3 = (eg << 1);
		e1 = e3 + eg;
		e2 = e3 + e1;
		e3 += e2;

		ep[0].eg += e1 >>= 4;
		ep[1].eg += e2 >>= 4;
		ep[2].eg += e3 >>= 4;
		eg -= e1 + e2 + e3;

		/* New transparent pixels require >= 75% transparency. */
		*t2++ = ((int) ((ep[1].at * 4) / (mult * 3)));
	    }
	    p3 += off2;
	    continue;
	}

	for (x = dw ; x-- != 0 ; ep++) {
	    /* decompose a pixel and add the error */
	    int r = ((int) (ep[1].ar / mult)) + ep[1].er;
	    int g = ((int) (ep[1].ag / mult)) + ep[1].eg;
	    int b = ((int) (ep[1].ab / mult)) + ep[1].eb;

	    /* bounds checking */
	    if (r < 0) e1 = 0;
	    else if (r > 255) e1 = 255;
	    else e1 = r;

	    if (g < 0) e2 = 0;
	    else if (g > 255) e2 = 255;
	    else e2 = g;

	    if (b < 0) e3 = 0;
	    else if (b > 255) e3 = 255;
	    else e3 = b;

	    /* Store the closest color in the destination pixel */
	    cp = colors + (*p2++ = RGBMAP(e1, e2, e3));

	    /* New transparent pixels require >= 75% transparency. */
	    *t2++ = ((int) ((ep[1].at * 4) / (mult * 3)));

	    /* Set the error from the previous lap */
	    ep[1].er = er;
	    ep[1].eg = eg;
	    ep[1].eb = eb;

	    /* compute the errors */
	    er = e1 - cp->r;
	    eg = e2 - cp->g;
	    eb = e3 - cp->b;

	    /* distribute the red error */
	    e3 = (er << 1);
	    e1 = e3 + er;
	    e2 = e3 + e1;
	    e3 += e2;

	    ep[0].er += e1 >>= 4;
	    ep[1].er += e2 >>= 4;
	    ep[2].er += e3 >>= 4;
	    er -= e1 + e2 + e3;

	    /* distribute the green error */
	    e3 = (eg << 1);
	    e1 = e3 + eg;
	    e2 = e3 + e1;
	    e3 += e2;

	    ep[0].eg += e1 >>= 4;
	    ep[1].eg += e2 >>= 4;
	    ep[2].eg += e3 >>= 4;
	    eg -= e1 + e2 + e3;

	    /* distribute the blue error */
	    e3 = (eb << 1);
	    e1 = e3 + eb;
	    e2 = e3 + e1;
	    e3 += e2;

	    ep[0].eb += e1 >>= 4;
	    ep[1].eb += e2 >>= 4;
	    ep[2].eb += e3 >>= 4;
	    eb -= e1 + e2 + e3;
	}
	p2 += off2;
    }

    free(err);
}

int
awt_retrieve_colors(u_char *rm, u_char *gm, u_char *bm)
{
    int i;
    ColorEntry *p = colors;

    for (i = 0; i < awt_num_colors; i++) {
	*rm++ = p->r;
	*gm++ = p->g;
	*bm++ = p->b;
	p++;
    }

    return awt_num_colors;
}
