/*
 * @(#)CompSupport_md.c	1.3 95/04/27 Frank Yellin
 *
 * Copyright (c) 1995 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */



#include <stdarg.h>
#include <stddef.h>
#include "oobj.h"
#include "interpreter.h"
#include "opcodes.h"
#include "threads.h"
#include "typecodes.h"
#include "tree.h"
#include "bool.h"

/* The following two opcodes are used since they don't ordinarily
 * require any patching.
 */

#define HERE       opc_jsr
#define HERE_AGAIN opc_ret


/* The various types of patches required by code. */

#define PATCH_CALL30 1
#define PATCH_HIGH22 2
#define PATCH_LOW10 3
#define PATCH_FIELD_OFFSET 4
#define PATCH_CODE_OFFSET 5


/* The name of the attribute that represents compiled code on this machine. */
char *CompiledCodeAttribute = "SparcCode";

static void *getConstantPatchAddress(int opcode);
static void *getCpoolPatchAddress(struct methodblock *mb, int index, int type);
static bool_t PatchCompiledCode(struct methodblock *, unsigned char *code, 
                                unsigned char *patch, int patchSize);
static long calculateFlags(struct methodblock *mb);


/* The methodtable for java.lang.object */
struct methodtable  *ObjectMethodTable;

/* Called from the initializer.  Initialize the methods in this class in
 * whatever way is needed by Compiled code.  In particular, we should
 *    1) Patch compiled code in whatever way necessary
 *    2) Initialize the mb->CompiledCode field for non-compiled functions.
 *    3) Initialize the mb->CompiledCodeFlags field.
 */
void InitializeForCompiler_md(ClassClass *cb) {
    struct methodblock *mb = cbMethods(cb);
    int size;
    long flags;
    for (size = 0; size < (int) cb->methods_count; size++, mb++) {
	if ((mb->fb.access & ACC_MACHINE_COMPILED) == 0) {
	    extern CallbackNativeMethod;
	    extern CallbackStaticMethod;
	    extern CallbackDynamicMethod;
	    if (mb->fb.access & ACC_NATIVE)
		mb->CompiledCode = &CallbackNativeMethod;
	    else if (mb->fb.access & ACC_STATIC)
		mb->CompiledCode = &CallbackStaticMethod;
	    else 
		mb->CompiledCode = &CallbackDynamicMethod;
	}
    }
    mb = cbMethods(cb);
    for (size = 0; size < (int) cb->methods_count; size++, mb++) {
	if (mb->fb.access & ACC_MACHINE_COMPILED) {
	    /* The mb->code field temporarily holds the patches.  We should 
	     * free them when we're done. */
	    PatchCompiledCode(mb, mb->CompiledCode, mb->code, mb->code_length);
	    free(mb->code);
	    mb->code = NULL;
	    mb->code_length = 0;
	} 
	mb->CompiledCodeFlags = calculateFlags(mb);
    }
    ObjectMethodTable = cbMethodTable(classobject);
}


/* Determine the value that should code in the CompiledCodeFlags field.
 * For sparc.
 *   Bits 0-7 indicate that the corresponding argument is a double or float
 *            and thus its value is on the stack rather than in a register.
 *   Bits 8-10 give the return type of the function.
 */

static long 
calculateFlags(struct methodblock *mb) { 
    char *signature = fieldsig(&mb->fb);
    char *p;
    long flags = 0;
    int offset = (mb->fb.access & ACC_STATIC) ? 0 : 1;
    for (p = signature + 1; *p != SIGNATURE_ENDFUNC; p++) {
	switch (*p) {
	    case SIGNATURE_FLOAT:
		if (offset <= 7) 
		    flags |= (1 << offset);
		/* Fall through */
	    default:
		offset++; break;

	    case SIGNATURE_ARRAY:
		while ((*p == SIGNATURE_ARRAY)) p++;
		if (*p != SIGNATURE_CLASS) {
		    offset++; break;
		} 
		/* else fall through */
	    case SIGNATURE_CLASS:
		while (*p != SIGNATURE_ENDCLASS) p++;
		offset++; 
		break;

	    case SIGNATURE_DOUBLE:
		if (offset <= 7) 
		    flags |= ((3 << offset) & 0xFF);
		/* Fall through */
	    case SIGNATURE_LONG:
		offset += 2; 
		break;
	}
    }
    /* On the Sparc, you get better code if you tell it that all arguments past
     * the real ones are already on the stack.  You may want different results
     * on other machines.
     */
    if (offset <= 7)
	flags |= ((-1 << offset) & 0xFF);

    switch (p[1]) {
        case SIGNATURE_LONG:
	    flags |= (1 << 8); break;
        case SIGNATURE_FLOAT:
	    flags |= (2 << 8); break;
        case SIGNATURE_DOUBLE:
	    flags |= (3 << 8); break;
        case SIGNATURE_VOID:
	    flags |= (4 << 8); break;
    }
    flags |= p[1] << 24;
    return flags;
}

/* Called by assembly language.  Assembly language shouldn't have to know
 * field offsets. 
 */
long getCompiledCodeInfo(struct methodblock *mb, unsigned char **code_ptr) {
    if ((mb->fb.access & ACC_NATIVE) && (mb->code == 0)) { 
	dynoLink(mb);
	if (mb->code == 0) {
	    /* FIX */
	}
    }
    *code_ptr = mb->code;
    return mb->CompiledCodeFlags;
}

int eeExceptionKind(ExecEnv *ee) {
    return ee->exceptionKind;
}



/* Perform runtime patches on the compiled code. */
static bool_t PatchCompiledCode(struct methodblock *mb, unsigned char *code, 
                                unsigned char *patch, int patchSize) {
    ClassClass *cb = fieldclass(&mb->fb);
    union cp_item_type *cpool = cbConstantPool(cb);
    char *type_table = cpool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;
    unsigned char *endPatch = patch + patchSize;
    void *here;

    printf("Method %s %s%s at %x\n", 
	   mb->fb.clazz->name, mb->fb.name, mb->fb.signature, code);

    for ( ; patch < endPatch; patch += 7) {
	int line = (patch[0]<<24) + (patch[1]<<16) + (patch[2]<<8) + patch[3];
	int patchType = patch[4];
	int index = (patchType == PATCH_CODE_OFFSET) ? 0 : 
	                        (patch[5] << 8) + patch[6];
	unsigned char *codePtr = &code[line * 4];
	unsigned instruction = (codePtr[0] << 24) + (codePtr[1] << 16) + 
		               (codePtr[2] << 8) + (codePtr[3]);

	if (patchType == PATCH_FIELD_OFFSET) {
	    /* The low 13 bits of the instruction should get this field or
	     * methods offset. */
	    int flag = (1 << CONSTANT_Fieldref) | (1 << CONSTANT_Methodref);
	    unsigned type = CONSTANT_POOL_TYPE_TABLE_GET_TYPE(type_table, 
							      index);
	    int offset;
	    if (!ResolveClassConstant(cpool, index, EE(), flag)) {
		fprintf(stderr, "Unable to resolve class constant\n");
		return FALSE;
	    }
	    if (type == CONSTANT_Fieldref) {
		struct fieldblock *fb = cpool[index].p;
		offset = fb->u.offset;
	    } else { 
		struct methodblock *mb = cpool[index].p;
		offset = offsetof(struct methodtable, methods[mb->fb.u.offset]);
	    }
	    offset += (instruction & 0x1FFF); /* add in old offset */
	    instruction &= ~0x1FFF;
	    instruction += (offset & 0x1FFF);
	} else if (patchType == PATCH_CODE_OFFSET) {
	    /* The low 13 bits need the constant value corresponding to
	     * the offset of the CompiledCode field in a method block */
	    int offset =  offsetof(struct methodblock, CompiledCode);
	    offset += (instruction & 0x1FFF); /* add in old offset */
	    instruction &= ~0x1FFF;
	    instruction += (offset & 0x1FFF);
	    patch -= 2;
	} else {
	    /* Determine the offset of the given index. */
	    void *address = (index < 0x10000 - 0x200) 
		                  ? getCpoolPatchAddress(mb, index, patchType)
		          : (index == 0x10000 - HERE) ? (here = codePtr) 
		          : (index == 0x10000 - HERE_AGAIN) ? here
		          : getConstantPatchAddress(0x10000 - index);
	    unsigned long laddress = (unsigned long)(address);
	    if (patchType == PATCH_CALL30) {
		/* A call to the indicated spot. */
		int delta = ((unsigned char *)address) - codePtr;
		delta += instruction << 2; /* add in old delta */
		instruction &= 0xC0000000;
		instruction += ((unsigned)delta) >> 2;
	    } else if (patchType == PATCH_HIGH22) {
		/* Put the high 22 bits of the address into the low 22 bits of
		 * the opcode.  We can't really deal with already existing 
		 * offsets in the opcode. */
		instruction &= 0xFFC00000; 
		instruction += (laddress >> 10);
	    } else {
		/* Put the low 10bits of the address into the low 13 bits of 
		 * the opcode.  Don't forget any offset already there. */
		laddress += (instruction & 0x3ff);
		instruction &= ~0x3FF;
		instruction += (laddress & 0x3ff);
	    }
	}
	codePtr[0] = instruction >> 24; 
	codePtr[1] = instruction >> 16;
	codePtr[2] = instruction >> 8;  
	codePtr[3] = instruction;
    }
}

extern CompSupport_imul;
extern CompSupport_idiv;
extern CompSupport_imod;
extern CompSupport_lmul;
extern CompSupport_ldiv;
extern CompSupport_lmod;
extern CompSupport_fmod;
extern CompSupport_dmod;
extern CompSupport_d2l;
extern CompSupport_f2l;
extern CompSupport_l2d;
extern CompSupport_l2f;
extern CompSupport_lshl;
extern CompSupport_lshr;
extern CompSupport_lushr;
extern CompSupport_invokeinterface;
extern CompSupport_new;
extern CompSupport_newarray;
extern CompSupport_anewarray;
extern CompSupport_multianewarray;
extern CompSupport_checkcast;
extern CompSupport_instanceof;
extern CompSupport_athrow;
extern CompSupport_monitorenter;
extern CompSupport_monitorexit;

double CompSupport_dconst_0 = 0.0;
double CompSupport_dconst_1 = 1.0;

float CompSupport_fconst_0 = 0.0;
float CompSupport_fconst_1 = 1.0;
float CompSupport_fconst_2 = 2.0;




/* Constant addresses are associated with opcodes. */
static void *getConstantPatchAddress(int opcode) {
    static bool_t inited = FALSE;
    static void *values[256];
    static void *result;
    if (!inited) {
	values[opc_imul] = (void *)&CompSupport_imul;
	values[opc_idiv] = (void *)&CompSupport_idiv;
	values[opc_imod] = (void *)&CompSupport_lmod;

	values[opc_lmul] = (void *)&CompSupport_lmul;
	values[opc_ldiv] = (void *)&CompSupport_ldiv;
	values[opc_lmod] = (void *)&CompSupport_lmod;

	values[opc_lshl] = (void *)&CompSupport_lshl;
	values[opc_lshr] = (void *)&CompSupport_lshr;
	values[opc_lushr] = (void *)&CompSupport_lushr;

	values[opc_fmod] = (void *)&CompSupport_fmod;
	values[opc_dmod] = (void *)&CompSupport_dmod;

	values[opc_d2l] = (void *)&CompSupport_d2l;
	values[opc_f2l] = (void *)&CompSupport_f2l;
	values[opc_l2d] = (void *)&CompSupport_l2d;
	values[opc_l2f] = (void *)&CompSupport_l2f;

	values[opc_invokevirtualobject_quick] = 
	       (void *)cbMethodTable(classobject);
	values[opc_invokeinterface] = (void *)&CompSupport_invokeinterface;

	values[opc_dconst_0] = (void *)&CompSupport_dconst_0;
	values[opc_dconst_1] = (void *)&CompSupport_dconst_1;
	values[opc_fconst_0] = (void *)&CompSupport_fconst_0;
	values[opc_fconst_1] = (void *)&CompSupport_fconst_1;
	values[opc_fconst_2] = (void *)&CompSupport_fconst_2;

	values[opc_new] = (void *)&CompSupport_new;
	values[opc_newarray] = (void *)&CompSupport_newarray;
	values[opc_anewarray] = (void *)&CompSupport_anewarray;
	values[opc_multianewarray] = (void *)&CompSupport_multianewarray;
	values[opc_checkcast] = (void *)&CompSupport_checkcast;
	values[opc_instanceof] = (void *)&CompSupport_instanceof;
	values[opc_athrow] = (void *)&CompSupport_athrow;
	values[opc_monitorenter] = (void *)&CompSupport_monitorenter;
	values[opc_monitorexit] = (void *)&CompSupport_monitorexit;

	inited = TRUE;
    }
    result = values[opcode];
    if (result == 0) 
	fprintf(stderr, "Unknown opcode %d\n", opcode);
    return result;
}

/* Get the address of a constant pool item */
static void *
getCpoolPatchAddress(struct methodblock *mb, int index, int patchType) {
    ClassClass *cb = fieldclass(&mb->fb);
    union cp_item_type *cpool = cbConstantPool(cb);
    char *type_table = cpool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;
    switch(CONSTANT_POOL_TYPE_TABLE_GET_TYPE(type_table, index)) {
        case CONSTANT_Integer: case CONSTANT_Long:
        case CONSTANT_Float: case CONSTANT_Double:
	    return &cpool[index];

        case CONSTANT_String:
	    ResolveClassConstantFromClass(cb, index, EE(), ~0);
	    return cpool[index].p;

	case CONSTANT_Class:
	    ResolveClassConstantFromClass(cb, index, EE(), ~0);
	    return cpool[index].p;

        case CONSTANT_Methodref:
	    ResolveClassConstantFromClass(cb, index, EE(), ~0);
	    if (patchType == PATCH_CALL30) 
		return ((struct methodblock *)(cpool[index].p))->CompiledCode;
	    else 
		return ((struct methodblock *)(cpool[index].p));
	    
        case CONSTANT_InterfaceMethodref: 
	    ResolveClassConstant(cpool, index, EE(), ~0);
	    {
		unsigned InterfaceMethodref = cpool[index].i;
		unsigned NameAndType_index = InterfaceMethodref & 0xFFFF;
		unsigned ID = cpool[NameAndType_index].i;
		return (void *)ID;
	    }

        case CONSTANT_Fieldref: {
	    struct fieldblock *fb;
	    char sigstart;
	    ResolveClassConstantFromClass(cb, index, EE(), ~0);
	    fb = cpool[index].p;
	    sigstart = (fieldsig(fb))[0];
	    if (sigstart == SIGNATURE_LONG || sigstart == SIGNATURE_DOUBLE) 
		return twoword_static_address(fb);
	    else 
		return normal_static_address(fb);
	}
    }	
}

