/*
 * @(#)clock.c	1.16 95/04/18  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * The clock manager for the Green threads package.  This is essentially
 * a thread running at maximum priority that manages a database of timeouts.
 *
 * There are two primary interfaces to this code:
 *  1. suspend a thread for a period of time (thread_sleep).
 *  2. notify a condition variable after a period of time.
 */

#include <threads.h>
#include <timeval_md.h>
#include <debug.h>
#include <clock.h>
#include <schedule.h>
#include <log.h>
#include <machapi.h>
#include <monitor_md.h>
#include <queue.h>

/*
 * Internal interfaces.
 */
TID clock_tid;			/* Can't be a sys_thread_t... */
static unsigned int clockMonKey;
extern sys_mon_t *asyncMon(async_mon_key_t key);
#define CLOCK_MID()  asyncMon(clockMonKey)


/*
 * A list of threads that are waiting on alarms.  This list is sorted in
 * order of timeout value.
 */
sys_thread_t *threadAlarmQ;


/*
 * The first thread on the alarm queue has expired.  Notify it, remove
 * it from the queue, and return.
 */
static void
deliverAlarm()
{
    sys_thread_t *t;
    sys_mon_t *mid;

    assert(CLOCK_MID()->monitor_owner == SYSTHREAD(clock_tid));
    assert(SCHED_LOCKED());

    Log(3,"deliver_alarm: checking if we need to do something.\n");
    t = threadAlarmQ;
    threadAlarmQ = t->timeoutQ;
    t->timeoutQ = SYS_THREAD_NULL;
    timerClear(&t->timeout);

    /*
     * It is possible that we got scheduled before the clock thread,
     * so don't do anything stupid if this isn't a waiting thread.
     */
    if (t->state == CONDVAR_WAIT) {
	mid = (sys_mon_t *)t->mon_wait;
	queueRemove(&mid->condvar_waitq, t);
	queueInsert(&mid->monitor_waitq, t);
	t->state = MONITOR_WAIT;
	Log2(4, "clock wakes up sys_mon_t 0x%x 0x%8x\n", mid, t);
	if (mid->monitor_owner == SYS_THREAD_NULL) {
	    Log2(4, "clock signals sys_mon_t 0x%x 0x%8x\n", mid, t);
	    (void) queueSignal(mid, &mid->monitor_waitq);
	} else {
	    monitorApplyInversion(mid, t);
	}
    }
}


/*
 * Register an alarm for this thread.  The caller has set the thread
 * state already.
 */
void
registerAlarm(sys_thread_t *t, timeval_t timeout)
{
    sys_thread_t *q;
    sys_thread_t *prev;
    sys_mon_t *mid;

    Log(3,"register_alarm..\n");
    mid = CLOCK_MID();

    assert(t->state == RUNNABLE);
    assert(SCHED_LOCKED());

    t->timeout = timeout;

    q = threadAlarmQ;
    prev = 0;
    while (q && (timerCmp(&q->timeout, &t->timeout, <))) {
	prev = q;
	q = q->timeoutQ;
    }
    if (prev) {
	prev->timeoutQ = t;
	t->timeoutQ = q;
    } else {
	t->timeoutQ = threadAlarmQ;
	threadAlarmQ = t;
    }
    (void) queueSignal(mid, &mid->condvar_waitq);
    if (!mid->monitor_owner)
	(void) queueSignal(mid, &mid->monitor_waitq);
    Log(3,"..register_alarm\n");
    return;
}


/*
 * If this thread is on the alarm queue, remove it.  This is called
 * during thread death (& cleanup), as well as during monitor wait
 * cleanup.
 */
void
cleanupPendingAlarm(sys_thread_t *t)
{
    sys_thread_t *q;
    sys_thread_t *prev;

    assert(sysThreadSelf()->state == RUNNABLE);
    assert(SCHED_LOCKED());
    assert(CLOCK_MID()->monitor_owner == SYS_THREAD_NULL);

    if (threadAlarmQ) {
	q = threadAlarmQ;
	prev = SYS_THREAD_NULL;
	while (q) {
	    if (q == t) {
		if (prev) {
		    prev->timeoutQ = q->timeoutQ;
		} else {
		    threadAlarmQ = q->timeoutQ;
		}
		q->timeoutQ = SYS_THREAD_NULL;
		timerClear(&q->timeout);
		break;
	    }
	    prev = q;
	    q = q->timeoutQ;
	}
    }

    assert(timerIsSet(&t->timeout) == 0);
    assert(t->timeoutQ == SYS_THREAD_NULL);
    return;
}


/*
 * This is the main loop of the clock thread.  It basically waits for
 * CV notification (i.e. it runs in a high priority thread context, NOT
 * at interrupt time).  It gets notified when one of two things happens:
 *
 *	1. An alarm expires
 *	2. somebody manipulates the database.
 *
 * This threads job is to sit in a tight loop, and notify other threads
 * when their timeouts occur.
 */

static void
clockHandler()
{
    timeval_t time, timeout, t;
    sys_mon_t *mid;

    sysMonitorEnter(asyncMon(clockMonKey));
    SCHED_LOCK();

    while (1) {
	time = currentTime();
	timerClear(&timeout);

	/*
	 * Calculate next timeout.  Also do any pending notifications.
	 */
	Log(3,"clockHandler() woke up...\n");
	while (threadAlarmQ) {
	    t = threadAlarmQ->timeout;
	    if (timerCmp(&t, &time, <) || timerEq(&t, &time)) {
		deliverAlarm();
		continue;
	    }
	    timeout = t;
	    break;
	}
	
	/*
	 * At this point, we have handled all of the pending timeouts, and
	 * have found the minimum timeout.  Wait.
	 */
	if (timerIsSet(&timeout)) {
	    t = timeout;
	    timerSub(&t, &time);
	    assert(timerIsSet(&t));
	    scheduleAlarm(t);
	}
	Log(3,"...clockHandler() woke up\n");
	mid = CLOCK_MID();
	queueWait(mid, &mid->condvar_waitq);

	/*
	 * Now we have either been notified because the database has
	 * changed, or because the timeout has expired.  We just loop
	 * around, and handle all the pending events while we calculate
	 * the new timeout.  We don't really care why we were scheduled.
	 */

    }
    /* NOTREACHED */
}

/*
 * At least for now the clock thread breaks the HPI abstraction.
 * There isn't really a notion of a thread that doesn't exist above
 * the HPI level -- all threads are assumed to have a TID, etc.  But
 * the clock is only present in Green systems and not native ones, so
 * wants to be below the HPI.  Otherwise, the clock is just like the
 * idle thread and GC thread, which are set up by threadruntime.c.
 */
void
InitializeClock()
{
    extern ClassClass *Thread_classblock;
    char *name = "clock handler";

    clock_tid = (TID) ObjAlloc(Thread_classblock, 0);
    memcpy((char *) unhand(clock_tid), 
	   (char *) unhand(cbPrototype(Thread_classblock)), 
	   cbInstanceSize(Thread_classblock));
    if (threadCreate(clock_tid, THR_SYSTEM, 10*1024, 
	(void * (*)(void *)) clockHandler) != 0) {
	SignalError(0, JAVAPKG "OutOfMemoryException", 0);
    } else {
	THREAD(clock_tid)->name = MakeString(name, strlen(name));
	threadSetPriority(clock_tid, MaximumPriority + 1);
	clockMonKey = hostClockInit();
	threadResume(clock_tid);
    }
}



/**
 ** External interfaces
 **/

void
threadSetTimeout(timeval_t *t, int Sec, int uSec)
{
    if (t) {
	t->tv_sec = Sec;
	t->tv_usec = uSec;
    }
    return;
}

void
threadGetTimeout(timeval_t *t, int *Sec, int *uSec)
{
    if (!t) {
	return;
    }

    if (Sec) {
	*Sec = t->tv_sec;
    }

    if (uSec) {
	*uSec = t->tv_usec;
    }

    return;
}

static void
timerFix(timeval_t *t1)
{
	if (t1->tv_usec < 0) {
		t1->tv_sec--;
		t1->tv_usec += 1000000;
	}
	if (t1->tv_usec >= 1000000) {
		t1->tv_sec++;
		t1->tv_usec -= 1000000;
	}
}

void
timerAdd(timeval_t *t1, timeval_t *t2)
{
	t1->tv_sec += t2->tv_sec;
	t1->tv_usec += t2->tv_usec;
	timerFix(t1);
}

void
timerSub(timeval_t *t1, timeval_t *t2)
{
	t1->tv_sec -= t2->tv_sec;
	t1->tv_usec -= t2->tv_usec;
	timerFix(t1);
}

void
DumpAlarmQ()
{
    timeval_t	 t, now;
    sys_thread_t *tid;

    SCHED_LOCK();
    fprintf(stderr, "Thread Alarm Q:\n");
    tid = threadAlarmQ;
    now = currentTime();
    while (tid != 0) {

	fprintf(stderr, "    sys_thread_t 0x%x", tid);
	t = tid->timeout;
	if (timerCmp(&t, &now, <) || timerEq(&t, &now)) {
	    fprintf(stderr, "   [Timeout expired]\n");
	} else {
	    fprintf(stderr, "   [Timeout in %d ms]\n",
		    (t.tv_sec - now.tv_sec) * 1000
		    + (t.tv_usec - now.tv_usec) / 1000);
	}
	tid = tid->timeoutQ;
    }
    SCHED_UNLOCK();
}

