/*
 * @(#)context.c	1.27 95/04/18
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Initialize the platform dependent parts of a thread context.
 */

#include <sys/types.h>
#include <stdlib.h>
#include <threads_md.h>
#include <interrupt.h>
#include <machapi.h>
#include <sparc_md.h>
#include <context.h>
#include <debug.h>

#include "sys_api.h"

/*
 * The list on which we put freed thread stacks pending reclamation.
 */
typedef struct free_stack_t {
    long   stack_size;
    struct free_stack_t *next;
} free_stack_t;

static free_stack_t *stackFreeList = (free_stack_t *) 0;

/*
 * Free tid and stack space for dead threads.  We do this whenever
 * we do sysThreadCreate().
 */
void freeStacks() 
{
    free_stack_t* stack;
    free_stack_t* next;

    if (stackFreeList != NULL) {
	SCHED_LOCK();
	stack = stackFreeList;
	stackFreeList = 0;
	SCHED_UNLOCK();
	while (stack) {
	    next = stack->next;
	    sysFreeMem((char *)stack, stack->stack_size);
	    stack = next;
	}
    }
}

/*
 * Allocate a single thread stack of specified size.
 *
 * Note that "stack base" here and elsewhere means the base of the
 * logical stack, growing from high addresses to low addresses, and
 * *not* the way you'd think of, say, a malloc'ed chunk.
 *
 * We let the low level stack allocator take care of any rounding or
 * other system-specific sizing of the stack request.  But if no stack
 * size is given we take the default one. 
 */
static int 
allocateStack(gstack_t *stack, long size)
{
    extern int ProcStackSize;

    freeStacks();		/* Free any stacks for dead threads */

    if (size == 0) {
        /* Allocate conservatively if we don't know better */
        size = ProcStackSize;
    } 
    stack->base = (stackp_t) ((char *) sysAllocateMem(size) + size);
    if (!stack->base) {
	return 0;		/* Allocation failed */
    } else {
	stack->size = size;
	return 1;
    }
}

/*
 * Returns the stack pointer for any threads but the current one.
 */
void *
currentStackPointer(sys_thread_t *tp)
{
    context_t   *context = &tp->mdcontext;
    ucontext_t	*uc = &(context->ucontext);
    return (void *) (uc->uc_mcontext.gregs[REG_O6]);
}

/*
 * This is the stack layout, as defined by the SPARC ABI:
 *
 *      HIGH ADDRESSES
 *	|-----------------------|
 *	| ThreadPrivate data	|
 *      |-----------------------|
 *      | 16 words for in and   |  death_func() stack frame
 *      | local save area       |
 *      |-----------------------|  --------
 *      | arguments past the 6th|
 *      |-----------------------|  startup func stack frame
 *	| 6 words for locals    |
 *	|-----------------------|
 *      | 1 word for aggregate  |
 *      | return value          |
 *      |-----------------------|
 *      | %i7 = startaddr       |
 *      | %i6 = sp              |
 *      | %i0-%i1 --> args 1-5  |
 *      |-----------------------|
 *      | 8 words for saved %l's|
 *      |-----------------------|       <--- sp(double aligned)
 *      LOW ADDRESSES
 */
void
initContext(sys_thread_t *tid, unsigned int pc,
	     void (*death_func)(), unsigned int arg)
{
    unsigned char *sp;
    ucontext_t *uc; 
    context_t *mdcontext = CONTEXT(tid);
    machdep_t *machdepP = MACHDEP(tid);

    uc = &(mdcontext->ucontext);
    sp = (unsigned char *) tid->stack_base;
    assert(sp == ((unsigned char *)tid+sizeof(sys_thread_t)));

    sp -= (sizeof(sys_thread_t) + WINDOWSIZE + SA(MINFRAME));

    /* Force sp to be double aligned! */
    sp = (unsigned char *)((unsigned long)(sp) & 0xfffffff8); 

    /* clear the context structure */
    memset(uc, 0, sizeof(ucontext_t)); 
    uc->uc_stack.ss_sp = (char *) sp;
    uc->uc_stack.ss_size = tid->stack_size;
    uc->uc_stack.ss_flags = 0; /* ? */

    /* initialially all signals are enabled */
    /* FIXME change to be initially disabled? */
    memset((char *)&(uc->uc_sigmask), 0, sizeof(sigset_t));
    machdepP->intrLockCount = 0;

    uc->uc_mcontext.gregs[REG_SP] = (unsigned int) sp;
    uc->uc_mcontext.gregs[REG_PC] = (unsigned int) death_func;
    uc->uc_mcontext.gregs[REG_nPC] = (unsigned int) death_func+4;
    uc->uc_mcontext.gregs[REG_O1] = arg; 
    uc->uc_mcontext.gregs[REG_O2] = (int) pc;

    uc->uc_flags = UC_ALL;
    return;
}

/*
 * This routine allocates enought memory for the thread context
 * and the stack.  This implementation puts the thread context
 * at the top of the stack.
 *
 * Must zero memory and initialize tid->PrivateInfo and
 * tid->PrivateInfo->stack.
 */
int
allocateContextAndStack(sys_thread_t **ptid, long requested_size)
{
    gstack_t stack;
    char *sp;

    if (!allocateStack(&stack, requested_size)) {
	return 0;
    }

    sp = (char *) stack.base;
    sp -= sizeof(sys_thread_t);
    memset(sp, 0, sizeof(struct sys_thread_t));

    *ptid = (sys_thread_t *) sp;
    (*ptid)->stack_base = stack.base;
    (*ptid)->stack_size = stack.size;	/* Actual size allocated */

    return 1;
}

/*
 * Delete the specified context, and the stack associated with it.
 * This does *not* free the memory associated with the stack: that
 * is handled elsewhere!
 *
 * WARNING: Called when SCHED_LOCKED(), so may not enter any monitors!
 */
void
deleteContextAndStack(sys_thread_t *tid)
{
    /*
     * If tid->primordial_thread (Unix main()), we don't want to free
     * the C stack, but we have malloc'ed the sys_thread_t.  This is
     * NOT the normal case!  But we don't want this going back to the
     * stack allocator, so we drop the sys_thread_t on the floor here.
     */
    if (!tid->primordial_thread) {
        free_stack_t *tmp;
	/*
	 * The normal case: queue the stack and the thread private data
	 * to be freed when safe.
	 */
	assert(0 <= tid->stack_size);
	tmp = (free_stack_t *)((char *) tid->stack_base - tid->stack_size);
	tmp->stack_size = tid->stack_size;
	tmp->next = stackFreeList;
	stackFreeList = tmp;
    }
    return;
}
