/*
 * @(#)interrupt.c	1.9 95/01/31  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Green threads interrupt dispatch
 */

/*
 * Header files.
 */
#include <stddef.h>
#include <stdlib.h>
#include <debug.h>
#include <interrupt.h>
#include <log.h>

/*
 * Type definintions.
 */

/* handler_entry_t is used to keep track of the registered handlers. */
typedef struct handler_entry {
    struct handler_entry	*next;
    intr_handler_t		handler;
    void			*handlerArg;
} handler_entry_t;


/*
 * Local Variables
 */
static handler_entry_t	*handlerList[N_INTERRUPTS];


/*
 * Exported Routines
 */

/* Initialize the interrupt system */
extern void
intrInit(void)
{
    int		i;

    /* Initialize the handler lists. */
    for (i = 0; i < N_INTERRUPTS; i++) {
	handlerList[i] = NULL;
    }

    /*
     * Target-dependant initialization.
     */
    intrInitMD();
}

/* Add/Remove a handler for a particular interrupt */
extern void
intrRegister(int interrupt, intr_handler_t handler, void *handlerArg)
{
    handler_entry_t	*newEntryP;

    /* Allocate and initialize a new handler entry */
    newEntryP = (handler_entry_t *) malloc(sizeof (*newEntryP));
    newEntryP->handler = handler;
    newEntryP->handlerArg = handlerArg;

    /* Add it to the list */
    intrLock();
    newEntryP->next = handlerList[interrupt];
    handlerList[interrupt] = newEntryP;

    /*
     * If this is the first handler, enable the interrupt at
     * the machine dependant level.
     */
    if (newEntryP->next == NULL) {
	intrEnableDispatch(interrupt);
    }
    intrUnlock();
}
    
extern void
intrUnregister(int interrupt, intr_handler_t handler, void *handlerArg)
{
    handler_entry_t	**entryPP;
    handler_entry_t	*oldEntryP;
    
    /*
     * Search the specified interrupt's list for a handler
     * entry that matches both handler and arg, remove
     * it from the list, and free it.
     */
    oldEntryP = NULL;
    intrLock();
    for (entryPP = &handlerList[interrupt]; *entryPP;
	 entryPP = &(*entryPP)->next) {
	if ((*entryPP)->handler == handler &&
	    (*entryPP)->handlerArg == handlerArg) {
	    oldEntryP = *entryPP;
	    *entryPP = oldEntryP->next;
	    break;
	}
    }

    /*
     * If there are no more handlers, disable the interrupt
     * at the machine-dependant level.
     */
    if (handlerList[interrupt] == NULL) {
	intrDisableDispatch(interrupt);
    }
    intrUnlock();
    if (oldEntryP != NULL) {
	free(oldEntryP);
    }
}

/*
 * intrDispatch -- Dispatch an interrupt.
 *
 * 		   This routine is called from the low-level handlers
 * 		   at interrupt time.
 */
extern void
intrDispatch(int interrupt, void *siginfo)
{
    handler_entry_t	**entryPP;
    handler_entry_t	*entryP;

    /*
     * Assumptions:
     *	- Each interrupt only has one priority level associated with
     *	  it.
     *	- Each handler will do enough work so that when it returns
     *	  the source of the interrupt is masked.
     */

    /*
     * Run through the handler list for a particular interrupt.
     * If the handler is successful, then put it at the beginning
     * of the list.
     *
     * Putting the successful handler at the beginning of the list is
     * an optimization that should leave the most frequent interrupts
     * near the beginning of their handler lists.
     */
    assert(interrupt >= 0 && interrupt < N_INTERRUPTS);
    for (entryPP = &handlerList[interrupt]; *entryPP;
	 entryPP = &((*entryPP)->next)) {
	entryP = *entryPP;
	if ((*entryP->handler)(interrupt, entryP->handlerArg, siginfo) != 0) {
	    /*
	     * This interrupt was handled, if it's not the first
	     * handler, then make it the first handler
	     */
	    if (entryPP != &handlerList[interrupt]) {
		*entryPP = entryP->next;
		entryP->next = handlerList[interrupt];
		handlerList[interrupt] = entryP;
	    }

	    /* We've done our duty, now return. */
	    return;
	}
    }

    /* No handler for this interrupt, log the error */
    Log1(0, "spurious interrupt %d\n", interrupt);
    return;
}
