/*
 * @(#)interrupt_md.c	1.26 95/04/27 Chuck McManis, Chris Warth
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Solaris 2.X dependant interrupt handling code.
 */


/*
 * Header files.
 */
#include <stdio.h>
#include <signal.h>
#include <oobj.h>
#include <debug.h>
#include <ucontext.h>
#include <interrupt.h>
#include <machapi.h>
#include <log.h>
#include <internal_md.h>
#include <threads_md.h>
#include <schedule.h>
#include <sys_api.h>

/* This is supposed to be in some include file, but where?? */
#define ILL_TRAP_FAULT(n) (128 + n)
#define TICC_NULL_POINTER_EXCEPTION  16
#define TICC_ARRAY_INDEX_OUT_OF_BOUNDS 17


/*
 * Local data.
 */
static char *signalNames[] = {
    "no signal",
    "SIGHUP    1    hangup",
    "SIGINT    2    interrupt",
    "SIGQUIT   3*   quit",
    "SIGILL    4*   illegal instruction",
    "SIGTRAP   5*   trace trap",
    "SIGABRT   6*   abort (generated by abort(3) routine)",
    "SIGEMT    7*   emulator trap",
    "SIGFPE    8*   arithmetic exception",
    "SIGKILL   9    kill (cannot be caught, blocked, or ignored)",
    "SIGBUS    10*  bus error",
    "SIGSEGV   11*  segmentation violation",
    "SIGSYS    12*  bad argument to system call",
    "SIGPIPE   13   write on a pipe or other socket with no one to read it",
    "SIGALRM   14   alarm clock",
    "SIGTERM   15   software termination signal",
    "SIGUSR1   16   user defined signal 1",
    "SIGUSR2   17   user defined signal 2",
    "SIGCHLD   18@  child status change",
    "SIGPWR    19   power-fail restart",
    "SIGWINCH  20@  window changed (see termio(4) and win(4S))",
    "SIGURG    21@  urgent condition present on socket",
    "SIGPOLL   22   pollable event occurred (also SIGIO)",
    "SIGSTOP   23  stop (cannot be caught, blocked, or ignored)",
    "SIGTSTP   24+  stop signal generated from keyboard",
    "SIGCONT   25@  continue after stop",
    "SIGTTIN   26+  background read attempted from control terminal",
    "SIGTTOU   27+  background write attempted to control terminal",
    "SIGVTALRM 28   virtual time alarm (see getitimer(2))",
    "SIGPROF   29   profiling timer alarm (see getitimer(2))",
    "SIGXCPU   30   cpu time limit exceeded (see getrlimit(2))",
    "SIGXFSZ   31   file size limit exceeded (see getrlimit(2))",
    
    "SIGWAITING 32  process' lwps are blocked",
    "SIGLWP    33   private to thread library",
    "SIGFREEZE 34   private to checkpoint/resume",
    "SIGTHAW   35   private to checkpoint/resume",
    "SIGRTMIN  36   first (high-priority) realtime signal",
    "SIGRTMAX  43   last (lowest-priority) realtime signal"
};

static char *shortSignalNames[] = {
    "0***", "HUP", "INT", "QUIT", "ILL", "TRAP", "ABRT", "EMT",
    "FPE", "KILL", "BUS", "SEGV", "SYS", "PIPE", "ALRM", "TERM",
    "USR1", "USR2", "CHLD", "PWR", "WINCH", "URG", "POLL", "STOP",
    "TSTP", "CONT", "TTIN", "TTOU", "VTALRM", "PROF", "XCPU",
    "XFSZ", "WAITING", "LWP", "FREEZE", "THAW", "RTMIN",
    "37**", "38**", "39**", "40**", "41**", "42**", "RTMAX"
};
    
/*
 * Local routine prototypes.
 */
void signalHandlerPanic(int, siginfo_t *, void *);
void signalHandlerDispatch(int, siginfo_t *, void *);

/*
 * Exported routines.
 */

/*
 * intrLock() -- Disable all interrupts.
 */
void
intrLock(void)
{
    sys_thread_t *cth = sysThreadSelf();
    machdep_t *machdepP;
    sigset_t newMask;
    
    if (cth) {
        machdepP = MACHDEP(cth);

        /* Sanity check */
	assert(machdepP->intrLockCount >= 0);
    
	if (machdepP->intrLockCount++ == 0) {
	    memset((char *)(&newMask), -1, sizeof (newMask));
	    sigprocmask(SIG_BLOCK, &newMask, &machdepP->intrLockOldMask);
	}
    } else {
	memset((char *)(&newMask), -1, sizeof (newMask));
	sigprocmask(SIG_BLOCK, &newMask, NULL);
    }
}

/*
 * intrUnlock() -- Enable all interrupts.
 */
void
intrUnlock(void)
{
    sys_thread_t *cth = sysThreadSelf();
    machdep_t *machdepP;
    sigset_t newMask;
    
    if (cth) {
        machdepP = MACHDEP(cth);

        /* Sanity check */
	assert(machdepP->intrLockCount > 0);
    
	if (--machdepP->intrLockCount == 0) {
	    sigprocmask(SIG_SETMASK, &machdepP->intrLockOldMask, NULL);
	}
    } else {
	memset((char *)(&newMask), 0, sizeof (newMask));
	sigprocmask(SIG_SETMASK, &newMask, NULL);
    }
}

/*
 * intrIsLocked() -- Return whether we have locked the interrupts
 *		     or not.
 */
bool_t
intrIsLocked(void)
{
    machdep_t *machdepP = MACHDEP(sysThreadSelf());
    
    return (machdepP->intrLockCount != 0);
}

/*
 * intrInitMD() -- Target-specific initialization.
 */
extern void
intrInitMD()
{
	struct sigaction	sigAct;
	stack_t		sigStack;
	int i;

	/*
	 * Install the panic handler for the signals that should cause us
	 * to panic.
	*/
	sigAct.sa_handler = NULL;
	sigAct.sa_sigaction = signalHandlerPanic;
	memset((char *)&(sigAct.sa_mask), 0, sizeof (sigset_t));
	sigaddset(&sigAct.sa_mask, SIGPIPE);
	sigAct.sa_flags = SA_SIGINFO | SA_RESTART;

	sigaction(SIGHUP, &sigAct, (struct sigaction *)0);
	sigaction(SIGQUIT, &sigAct, (struct sigaction *)0);
	sigaction(SIGILL, &sigAct, (struct sigaction *)0);  /* XXX: Exception */
	sigaction(SIGTRAP, &sigAct, (struct sigaction *)0); /* XXX: Exception */
	sigaction(SIGEMT, &sigAct, (struct sigaction *)0);  /* XXX: Exception */
	sigaction(SIGFPE, &sigAct, (struct sigaction *)0);  /* XXX: Exception */
	sigaction(SIGBUS, &sigAct, (struct sigaction *)0);  /* XXX: Exception */
	sigaction(SIGSEGV, &sigAct, (struct sigaction *)0); /* XXX: Exception */
	sigaction(SIGSYS, &sigAct, (struct sigaction *)0);  /* XXX: Exception */
	sigaction(SIGTERM, &sigAct, (struct sigaction *)0);
	sigaction(SIGXCPU, &sigAct, (struct sigaction *)0);
	sigaction(SIGXFSZ, &sigAct, (struct sigaction *)0);

	sigignore(SIGPIPE);

	if (sigprocmask(SIG_SETMASK, &(sigAct.sa_mask), NULL))
		perror("intrinitMD  - 2");
}

/*-
 * intrEnableDispatch() -- Enable a particular interrupt for dispatch
 *			   by installing signalDispatchHandler in its
 *			   sigvec.
 */
void
intrEnableDispatch(int interrupt)
{
	struct sigaction	sigAct;
    
	sigAct.sa_handler = 0;
	sigAct.sa_sigaction = signalHandlerDispatch;
	memset((char *)&(sigAct.sa_mask), -1, sizeof (sigset_t));
	sigAct.sa_flags = SA_SIGINFO | SA_RESTART;

	sigaction(interrupt, &sigAct, (struct sigaction *)0);
}

/*-
 * intrDisableDispatch() -- Disable a particular interrupt for dispatch
 *			    by setting its sigvec back to the default.
 */
void
intrDisableDispatch(int interrupt)
{
	struct sigaction	sigAct;

    /*
     * Set to the panic handler in case we get one when we shouldn't
     */
	memset((char *)&sigAct, -1, sizeof (sigAct));
	sigAct.sa_handler = NULL;
	sigAct.sa_sigaction = signalHandlerPanic;
	sigAct.sa_flags = SA_SIGINFO;
	sigaction(interrupt, &sigAct, (struct sigaction *)0);
}

/*
 * Local routines.
 */

/*
 * Decode a siginfo_t structure, and print out something interesting
 * about it.
 */
static char *
decode_si_code(int sig, siginfo_t *info)
{
    char *unknown = "unknown siginfo";
    static char msg[1024];
    int si_code;
    si_code = info->si_code;
    if (si_code == SI_NOINFO)
	return "no information";
    else if (SI_FROMUSER(info)) {
	char *si_user[] = {
	    "SI_USER", "SI_LWP", "SI_QUEUE",
	    "SI_TIMER", "SI_ASYNCIO", "SI_MESGQ"
	};
	if (si_code < SI_MESGQ) {
	    return unknown;
	}
	si_code = abs(si_code);
	sprintf(msg, "%s [pid: %d, uid: %d]", si_user[si_code],
	    info->si_pid, info->si_uid);
	return msg;
    } else {	/* SI_FROMKERNEL */
	switch(sig) {
	    case SIGILL: {
		char *si_code_name[] = {
		    "ILL_ILLOPC", "ILL_ILLOPN", "ILL_ILLADR",
		    "ILL_ILLTRP", "ILL_PRVOPC", "ILL_PRVREG",
		    "ILL_COPROC", "ILL_BADSTK"
		};
		if (si_code > NSIGILL)
		    return unknown;
		sprintf(msg, "%s [addr: 0x%x]\n",
			si_code_name[si_code], info->si_addr);
	    }
	    case SIGFPE: {
		char *si_code_name[] = {
		    "FPE_INTDIV", "FPE_INTOVF", "FPE_FLTDIV",
		    "FPE_FLTOVF", "FPE_FLTUND", "FPE_FLTRES",
		    "FPE_FLTINV", "FPE_FLTSUB"
		};
		if (si_code > NSIGFPE)
		    return unknown;
		sprintf(msg, "%s [addr: 0x%x]\n",
			si_code_name[si_code], info->si_addr);
		break;
	    }
	    case SIGSEGV: {
		char *si_code_name[] = {
		    "SEGV_MAPERR", "SEGV_ACCERR"
		};
		if (si_code > NSIGSEGV)
		    return unknown;
		sprintf(msg, "%s [addr: 0x%x]\n",
			si_code_name[si_code], info->si_addr);
		break;
	    }
	    case SIGBUS: {
		char *si_code_name[] = {
		    "BUS_ADRALN", "BUS_ADRERR", "BUS_OBJERR"
		};
		if (si_code > NSIGBUS)
		    return unknown;
		sprintf(msg, "%s [addr: 0x%x]\n",
			si_code_name[si_code], info->si_addr);
		break;
	    }
	    case SIGTRAP: {
		char *si_code_name[] = {
		    "TRAP_BRKPT", "TRAP_TRACE"
		};
		if (si_code > NSIGTRAP)
		    return unknown;
		sprintf(msg, "%s\n", si_code_name[si_code]);
		break;
	    }
	    case SIGCHLD: {
		char *si_code_name[] = {
		    "CLD_EXITED", "CLD_KILLED", "CLD_DUMPED",
		    "CLD_TRAPPED", "CLD_STOPPED", "CLD_CONTINUED"
		};
		if (si_code > NSIGCLD)
		    return unknown;
		sprintf(msg, "%s [pid: %d, status: %d]\n",
			si_code_name[si_code], info->si_pid, info->si_status);
		break;
	    }
	    case SIGPOLL: {
		char *si_code_name[] = {
		    "POLL_IN", "POLL_OUT", "POLL_MSG",
		    "POLL_ERR", "POLL_PRI", "POLL_HUP"
		};
		if (si_code > NSIGPOLL)
		    return unknown;
		sprintf(msg, "%s [band: %d]\n",
			si_code_name[si_code], info->si_band);
		break;
	    }
	    default:
		return unknown;
	}
	return msg;
    }
}

/*-
 * signalHandlerPanic() --  Deal with unexpected signals by dumping
 *			    out the state of the world, and then
 *			    panicing.
 */
void
signalHandlerPanic(int sig, siginfo_t *info, void *uc)
{
    extern int end, etext, edata;
    extern int _sbrk(int);
    extern void DumpMonitors();
    extern void DumpAlarmQ();
    static int	count = 0;

    if (sig != SIGQUIT && ++count > 1) {
	abort();
    }

    if (sig == SIGILL && info->si_code == ILL_ILLTRP) {
	int trapno = info->si_trapno;
	if ((trapno == ILL_TRAP_FAULT(TICC_NULL_POINTER_EXCEPTION)) || 
	    (trapno == ILL_TRAP_FAULT(TICC_ARRAY_INDEX_OUT_OF_BOUNDS))) {
	    extern CompiledCodeError_md();
	    ucontext_t *cp = uc;
	    if (trapno == 128 + TICC_NULL_POINTER_EXCEPTION)
		SignalError(0, JAVAPKG "NullPointerException", 0);
	    else 
		SignalError(0, JAVAPKG "ArrayIndexOutOfBoundsException", 0);
	    cp->uc_mcontext.gregs[REG_PC] = 
		    (long)(void *)(CompiledCodeError_md);
	    cp->uc_mcontext.gregs[REG_nPC] = cp->uc_mcontext.gregs[REG_PC]+4;
	    count = 0;
	    setcontext(cp);
	}
    }

    if (sig > MAXSIG) {
	fprintf(stderr, "Unknown signal: %d\n", sig);
    } else {
	fprintf(stderr, "%s\n", signalNames[sig]);
    }

    if (info) {
	/* print out signal info */
	if (info->si_code != SI_NOINFO) {
	    fprintf(stderr, "    si_signo [%d]: %s\n",
		    info->si_signo, signalNames[info->si_signo]);
	    fprintf(stderr, "    si_errno [%d]: %s\n",
		    info->si_errno, strerror(info->si_errno));
	    fprintf(stderr, "    si_code [%d]: %s\n",
		    info->si_code, decode_si_code(sig, info));
	}

        if (sysThreadSelf()) {
            fprintf(stderr, "\tstackbase=%X, stackpointer=%X\n",
                    sysThreadStackBase(sysThreadSelf()),
                    sysThreadStackPointer(sysThreadSelf()));
        }
    }

    DumpThreads();
    DumpMonitors();
    DumpAlarmQ();
    if (sig != SIGQUIT) {
	abort();
    }
}

/*-
 * signalHandlerDispatch() -- Turn our signal into an intrDispatch().
 */
void
signalHandlerDispatch(int sig, siginfo_t *info, void *uc)
{
    int		old_errno;
    static	int recursive;

    Log1(1, "signalHandlerDispatch(sig=%d)\n", sig);
    if (sig == 11)
	    exit(1);
    old_errno = errno;

    /*
     * Can't intrLock() if gc is running because TID is dereferenced!
     * So just process the interrupt without updating recursive or
     * responding to needReschedule.
     *
     * FIXME: this is too strict -- fix when gc becomes incremental.
     */
    if (SCHED_LOCKED()) {
	intrDispatch(sig, (void *) info);
	errno = old_errno;
    	return;
    }

    intrLock();
    recursive++;
    intrUnlock();

    intrDispatch(sig, (void *) info);

    errno = old_errno;
    
    intrLock();
    recursive--;
    intrUnlock();

    /*
     * We are not preempting the current thread.  Just return
     * and let the signal trampoline restore state.
     * Also, can't yield here until the entire interrupt stack
     * is unwound (hence the test of "recursive").
     */
    if (recursive || MACHDEP(sysThreadSelf())->needReschedule == 0) {
    	return;
    }

    MACHDEP(sysThreadSelf())->needReschedule = 0;

    /*
     * Remember that this thread needs to be resumed later
     * via fullSwitchContext().
     */
    sysThreadSelf()->full_switch = 1;

    SCHED_LOCK();
    YIELD(ASYNCHRONOUS);
    SCHED_UNLOCK();
}
