/*
 * @(#)iomgr.c	1.34 95/05/11  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * I/O management, including non-blocking io.
 *
 * Eventually this module needs to have a formal interface, and
 * be replacable (e.g. it needs to be a porting layer).
 */

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/fcntl.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <errno.h>
#include <sys/filio.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <sys/syscall.h>
#include <limits_md.h>
#include <stdio.h>

#include <stropts.h>
#include <sys/conf.h>
#include <poll.h>

#include <threads.h>
#include <threads_md.h>
#include <monitor_md.h>
#include <machapi.h>
#include <debug.h>
#include <timeval.h>
#include <iomgr.h>

extern int errno;

/* This isn't in syscall.h, for some reason. */
int syscall (int, ...);
int _select (int width, fd_set *, fd_set *, fd_set *, struct timeval *);


/*
 * nb file descriptor state
 */
static char nb_flags[OPEN_MAX];
static int max_nbfd;
#define FD_NBINIT       0x1
#define FD_IOPENDING    0x2

/*
 * SYS_ASYNC_MON_IO notification
 */
static int nb_init;


static void
nbInit()
{
    asyncIORegister();
    nb_init = 1;
    return;
}

/*
 * Returns non-zero if an exception is pending for this ee.
 */
int pendingException() { 
    struct execenv *ee = EE(); 
    return (ee && exceptionOccurred(ee));
}

/*
 *  If the second argument is true the fd is put into non-blocking
 *  mode.  That is,  a read when there is no data will not stop
 *  the entire heavyweight process.
 *  Frankly I can't imagine a situation where blocking IO is what is
 *  desired.  Should all fd's opened in a multithreaded environment be
 *  nonblocking? -csw
 *
 */
int
nonblock_io(int desc, int onoff)
{
    long flgs;

    if (!nb_init)
	nbInit();

    if (desc > OPEN_MAX) {
	fprintf(stderr, "out of file descriptors!\n");
	abort();
    }

    if (onoff == IO_DONTBLOCK) {  /* put into non-blocking mode... */

	/* if it's already non-blocking, do nothing... */
	if (nb_flags[desc] & FD_NBINIT)
	    return 0;

	if ((flgs = fcntl(desc, F_GETFL, 0)) == -1)
	    return -1;
	fcntl(desc, F_SETFL, flgs | O_NONBLOCK);
	ioctl(desc, I_SETSIG, S_INPUT | S_OUTPUT);

	nb_flags[desc] = FD_NBINIT;
	if (desc > max_nbfd)
	    max_nbfd = desc;
    } else {           /* put into blocking mode... */

	/* if it's already blocking, do nothing... */
	if (!(nb_flags[desc] & FD_NBINIT))
	    return 0;

	if ((flgs = fcntl(desc, F_GETFL, 0)) == -1)
	    return -1;
	fcntl(desc, F_SETFL, flgs & ~O_NONBLOCK);
	ioctl(desc, I_SETSIG, 0);

	nb_flags[desc] &= ~FD_NBINIT;
	if (desc > max_nbfd)
	    max_nbfd = desc;
    }
    return 0;
}

int
close(int fd)
{
    if (nb_flags[fd] & FD_NBINIT)
	nb_flags[fd] = 0;
    
    return syscall(SYS_close, fd);
}

#define	check_single_fd(fd)		    \
    if ((nb_flags[fd] & FD_NBINIT) == 0) {    \
	nonblock_io(fd, IO_DONTBLOCK);	    \
    }

int
write_noblock(int fd, caddr_t buf, int nbytes)
{
    register int nwrote;

    if (nbytes < 0)
	return nbytes;
    check_single_fd(fd);
    while ((nwrote = syscall(SYS_write, fd, buf, nbytes)) < 0) {
	if (errno == EAGAIN)
	    return 0;
	if (errno != EINTR)
	    return -1;
    }
    return nwrote;
}

ssize_t
recv(int fd, char *buf, int nbytes, int flags)
{
    return read(fd, (void *) buf, (uint) nbytes);
}


ssize_t
send(int fd, const char *buf, int nbytes, int flags)
{
    return write(fd, buf, (uint) nbytes);
}

ssize_t
read(int fd, void *buf, unsigned int nbytes)
{
    register int nread;

    check_single_fd(fd);
    sysMonitorEnter(asyncMon(SYS_ASYNC_MON_INPUT));
    nread = syscall(SYS_read, fd, buf, nbytes);
    while (nread == -1 && !pendingException() 
	   && (errno == EAGAIN || errno == EINTR)) {
	if (errno == EAGAIN)
	    sysMonitorWait(asyncMon(SYS_ASYNC_MON_INPUT), 2000);
	nread = syscall(SYS_read, fd, buf, nbytes);
    }
    sysMonitorExit(asyncMon(SYS_ASYNC_MON_INPUT));
    return nread;
}

ssize_t
write(int fd, const void *buf, unsigned int nbytes)
{
    register int nwrote;
    register caddr_t next = (caddr_t)buf;
    register int total_written = 0;

    /* don't muck with stdin, stdout and stderr, since it breaks
       many shells (when java exits) */
    if (fd > 2) {
	check_single_fd(fd);
    }
    sysMonitorEnter(asyncMon(SYS_ASYNC_MON_OUTPUT));

    while (total_written < nbytes && !pendingException()) {
	nwrote = syscall(SYS_write, fd,
			 next + total_written,
			 nbytes - total_written);

	while (nwrote == -1 && !pendingException() ) {
	    if (errno == EAGAIN)
		sysMonitorWait(asyncMon(SYS_ASYNC_MON_OUTPUT), 250);
	    else if (errno != EINTR) {
		total_written = -1;
		goto unlock;
	    }		
	    nwrote = syscall(SYS_write, fd,
			     next + total_written,
			     nbytes - total_written);
	}
	total_written += nwrote;
    }
unlock:
    sysMonitorExit(asyncMon(SYS_ASYNC_MON_OUTPUT));
    return total_written;
}



/* AVH -- added the writev system call, it is used
 *        by some libraries like X11.
 */

int
writev(int fd, const struct iovec *iovec, int iovcnt)
{
    int total_written = 0;
    int nwrote, ndone = 0;
    struct iovec *iov = (struct iovec *)iovec;

    check_single_fd(fd);
    sysMonitorEnter(asyncMon(SYS_ASYNC_MON_OUTPUT));

    while (iovcnt > 0 && !pendingException() ) {
	iov->iov_len -= ndone;
	iov->iov_base += ndone;
	nwrote = syscall(SYS_writev, fd, iov, iovcnt);

	while (nwrote == -1) {
	    if (errno == EAGAIN) {
		sysMonitorWait(asyncMon(SYS_ASYNC_MON_OUTPUT), 250);
	    } else if (errno != EINTR) {
		total_written = -1;
		goto unlock;
	    }

	    nwrote = syscall(SYS_writev, fd, iov, iovcnt);
	}
	total_written += nwrote;
	iov->iov_len += ndone;
	iov->iov_base -= ndone;
	ndone += nwrote;
	for (; (iovcnt > 0) && (ndone >= iov->iov_len) ; iovcnt--, iov++) {
	    ndone -= iov->iov_len;
	}
    }
unlock:
    sysMonitorExit(asyncMon(SYS_ASYNC_MON_OUTPUT));
    return total_written;
}

/* AVH -- added the readv system call, it is used
 *        by some libraries like X11.
 */
int
readv(int fd, struct iovec *iov, int iovcnt)
{
    int nread;
    int	fds[1];

    check_single_fd(fd);
    sysMonitorEnter(asyncMon(SYS_ASYNC_MON_INPUT));

    nread = syscall(SYS_readv, fd, iov, iovcnt);
    while ((nread == -1) && !pendingException() && ((errno == EAGAIN) || (errno == EINTR))) {
	if (errno == EAGAIN) {
	    sysMonitorWait(asyncMon(SYS_ASYNC_MON_INPUT), 2000);
	}
	nread = syscall(SYS_readv, fd, iov, iovcnt);
    }
    sysMonitorExit(asyncMon(SYS_ASYNC_MON_INPUT));
    return nread;
}

int
getmsg(int fd, struct strbuf *ctlptr, struct strbuf *dataptr, int *flags)
{
    int retval;

    sysMonitorEnter(asyncMon(SYS_ASYNC_MON_INPUT));
    retval = syscall(SYS_getmsg, fd, ctlptr, dataptr, flags);
    while (retval == -1 && !pendingException() && (errno == EAGAIN || errno == EINTR)) {
	if (errno == EAGAIN)
	    sysMonitorWait(asyncMon(SYS_ASYNC_MON_INPUT), SYS_TIMEOUT_INFINITY);
	retval = syscall(SYS_getmsg, fd, ctlptr, dataptr, flags);
    }
    sysMonitorExit(asyncMon(SYS_ASYNC_MON_INPUT));
    return retval;
}

int
putmsg(int fd, const struct strbuf *ctlptr,
       const struct strbuf *dataptr, int flags)
{
    int retval;

    sysMonitorEnter(asyncMon(SYS_ASYNC_MON_OUTPUT));
    retval = syscall(SYS_putmsg, fd, ctlptr, dataptr, flags);
    while (retval == -1 && !pendingException() && (errno == EAGAIN || errno == EINTR)) {
	if (errno == EAGAIN)
	    sysMonitorWait(asyncMon(SYS_ASYNC_MON_OUTPUT), SYS_TIMEOUT_INFINITY);
	retval = syscall(SYS_putmsg, fd, ctlptr, dataptr, flags);
    }
    sysMonitorExit(asyncMon(SYS_ASYNC_MON_OUTPUT));
    return retval;
}




/*
 * On SunOS 2.3 the connect routine (really a library call) can
 * sometimes return an error and errno == EPROTO.  in this case it
 * is pointless to restart the connect call as it will just fail
 * again with EPROTO.  I don't know what causes this but it seem
 * like for now that the only way to fix the problem is to shut the
 * socket down and open it up again.  We report the error to Java and
 * it can try to restart the connection.  Also, if the connection is
 * refused, it seems that select() returns 1, as if everything is
 * OK, but there's no way to find out that the connection failed
 * until you try something!  So, now we try something immediately,
 * and see what we get.  What this boils down to is socket "system
 * calls" do not work when the socket is in nonblocking I/O mode.
 */

int
sysConnect(int s, struct sockaddr *him, int len)
{
    int	    result;
    int	    delay = 50;	/* sleep 50 ms and try again */
    int	    totalTime = 0;
    long    flgs;
    fd_set  write_fds;
    struct timeval  poll;
    extern int _select (int width, fd_set *, fd_set *, fd_set *, struct timeval *);

    FD_ZERO(&write_fds);
    FD_SET(s, &write_fds);
    poll.tv_sec = poll.tv_usec = 0;

    sysMonitorEnter(asyncMon(SYS_ASYNC_MON_IO));
    result = connect(s, him, len);
    while (result < 0 && !pendingException() ) {
	if (errno == EINPROGRESS) {
	    int	cnt = _select(s+1, 0, &write_fds, 0, &poll);

	    FD_SET(s, &write_fds);
	    if (cnt == 1) {
		int bytes;

		if (ioctl(s, FIONREAD, &bytes) == -1) {
		    if (errno == EPIPE) {
			errno = ECONNREFUSED;
		    }
		    /* If this fails with EPIPE, then we assume
		       connection is refused - who knows what's
		       right?  Not I. */
		    assert(result < 0);
		    break;
		}
		result = 0;
		break;
	    }
	    sysMonitorWait(asyncMon(SYS_ASYNC_MON_IO), delay);
	} else {
	    break;
	}
    }
    sysMonitorExit(asyncMon(SYS_ASYNC_MON_IO));
    return result;
}


/* Poll system call */

int
poll(struct pollfd *fds, unsigned long nfds, int timeout)
{
    timeval_t time_limit;
    int	    poll_timeout = 0;
    int	    retval;
    int	    fdArray[OPEN_MAX];
    int	    i, fd, fdi = 0;

    for (i = nfds; --i >= 0; ) {
	if ((fd = fds[i].fd) < OPEN_MAX) {
	    check_single_fd(fd);
	    fdArray[fdi] = fd;
	    if ((fdi += 1) >= OPEN_MAX)
		break;
	}
    }

    if (timeout > 0) {
	timeval_t tv;
	tv.tv_sec = timeout / 1000;
	tv.tv_usec = timeout % 1000;
	time_limit = currentTime();
	timerAdd(&time_limit, (timeval_t *) &tv);
    }

    sysMonitorEnter(asyncMon(SYS_ASYNC_MON_IO));
    while (!pendingException() ) {
	timeval_t delay;

	errno = 0;
	retval = syscall(SYS_poll, fds, nfds, poll_timeout);

	if (retval == -1)
	    fprintf(stderr, "poll => -1; errno = %d\n", errno);

	if (retval != 0)
	    break;
	if (timeout != 0) {
	    /* would have blocked, see whether timer has expired */
	    timeval_t now = currentTime();

	    if (timerCmp(&now, &time_limit, >) || timerEq(&now, &time_limit)) {
		retval = 0;
		break;
	    }
	    delay = time_limit;
	    timerSub(&delay, &now);
	    timeout = (delay.tv_sec * 1000) + (delay.tv_usec / 1000);
	}
	sysMonitorWait(asyncMon(SYS_ASYNC_MON_IO), timeout > 0 ? timeout : SYS_TIMEOUT_INFINITY);
    }
    sysMonitorExit(asyncMon(SYS_ASYNC_MON_IO));
    return retval;
}

/* Select system call in a multi-threaded environment.

   This works by affecting a poll.  If _select returns
   non-negative, then we return that value.  Otherwise, we block
   until *any* IO is ready, and then try again.  If we still come
   up empty then we block for a smaller amount of time this time.
   If we reach the time limit, then we just return 0, as select
   normally would. */

int
select(int width, fd_set *readfds, fd_set *writefds,
       fd_set *exceptfds, struct timeval *tv)
{
    struct timeval  poll;
    int             nfds = -1;
    timeval_t       time_limit;
    fd_set          reads, writes, excepts;
    int		    fdArray[OPEN_MAX];
    int		    i, fdi = 0;

    for (i = 0; i < width; i++) {
	if (i >= OPEN_MAX)
	    break;
	if ((readfds != 0 && FD_ISSET(i, readfds))
	    || (writefds != 0 && FD_ISSET(i, writefds))
	    || (exceptfds != 0 && FD_ISSET(i, exceptfds))) {
	    fdArray[fdi] = i;
	    check_single_fd(i);
	}
    }

    poll.tv_sec = poll.tv_usec = 0;
    if (tv != 0) {
	time_limit = currentTime();
	timerAdd(&time_limit, (timeval_t*)tv);
    }

    if (readfds != 0)
	reads = *readfds;
    if (writefds != 0)
	writes = *writefds;
    if (exceptfds != 0)
	excepts = *exceptfds;

    sysMonitorEnter(asyncMon(SYS_ASYNC_MON_IO));
    while (!pendingException()) {
	timeval_t delay;
	int timeout = SYS_TIMEOUT_INFINITY;

	nfds = _select(width, readfds, writefds, exceptfds, &poll);

	if (nfds != 0) {
	    break;
	}
	if (tv != 0) {
	    timeval_t now = currentTime();

	    if (timerCmp(&now, &time_limit, >) || timerEq(&now, &time_limit)) {
		nfds = 0;
		break;
	    }
	    delay = time_limit;
	    timerSub(&delay, &now);
	    timeout = (delay.tv_sec * 1000) + (delay.tv_usec / 1000);
	}
	sysMonitorWait(asyncMon(SYS_ASYNC_MON_IO), timeout);
	if (readfds != 0)
	    *readfds = reads;
	if (writefds != 0)
	    *writefds = writes;
	if (exceptfds != 0)
	    *exceptfds = excepts;
    }
    sysMonitorExit(asyncMon(SYS_ASYNC_MON_IO));
    return nfds;
}


/*
 * Although we set the SA_RESTART flag on all our signal handlers we
 * still see open() occasionally return with errno == EINTR, at
 * least on SunOS 2.3.  This routine is an attempt to get aorund
 * this bug by making sure open doesnot return until it either
 * succeeds or errno != EINTR.
 */
int
open(const char *fname, int flags, ...)
{
    va_list ap;
    register int fd;
    unsigned int mode;

    va_start(ap, &flags);
    if (flags & O_CREAT) {
	mode = va_arg(ap, unsigned int);
    } else {
	mode = 0;
    }
    va_end(ap);
    fd = syscall(SYS_open, fname, flags, mode);
    while (fd == -1 && !pendingException() && (errno == EAGAIN || errno == EINTR)) {
	fd = syscall(SYS_open, fname, flags, mode);
    }
    return fd;
}

