/*
 * @(#)mem.c	1.4 95/04/28
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Implementation of the basic chunk (called a "mem", used for e.g.
 * a thread stack or the heap) memory allocator.  This is a simpler
 * rip-off of the earlier range allocator.
 *
 * Mems are allocated using a first fit algorithm.  As needed, mems
 * are split and changed (i.e. allocated, freed, etc).
 *
 * The only thing machine dependent about this allocator is how it 
 * initially finds all of the possible memory, and how it implements
 * mapChunk() and unmapChunk().
 */

#include <sys_api.h>

#include <stdio.h>	/* GET RID OF THIS! */

/*
 * Memory allocation lock: just one shared by allocate and free, as
 * the mapped memory allocator is currently only used to allocate
 * thread stacks, i.e. relatively rarely.
 */
sys_mon_t _alloc_lock;
#define ALLOC_LOCK_INIT()    monitorRegister(&_alloc_lock, "Allocation lock")
#define ALLOC_LOCK()         sysMonitorEnter(&_alloc_lock)
#define ALLOC_UNLOCK()       sysMonitorExit(&_alloc_lock)

/*
 * The pre-defined mem owners.  More can be added (up to 2^16 cases).
 * It's not clear whether this concept should be retained.
 */
#define MEM_NONE 	0	/* Unowned */
#define MEM_ALLOC 	1	/* Owned by memory allocator itself */
#define MEM_JAVA 	2	/* Owned by Java */

/*
 * A description of a basic block of memory that the system provides.
 */
typedef struct {
    char *base;
    unsigned int length;
    union {
	struct {
    	    unsigned int owner:16;	/* zero is reserved */
	    unsigned int allocated:1;
	    unsigned int :0;
	} flags;
	unsigned int mask; /* allows faster comparisons of all flags */ 
    } u;
} Mem;

/*
 * The MemNode is a linked list node that contains a Mem.  They
 * are allocated out of a MemBuffer.
 */
typedef struct _MemNode {
    struct _MemNode *next;
    struct _MemNode *prev;
    Mem mem;
} MemNode;

/*
 * The MemBuffer is the structure in which we keep track of allocation.
 * It is an array of allocated mem nodes, with a small header at the top
 * to allow chaining of multiple buffers.  This allows us to grow the
 * number of Mems that we keep track of dynamically.
 *
 * The current implementation stores mems in a sorted (by address)
 * linear array.
 *
 * MEMBUFLENGTH defines the maximum number of Mems which can exist in
 * a MemBuffer.
 */
typedef struct _MemBuffer {
    struct _MemBuffer *next;
    unsigned int used;
    MemNode mem_node[1];
} MemBuffer;

#define MEMBUFLENGTH \
  ((mem_grain_size - sizeof(MemBuffer) + sizeof(MemNode)) \
	/ sizeof(MemNode))

static MemBuffer *memBufferHead;
static MemNode *memNodeHead;
static unsigned int mem_nodes_free;
unsigned int mem_grain_size;
unsigned int mems_length;
char *mems_start;

/*
 * Find the first MemNode which will completely encompass the specified
 * memory span, and is owned by owner.
 */
static MemNode *
findMemNode(char *addr, unsigned int length, unsigned short owner,
	    int allocated)
{
    MemNode *rn;

    rn = memNodeHead;

    if (addr) {
	if (rn->mem.base > addr) {
	    return 0;
	}
	while (rn) {
	    if ((rn->mem.base + rn->mem.length) >= (addr + length)) {
		if (rn->mem.base <= addr
		    && rn->mem.u.flags.owner == owner
		    && rn->mem.u.flags.allocated == allocated) {
		    return rn;
		} else {
		    return 0;
		}
	    }
	    rn = rn->next;
	}

    } else {
	while (rn) {
	    if (rn->mem.length >= length
		&& rn->mem.u.flags.allocated == allocated
		&& rn->mem.u.flags.owner == owner) {
		return rn;
	    }
	    rn = rn->next;
	}
    }
    return 0;
}

/*
 * Allocate a single mem node from any of the currently allocated
 * mem buffers.  If there is no space, allocate a new mem buffer.
 */
static MemNode *
allocateMemNode()
{
    MemBuffer *rb;
    MemNode *rn;

    rb = memBufferHead;
    while (rb) {
	if (rb->used < MEMBUFLENGTH) {
	    rb->used++;
	    mem_nodes_free--;
	    rn = &(rb->mem_node[0]);
	    do {
		if (rn->next == 0 && rn->prev == 0) {
		    rn->next = (MemNode *) 0x1;
		    return rn;
		}
		rn++;
	    } while (1);
	} else {
	    rb = rb->next;
	}
    }

    /* This should never happen unless we are really out of memory */
    return 0;
}

/*
 * Free the specified mem node, and mark the mem buffer as having
 * another free slot.
 *
 * XXX - do we ever free a MemBuffer once it is empty?
 */
static void
freeMemNode(MemNode *rn)
{
    MemBuffer *rb;

    rn->next = 0;
    rn->prev = 0;

    /*
     * We assume that mem buffers are allocated on 'mem_grain_size'
     * aligned blocks of memory, and calculate this mem nodes
     * mem buffer by just rounding down to the nearest min_size
     * aligned block.
     */
    rb = (MemBuffer *) ((unsigned int) rn & ~(mem_grain_size - 1));
    rb->used--;
    mem_nodes_free++;
}

/*
 * Coalesce a given mem node into its neighbors.  It may actually
 * be possible to merge it into both its next and prev.
 *
 * If the mem is empty (length == 0), just remove it from the 
 * queue, and try to merge its neighbors.
 */
static void
coalesceMemNode(MemNode *rn)
{
    MemNode *prev, *next;

    /* try to merge with previous mem node */
    if (rn->prev
	&& rn->mem.u.mask == rn->prev->mem.u.mask
	&& rn->prev->mem.base + rn->prev->mem.length == rn->mem.base) {
	prev = rn->prev;
	prev->mem.length += rn->mem.length;
	prev->next = rn->next;
	if (rn->next) {
	    rn->next->prev = prev;
	}
	freeMemNode(rn);
	rn = prev;
    }

    /* if length == 0, remove from queue */
    if (rn->mem.length == 0) {
	if (rn->prev) {
	    prev = rn->prev;
	    rn->prev->next = rn->next;
	} else {
	    prev = rn->next;
	    memNodeHead = rn->next;
	}
	if (rn->next) {
	    rn->next->prev = rn->prev;
	} /* we don't keep track of the mem node tail */

	freeMemNode(rn);
	rn = prev;
    }

    /* try to merge with next mem node */
    if (rn->next
	&& rn->mem.u.mask == rn->next->mem.u.mask
	&& rn->next->mem.base == rn->mem.base + rn->mem.length) {
	next = rn->next;
	rn->mem.length += next->mem.length;
	rn->next = next->next;
	if (next->next) {
	    next->next->prev = rn;
	}
	freeMemNode(next);
    }
}

/*
 * This routine is called after every mem property change, and 
 * is responsible for maintaining enough mem buffer space to handle
 * any future allocations.  At this point, all this means is ensuring
 * that there is at least three free mem nodes in the system (two
 * to allow change_mem_property() to allocate head and tail, and one
 * to allow a new mem buffer immediately after the change_mem_property).
 *
 * We always keep at least two free so that we don't have to do allocation
 * in the middle of change_mem_properties(), which be overly complex.
 *
 * If there are not two free mem nodes, allocate a new mem buffer.
 */
static void
maintainBufferSpace()
{
    MemBuffer *rb;
    MemNode *rn, *rntail;

    if (mem_nodes_free < 3) {
	rn = findMemNode(0, mem_grain_size, MEM_NONE, 0);
	if (!rn) {    /* out of memory */
	    return;
	}

	/* out of backing store */
	if (!mapChunk(rn->mem.base, mem_grain_size)) {
	    return;
	}

	rntail = 0;
	if (rn->mem.length > mem_grain_size) {
	    /* Split this mem */
	    rntail = allocateMemNode();
	    if (!rntail) {
		unmapChunk(rn->mem.base, mem_grain_size);
		return;
	    }

	    memcpy((char *)rntail, (char *)rn, sizeof(MemNode));
	    rntail->mem.length -= mem_grain_size;
	    rntail->mem.base += mem_grain_size;
	    rn->mem.length = mem_grain_size;
	    rntail->prev = rn;
	    rntail->next = rn->next;
	    if (rn->next) {
		rn->next->prev = rntail;
	    }
	    rn->next = rntail;
	}

	rn->mem.u.flags.allocated = 1;
	memset(rn->mem.base, 0, mem_grain_size);
	rb = (MemBuffer *) rn->mem.base;

	rb->next = memBufferHead;
	memBufferHead = rb;
	mem_nodes_free += MEMBUFLENGTH;

	if (rntail) {
	    coalesceMemNode(rntail);
	}
	coalesceMemNode(rn);
    }
}

/*
 * Change the properties on the specified memory mem.  This may
 * involve splitting and coallescing mems.   If this memory span
 * intersects multiple mems, fail (return 0).
 *
 * If any of the old properties are incorrect, fail.  addr is normally
 * zero, meaning that we don't care where the mem sits, except when
 * we have made a change that we want to change *back*.
 */
static char *
changeMemProperty(char * addr, unsigned int length,
		      unsigned short old_owner, unsigned short new_owner,
		      int allocated)
{
    MemNode *rn, *rnhead, *rntail;
    MemNode *next, *prev;

    rn = findMemNode(addr, length, old_owner, !allocated);
    if (!rn) {
	return 0;
    }

    if (addr == 0) {
        addr = rn->mem.base;
    }

    /* Pessimistically preallocate two new mem nodes - worst case */
    if (!(rnhead = allocateMemNode())) {
	return 0;
    }
    if (!(rntail = allocateMemNode())) {
	freeMemNode(rnhead);
	return 0;
    }

    /*
     * Now we have the mem which completely encompasses the [addr,length]
     * mem we want to change the property upon.  Changing the properties on
     * this span may require that the parent be split into multiple mems.
     */
    memcpy((char *)&(rntail->mem), (char *)&(rn->mem), sizeof(Mem));
    memcpy((char *)&(rnhead->mem), (char *)&(rn->mem), sizeof(Mem));
    rnhead->mem.length = addr - rn->mem.base;
    rntail->mem.length -= (rnhead->mem.length + length);
    rntail->mem.base += (rnhead->mem.length + length);

    rn->mem.base += rnhead->mem.length;
    rn->mem.length = length;
    rn->mem.u.flags.owner = new_owner;
    rn->mem.u.flags.allocated = allocated;

    rntail->next = rn->next;
    rntail->prev = rn;
    rnhead->next = rn;
    rnhead->prev = rn->prev;
    if (rn->next) {
	rn->next->prev = rntail;
    }
    if (rn->prev) {
	rn->prev->next = rnhead;
    }
    rn->next = rntail;
    rn->prev = rnhead;

    /*
     * Now coalesce all mems.  If we can't coalesce a mem,
     * insert it in the mem list.
     */
    coalesceMemNode(rntail);
    coalesceMemNode(rnhead);
    coalesceMemNode(rn);
    maintainBufferSpace();
    return addr;
}

/*
 * Mem size rounding is done at this level.  The thread/context
 * creation code just asks for what it thinks it wants.
 */
static int
roundMemSize(int requested_size)
{
    if (requested_size < mem_grain_size) {
	requested_size = mem_grain_size;
    } 

    return (requested_size + mem_grain_size - 1) & ~(mem_grain_size - 1);
}

void
InitializeMems()
{
    static int init = 0;
    MemNode *rn;

    if (init) {
	return;		/* Subsequent calls are no-ops */
    }

    InitMapping();

    ALLOC_LOCK_INIT();

    if (!mapChunk(mems_start, mem_grain_size)) {
	out_of_memory();
    }
    memset(mems_start, 0, mem_grain_size);

    /* Initialize the mem buffer */
    memBufferHead = (MemBuffer *) mems_start;
    memBufferHead->used = 2;
    memBufferHead->next = 0;

    /* Initialize the mem nodes */
    memNodeHead = &(memBufferHead->mem_node[0]);
    rn = memNodeHead;
    mem_nodes_free = MEMBUFLENGTH - 2;

    rn->mem.base = mems_start;
    rn->mem.length = mem_grain_size;
    rn->mem.u.flags.owner = MEM_ALLOC;
    rn->mem.u.flags.allocated = 1;
    rn->prev = 0;
    rn->next = (rn + 1);

    rn++;
    rn->mem.base = mems_start + mem_grain_size;
    rn->mem.length = mems_length - mem_grain_size;
    rn->mem.u.flags.owner = MEM_NONE;
    rn->mem.u.flags.allocated = 0;
    rn->prev = memNodeHead;
    rn->next = 0;

    init = 1;		/* We're initialized now */
}

/*
 * Allocate a mem of memory.  If any of the mem is already allocated,
 * return NULL.  Else return a pointer to the memory.  Note that the size
 * asked for here is precisely what the upper level has asked for.  We
 * need to do any rounding, etc. here.
 *
 * I suppose we could try to be smarter here and wait if we can't get
 * memory.
 */
void *
sysAllocateMem(long requested_size)
{
    long size;
    char * addr;

    size = roundMemSize(requested_size);
    ALLOC_LOCK();
    addr = changeMemProperty(0, size, MEM_NONE, MEM_JAVA, 1);
    if (addr && addr != (char *) -1) {
	if (!mapChunk(addr, size)) {
	    changeMemProperty(addr, size, MEM_JAVA, MEM_NONE, 0);
	    addr = 0;
	}
    }
    ALLOC_UNLOCK();
    return addr;
}

/*
 * Free the specified mem. If all of this mem is not already allocated,
 * return -1.  Otherwise, free it and return 0.  Note that the size asked
 * for here is what the upper level has asked for.  We need to determine
 * what was actually allocated.
 */
int
sysFreeMem(char *addr, long requested_size)
{
    long size;
    int ret;

    size = roundMemSize(requested_size);
    ALLOC_LOCK();
    if (changeMemProperty(addr, size, MEM_JAVA, MEM_NONE, 0) == addr) {
	unmapChunk(addr, size);
	ret = 0;
    } else {
	ret = -1;
    }
    ALLOC_UNLOCK();
    return ret;
}

#ifdef	DEBUG

#include <stdio.h>

/*
 * Return the total number of mems in existance
 */
int
memNumMems(void)
{
    MemBuffer *rb;
    int total_slots = 0;

    rb = memBufferHead;
    while (rb) {
	total_slots += rb->used;
	rb = rb->next;
    }
    return total_slots;
}

/*
 * Dump the mem tables, in address order.
 */
void
DumpMemTable()
{
    MemNode *rn = memNodeHead;

    while (rn) {
	fprintf(stderr, "[0x%x, 0x%x] %s",
	    rn->mem.base, rn->mem.length,
	    rn->mem.u.flags.allocated ? "allocated" : "unallocated");
	switch (rn->mem.u.flags.owner) {
	    case MEM_NONE:
		fprintf(stderr, " (owner: none)\n");
		break;
	    case MEM_JAVA:
		fprintf(stderr, " (owner: java)\n");
		break;
	    case MEM_ALLOC:
		fprintf(stderr, " (owner: allocator)\n");
		break;
	    default:
		fprintf(stderr, " (owner: %d)\n",
			(int) rn->mem.u.flags.owner);
		break;
	}

	rn = rn->next;
    }
}

void
verifyMems()
{
    MemBuffer *rb = memBufferHead;
    MemNode *rn = memNodeHead;
    int i, count, total, alloced;

    while (1) {
	if (rn->next) {
	    rn = rn->next;
	} else {
	    break;
	}
    }

    total = alloced = 0;
    while (rb) {
	count = 0;
	for (i = 0; i < MEMBUFLENGTH; i++) {
	    if (rb->mem_node[i].next || rb->mem_node[i].prev) {
		count++;
	    }
	}
	total += count;
	alloced += MEMBUFLENGTH;
	rb = rb->next;
    }
}
#endif
