/*
 * @(#)monitor_md.c	1.25 95/02/16
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Monitor implementation for Green threads
 */

#include <typedefs.h>
#include <timeval.h>
#include <exceptions.h>
#include <interrupt.h>
#include <machapi.h>
#include <schedule.h>
#include <queue.h>
#include <clock.h>
#include <log.h>

#include <threads_md.h>
#include <monitor_md.h>

#include <sys_api.h>

/**
 ** Queue maniplation code
 **/

void
queueInsert(queue_t *q, sys_thread_t *waiter)
{
    sys_thread_t *t;
    sys_thread_t *p;

    t = *q;
    p = 0;
    while (t && (t->priority >= waiter->priority)) {
	p = t;
	t = t->waitq;
    }

    if (p) {
	p->waitq = waiter;
    } else {
	*q = waiter;
    }

    waiter->waitq = t;
    return;
}

int
queueRemove(queue_t *q, sys_thread_t *waiter)
{
    sys_thread_t *t;
    sys_thread_t *p;

    t = *q;
    p = 0;
    while (t && t != waiter) {
	p = t;
	t = t->waitq;
    }

    if (t) {
	if (p) {
	    p->waitq = t->waitq;
	} else {
	    *q = t->waitq;
	}
	return 1;
    } else {
	return 0;
    }
}


/*
 * This function signals a specific queue.  If the queue is the monitor's wait
 * queue, the thread is put on the runnable queue.  If the queue is not
 * the monitor's wait queue, it is put on the wait queue.
 */
int
queueSignal(sys_mon_t *mid, queue_t *q)
{
    sys_thread_t *waiter;
    int need_to_reschedule = 0;
    sys_thread_t *self = sysThreadSelf();

    Log3(4, "queueSignal: mid 0x%x q 0x%x sys_thread_t 0x%x\n", mid, q, self);
    while (1) {
	if (waiter = queueRemoveHead(q, waiter)) {
	    if (q == &mid->monitor_waitq) {
		if (waiter->pending_suspend) {
		    waiter->state = MONITOR_SUSPENDED;
		    queueInsert(&mid->suspend_waitq, waiter);
		    waiter->pending_suspend = 0;
		    Log1(4, "queueSignal: making waiter 0x%x suspended\n",
			 waiter);
		    continue;
		} else {
		    Log1(4, "queueSignal: making waiter 0x%x runnable\n",
			 waiter);
		    mid->monitor_owner = waiter;
		    mid->entry_count = waiter->monitor_entry_count;
		    waiter->state = RUNNABLE;
		    waiter->mon_wait = 0;
		    waiter->monitor_entry_count = 0;
		    queueInsert(&runnable_queue, waiter);
		    if (runnable_queue->priority > self->priority) {
			need_to_reschedule = 1;
		    }
		}
	    } else {
	        Log(4, "queueSignal: CONDVAR_WAIT -> MONITOR_WAIT\n");
		waiter->state = MONITOR_WAIT;
		queueInsert(&mid->monitor_waitq, waiter);
		
		/* if nothing else in the waitq then wake it up */
		if (mid->monitor_owner == 0) {
		    q = &mid->monitor_waitq;
		    continue;
		}
	    }
	} else if (q == &mid->monitor_waitq) {
	
	    /* assert below not valid given how monitor_cleanup works!
	    assert(mid->monitor_owner == self
		    || mid->monitor_owner == SYS_THREAD_NULL); */
	    Log(4, "queueSignal: no waiter, set no monitor owner\n");
	    mid->monitor_owner = SYS_THREAD_NULL;
	    mid->entry_count = 0;
	}
	break;
    }
    return need_to_reschedule;
}

/*
 * signal all waiters on a queue.  We assume that the queue is the CV,
 * as that is the only queue this operation is permitted on.
 */
void
queueBroadcast(sys_mon_t *mid)
{
    sys_thread_t *waiter;

    while (waiter = queueRemoveHead(&mid->condvar_waitq, waiter)) {
        Log1(4, "queueBroadcast: waiter 0x%x CONDVAR_WAIT -> MONITOR_WAIT\n",
	    waiter);
	waiter->state = MONITOR_WAIT;
	queueInsert(&mid->monitor_waitq, waiter);
    }
    return;
}


/*
 * Wait on a specific queue until signalled
 */
void
queueWait(sys_mon_t *mid, queue_t *q)
{
    sys_thread_t *next;
    sys_thread_t *self = sysThreadSelf();

    Log3(4, "queueWait: sys_thread_t 0x%x q 0x%x mid 0x%x\n",
    	self, q, mid);
    assert(SCHED_LOCKED());
    queueInsert(q, self);
    self->mon_wait = (struct sys_mon_t *) mid;

    if (q == &mid->monitor_waitq) {
        Log(4, "queueWait: MONITOR_WAIT\n");
	self->state = MONITOR_WAIT;
	monitorApplyInversion(mid, self);
    } else {
        Log(4, "queueWait: CONDVAR_WAIT\n");
	self->state = CONDVAR_WAIT;
	self->monitor_entry_count = mid->entry_count;
	monitorUndoInversion(mid, self);
	(void) queueSignal(mid, &mid->monitor_waitq);
    }

    yieldContext(CONTEXT(self), reschedule, SYNCHRONOUS);
    return;
}

/*
 * handle priority inversion
 */
void
monitorApplyInversion(sys_mon_t *mid, sys_thread_t *t)
{
    int threadPriority;
    int monitorOwnerPriority;
    sys_mon_t    *mon;
    sys_thread_t *monitor_owner;
    sys_thread_t *self = sysThreadSelf();

    monitor_owner = mid->monitor_owner;
    if (monitor_owner == SYS_THREAD_NULL) {
        return;
    }
    
    assert(SCHED_LOCKED());
    assert(monitor_owner != t);
    assert(monitor_owner != self);
    threadPriority = t->priority;
    monitorOwnerPriority = monitor_owner->priority;
    if (monitorOwnerPriority < threadPriority) {
        intrLock();
	if ((mid->flags & SYS_MON_INVERTED_PRIORITY) == 0) {
	    mid->oldPriority = monitorOwnerPriority;
	    mid->flags |= SYS_MON_INVERTED_PRIORITY;
	}

	/*
	 * The priority is changing, which means we have to reorder the
	 * queue that the thread is queued upon.
	 * Don't use thread_setpriority() here, it reschedules.
	 */
	monitor_owner->priority = threadPriority;

	switch(monitor_owner->state) {
	    case RUNNABLE:
		queueRemove(&runnable_queue, monitor_owner);
		queueInsert(&runnable_queue, monitor_owner);
		break;

	    case MONITOR_WAIT:
		mon = (sys_mon_t *) monitor_owner->mon_wait;
		queueRemove(&mon->monitor_waitq, monitor_owner);
		queueInsert(&mon->monitor_waitq, monitor_owner);
		break;

	    case CONDVAR_WAIT:
		mon = (sys_mon_t *) monitor_owner->mon_wait;
		queueRemove(&mon->condvar_waitq, monitor_owner);
		queueInsert(&mon->condvar_waitq, monitor_owner);
		break;

	    case MONITOR_SUSPENDED:
		mon = (sys_mon_t *) monitor_owner->mon_wait;
		queueRemove(&mon->suspend_waitq, monitor_owner);
		queueInsert(&mon->suspend_waitq, monitor_owner);
		break;
	}
	
        intrUnlock();
	Log4(3, "monitorApplyInversion: sys_thread_t = 0x%x"
	     "\n\tsys_mon_t 0x%x"
	     "priority %d -> %d\n", monitor_owner,
	     mid, monitorOwnerPriority, threadPriority);
    }
    return;
}

void
monitorUndoInversion(sys_mon_t *mid, sys_thread_t *t)
{
    if (mid->flags & SYS_MON_INVERTED_PRIORITY) {
	Log4(3, "monitorUndoInversion: sys_thread_t = 0x%x,\n\t"
	     "sys_mon_t 0x%x " "priority %d -> %d\n", t, mid,
	     t->priority, mid->oldPriority);

	/* Don't use thread_setpriority() here, it reschedules */
	intrLock();
	t->priority = mid->oldPriority;
	mid->flags &= ~SYS_MON_INVERTED_PRIORITY;
	intrUnlock();
    }
}

/*
 * Operations on monitors
 */

int
sysMonitorInit(sys_mon_t *mid)
{
    return SYS_OK;		/* Trivial on Green threads */
}

/*
 * Enter a monitor
 *
 * A subtlety of the Solaris/Green threads package is the interaction
 * between monitor entry and scheduler locking for single-threaded execution.
 * If care is not taken in the low-level Java thread package support, race
 * conditions can unpredictably cause other threads to run at precisely the
 * times that you may expect to be single threaded.  This is often disasterous.
 *
 * The simple rule is that monitors must not be (newly) entered when the
 * scheduler is locked.  If that were to be attempted, there is some chance
 * that another thread might be holding the monitor.  In this case the running
 * thread waits for the monitor, which means yielding, the alternative being
 * deadlock.  So another thread may run.
 *
 * In the Solaris/Green threads, monitors are entered in two different
 * ways.  They are explicitly entered via monitorEnter(), sysMonitorEnter()
 * etc.  They are also implicitly entered when certain libc functions are
 * called (see the comment in synch.c).  If entering a monitor explicitly
 * while scheduling is locked can't be avoided in all contexts (e.g. when
 * enumerating threads or monitors), it is safe to acquire the necessary
 * monitor prior to locking the scheduler, so we will be guaranteed not to
 * wait once the scheduler is locked.  This doesn't really work for implicit
 * monitor use, where we simply have to avoid making certain libc calls when
 * scheduling is locked.
 *
 * To police monitor use, assertions may be placed in sysMonitorEnter()
 * and synch.c to abort if a monitor is entered when scheduling is locked.
 * Unfortunately, those in sysMonitorEnter() break the current implemen-
 * tation of ^\, so are commented out for now.
 */
int
sysMonitorEnter(sys_mon_t *mid)
{
    sys_thread_t *self = sysThreadSelf();

    SCHED_LOCK();
    if (mid->monitor_owner) {
	if (mid->monitor_owner != self) {
	    /* assert(_scheduling_lock == 1); */
	    queueWait(mid, &mid->monitor_waitq);
	}
    } else {
	/* assert(_scheduling_lock == 1); */
	mid->monitor_owner = self;
    }
    mid->entry_count++;
    SCHED_UNLOCK();

    return SYS_OK;
}

/*
 * Return true if we currently own this monitor (and threads have been
 * initialized.  This *can* get called before threads are initialized.
 */
bool_t
sysMonitorEntered(sys_mon_t *mid)
{
    sys_thread_t *self = sysThreadSelf();
    extern sys_thread_t *ThreadQueue;

    /* We can only have locked monitors if threads have been initialized */
    return (!ThreadQueue || 
	    (mid->monitor_owner != SYS_THREAD_NULL &&
	    	     mid->monitor_owner == self));
}

/*
 * Note that sysMonitorExit() essentially replicates the functionality
 * of sysMonitorDestroy(), and that monitorExit(), using sysMonitorExit(),
 * doesn't do sysMonitorDestroy().
 */
int
sysMonitorExit(sys_mon_t *mid)
{
    int ret = SYS_OK;
    sys_thread_t *self = sysThreadSelf();

    SCHED_LOCK();
    if (!mid || mid->monitor_owner != self) {
	ret = SYS_ERR;
	goto unlock;
    } else {
        /* exiting the monitor, so don't care about asynch notifications */
	mid->flags &= ~SYS_MON_STICKY_NOTIFICATION;

        --mid->entry_count;
	assert(mid->entry_count >= 0);
        if (mid->entry_count == 0) {
            monitorUndoInversion(mid, self);
	    if (!mid->monitor_waitq && !mid->condvar_waitq &&
		!mid->suspend_waitq) {
		/* Nobody on queues, owner exiting, destroy if in cache */
	        mid->monitor_owner = SYS_THREAD_NULL;
		ret = SYS_DESTROY;
	    } else if (queueSignal(mid, &mid->monitor_waitq)) {
	        assert(mid->monitor_owner != SYS_THREAD_NULL);
	        YIELD(SYNCHRONOUS);
	    }
	}

        /* self might be null, else: assert(self->monitor_entry_count == 0); */
    }

unlock:
    SCHED_UNLOCK();
    return ret;
}

/*
 * When the scheduler is locked, we need to avoid exiting monitors
 * using the normal means.  If we do (using sysMonitorExit()), we relock
 * the scheduler and then may yield and reschedule.  Rescheduling with
 * the scheduler locked more than once is illegal, and is caught
 * #ifdef DEBUG by an assertion in reschedule().  This version does
 * not reschedule, but rather returns a flag that says whether it
 * would like to.  We also avoid relocking the scheduler again, as
 * the code for unlocking it provides another opportunity to yield.
 *
 * This is a Green threads-only function, and is NOT considered part
 * of the HPI!  It should not be needed on ports using native threads.
 */
int
greenMonitorExitLocked(sys_mon_t *mid)
{
    int resched = 0;
    sys_thread_t *self = sysThreadSelf();

    assert(SCHED_LOCKED());
    /* Exiting the monitor, so don't care about asynch notifications */
    mid->flags &= ~SYS_MON_STICKY_NOTIFICATION;

    --mid->entry_count;
    assert(mid->entry_count >= 0);
    if (mid->entry_count == 0) {
        monitorUndoInversion(mid, self);
	if (!mid->monitor_waitq && !mid->condvar_waitq &&
	    !mid->suspend_waitq) {
	    /* Nobody on queues, owner exiting, destroy if in cache */
	    mid->monitor_owner = SYS_THREAD_NULL;
	} else {
	    resched = queueSignal(mid, &mid->monitor_waitq);
	}
    }

    return resched;
}

int
sysMonitorNotify(sys_mon_t *mid)
{
    int ret = SYS_OK;
    sys_thread_t *self = sysThreadSelf();

    SCHED_LOCK();
    if (self != mid->monitor_owner) {
	ret = SYS_ERR;
    } else {
	(void) queueSignal(mid, &mid->condvar_waitq);
    }
    
    /* don't need to signal the waitq since current thread has monitor */
    SCHED_UNLOCK();
    return ret;
}

int
sysMonitorNotifyAll(sys_mon_t *mid)
{
    int ret = SYS_OK;
    sys_thread_t *self = sysThreadSelf();

    SCHED_LOCK();
    if (self != mid->monitor_owner) {
	ret = SYS_ERR;
    } else {
        queueBroadcast(mid);
    }

    /* don't need to signal the waitq since current thread has monitor */
    SCHED_UNLOCK();
    return ret;
}

int
sysMonitorWait(sys_mon_t *mid, int millis)
{
    int ret = SYS_OK;
    sys_thread_t *self = sysThreadSelf();
    timeval_t tm, timeout;

    SCHED_LOCK();
    if (self != mid->monitor_owner) {
	ret = SYS_ERR;
	goto unlock;
    }

    assert(self->state == RUNNABLE);
    assert(self->mon_wait == 0);
    assert(self->monitor_entry_count == 0);
    if ((mid->flags & SYS_MON_STICKY_NOTIFICATION) != 0) {
	mid->flags &= ~SYS_MON_STICKY_NOTIFICATION;
	goto unlock;
    }

    if (millis != SYS_TIMEOUT_INFINITY) {
        threadSetTimeout(&tm, millis/1000, (millis % 1000) * 1000);
	timeout = currentTime();
	timerAdd(&timeout, &tm);
	registerAlarm(self, timeout);
    }

    queueWait(mid, &mid->condvar_waitq);

    assert(mid->monitor_owner == self);
    assert(SCHED_LOCKED());

    if (millis != SYS_TIMEOUT_INFINITY) {
	cleanupPendingAlarm(self);
    }

unlock:
    SCHED_UNLOCK();
    return ret;
}

void dumpQueue(sys_thread_t *t, char *mesg) {

    if (t == 0) {
	return;
    }

    fprintf(stderr, "\t%s:\n", mesg);
    for (; t != 0; t = t->waitq) {
	fprintf(stderr,"\t");
	threadDumpInfo((HThread *) t->cookie);
    }
}

void
sysMonitorDumpInfo(sys_mon_t *mon)
{
    TID tid;

    if (mon->monitor_owner) {
	tid = (TID) (mon->monitor_owner->cookie);
	fprintf(stderr, "    monitor owner: %s (0x%x)\n",
		Object2CString((Handle *) THREAD(tid)->name),
		mon->monitor_owner);
    } else {
	fprintf(stderr, "    unowned\n");
    }
    dumpQueue(mon->monitor_waitq, "Waiting to enter");
    dumpQueue(mon->condvar_waitq, "Waiting to be notified");
    return;
}

/*
 * Return SYS_OK if the monitor was owned by the tid, otherwise
 * retrun SYS_ERR.  If the monitor was owned by the tid reset the
 * monitor and notify anyone who is waiting to aquire the monitor.
 */
int
sysMonitorDestroy(sys_mon_t *mid, sys_thread_t *tid)
{
    int ret = SYS_OK;

    SCHED_LOCK();
    if (mid->monitor_owner == tid) {
	mid->entry_count = 0;   /* for this thread, anyway! */
	if (!mid->monitor_waitq && !mid->condvar_waitq && !mid->suspend_waitq) {
	    mid->monitor_owner = SYS_THREAD_NULL;
	    ret = SYS_DESTROY;	/* remove from hash table */
	} else {
	    mid->flags &= ~SYS_MON_STICKY_NOTIFICATION;
	    monitorUndoInversion(mid, tid);
	    queueSignal(mid, &mid->monitor_waitq);
	    /*
	     * do not remove from the hash table because someone has
	     * already been assigned this monitor and is trying to
	     * enter it.
	     */
	}
    }	/* don't remove if not owner, but it's not an error */
    SCHED_UNLOCK();
    return ret;
}


void
resumeMonitor(sys_thread_t *tid)
{
    sys_mon_t *mid;
    int key;

    assert(SCHED_LOCKED());

    mid = (sys_mon_t *) tid->mon_wait;
    assert(mid);
    tid->state = MONITOR_WAIT;
    queueRemove(&mid->suspend_waitq, tid);
    queueInsert(&mid->monitor_waitq, tid);

    /* AVH: Signal if this thread was put in the wait queue
     * and the monitor is not blocked.
     */
    if ((mid->monitor_waitq == tid) && (mid->monitor_owner == 0)) {
	queueSignal(mid, &mid->monitor_waitq);
    }
}


#ifdef notdef
/*
 * returns true if the monitor is in use in any way.
 */
bool_t
sysMonitorInUse(sys_mon_t *mon) 
{
    int ret;

    SCHED_LOCK();
    ret = (mid->monitor_waitq 
	   || mid->condvar_waitq 
	   || mid->suspend_waitq
	   || mid->entry_count != 0);
    SCHED_UNLOCK();
    return ret;
}
	
#endif
