/*
 * @(#)signals.c	1.25 95/01/31  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * A spl and interrupt implementation to run on top of Unix.
 */

/*
 * Header files.
 */
#include <ucontext.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>

#include <threads_md.h>
#include <monitor_md.h>
#include <interrupt.h>
#include <debug.h>
#include <ucontext.h>
#include <log.h>
#include <machapi.h>
#include <schedule.h>
#include <monitor.h> 	/* For monitorRegister() */

#include <sys_api.h>

/*
 * Local variables.
 */
static sys_mon_t	handlerMonitors[SYS_ASYNC_MON_MAX];

sys_mon_t *PendingNotifyQ;

/*
 * Local routine prototypes.
 */

/*
 * intrNotifyHandler() --
 *	Interrim routine that delivers signals in BKM's original fashion.
 */
static int intrNotifyHandler(int interrupt, void *arg, void *siginfo);

/*
 * External routines.
 */

void
InitializeAsyncMonitors()
{
    int i;

    /* Async monitor numbering starts at 1 */
    for (i = 1; i < SYS_ASYNC_MON_MAX; i++) {
	switch (i) {	/* Have to keep in sync with the enum... */
	    case SYS_ASYNC_MON_ALARM:
		monitorRegister(&handlerMonitors[i], "Alarm monitor");
		break;
	    case SYS_ASYNC_MON_IO:
		monitorRegister(&handlerMonitors[i], "I/O monitor");
		break;
	    case SYS_ASYNC_MON_EVENT:
		monitorRegister(&handlerMonitors[i], "Event monitor");
		break;
	    default:
		monitorRegister(&handlerMonitors[i], "Unknown monitor");
		break;
	}
    }
}

/*
 * This function is used to return to any context that was interrupted
 * by a signal delivery.  We assume that the signal trampoline has saved
 * all state, and all we need to do is execute a signal cleanup (the
 * backend of the signal trampoline) to return control to the interrup-
 * ted thread.
 */
void
fullSwitchContext(context_t *c, preemption_t preemption)
{
    /*
     * This thread was interrupted, which means that it was NOT inside
     * a critical section.  This also means that the thread does not know
     * that it needs to clear the critical section lock when control
     * returns to it.  We have to clear the lock for it, and there may
     * be pending interrupts.
     */

    /*
     * If the scheduler was interrupted, we process the events, and call
     * it with interrupts disabled.  This prevents recursion.
     */
    /* intrLock(); - removed by csw */
    if (PendingNotifyQ) {
	if (processPendingNotification()) {
	    sys_thread_t    *self = sysThreadSelf();

	    if (self->state == RUNNABLE) {
		queueInsert(&runnable_queue, self);
	    }
	    reschedule(preemption);
	}
    }
    /* intrUnlock(); - removed by csw */

    /*
     * Don't have to call return_from_trap instead of switch_context
     * if preemption is set because we are continuing the thread that
     * was interrupted when the signal came in and the signal trampoline
     * will restore the original signal mask.
     */
    sysThreadSelf()->full_switch = 0;
    switchContext(c);
    /* NOTREACHED */
    panic("return from switch_context");
}

/*
 * return a registered asynchronous notifier
 */
sys_mon_t *
asyncMon(async_mon_key_t key)
{
    if (key >= SYS_ASYNC_MON_MAX || key == 0) {
	return SYS_MID_NULL;
    } else {
	return &handlerMonitors[key];
    }
}

int asyncKeyToSig[SYS_ASYNC_MON_MAX] = {
        0,		/* Placeholder.  async_mon_key_t starts at 1 */
	SIGALRM,	/* SYS_ASYNC_MON_ALARM */
	SIGIO,		/* SYS_ASYNC_MON_IO */
	SIGUSR1,	/* SYS_ASYNC_MON_EVENT */
};

/*
 * This interface is used by client programs to register monitors that
 * they will recieve asynchronous notification on.  The async monitor key
 * is translated into a machine dependent Unix signal number.
 */
int
asyncNotifier(async_action_t action, async_mon_key_t key)
{
    int sig;
    sys_mon_t *mid;

    if (key >= SYS_ASYNC_MON_MAX) {
	return -1;
    }

    sig = asyncKeyToSig[key];
    mid = &handlerMonitors[key];
    memset((void *)mid, 0, sizeof (*mid));

    if (action == ASYNC_REGISTER) {
	mid->flags |= SYS_MON_IN_USE;

	/* Register our interrrupt handler */
	intrRegister(sig, intrNotifyHandler, &handlerMonitors[key]);
    } else {
	intrUnregister(sig, intrNotifyHandler, &handlerMonitors[key]);
    }

    return 0;
}

static int
sigioNotifier(int interrupt, void *arg, void *siginfo)
{
    siginfo_t *si = (siginfo_t *) siginfo;
    sys_mon_t *mon;
    int resched = 0;

    /* We always notify the most general sigio monitor. */
    mon = &handlerMonitors[SYS_ASYNC_MON_IO];
    resched |= asyncEventNotify(mon);
    if (resched) {
	MACHDEP(sysThreadSelf())->needReschedule = 1;
    }
    return 1;
}

asyncIORegister()
{
    intrRegister(SIGIO, sigioNotifier, 0);
}

/*
 * intrNotifyHandler() --
 *	Handle a unix signal.  This is called from the interrupt
 *	routine, which saves all pending thread state.  If there is
 *	a registered monitor, notify it (and potentially yield).
 *
 *	There can be no signal which needs both async and sync
 *	notification.
 */
static int
intrNotifyHandler(int interrupt, void *arg, void *siginfo)
{
    sys_mon_t	*mon = (sys_mon_t *)arg;

    Log2(2, "Received signal %d interrupting sys_thread_t 0x%x\n", interrupt,
	 sysThreadSelf());

    if ((mon->flags & SYS_MON_IN_USE) != 0
	&& asyncEventNotify(mon)) {
	Log(2, " (Not Locked)\n");
	MACHDEP(sysThreadSelf())->needReschedule = 1;
    }
#ifdef DEBUG
    else {
	Log1(2, " (Locked: %d)\n", _scheduling_lock);
    }
#endif
    return (1);
}


/**
 ** Asynchronous Interface
 **/

/*
 * Asynchronous notification of a condition variable
 *
 * NotifyAll the condition variable, and perform any scheduling updates
 * required.   Return TRUE if the caller needs to immediately
 * reschedule.
 *
 * ASSUMPTION: the current threads state is already saved, and we
 * are free to do anything we wish.
 */
int
asyncEventNotify(sys_mon_t *mid)
{
    int need_to_switch = 0;
    sys_thread_t *tid;

    if (SCHED_LOCKED()) {
	/*
	 * Queue this notification
	 */
	if ((mid->flags & SYS_MON_PENDING_NOTIFICATION) == 0) {
	    assert(mid->pendingq == SYS_MID_NULL);
	    mid->pendingq = PendingNotifyQ;
	    PendingNotifyQ = mid;
	}
	mid->flags |= SYS_MON_PENDING_NOTIFICATION;
    } else {
	/*
	 * If a thread is currently inside the monitor, there is a potential
	 * race condition.  It may have already checked its test condition,
	 * and decided it wants to wait.  It will miss the notification,
	 * assuming that none can arrive while it owns the monitor.
	 *
	 * Solve this race by setting the stick_notification boolean, 
	 * indicating that a CV notify was done while some thread owned
	 * the monitor.  monitorExit() and monitorWait() are responsible
	 * for clearing this boolean.  monitorWait() should also fail
	 * to wait if the boolean is set.
	 */
	if (mid->monitor_owner != SYS_THREAD_NULL) {
	    mid->flags |= SYS_MON_STICKY_NOTIFICATION;
	}

	/* Actually handle the notification. */
	queueBroadcast(mid);

	/* If the monitor isn't locked, give it to the first waiter */
	if (!mid->monitor_owner && mid->monitor_waitq) {
	    need_to_switch = queueSignal(mid, &mid->monitor_waitq);
	}
    }
    return need_to_switch;
}



/*
 * Thread all outstanding CV notification.  Don't mess with the currently
 * running context, because we have no idea what state we are in.
 */
int
processPendingNotification()
{
    sys_mon_t *mid;
    sys_mon_t *midq;
    sys_thread_t *tid;
    int need_to_reschedule = 0;

    /*
     * Perform the notifyall's that couldn't be done because of locking.
     * Pull all of the pending CV notifies off of the queue, and process
     * them.  The notification is completely normal, except for the locking.
     * If a yield is necessary, it is done by the caller.
     */
    assert(PendingNotifyQ);
    midq = PendingNotifyQ;
    while (midq) {
	mid = midq;
	midq = midq->pendingq;
	mid->pendingq = SYS_MID_NULL;
	mid->flags &= ~SYS_MON_PENDING_NOTIFICATION;

	/* See asyncEventNotify for explanation of this. */
	if (mid->monitor_owner != SYS_THREAD_NULL) {
	    mid->flags |= SYS_MON_STICKY_NOTIFICATION;
	}

	/* Now notify all of the waiters. */
	queueBroadcast(mid);

	/* If the monitor isn't locked, give it to the first waiter */
	if (!mid->monitor_owner && mid->monitor_waitq) {
	    need_to_reschedule = queueSignal(mid, &mid->monitor_waitq);
	}
    }
    PendingNotifyQ = SYS_MID_NULL;
    return need_to_reschedule;
}


/*
 * Thread all of the pending condition variable notifications that
 * couldn't be handled at interrupt time.
 *
 * ASSUMPTION: this is being called from a thread context.
 */
void
_sched_unlock()
{
    int need_to_reschedule;

restart:
    intrLock();
    SCHED_LOCK();

    /*
     * Perform the notifyall's that couldn't be done because of locking.
     * The notification is completely normal, except for the locking.
     */
    if (PendingNotifyQ) {
	need_to_reschedule = processPendingNotification();

	intrUnlock();
	if (need_to_reschedule) {
	    YIELD(SYNCHRONOUS);
	}

	/*
	 * We may have gotten an interrupt in between the last two
	 * statements (intrUnlock & YIELD).  Check, and if so, 
	 * loop.  This also handles the case were we really do yield
	 * to somebody else, and they get an interrupt while yielding
	 * back to us.
	 *
	 * This is gross, but it keeps this thread from recursing in
	 * this routine.
	 */
	_scheduling_lock--;
	if (PendingNotifyQ) {
	    goto restart;
	}

    } else {
	_scheduling_lock--;
	intrUnlock();
    }

    return;
}
