/*
 * @(#)synch.c	1.8 95/01/31  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * This is an attempt to provide synchronization primitives for the
 * solaris version of libc.  Because we use our own threads package,
 * instead of solaris threads, so we need a synchronization
 * mechanism that works with our thread package.
 */
#include <monitor.h>
#include <synch.h>
#include <schedule.h>
#include <debug.h>

extern bool_t	monitorsInitialized;

/*
 * WARNING: It is crucial for the Solaris/Green threads implementation
 * of JAVA that we never try to acquire a monitor while scheduling is
 * locked.  If we did, and the monitor is already held by someone else,
 * then we would wait for it.  If we wait, we will switch contexts and
 * another thread will run at precisely the time when we don't want that.  
 * (Note that reacquiring a monitor that we already own isn't a problem:
 * we are always guaranteed to get it!)  We take pains throughout the
 * code to ensure that we don't enter a monitor when scheduling is locked.
 *
 * However, the Solaris implementation plays another trick with monitors.
 * libc calls (e.g. malloc()) must be made Green thread-safe.  libc is
 * normally made Solaris thread-safe by libthread providing definitions
 * for _mutex_lock() and _mutex_unlock().  Without libthread, libc's calls
 * to these functions are effectively noops.
 *
 * We make libc Green thread-safe by defining these functions to use our
 * monitors, taking them from the monitor cache using as the cache key
 * the mutex_t pointer that libc passes in.  Note, though, that this
 * leaves us with a new, hidden way to enter a monitor with scheduling
 * locked.  Because several threads may simultaneously want to call, say,
 * malloc(), passing _mutex_lock() the same pointer, we are subject
 * to the same race conditions that can defeat the intended behavior.
 *
 * This situation can only be managed by avoiding libc calls when the
 * scheduler is locked.  To ensure that this doesn't happen, assertions
 * have been placed in _mutex_lock() and _mutex_unlock() that will die
 * if those functions are called with scheduling locked.  (monitorExit()
 * acquires the monitor cache lock, so needs to be checked too.)  If
 * you find these assertions blowing up on you, look in the stack trace
 * for libc or other calls that are calling these mutex functions with
 * scheduling locked.
 */

_mutex_lock(mutex_t *mp)
{
    assert(_scheduling_lock == 0);

    if (monitorsInitialized) {
	monitorEnter((int) mp);
    }
    return 0;
}

_mutex_unlock(mutex_t *mp)
{
    assert(_scheduling_lock == 0);

    if (monitorsInitialized) {
	monitorExit((int) mp);
    }
    return 0;
}
