/*
 * @(#)network.c	1.27 95/05/15 Jonathan Payne
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/systeminfo.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <stdio.h>
#include <oobj.h>
#include <interpreter.h>
#include <javaString.h>
#include <iomgr.h>
#include <net_InetAddress.h>
#include <net_Socket.h>
#include <net_SocketOutputStream.h>
#include <net_SocketInputStream.h>

#include <fcntl.h>
#include <sys/fcntl.h>
#include <sys/filio.h>
#include <sys/ioctl.h>

#define TRUE	1
#define FALSE	0

#define NETPKG "net/"


HString *
net_InetAddress_getLocalHostName(Hnet_InetAddress *Haddr)
{
    HString *localHostName;
    char    hostname[MAXHOSTNAMELEN];

    if (sysinfo(SI_HOSTNAME, hostname, sizeof(hostname)) == -1) {
	strcpy(hostname, "localhost");
    }
    return makeJavaString(hostname, strlen(hostname));
}

long
net_InetAddress_getPortByName(Hnet_InetAddress *Hnull, HString *name)
{
    struct servent  *sp;
    char	    serviceName[128];

    if (name == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return;
    }

    javaString2CString(name, serviceName, sizeof (serviceName));
    if ((sp = getservbyname(serviceName, "tcp")) == NULL) {
	SignalError(0, NETPKG "UnknownServiceException", serviceName);
	return 0;
    }
    return (long) ntohs((short) sp->s_port);
}    

static int
hostnameToAddress(struct sockaddr_in *sin, char *host_name)
{
    struct hostent *hp;

    memset((char *) sin, 0, sizeof(*sin));
    if (isdigit(host_name[0])) {
	unsigned long addr;

	addr = inet_addr(host_name);
	if ((int) addr != -1) {
	    memcpy((char *) &sin->sin_addr, (char *) &addr, sizeof(addr));
	    sin->sin_family = htons(AF_INET);
	    return TRUE;
	}
    }
    if ((hp = gethostbyname(host_name)) == NULL)
	return FALSE;
    memcpy((char *) &sin->sin_addr, hp->h_addr, hp->h_length);
    sin->sin_family = hp->h_addrtype;

    return TRUE;
}

/*
 * This is a version of hostnameToAddress that bypasses the system
 * facility and talks directly to DNS to a particular name server.
 * It is used with SOCKS is being used to talk to a name server
 * that can resolve names on the other side of the firewall.
 */
static int
hostnameToAddressDNS(struct sockaddr_in *sin, char *host_name, 
	long DNSServer) {
    struct hostent *hp;
    int savecnt;
    long saveAddr;

#ifdef SOCKS_DEBUG
    printf("DNS Direct Lookup: %s\n", host_name);
#endif
    if ((_res.options & RES_INIT) == 0)
	res_init();

    saveAddr = _res.nsaddr.sin_addr.s_addr;
    savecnt = _res.nscount;

    _res.nsaddr.sin_addr.s_addr = DNSServer;
    _res.nscount = 1;

    /* Here we "cheat" and use the DNS gethostbyname call directly. */
    hp = (struct hostent *)res_gethostbyname(host_name);
    if (hp != NULL) {
	memcpy((char *) &sin->sin_addr, hp->h_addr, hp->h_length);
	sin->sin_family = hp->h_addrtype;
    }
    _res.nsaddr.sin_addr.s_addr = saveAddr;
    _res.nscount = savecnt;
#ifdef SOCKS_DEBUG
    if (hp) {
        printf("Address: %s\n", inet_ntoa(*(struct in_addr *)hp->h_addr));
    }
#endif
    return (hp != NULL);
}

static int makeInetSocket(int type)
{
    return socket(AF_INET, type, 0);
}

static int makeStreamSocket()
{
    return makeInetSocket(SOCK_STREAM);
}

static int makeDgramSocket()
{
    return makeInetSocket(SOCK_DGRAM);
}

static char *
addressToHostname(struct sockaddr_in *addr, int len)
{
    struct hostent *hp;

    hp = gethostbyaddr((char *) &addr->sin_addr, len, addr->sin_family);
    if (hp == NULL)
	return NULL;
    return hp->h_name;
}

static void
setPort(struct sockaddr_in *addr, int port)
{
    addr->sin_port = htons((short) port);
}

void
net_Socket_create(Hnet_Socket *Hs, long isStream)
{
    int	s;

    s = (isStream != 0) ? makeStreamSocket() : makeDgramSocket();
    if (s == -1) {
	SignalError(0, NETPKG "SocketException", strerror(errno));
	return;
    }

    /* make the socket address immediatly available for reuse */
    setsockopt(s, SOL_SOCKET, SO_REUSEADDR, "yes", 4);

    nonblock_io(s, IO_DONTBLOCK);
    unhand(Hs)->s = s;
}

static void
bindSocket(Classnet_Socket *s, Hnet_InetAddress *Haddress, int port)
{
    Classnet_InetAddress *address;
    struct sockaddr_in	addr;

    if (Haddress == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return;
    }


    address = unhand(Haddress);

    memset((char *) &addr, 0, sizeof(addr));
    setPort(&addr, port);
    addr.sin_addr.s_addr = (unsigned long) ntohl(address->address);
    addr.sin_family = address->family;
    if (bind(s->s, (struct sockaddr *) &addr, sizeof(addr)) == -1) {
	SignalError(0, NETPKG "SocketException", "bind");
	return;
    }
    s->address = Haddress;
}

void
net_Socket_bindToPort(Hnet_Socket *Hs, Hnet_InetAddress *Haddr, long port)
{
    if (Haddr == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return;
    }

    bindSocket(unhand(Hs), Haddr, port);
}

static void
getPortFromConnectedSocket(Hnet_Socket *Hs)
{
    /* Now that we're a connected socket, let's extract the port number
       that the system chose for us and store it in the Socket object. */
    struct sockaddr_in  me;
    int			len = sizeof(me);

    if (getsockname(unhand(Hs)->s, (struct sockaddr *) &me, &len) == -1) {
	SignalError(0, NETPKG "SocketException", strerror(errno));
	return;
    }
    unhand(Hs)->port = htons(me.sin_port);
}

void
net_Socket_bindAnonymously(Hnet_Socket *Hs, Hnet_InetAddress *Haddr)
{
    if (Haddr == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return;
    }


    bindSocket(unhand(Hs), Haddr, 0);
    getPortFromConnectedSocket(Hs);
}

void
net_Socket_connectUnchecked(Hnet_Socket *Hs, Hnet_InetAddress *Hdest, long port)
{
    struct sockaddr_in	him;
    char		*msg;
    int			socket;
    Classnet_InetAddress	*dest;

    if (Hdest == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return;
    }

    dest = unhand(Hdest);

    him.sin_family = dest->family;
    setPort(&him, port);
    him.sin_addr.s_addr = dest->address;

    if (sysConnect(unhand(Hs)->s, (struct sockaddr *) &him, sizeof(him)) < 0) {
	if (errno == EPROTO) {
	    SignalError(0, NETPKG "ProtocolException", strerror(errno));
	} else {
	    SignalError(0, NETPKG "SocketException", strerror(errno));
	}
	return;
    }
    getPortFromConnectedSocket(Hs);
}

void
net_Socket_listen(Hnet_Socket *Hsrc, long count)
{
    listen(unhand(Hsrc)->s, count);
}

long
net_Socket_dup(Hnet_Socket *Hsrc)
{
    Classnet_Socket *src = unhand(Hsrc);
    return (dup(src->s));
}

void
net_Socket_accept(Hnet_Socket *Hsrc, Hnet_Socket *Hdst)
{
    struct sockaddr_in	him;
    char		*msg;
    int			socket;
    Classnet_Socket		*dst;
    Classnet_Socket		*src = unhand(Hsrc);
    int			len;

    if (Hdst == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return;
    }

    dst = unhand(Hdst);

    len = sizeof(him);
    dst->s = accept(src->s, (struct sockaddr *) &him, &len);
    /*
     * This is a gross temporary workaround for a bug in libsocket
     * of solaris2.3.  They exit a mutex twice without entering it
     * and our monitor stuff throws an exception.  The workaround is
     * to simply ignore InternalExceptions from accept.  If we get
     * an exception, and it has an exception object, and we can find
     * the class InternalException in the java pakage, and the
     * exception that was thrown was an instance of
     * InternalException, then ignore it.
     */
    if (EE()->exceptionKind != EXCKIND_NONE) {
	Handle *h = EE()->exception.exc;
	if (h != NULL) {
	    ClassClass *cb;
	    cb = FindClassFromClass(ee, JAVAPKG "InternalException", TRUE, NULL);
	    if (cb != NULL && is_instance_of(h, cb, ee)) {
		exceptionClear(EE());	
	    }
	}
    }

    if (dst->s < 0) {
	SignalError(0, NETPKG "SocketException", strerror(errno));
	return;
    }

    unhand(dst->address)->family = him.sin_family;
    unhand(dst->address)->address = htonl(him.sin_addr.s_addr);
    dst->port = htons(him.sin_port);
}
void
net_InetAddress_lookupWithDNS(Hnet_InetAddress *Hnull, Hnet_InetAddress *Haddr,
	Hnet_InetAddress *server)
{
    Classnet_InetAddress *addr;
    Classnet_InetAddress *DNSServer;
    struct sockaddr_in	saddr;
    char		hostname[256];

    if (Haddr == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return;
    }

    DNSServer = unhand(server);
    addr = unhand(Haddr);
    if (addr->hostName == 0) {
	addr->address = htonl(INADDR_ANY);
	addr->family = AF_INET;
	return;
    }

    javaString2CString(addr->hostName, hostname, sizeof (hostname));
    if (!hostnameToAddressDNS(&saddr, hostname, DNSServer->address)) {
	SignalError(0, NETPKG "UnknownHostException", hostname);
	return;
    }
    addr->address = htonl(saddr.sin_addr.s_addr);
    addr->family = saddr.sin_family;
}


void
net_InetAddress_lookup(Hnet_InetAddress *Hnull, Hnet_InetAddress *Haddr)
{
    Classnet_InetAddress *addr;
    struct sockaddr_in	saddr;
    char		hostname[256];

    if (Haddr == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return;
    }

    addr = unhand(Haddr);
    if (addr->hostName == 0) {
	addr->address = htonl(INADDR_ANY);
	addr->family = AF_INET;
	return;
    }

    javaString2CString(addr->hostName, hostname, sizeof (hostname));

    if (!hostnameToAddress(&saddr, hostname)) {
	SignalError(0, NETPKG "UnknownHostException", hostname);
	return;
    }
    addr->address = htonl(saddr.sin_addr.s_addr);
    addr->family = saddr.sin_family;
}

void
net_SocketOutputStream_write(Hnet_SocketOutputStream *this,
			     HArrayOfByte *data,
			     long off, long len)
{
    Classnet_SocketOutputStream *thisptr = unhand(this);
    char *dataptr;
    int datalen, n;

    if (data == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return;
    }

    dataptr = unhand(data)->body;
    datalen = obj_length(data);

    if ((len < 0) || (off < 0) || (len + off > datalen)) {
        SignalError(0, JAVAPKG "ArrayIndexOutOfBoundsException", 0);
        return;
    }

    while (len > 0) {
	n = send(thisptr->fd, dataptr + off, len, 0);
	if (n == -1) {
	    SignalError(0, JAVAPKG "IOException", strerror(errno));
	    break;
	}
        off += n;
	len -= n;
    }
}

long 
net_SocketInputStream_recv(Hnet_SocketInputStream *this,
			   HArrayOfByte *data,
			   long off, long len)
{
    Classnet_SocketInputStream *thisptr = unhand(this);
    char *dataptr;
    int datalen, n;

    if (data == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return;
    }

    dataptr = unhand(data)->body;
    datalen = obj_length(data);

    if (len < 0 || len + off > datalen) {
        SignalError(0, JAVAPKG "ArrayIndexOutOfBoundsException", 0);
        return 0;
    }

    n = recv(thisptr->fd, dataptr + off, len, 0);
    if (n == -1) {
        SignalError(0, JAVAPKG "IOException", strerror(errno));
    }
    /* AVH: this is bogus but it stops the pointer from being gc'd */
    if (dataptr == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
    }
    return n;
}
