/*
 * @(#)check_code.c	1.21 95/02/27  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */
 
/*-
 *      Verify that the code within a method block doesn't exploit any 
 *      security holes.
 *
 *      This code is still a work in progress.  All currently existing code
 *      passes the test, but so does a lot of bad code.
 */

#include <oobj.h>
#include <interpreter.h>
#include <opcodes.h>
#include <opcodes.length>
#include <opcodes.in_out>
#include <bool.h>
#include <tree.h>
#include <byteorder.h>
#include <limits.h>

#ifdef DEBUG
    int verify_verbose = 0;
    static struct context_type *GlobalContext;
#else
#   define verify_verbose 0
    static char *opnames[1];	    /* so compiler doesn't give error */
#endif

enum {
    ITEM_Bogus,
    ITEM_Void,			/* only as a function return value */
    ITEM_Integer,
    ITEM_Float,
    ITEM_Double,
    ITEM_Double_2,		/* 2nd word of double in register */
    ITEM_Long,
    ITEM_Long_2,		/* 2nd word of long in register */
    ITEM_Array,
    ITEM_Object,		/* Extra info field gives name. */
    ITEM_ReturnAddress,		/* Extra info gives instr # of start pc */
    /* The following three are only used within array types. 
     * Normally, we use ITEM_Integer, instead. */
    ITEM_Byte,
    ITEM_Short,
    ITEM_Char
};


#define UNKNOWN_STACK_SIZE -1
#define UNKNOWN_REGISTERS_USED -1

#define MAX(a, b) ((a) > (b) ? (a) : (b))
#define MIN(a, b) ((a) < (b) ? (a) : (b))

#define BITS_PER_INT   (CHAR_BIT * sizeof(int)/sizeof(char))
#define SET_BIT(flags, i)  (flags[(i)/BITS_PER_INT] |= \
			               ((unsigned)1 << ((i) % BITS_PER_INT)))
#define	IS_BIT_SET(flags, i) (flags[(i)/BITS_PER_INT] & \
			               ((unsigned)1 << ((i) % BITS_PER_INT)))

typedef unsigned long fullinfo_type;
typedef unsigned int *bitvector;

#define GET_ITEM_TYPE(thing) ((thing) & 0x1F)
#define GET_INDIRECTION(thing) (((thing) & 0xFFFF) >> 5)
#define GET_EXTRA_INFO(thing) ((thing) >> 16)
#define WITH_ZERO_INDIRECTION(thing) ((thing) & ~(0xFFE0))
#define WITH_ZERO_EXTRA_INFO(thing) ((thing) & 0xFFFF)

#define MAKE_FULLINFO(type, indirect, extra) \
     ((type) + ((indirect) << 5) + ((extra) << 16))
#define MAKE_CLASSNAME_INFO(classname, addr) \
       MAKE_FULLINFO(ITEM_Object, 0, \
                      Str2ID(&context->classHash, (classname), (addr), FALSE))
#define MAKE_CLASSNAME_INFO_WITH_COPY(classname) \
       MAKE_FULLINFO(ITEM_Object, 0, \
                      Str2ID(&context->classHash, (classname), 0, TRUE))


#define NULL_FULLINFO MAKE_FULLINFO(ITEM_Object, 0, 0)


struct context_type {
    /* these fields are per class */
    ClassClass *class;		/* current class */
    struct StrIDhash *classHash;
    fullinfo_type object_info;	/* fullinfo for java/lang/Object */
    fullinfo_type string_info;	/* fullinfo for java/lang/Stringn */
    fullinfo_type currentclass_info; /* fullinfo for context->class */

    /* these fields are per method */
    struct methodblock *mb;	/* current method */
    unsigned char *code;	/* current code object */
    short *code_data;		/* offset to instruction number */
    struct instruction_data_type *instruction_data; /* info about each */
    struct handler_info_type *handler_info;
    int instruction_count;	/* number of instructions */
    fullinfo_type return_type;	/* function return type */
    fullinfo_type swap_table[4]; /* used for passing information */

    /* Used by the space allocator */
    struct CCpool *CCroot;
    char *CCfree_ptr;
    int CCfree_size;
};


struct instruction_data_type {
    opcode_type opcode;		/* may turn into "canonical" opcode */
    bool_t changed;		/* has it changed? */
    union {
	int i;			/* operand to the opcode */
	int *ip;
	fullinfo_type fi;
    } operand;
    int operand2;
    int stack_size;		/* stack_size at this instruction */
    struct stack_item_type *stack; /* the stack */
    int registers_used;		/* number of registers used */
    fullinfo_type *registers;

    /* Handling of JSR.  Is this even right? */
    bitvector register_mask;
};

struct handler_info_type {
    int start, end, handler;
    struct stack_item_type *stack;
};

struct stack_item_type {
    fullinfo_type item;
    struct stack_item_type *next;
};

typedef struct context_type context_type;
typedef struct instruction_data_type instruction_data_type;
typedef struct stack_item_type stack_item_type;

static bool_t verify_code(context_type *context, struct methodblock *mb);

static int instruction_length(unsigned char *iptr);
static bool_t verify_opcode_operands (context_type *, int inumber, int offset);
static bool_t verify_target(context_type *, int offset);
static bool_t verify_constant_pool_type(context_type *, int, unsigned);
static bool_t verify_exception_table(context_type *);

static bool_t initialize_verifier_data_flow(context_type *);
static bool_t run_verifier_data_flow(context_type *context);


static bool_t verify_registers(context_type *context, int inumber);
static bool_t verify_and_pop_stack(context_type *, int, 
				   int *, stack_item_type **);

static void update_registers(context_type *context, int inumber,
			     int *new_registers_used_p, 
			     fullinfo_type **new_registers_p,
			     bitvector *new_register_mask_p);

static void update_push_stack(context_type *, int inumber,
			      int *new_stack_size_p, 
			      stack_item_type **new_stack_p);

static bool_t verify_merge_into_successors(context_type *, int, 
					   int, fullinfo_type *, bitvector,
					   int, stack_item_type *);
static bool_t verify_merge_into_one_successor(context_type *context, 
					      int inumber, 
					      int new_registers_used, 
					      fullinfo_type *new_registers,
					      bitvector new_register_mask,
					      int new_stack_size, 
					      stack_item_type *stack);
static stack_item_type *copy_stack(context_type *, stack_item_type *);
static bitvector copy_register_mask(context_type *, bitvector);
static fullinfo_type decrement_indirection(fullinfo_type);



static fullinfo_type merge_fullinfo_types(context_type *context, 
					  fullinfo_type a, fullinfo_type b,
					  bool_t assignment);
static bool_t is_assignable_to(context_type *,fullinfo_type a, fullinfo_type b);
static ClassClass *object_fullinfo_to_classclass(context_type *, fullinfo_type);


#define NEW(type, count) \
        ((type *)CCalloc(context, (count)*(sizeof(type)), FALSE))
#define ZNEW(type, count) \
        ((type *)CCalloc(context, (count)*(sizeof(type)), TRUE))

static void CCinit(context_type *context);
static void CCdestroy(context_type *context);
static void *CCalloc(context_type *context, int size, bool_t zero);

static char *cp_index_to_signature(context_type *context, int cp_index);
static fullinfo_type cp_index_to_class_fullinfo(context_type *, int, bool_t);

static char signature_to_fieldtype(context_type *context, 
				   char **signature_p, fullinfo_type *info);

static void CCerror (context_type *, char *format, ...);

static void print_stack (context_type *, stack_item_type *stack);
static void print_registers(context_type *, int,  fullinfo_type *, bitvector);
static void print_formatted_fieldname(context_type *context, int index);



/* Called by verify_class.  Verify the code of each of the methods
 * in a class.
 */

bool_t verify_class_codes(ClassClass *cb) {
    context_type context_structure;
    context_type *context = &context_structure;
    bool_t result = TRUE;
    struct methodblock *mb;
    static ClassClass *string_class = 0;
    void **addr;
    int i;

    if (string_class == 0) 
	string_class = FindClass(0, JAVAPKG "String", FALSE);
#ifdef DEBUG
    GlobalContext = context;
#endif

    /* Initialize the class-wide fields of the context. */
    context->class = cb;
    context->classHash = 0;
    context->object_info = MAKE_CLASSNAME_INFO(JAVAPKG "Object", &addr); 
    *addr = classobject;
    context->string_info = MAKE_CLASSNAME_INFO(JAVAPKG "String", &addr);
    *addr = string_class;
    context->currentclass_info = MAKE_CLASSNAME_INFO(cb->name, &addr);
    *addr = cb;
    
    /* Look at each method */
    for (i = cb->methods_count, mb = cb->methods; --i >= 0; mb++) {
	if ((mb->fb.access & (ACC_NATIVE | ACC_ABSTRACT)) == 0)
	    if (!verify_code(context, mb))
		result = FALSE;
	}
    /* Cleanup */
    Str2IDFree(&context->classHash);
#ifdef DEBUG
    GlobalContext = 0;
#endif
    return result;
}



/* Verify the code of one method */

static bool_t
verify_code(context_type *context, struct methodblock *mb)
{
    unsigned char *code = mb->code;
    int code_length = mb->code_length;
    short *code_data = (short *)malloc(code_length * sizeof(short));
    instruction_data_type *idata = 0;
    int instruction_count;
    bool_t result = FALSE;	/* default */

    if (verify_verbose)
	printf("Looking at %s.%s%s 0x%x\n", 
	       fieldclass(&mb->fb)->name, mb->fb.name, mb->fb.signature, mb);
    /* Run through the code.  Mark the start of each instruction, and give
     * the instruction a number */
    {
	int i;
	int offset;
	
	for (i = 0, offset = 0; offset < code_length; i++) {
	    int length = instruction_length(&code[offset]);
	    int next_offset = offset + length;
	    if (length <= 0) {
		CCerror(context, 
			"Illegal instruction found at offset %d", offset);
		goto failure_return;
	    } else if (next_offset > code_length) {
		CCerror(context, "Code stops in the middle of instruction "
			" starting at offset %d", offset);
		goto failure_return;
	    } else {
		code_data[offset] = i;
		while (++offset < next_offset)
		    code_data[offset] = -1; /* illegal location */
	    }
	}
	instruction_count = i;	/* number of instructions in code */
    }
    
    /* Allocate a structure to hold info about each instruction. */
    idata = (instruction_data_type *)
             malloc(instruction_count * sizeof(instruction_data_type));

    /* Initialize the heap, and other information in the context structure. */
    CCinit(context);		/* initial heap */
    context->mb = mb;
    context->code = code;
    context->instruction_data = idata;
    context->code_data = code_data;
    context->instruction_count = instruction_count;
    context->handler_info = NEW(struct handler_info_type, 
				  mb->exception_table_length);
    
    if (instruction_count == 0) {
	CCerror(context, "Empty code");
	goto failure_return;
    }

    /* Run through the code again, filling in instruction_data */
    {
	int inumber;
	int offset;
	bool_t error = FALSE;
	for (inumber = 0, offset = 0; offset < code_length; inumber++) {
	    int length = instruction_length(&code[offset]);
	    instruction_data_type *this_idata = &idata[inumber];
	    this_idata->opcode = code[offset];
#ifdef DEBUG
	    this_idata->operand.i = 0;
#endif
	    this_idata->stack = NULL;
	    this_idata->stack_size  = UNKNOWN_STACK_SIZE;
	    this_idata->registers_used = UNKNOWN_REGISTERS_USED;
	    this_idata->changed = FALSE;  /* no need to look at it yet. */
	    /* This also sets up this_data->operand.  It also makes the 
	     * xload_x and xstore_x instructions look like the generic form. */
	    if (!verify_opcode_operands(context, inumber, offset)) {
		CCerror(context, 
			"Illegal operand found in instruction at offset %d", 
			offset);
		error = TRUE;
	    }
	    offset += length;
	}
	if (error) 
	    goto failure_return;
    }

    if (   /* make sure exception table is reasonable. */
	   verify_exception_table(context) && 
	   /* Set up first instruction, and start of exception handlers. */
	   initialize_verifier_data_flow(context) &&
	   /* Run data flow analysis on the instructions. */
	   run_verifier_data_flow(context)) {
	result = TRUE;
    }
    
  failure_return:
    if (idata) 
	free(idata);
    if (code_data) 
	free(code_data);
    CCdestroy(context);		/* destory heap */
    if (verify_verbose)
	printf("%s\n", result ? "Okay" : "Failure");
    return result;
}

/* Look at a single instruction, and verify its operands.  Also, for 
 * simplicity, move the operand into the ->operand field.
 * Make sure that branches don't go into the middle of nowhere.
 */

static bool_t
verify_opcode_operands(context_type *context, int inumber, int offset)
{
    instruction_data_type *idata = context->instruction_data;
    instruction_data_type *this_idata = &idata[inumber];
    short *code_data = context->code_data;
    struct methodblock *mb = context->mb;
    unsigned char *code = context->code;
    opcode_type opcode = this_idata->opcode;
    int var; 
    
    switch (opcode) {

    case opc_ifeq: case opc_ifne: case opc_iflt: 
    case opc_ifge: case opc_ifgt: case opc_ifle:
    case opc_if_icmpeq: case opc_if_icmpne: case opc_if_icmplt: 
    case opc_if_icmpge: case opc_if_icmpgt: case opc_if_icmple:
    case opc_if_acmpeq: case opc_if_acmpne:   
    case opc_goto: case opc_jsr: {
	/* Set the ->operand to be the instruction number of the target. */
	int jump = (((signed char)(code[offset+1])) << 8) + code[offset+2];
	int target = offset + jump;
	if (!verify_target(context, target)) {
	    CCerror(context, "Illegal target of jump or branch");
	    return FALSE;
	} else {
	    this_idata->operand.i = code_data[target];
	    return TRUE;
	}
    }
	
    case opc_tableswitch: 
    case opc_lookupswitch: {
	/* Set the ->operand to be a table of possible instruction targets. */
	long *lpc = (long *) UCALIGN(code + offset + 1);
	long *lptr;
	int *saved_operand;
	int keys;
	int k, delta;
	if (opcode == opc_tableswitch) {
	    keys = ntohl(lpc[2]) -  ntohl(lpc[1]) + 1;
	    delta = 1;
	} else { 
	    keys = ntohl(lpc[1]); /* number of pairs */
	    delta = 2;
	}
	saved_operand = NEW(int, keys + 2);
	if (!verify_target(context, offset + ntohl(lpc[0]))) {
	    CCerror(context, "Illegal default target in switch");
	    return FALSE;
	}
	saved_operand[keys + 1] = code_data[offset + ntohl(lpc[0])];
	for (k = keys, lptr = &lpc[3]; --k >= 0; lptr += delta) {
	    int target = offset + ntohl(lptr[0]);
	    if (!verify_target(context, target)) {
		CCerror(context, "Illegal branch in opc_tableswitch");
		return FALSE;
	    }
	    saved_operand[k + 1] = code_data[target];
	}
	saved_operand[0] = keys + 1; /* number of successors */
	this_idata->operand.ip = saved_operand;
	return TRUE;
    }
	
    case opc_ldc1: {   
	/* Make sure the constant pool item is the right type. */
	int key = code[offset + 1];
	int types = (1 << CONSTANT_Integer) | (1 << CONSTANT_Float) |
	                    (1 << CONSTANT_String);
	this_idata->operand.i = key;
	return verify_constant_pool_type(context, key, types);
    }
	  
    case opc_ldc2: {   
	/* Make sure the constant pool item is the right type. */
	int key = (code[offset + 1] << 8) + code[offset + 2];
	int types = (1 << CONSTANT_Integer) | (1 << CONSTANT_Float) |
	    (1 << CONSTANT_String);
	this_idata->operand.i = key;
	return verify_constant_pool_type(context, key, types);
    }
	  
    case opc_ldc2w: { 
	/* Make sure the constant pool item is the right type. */
	int key = (code[offset + 1] << 8) + code[offset + 2];
	int types = (1 << CONSTANT_Double) | (1 << CONSTANT_Long);
	this_idata->operand.i = key;
	return verify_constant_pool_type(context, key, types);
    }

    case opc_getfield: case opc_putfield:
    case opc_getstatic: case opc_putstatic: {
	/* Make sure the constant pool item is the right type. */
	int key = (code[offset + 1] << 8) + code[offset + 2];
	this_idata->operand.i = key;
	return verify_constant_pool_type(context, key, 1 << CONSTANT_Fieldref);
    }
	
    case opc_invokevirtual:
    case opc_invokenonvirtual:
    case opc_invokestatic: {
	/* Make sure the constant pool item is the right type. */
	int key = (code[offset + 1] << 8) + code[offset + 2];
	this_idata->operand.i = key;
	return verify_constant_pool_type(context, key, 1 << CONSTANT_Methodref);
    }
	
    case opc_invokeinterface: {
	/* Make sure the constant pool item is the right type. */
	int key = (code[offset + 1] << 8) + code[offset + 2];
	this_idata->operand.i = key;
	return verify_constant_pool_type(context, key, 
					 1 << CONSTANT_InterfaceMethodref);
    }
	

    case opc_instanceof: 
    case opc_checkcast: 
    case opc_new:
    case opc_anewarray: 
    case opc_multianewarray: {
	/* Make sure the constant pool item is a class */
	int key = (code[offset + 1] << 8) + code[offset + 2];
	if (!verify_constant_pool_type(context, key, 1 << CONSTANT_Class)) {
	    return FALSE;
	} else {
	    fullinfo_type target = 
		cp_index_to_class_fullinfo(context, key, FALSE);
	    if (GET_ITEM_TYPE(target) == ITEM_Bogus) {
		/* Some error has happened */
		return FALSE;
	    } else if (opcode == opc_anewarray)
		this_idata->operand.fi = MAKE_FULLINFO(GET_ITEM_TYPE(target),
						      GET_INDIRECTION(target)+1,
						       GET_EXTRA_INFO(target));
	    else 
		this_idata->operand.fi = target;

	    if (opcode == opc_multianewarray) {
		this_idata->operand2 = code[offset + 3];
		if (GET_INDIRECTION(target) < this_idata->operand2)
		    return FALSE;
	    } else if (opcode == opc_new) {
		if (GET_INDIRECTION(target) != 0) 
		    return FALSE;
	    }
	    return TRUE;
	}
    }
	
    case opc_newarray: {
	/* Cache the result of the opc_newarray into the operand slot */
	fullinfo_type full_info;
	switch (code[offset + 1]) {
	    case T_INT:    
	        full_info = MAKE_FULLINFO(ITEM_Integer, 1, 0); break;
	    case T_LONG:   
		full_info = MAKE_FULLINFO(ITEM_Long, 1, 0); break;
	    case T_FLOAT:  
		full_info = MAKE_FULLINFO(ITEM_Float, 1, 0); break;
	    case T_DOUBLE: 
		full_info = MAKE_FULLINFO(ITEM_Double, 1, 0); break;
	    case T_BYTE:   
		full_info = MAKE_FULLINFO(ITEM_Byte, 1, 0); break;
	    case T_CHAR:   
		full_info = MAKE_FULLINFO(ITEM_Char, 1, 0); break;
	    case T_SHORT:  
		full_info = MAKE_FULLINFO(ITEM_Short, 1, 0); break;
	    case T_ARRAY:
		/* Treat this as Object[], an array that can hold anything. */
		/* This should become obsolete at some point */
		full_info = MAKE_FULLINFO(ITEM_Object, 1, 
					  GET_EXTRA_INFO(context->object_info));
		break;
	    default:
		CCerror(context, "Bad type passed to opc_newarray");
		return FALSE;
	}
	this_idata->operand.fi = full_info;
	return TRUE;
    }
	  
    /* Fudge iload_x, aload_x, etc to look like their generic cousin. */
    case opc_iload_0: case opc_iload_1: case opc_iload_2: case opc_iload_3:
	this_idata->opcode = opc_iload;
	var = opcode - opc_iload_0;
	goto check_local_variable;
	  
    case opc_fload_0: case opc_fload_1: case opc_fload_2: case opc_fload_3:
	this_idata->opcode = opc_fload;
	var = opcode - opc_fload_0;
	goto check_local_variable;

    case opc_aload_0: case opc_aload_1: case opc_aload_2: case opc_aload_3:
	this_idata->opcode = opc_aload;
	var = opcode - opc_aload_0;
	goto check_local_variable;

    case opc_lload_0: case opc_lload_1: case opc_lload_2: case opc_lload_3:
	this_idata->opcode = opc_lload;
	var = opcode - opc_lload_0;
	goto check_local_variable2;

    case opc_dload_0: case opc_dload_1: case opc_dload_2: case opc_dload_3:
	this_idata->opcode = opc_dload;
	var = opcode - opc_dload_0;
	goto check_local_variable2;

    case opc_istore_0: case opc_istore_1: case opc_istore_2: case opc_istore_3:
	this_idata->opcode = opc_istore;
	var = opcode - opc_istore_0;
	goto check_local_variable;
	
    case opc_fstore_0: case opc_fstore_1: case opc_fstore_2: case opc_fstore_3:
	this_idata->opcode = opc_fstore;
	var = opcode - opc_fstore_0;
	goto check_local_variable;

    case opc_astore_0: case opc_astore_1: case opc_astore_2: case opc_astore_3:
	this_idata->opcode = opc_astore;
	var = opcode - opc_astore_0;
	goto check_local_variable;

    case opc_lstore_0: case opc_lstore_1: case opc_lstore_2: case opc_lstore_3:
	this_idata->opcode = opc_lstore;
	var = opcode - opc_lstore_0;
	goto check_local_variable2;

    case opc_dstore_0: case opc_dstore_1: case opc_dstore_2: case opc_dstore_3:
	this_idata->opcode = opc_dstore;
	var = opcode - opc_dstore_0;
	goto check_local_variable2;

    case opc_iinc:
	this_idata->operand2 = (signed char)code[offset + 2];
    case opc_ret: 
    case opc_aload: case opc_iload: case opc_fload:
    case opc_astore: case opc_istore: case opc_fstore:
	var = code[offset + 1];
    check_local_variable:
	/* Make sure that the variable number isn't illegal. */
	this_idata->operand.i = var;
	if (var >= (int)mb->nlocals) {
	    CCerror(context, "Illegal local variable number");
	    return FALSE;
	} 
	return TRUE;
	    
    case opc_lload: case opc_dload: case opc_lstore: case opc_dstore: 
	var = code[offset + 1];
    check_local_variable2:
	/* Make sure that the variable number isn't illegal. */
	this_idata->operand.i = var;
	if ((var + 1) >= (int)mb->nlocals) {
	    CCerror(context, "Illegal local variable number");
	    return FALSE;
	} 
	return TRUE;
	
    default:
	if (opcode > opc_breakpoint) {
	    CCerror(context, "Quick instructions shouldn't appear yet.");
	    return FALSE;
	    }
	return TRUE;
    } /* of switch */
}


/* Look through each item on the exception table.  Each of the fields must 
 * refer to a legal instruction. 
 */
bool_t verify_exception_table(context_type *context)
{
    struct methodblock *mb = context->mb;
    struct CatchFrame *exception_table = mb->exception_table;
    struct handler_info_type *handler_info = context->handler_info;
    short *code_data = context->code_data;
    int i;
    for (i = mb->exception_table_length; --i >= 0; 
	 exception_table++, handler_info++) {
	int start = exception_table->start_pc;
	int end = exception_table->end_pc;
	int handler = exception_table->handler_pc;
	int catchType = exception_table->catchType;
	stack_item_type *stack_item = NEW(stack_item_type, 1);
	if (!(  start <= end && start >= 0 
	      && verify_target(context, start) 
	      && verify_target(context, end))) {
	    CCerror(context, "Illegal exception table range");
	    return FALSE;
	}
	if (!((handler > 0) && verify_target(context, handler))) {
	    CCerror(context, "Illegal exception table handler");
	    return FALSE;
	}
	handler_info->start = code_data[start];
	handler_info->end = code_data[end];
	handler_info->handler = code_data[handler];
	handler_info->stack = stack_item;
	stack_item->next = NULL;
	if (catchType != 0) {
	    union cp_item_type *cp = context->class->constantpool;
	    char *classname = GetClassConstantClassName(cp, catchType);
	    stack_item->item = MAKE_CLASSNAME_INFO(classname, 0);
	} else {
	    stack_item->item = context->object_info;
	}
    }
    return TRUE;
}


/* Given a pointer to an instruction, return its length.  Use the table
 * opcode_length[] which is automatically built.
 */
static int instruction_length(unsigned char *iptr)
{
    int instruction = *iptr;
    switch (instruction) {
        case opc_tableswitch: {
	    long *lpc = (long *) UCALIGN(iptr + 1);
	    int low = ntohl(lpc[1]);
	    int high = ntohl(lpc[2]);
	    if ((low > high) || (low + 1000 < high)) 
		return -1;	/* illegal */
	    else 
		return (unsigned char *)(&lpc[(high - low + 1) + 3]) - iptr;
	}
	    
	case opc_lookupswitch: {
	    long *lpc = (long *) UCALIGN(iptr + 1);
	    int npairs = ntohl(lpc[1]);
	    if (npairs < 0 || npairs >= 1000) 
		return  -1;
	    else 
		return (unsigned char *)(&lpc[2 * (npairs + 1)]) - iptr;
	}

	default: {
	    /* A length of 0 indicates an error. */
	    int length = opcode_length[instruction];
	    return (length <= 0) ? -1 : length;
	}
    }
}


/* Given the target of a branch, make sure that it's a legal target. */
static bool_t 
verify_target(context_type *context, int offset)
{
    struct methodblock *mb = context->mb;
    int code_length = mb->code_length;
    short *code_data = context->code_data;
    return (offset >= 0 && offset < code_length && code_data[offset] >= 0);
}


/* Make sure that an element of the constant pool really is of the indicated 
 * type.
 */
static bool_t 
verify_constant_pool_type(context_type *context, int index, unsigned mask)
{
    union cp_item_type *constant_pool = context->class->constantpool;
    unsigned char *type_table = 
           constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;
    unsigned type = CONSTANT_POOL_TYPE_TABLE_GET_TYPE(type_table, index);
    if ((mask & (1 << type)) == 0) {
	CCerror(context, "Illegal type in constant pool");
	return FALSE;
    }  else {
	return TRUE;
    }
}
        


static bool_t
initialize_verifier_data_flow(context_type *context) 
{
    instruction_data_type *idata = context->instruction_data;
    struct methodblock *mb = context->mb;
    fullinfo_type *reg_ptr;
    fullinfo_type full_info;
    char *p;

    /* Initialize the function entry, since we know everything about it. */
    idata[0].stack_size = 0;
    idata[0].stack = NULL;
    idata[0].registers_used = mb->args_size;
    idata[0].registers = NEW(fullinfo_type, mb->args_size);
    idata[0].register_mask = 0;
    reg_ptr = idata[0].registers;
    if ((mb->fb.access & ACC_STATIC) == 0) 
	*reg_ptr++ = context->currentclass_info;
    for (p = mb->fb.signature + 1; *p != SIGNATURE_ENDFUNC; ) {
	char fieldchar = signature_to_fieldtype(context, &p, &full_info);
	switch (fieldchar) {
	    case 'D': case 'L': 
	        *reg_ptr++ = full_info;
	        *reg_ptr++ = full_info + 1;
		break;
	    default:
	        *reg_ptr++ = full_info;
		break;
	}
    }
    p++;			/* skip over right parenthesis */
    if (*p == 'V')
	context->return_type = MAKE_FULLINFO(ITEM_Void, 0, 0);
    else {
	signature_to_fieldtype(context, &p, &full_info);
	context->return_type = full_info;
    }

    /* Indicate that we need to look at this first instruction. */
    idata[0].changed = TRUE;
    return TRUE;
}	


/* Run the data flow analysis, as long as there are things to change. */
static bool_t 
run_verifier_data_flow(context_type *context) {
    struct methodblock *mb = context->mb;
    int max_stack_size = mb->maxstack;
    instruction_data_type *idata = context->instruction_data;
    int icount = context->instruction_count;
    bool_t work_to_do = TRUE;
    int inumber;

    /* Run through the loop, until there is nothing left to do. */
    while (work_to_do) {
	work_to_do = FALSE;
	for (inumber = 0; inumber < icount; inumber++) {
	    instruction_data_type *this_idata = &idata[inumber];
	    if (this_idata->changed) {
		int new_registers_used, new_stack_size;
		fullinfo_type *new_registers;
		bitvector new_register_mask;
		stack_item_type *new_stack; 
		
		this_idata->changed = FALSE;
		work_to_do = TRUE;
		if (verify_verbose) {
		    printf("  Instruction %d: ", inumber);
		    print_stack(context, this_idata->stack);
		    print_registers(context, this_idata->registers_used, 
				    this_idata->registers, 
				    this_idata->register_mask);
		    printf("  %s(%d)", opnames[this_idata->opcode], 
			   this_idata->operand.i);
		    fflush(stdout);
		}
		/* Make sure the registers can deal with this instruction */
		if (!verify_registers(context, inumber))
		    return FALSE;
		/* Make sure the stack can deal with this instruction */
		if (!verify_and_pop_stack(context, inumber, 
					  &new_stack_size, &new_stack))
		    return FALSE;
		/* Update the registers and stack. */
		update_registers(context, inumber, 
				 &new_registers_used, &new_registers, 
				 &new_register_mask);
		update_push_stack(context, inumber, 
				  &new_stack_size, &new_stack);
		if (new_stack_size > max_stack_size) {
		    CCerror(context, "Stack size too large");
		    return FALSE;
		}
		/* Add the new stack and register information to any
		 * instructions that can follow this instruction.     */
		if (!verify_merge_into_successors(
			context, inumber, 
                        new_registers_used, new_registers, new_register_mask,
		        new_stack_size, new_stack))
		    return FALSE;
		if (verify_verbose) {
		    printf("  ");
		    print_stack(context, new_stack);
		    print_registers(context, new_registers_used, new_registers,
				    new_register_mask);
		    printf("\n");
		}
	    }
	}
    }
    return TRUE;
}


/* Make sure that the registers contain a legitimate value for the given
 * instruction.
*/

static bool_t
verify_registers(context_type *context, int inumber)
{
    instruction_data_type *idata = context->instruction_data;
    instruction_data_type *this_idata = &idata[inumber];
    opcode_type opcode = this_idata->opcode;
    int operand = this_idata->operand.i;
    int registers_used = this_idata->registers_used;
    fullinfo_type *registers = this_idata->registers;
    bool_t double_word = FALSE;	/* default value */
    int type;
    
    switch (opcode) {
        default:
	    return TRUE;
        case opc_iload: case opc_iinc:
	    type = ITEM_Integer; break;
        case opc_fload:
	    type = ITEM_Float; break;
        case opc_aload:
	    type = ITEM_Object; break;
        case opc_ret:
	    type = ITEM_ReturnAddress; break;
        case opc_lload:	
	    type = ITEM_Long; double_word = TRUE; break;
        case opc_dload:
	    type = ITEM_Double; double_word = TRUE; break;
    }
    if (!double_word) {
	/* Make sure we don't have an illegal register or one with wrong type */
	if (operand >= registers_used) {
	    CCerror(context, 
		    "Accessing value from uninitialized register %d", operand);
	    return FALSE;
	} else if (WITH_ZERO_EXTRA_INFO(registers[operand]) == 
	                          MAKE_FULLINFO(type, 0, 0)) {
	    return TRUE;
	} else if (GET_INDIRECTION(registers[operand]) > 0 
		                         && type == ITEM_Object) {
	    /* address type stuff be used on all arrays */
	    return TRUE;
	} else if (GET_ITEM_TYPE(registers[operand]) == ITEM_ReturnAddress
		   && (type == ITEM_Object || type == ITEM_Integer)) {
	    /* We might use iload or aload on an address.  Oh well. */
	    return TRUE;
	} else {
	    CCerror(context, "Register %d contains wrong type", operand);
	    return FALSE;
	}
    } else {
	/* Make sure we don't have an illegal register or one with wrong type */
	if ((operand + 1) >= registers_used) {
	    CCerror(context, 
		    "Accessing value from uninitialized register pair %d/%d", 
		    operand, operand+1);
	    return FALSE;
	} else {
	    if ((registers[operand] == MAKE_FULLINFO(type, 0, 0)) &&
		(registers[operand + 1] == MAKE_FULLINFO(type + 1, 0, 0))) {
		return TRUE;
	    } else {
		CCerror(context, "Register pair %d/%d contains wrong type", 
		        operand, operand+1);
		return FALSE;
	    }
	}
    } 
}


/* Make sure that the top of the stack contains reasonable values for the 
 * given instruction.  The post-pop values of the stack and its size are 
 * returned in *new_stack_size_p  and *new_stack_p.
 */

static bool_t 
verify_and_pop_stack(context_type *context, int inumber, 
		     int *new_stack_size_p, stack_item_type **new_stack_p)
{
    instruction_data_type *idata = context->instruction_data;
    instruction_data_type *this_idata = &idata[inumber];
    opcode_type opcode = this_idata->opcode;
    stack_item_type *stack = this_idata->stack;
    int stack_size = this_idata->stack_size;
    char *stack_operands, *p;
    char buffer[128];
    fullinfo_type stack_extra_info_buffer[128]; /* save info popped off stack */
    fullinfo_type *stack_extra_info = &stack_extra_info_buffer[128];
    fullinfo_type full_info;
    
    switch(opcode) {
        default:
	    /* For most instructions, we just use a built-in table */
	    stack_operands = opcode_in_out[opcode][0];
	    break;

        case opc_putstatic: case opc_putfield: {
	    /* The top thing on the stack depends on the signature of
	     * the object.                         */
	    char *signature = 
		cp_index_to_signature(context, this_idata->operand.i);
	    char *ip = buffer;
	    if (verify_verbose) 
		print_formatted_fieldname(context, this_idata->operand.i);
	    if (opcode == opc_putfield)
		*ip++ = 'A';	/* object for putfield */
	    *ip++ = signature_to_fieldtype(context, &signature, &full_info);
	    *ip = '\0';
	    stack_operands = buffer;
	    break;
	}

	case opc_invokevirtual: case opc_invokenonvirtual:
	case opc_invokestatic: case opc_invokeinterface: {
	    /* The top stuff on the stack depends on the method signature */
	    char *signature = 
		cp_index_to_signature(context, this_idata->operand.i);
	    char *ip = buffer;
	    char *p;
	    if (verify_verbose) 
		print_formatted_fieldname(context, this_idata->operand.i);
	    if (opcode != opc_invokestatic) 
		/* First, push the object */
		*ip++ = 'A';
	    for (p = signature + 1; *p != SIGNATURE_ENDFUNC; )
		*ip++ = signature_to_fieldtype(context, &p, &full_info);
	    *ip = 0;
	    stack_operands = buffer;
	    break;
	}

	case opc_multianewarray: {
	    /* the count field gives the number of integers on the stack. */
	    int count = this_idata->operand2;
	    memset(buffer, 'I', count);
	    buffer[count] = '\0';
	    stack_operands = buffer;
	    break;
	}

        case opc_verifystack:
	    /* Verify statically that the stack is empty. */
	    if (stack_size != 0) {
		CCerror(context, "opc_verifystack called when stack non-zero");
		return FALSE;
	    }
	    stack_operands = "";
	    break;
    } /* of switch */

    /* Run through the list of operands >>backwards<< */
    for (   p = stack_operands + strlen(stack_operands);
	    p > stack_operands; 
	    stack = stack->next) {
	int type = *--p;
	fullinfo_type top_type = stack ? stack->item : 0;
	int size = (type == 'D' || type == 'L') ? 2 : 1;
	*--stack_extra_info = top_type;
	if (stack == NULL) {
	    CCerror(context, "Unable to pop operand off an empty stack");
	    return FALSE; 
	} 
	switch (type) {
	    case 'I': 
	        if (top_type == MAKE_FULLINFO(ITEM_Integer, 0, 0))
		    break;
		if ((opcode == opc_istore) && 
		    (GET_ITEM_TYPE(top_type) == ITEM_ReturnAddress) &&
		    (GET_INDIRECTION(top_type) == 0))
		    break;
		CCerror(context, "Expecting to find integer on stack");
		return FALSE;
	    case 'F': 
		if (top_type == MAKE_FULLINFO(ITEM_Float, 0, 0))
		    break;
		CCerror(context, "Expecting to find float on stack");
		return FALSE;
	    case 'A':		/* object or array */
		if (GET_ITEM_TYPE(top_type) == ITEM_Object) 
		    break;
		if (GET_INDIRECTION(top_type) > 0)
		    break;
		if ((opcode == opc_astore) && 
		    (GET_ITEM_TYPE(top_type) == ITEM_ReturnAddress) && 
		    (GET_INDIRECTION(top_type) == 0))
		    break;
		CCerror(context, "Expecting to find object/array on stack");
		return FALSE;
	    case 'O':		/* object */
		if (GET_ITEM_TYPE(top_type) == ITEM_Object &&
		       GET_INDIRECTION(top_type) == 0) 
		    break;
		CCerror(context, "Expecting to find object on stack");
		return FALSE;
	    case 'a':		/* integer, object, or array */
	        if (top_type == MAKE_FULLINFO(ITEM_Integer, 0, 0))
		    break;
		if (GET_ITEM_TYPE(top_type) == ITEM_Object)
		    break;
		if (GET_INDIRECTION(top_type) > 0)
		    break;
		CCerror(context, "Expecting to find object, array, or int on stack");
		return FALSE;
	    case 'D':		/* double */
		if (top_type == MAKE_FULLINFO(ITEM_Double, 0, 0))
		    break;
		CCerror(context, "Expecting to find double on stack");
		return FALSE;
	    case 'L':		/* long */
		if (top_type == MAKE_FULLINFO(ITEM_Long, 0, 0))
		    break;
		CCerror(context, "Expecting to find long on stack");
		return FALSE;
	    case ']':		/* array of some type */
		switch(p[-1]) {
		    case 'I':	/* array of integers */
		        if (top_type == MAKE_FULLINFO(ITEM_Integer, 1, 0))
			    break;
			CCerror(context, 
				"Expecting to find array of ints on stack");
			return FALSE;
		    case 'L':	/* array of longs */
		        if (top_type == MAKE_FULLINFO(ITEM_Long, 1, 0))
			    break;
			CCerror(context, 
				"Expecting to find array of longs on stack");
			return FALSE;
		    case 'F':	/* array of floats */
		        if (top_type == MAKE_FULLINFO(ITEM_Float, 1, 0))
			    break;
			CCerror(context, 
				"Expecting to find array of floats on stack");
			return FALSE;
		    case 'D':	/* array of doubles */
		        if (top_type == MAKE_FULLINFO(ITEM_Double, 1, 0))
			    break;
			CCerror(context, 
				"Expecting to find array of doubles on stack");
			return FALSE;
		    case 'A':	/* array of addresses (arrays or objects) */
			if (GET_INDIRECTION(top_type) > 1 || 
			    (GET_ITEM_TYPE(top_type) == ITEM_Object &&
			     GET_INDIRECTION(top_type) >= 1))
			    break;
			CCerror(context, 
				"Expecting to find array of objects or arrays "
				"on stack");
			return FALSE;
		    case 'B':	/* array of bytes */
		        if (top_type == MAKE_FULLINFO(ITEM_Byte, 1, 0))
			    break;
			CCerror(context, 
				"Expecting to find array of bytes on stack");
			return FALSE;
		    case 'C':	/* array of characters */
		        if (top_type == MAKE_FULLINFO(ITEM_Char, 1, 0))
			    break;
			CCerror(context, 
				"Expecting to find array of chars on stack");
			return FALSE;
		    case 'S':	/* array of characters */
		        if (top_type == MAKE_FULLINFO(ITEM_Short, 1, 0))
			    break;
			CCerror(context, 
				"Expecting to find array of shorts on stack");
			return FALSE;
		    case '?':	/* any type of array is okay */
		        if (GET_INDIRECTION(top_type) >= 1) 
			    break;
			CCerror(context, 
				"Expecting to find array of floats on stack");
			return FALSE;
		    default:
			CCerror(context, "Internal error #1");
			assert(FALSE);
		}
		p -= 2;		/* skip over char and [ */
		break;

	    case '1': case '2': case '3': case '4': /* stack swapping */
		if (top_type == MAKE_FULLINFO(ITEM_Double, 0, 0) 
		    || top_type == MAKE_FULLINFO(ITEM_Long, 0, 0)) {
		    if ((p > stack_operands) && (p[-1] == '+')) {
			context->swap_table[type - '1'] = top_type + 1;
			context->swap_table[p[-2] - '1'] = top_type;
			size = 2;
			p -= 2;
		    } else {
			CCerror(context, 
				"Attempt to split long or double on the stack");
			return FALSE;
		    }
		} else {
		    context->swap_table[type - '1'] = stack->item;
		    if ((p > stack_operands) && (p[-1] == '+')) 
			p--;	/* ignore */
		}
		break;
	    case '+':		/* these should have been caught. */
	    default:
		CCerror(context, "Internal error #2");
		assert(FALSE);
	}
	stack_size -= size;
    }
    
    /* For many of the opcodes that had an "A" in their field, we really 
     * need to go back and do a little bit more accurate testing.  We can, of
     * course, assume that the minimal type checking has already been done. 
     */
    switch (opcode) {
	case opc_aastore: {	/* array index object  */
	    fullinfo_type array_type = stack_extra_info[0];
	    fullinfo_type object_type = stack_extra_info[2];
	    fullinfo_type target_type = decrement_indirection(array_type);
	    if ((WITH_ZERO_EXTRA_INFO(target_type) == 
		             MAKE_FULLINFO(ITEM_Object, 0, 0)) &&
		 (WITH_ZERO_EXTRA_INFO(object_type) == 
		             MAKE_FULLINFO(ITEM_Object, 0, 0))) {
		/* I disagree.  But other's seem to think that we should allow
		 * an assignment of any Object to any array of any Object type.
		 * There will be an runtime error if the types are wrong.
		 */
		break;
	    }
	    if (is_assignable_to(context, object_type, target_type))
		break;
	    CCerror(context, "Incompatible types for storing into array of "
		              "arrays or objects");
	    return FALSE;
	}

	case opc_newfromname:	/* name */
	    if (!is_assignable_to(context, 
				  stack_extra_info[0], context->string_info)) {
		CCerror(context, "String required");
		return FALSE;
	    }
	    break;

	case opc_putfield: 
	case opc_getfield: {	
	    int operand = this_idata->operand.i;
	    if (!is_assignable_to(context, 
				  stack_extra_info[0],
				  cp_index_to_class_fullinfo(context, operand,
							     TRUE))) {
		CCerror(context, 
			"Incompatible type for getting or setting field");
		return FALSE;
	    }
	    break;
	}

	case opc_aaload: {	/* array index */
	    /* We need to pass the information to the stack updater */
	    fullinfo_type array_type = stack_extra_info[0];
	    context->swap_table[0] = decrement_indirection(array_type);
	    break;
	}
	    
        case opc_invokevirtual: case opc_invokenonvirtual:
	case opc_invokeinterface: 
	case opc_invokestatic: {
	    char *signature = 
		cp_index_to_signature(context, this_idata->operand.i);
	    int item;
	    char *p;
	    if (opcode == opc_invokestatic) {
		item = 0;
	    } else {
		fullinfo_type target_type = 
		    cp_index_to_class_fullinfo(context, this_idata->operand.i, TRUE);
		if (!is_assignable_to(context, 
				      stack_extra_info[0], target_type)) {
		    CCerror(context, 
			    "Incompatible object argument for function call");
		    return FALSE;
		}
		item = 1;
	    }
	    for (p = signature + 1; *p != SIGNATURE_ENDFUNC; item++)
		if (signature_to_fieldtype(context, &p, &full_info) == 'A') {
		    if (!is_assignable_to(context, 
					  stack_extra_info[item], full_info)) {
			CCerror(context, "Incompatible argument to function");
			return FALSE;
		    }
		}

	    break;
	}
	    
	case opc_ireturn: case opc_lreturn: case opc_freturn: case opc_dreturn: 
	case opc_areturn: case opc_return: {
	    fullinfo_type target_type = context->return_type;
	    fullinfo_type object_type = (opcode == opc_return) 
		                          ? MAKE_FULLINFO(ITEM_Void, 0, 0)
					  : stack_extra_info[0];
	    if (!is_assignable_to(context, object_type, target_type)) {
		CCerror(context, "Wrong return type in function");
		return FALSE;
	    }
	    break;
	}
    }

    *new_stack_p = stack;
    *new_stack_size_p = stack_size;
    return TRUE;
}


/* We've already determined that the instruction is legal.  Perform the 
 * operation on the registers, and return the updated results in
 * new_registers_used_p and new_registers.
 */

static void
update_registers(context_type *context, int inumber,
		 int *new_registers_used_p, 
		 fullinfo_type **new_registers_p,
		 bitvector *new_register_mask_p
		 )
{
    instruction_data_type *idata = context->instruction_data;
    instruction_data_type *this_idata = &idata[inumber];
    opcode_type opcode = this_idata->opcode;
    int operand = this_idata->operand.i;
    int registers_used = this_idata->registers_used;
    fullinfo_type *registers = this_idata->registers;
    bitvector register_mask = this_idata->register_mask;
    stack_item_type *stack = this_idata->stack;

    /* Use these as default new values. */
    int new_registers_used = registers_used;
    fullinfo_type *new_registers = registers;
    bitvector new_register_mask = register_mask;
    enum { NONE, SINGLE, DOUBLE } access;
    
    /* Remember, we've already verified the type at the top of the stack. */
    switch (opcode) {
        default:
	    access = NONE;
	    break;

        case opc_istore: case opc_fstore: case opc_astore:
	    access = SINGLE;
	    goto continue_store;

        case opc_lstore: case opc_dstore:
	    access = DOUBLE;
	    goto continue_store;

	continue_store: {
	    /* We have a modification to the registers.  Copy them. */
	    fullinfo_type stack_top_type = stack->item;
	    int max_operand = operand + (access == DOUBLE);
	    int i;
	    
	    if (     max_operand < registers_used 
		  && registers[operand] == stack_top_type
		  && ((access == SINGLE) || 
		         (registers[operand + 1]== stack_top_type + 1))) 
		break;
	    new_registers_used = MAX(max_operand + 1, registers_used);
	    new_registers = NEW(fullinfo_type, new_registers_used);
	    for (i = 0; i < registers_used; i++) 
		new_registers[i] = registers[i];
	    for (i = registers_used; i < new_registers_used; i++) 
		new_registers[i] = MAKE_FULLINFO(ITEM_Bogus, 0, 0);
	    new_registers[operand] = stack_top_type;
	    if (access == DOUBLE)
		new_registers[operand + 1] = stack_top_type + 1;
	    break;
	}
	 
	
        case opc_iload: case opc_fload: case opc_aload:
        case opc_ret: case opc_iinc:
	    access = SINGLE;
	    break;

        case opc_lload: case opc_dload:	
	    access = DOUBLE;
	    break;
    }

    if ((access != NONE) && register_mask) {
	if (!IS_BIT_SET(register_mask, operand) ||
	      (access == DOUBLE && !IS_BIT_SET(register_mask, operand + 1))) {
	    new_register_mask = copy_register_mask(context, register_mask);
	    SET_BIT(new_register_mask, operand);
	    SET_BIT(new_register_mask, operand + 1);
	}
    }

    *new_registers_used_p = new_registers_used;
    *new_registers_p = new_registers;
    *new_register_mask_p = new_register_mask;
}



/* We've already determined that the instruction is legal.  Perform the 
 * operation on the stack;
 *
 * new_stack_size_p and new_stack_p point to the results after the pops have
 * already been done.  Do the pushes, and then put the results back there. 
 */

static void
update_push_stack(context_type *context, int inumber,
		  int *new_stack_size_p, stack_item_type **new_stack_p)
{
    instruction_data_type *idata = context->instruction_data;
    instruction_data_type *this_idata = &idata[inumber];
    opcode_type opcode = this_idata->opcode;
    int operand = this_idata->operand.i;

    int stack_size = *new_stack_size_p;
    stack_item_type *stack = *new_stack_p;
    char *stack_results = opcode_in_out[opcode][1]; 
	
    fullinfo_type full_info = 0;
    char buffer[5], *p;		/* actually [2] is big enough */

    /* We need to look at all those opcodes in which either we can't tell the
     * value pushed onto the stack from the opcode, or in which the value
     * pushed onto the stack is an object or array.  For the latter, we need
     * to make sure that full_info is set to the right value.
     */
    switch(opcode) {
        default:
	    break;

	case opc_ldc1: case opc_ldc2: case opc_ldc2w: {
	    /* Look to constant pool to determine correct result. */
	    union cp_item_type *cp = context->class->constantpool;
	    unsigned char *type_table = cp[CONSTANT_POOL_TYPE_TABLE_INDEX].p;
	    switch (CONSTANT_POOL_TYPE_TABLE_GET_TYPE(type_table, operand)) {
		case CONSTANT_Integer:   
		    stack_results = "I"; break;
		case CONSTANT_Float:     
		    stack_results = "F"; break;
		case CONSTANT_Double:    
		    stack_results = "D"; break;
		case CONSTANT_Long:      
		    stack_results = "L"; break;
		case CONSTANT_String: 
		    stack_results = "A"; 
		    full_info = context->string_info;
		    break;
		default:
		    CCerror(context, "Internal error #3");
		    assert(FALSE);
	    }
	    break;
	}

        case opc_getstatic: case opc_getfield: {
	    /* Look to signature to determine correct result. */
	    char *signature = 
		cp_index_to_signature(context, this_idata->operand.i);
	    if (verify_verbose) 
		print_formatted_fieldname(context, this_idata->operand.i);
	    buffer[0] = signature_to_fieldtype(context, &signature, &full_info);
	    buffer[1] = '\0';
	    stack_results = buffer;
	    break;
	}

	case opc_invokevirtual: case opc_invokenonvirtual:
	case opc_invokestatic: case opc_invokeinterface: {
	    /* Look to signature to determine correct result. */
	    char *signature = 
		cp_index_to_signature(context, this_idata->operand.i);
	    char *result_signature = strchr(signature, SIGNATURE_ENDFUNC) + 1;
	    if (result_signature[0] == SIGNATURE_VOID) {
		stack_results = "";
	    } else {
		buffer[0] = signature_to_fieldtype(context, &result_signature, &
						   full_info);
		buffer[1] = '\0';
		stack_results = buffer;
	    }
	    break;
	}
	    
	case opc_aconst_null:
	    full_info = NULL_FULLINFO; /* special NULL */
	    break;

	case opc_newfromname:
	    full_info = context->object_info;
	    break;

	case opc_new: 
	case opc_checkcast: 
	case opc_newarray: 
	case opc_anewarray: 
        case opc_multianewarray:
	    /* Conventiently, this result type is stored here */
	    full_info = this_idata->operand.fi;
	    break;
			
	case opc_aaload:
	    /* verify_and_pop_stack saved value for us. */
	    full_info = context->swap_table[0];
	    break;
		
	case opc_aload:
	    /* The register hasn't been modified, so we can use its value. */
	    full_info = this_idata->registers[operand];
	    break;
    } /* of switch */
    
    for (p = stack_results; *p != 0; p++) {
	int type = *p;
	stack_item_type *new_item = NEW(stack_item_type, 1);
	new_item->next = stack;
	stack = new_item;
	switch (type) {
	    case 'I': 
	        stack->item = MAKE_FULLINFO(ITEM_Integer, 0, 0); break;
	    case 'F': 
		stack->item = MAKE_FULLINFO(ITEM_Float, 0, 0); break;
	    case 'D': 
		stack->item = MAKE_FULLINFO(ITEM_Double, 0, 0); 
		stack_size++; break;
	    case 'L': 
		stack->item = MAKE_FULLINFO(ITEM_Long, 0, 0); 
		stack_size++; break;
	    case 'R': 
		stack->item = MAKE_FULLINFO(ITEM_ReturnAddress, 0, operand);
		break;
	    case '1': case '2': case '3': case '4': {
		/* Get the info saved in the swap_table */
		fullinfo_type stype = context->swap_table[type - '1'];
		stack->item = stype;
		if (stype == MAKE_FULLINFO(ITEM_Long, 0, 0) || 
		    stype == MAKE_FULLINFO(ITEM_Double, 0, 0)) {
		    stack_size++; p++;
		}
		break;
	    }
	    case 'A': 
		/* full_info should have the appropriate value. */
		assert(full_info != 0);
		stack->item = full_info;
		break;
	    default:
		CCerror(context, "Internal error #4");
		assert(FALSE);

	    } /* switch type */
	stack_size++;
    } /* outer for loop */
    *new_stack_size_p = stack_size;
    *new_stack_p = stack;
}


/* We've performed an instruction, and determined the new registers and stack 
 * value.  Look at all of the possibly subsequent instructions, and merge
 * this stack value into theirs. 
 */

static bool_t 
verify_merge_into_successors(context_type *context, 
			     int inumber, 
			     int registers_used, 
			     fullinfo_type *registers,
			     bitvector register_mask,
			     int stack_size, stack_item_type *stack)
{
    instruction_data_type *idata = context->instruction_data;
    instruction_data_type *this_idata = &idata[inumber];
    opcode_type opcode = this_idata->opcode;
    int operand = this_idata->operand.i;
    struct handler_info_type *handler_info = context->handler_info;
    int handler_info_length = context->mb->exception_table_length;


    int buffer[2];		/* default value for successors */
    int *successors = buffer;	/* table of successors */
    int successors_count;
    int i;
    
    switch (opcode) {
    default:
	successors_count = 1; 
	buffer[0] = inumber + 1;
	break;

    case opc_ifeq: case opc_ifne: case opc_ifgt: 
    case opc_ifge: case opc_iflt: case opc_ifle:
    case opc_if_icmpeq: case opc_if_icmpne: case opc_if_icmpgt: 
    case opc_if_icmpge: case opc_if_icmplt: case opc_if_icmple:
    case opc_if_acmpeq: case opc_if_acmpne:
	successors_count = 2;
	buffer[0] = inumber + 1; buffer[1] = operand;
	break;

    case opc_jsr: 
/*	register_mask = copy_register_mask(context, 0); */
    case opc_goto:
	successors_count = 1;
	buffer[0] = operand;
	break;


    case opc_ireturn: case opc_lreturn: case opc_return:	
    case opc_freturn: case opc_dreturn: case opc_areturn: 
    case opc_athrow:
	/* The testing for the returns is handled in verify_and_pop_stack */
	successors_count = 0;
	break;

    case opc_ret: {
	/* This is slightly slow, but good enough for a seldom used instruction.
	 * The EXTRA_ITEM_INFO of the ITEM_ReturnAddress indicates the
	 * address of the first instruction of the subroutine.  We can return
	 * to 1 after any instruction that jsr's to that instruction.
	 */
	fullinfo_type *registers = this_idata->registers;
	int called_instruction = GET_EXTRA_INFO(registers[operand]);
	int i;
	for (i = context->instruction_count, successors_count = 0; --i >= 0; ) {
	    if ((idata[i].opcode == opc_jsr) && 
		(idata[i].operand.i == called_instruction)) {
		/* Put into buffer, if we haven't overflowed.  Keep count. */
		if (successors_count < sizeof(buffer)/sizeof(buffer[0]))
		    successors[successors_count] = (i + 1);
		successors_count++;
	    }
	}
	if (successors_count >= sizeof(buffer)/sizeof(buffer[0])) {
	    /* We've overflowed the buffer, malloc a bigger one. */
	    successors = NEW(int, successors_count);
	    successors_count = 0; 
	    for (i = context->instruction_count; --i >= 0; ) {
		if ((idata[i].opcode == opc_jsr) && 
		    (idata[i].operand.i == called_instruction))
		    successors[successors_count++] = (i + 1);
	    }
	}
	/* THIS NEEDS TO BE FIXED!!! */
	successors_count = 0;
	/* !!!DEXIF EB OT SDEEN SIHT */
	break;
    }

    case opc_tableswitch:
    case opc_lookupswitch: 
	successors = this_idata->operand.ip; /* use this instead */
	successors_count = *successors++;
	break;
    }

    if (successors_count > 0) {
	if (verify_verbose) 
	    printf("  [");
	for (i = 0; i < successors_count; i++) {
	    int target = successors[i];
	    if (target >= context->instruction_count) {
		CCerror(context, "Falling off the end of the code");
		return FALSE;
	    }
	    if (verify_verbose) 
		printf("%d", target);
	    if (!verify_merge_into_one_successor(
			 context, target, 
			 registers_used, registers, register_mask,
			 stack_size, stack))
		return FALSE;
	    if (verify_verbose) 
		printf(idata[target].changed ? "! " : " ");
	}
    }

    for (i = handler_info_length; --i >= 0; handler_info++) {
	if (handler_info->start <= inumber &&
	          handler_info->end > inumber) {
	    int handler = handler_info->handler;
	    if (verify_verbose) 
		printf("%d*", handler);
	    if (!verify_merge_into_one_successor(
		      context, handler, 
		      registers_used, registers, register_mask,
		      1, handler_info->stack))
		return FALSE;
	    if (verify_verbose) 
		printf(idata[handler].changed ? "! " : " ");
	}
    }
    if (verify_verbose) 
	printf(  "]");
    return TRUE;
}

/* We have a new set of registers and stack values for a given instruction.
 * Merge this new set into the values that are already there. 
 */

static bool_t 
verify_merge_into_one_successor(context_type *context, 
				int inumber, 
				int new_registers_used, 
				fullinfo_type *new_registers,
				bitvector new_register_mask,
				int new_stack_size, stack_item_type *new_stack)
{
    instruction_data_type *idata = context->instruction_data;
    instruction_data_type *this_idata = &idata[inumber];
    int stack_size = this_idata->stack_size;
    stack_item_type *stack = this_idata->stack;

    if (stack_size == UNKNOWN_STACK_SIZE) {
	this_idata->stack_size = new_stack_size;
	this_idata->stack = new_stack;
	this_idata->changed = TRUE;
    } else if (new_stack_size != stack_size) {
	CCerror(context, "Inconsistent stack height %d != %d",
		new_stack_size, stack_size);
	return FALSE;		/* should never happen */
    } else {
	stack_item_type *old, *new;
	for (old = stack, new = new_stack; old != NULL; 
	           old = old->next, new = new->next) {
	    if (!is_assignable_to(context, new->item, old->item)) {
		stack = copy_stack(context, stack);
		for (old = stack, new = new_stack; old != NULL; 
		          old = old->next, new = new->next) {
		    old->item = 
			merge_fullinfo_types(context, old->item, new->item, 
					     FALSE);
		}
		this_idata->stack = stack;
		this_idata->changed = TRUE;
		break;		/* out of copying loop */
	    }
	}
    }
    
    if (this_idata->registers_used == UNKNOWN_REGISTERS_USED) {
	/* The first time we've seen this instruction's registers */
	this_idata->registers_used = new_registers_used;
	this_idata->registers = new_registers;
	this_idata->register_mask = new_register_mask;
	this_idata->changed = TRUE;
    } else {
#define REG_IN_USE(flags, i) ((flags) == 0 || IS_BIT_SET((flags), i))
	/* See if we've got new information on the register set. */
	int registers_used = this_idata->registers_used;
	fullinfo_type *registers = this_idata->registers;
	bitvector register_mask = this_idata->register_mask;
	bool_t copy = FALSE;
	int i, j;
	if ((registers_used > new_registers_used) && !new_register_mask) {
	    /* Any register larger than new_registers_used is now bogus */
	    registers_used = new_registers_used;
	    this_idata->registers_used = new_registers_used;
	    this_idata->changed = TRUE;
	}

	for (i = 0; i < registers_used; i++) {
	    if (REG_IN_USE(new_register_mask, i)) {
		if ((i < new_registers_used) 
		    ? !is_assignable_to(context, new_registers[i], registers[i])
		    : (registers[i] != MAKE_FULLINFO(ITEM_Bogus, 0, 0))) {
		    copy = TRUE;
		    break;
		}
	    }
	}
	if (copy) {
	    /* We need a copy.  So do it. */
	    fullinfo_type *new_set = NEW(fullinfo_type, registers_used);
	    for (j = 0; j < i; j++) 
		new_set[j] =  registers[j];
	    for (j = i; j < registers_used; j++) {
		if (!REG_IN_USE(new_register_mask, j))
		    new_set[j] = registers[j];
		else if (i >= new_registers_used) 
		    new_set[j] = MAKE_FULLINFO(ITEM_Bogus, 0, 0);
		else 
		    new_set[j] = merge_fullinfo_types(context, 
						      new_registers[j], 
						      registers[j], FALSE);
	    }
	    /* Some of the end items might now be bogus. This step isn't 
	     * necessary, but it may save work later. */
	    while (   registers_used > 0
		   && GET_ITEM_TYPE(new_set[registers_used-1]) == ITEM_Bogus) 
		registers_used--;
	    this_idata->registers_used = registers_used;
	    this_idata->registers = new_set;
	    this_idata->changed = TRUE;
	}

	if (register_mask) {
	    if (new_register_mask == 0) {
		this_idata->register_mask = 0;
		this_idata->changed = TRUE;
	    } else {
		int bits = context->mb->nlocals;
		int words = (bits + BITS_PER_INT - 1)/BITS_PER_INT;
		for (i = words - 1; i >= 0; i--) {
		    if (~register_mask[i] ^ new_register_mask[i]) {
			/* need to copy.  register_mask isn't a superset */
			bitvector new_set = 
			        copy_register_mask(context, register_mask);
			for (j = i; j >= 0; j--) 
			    new_set[j] |= new_register_mask[j];
			this_idata->register_mask = new_set;
			this_idata->changed = TRUE;
			break;
		    }
		}
	    }
	}
    }
    return TRUE;
}

/* Make a copy of a stack */

static stack_item_type *
copy_stack(context_type *context, stack_item_type *stack)
{
    int length;
    stack_item_type *ptr;
    for (ptr = stack, length = 0; ptr != NULL; ptr = ptr->next, length++);
    
    if (length > 0) { 
	stack_item_type *new_stack = NEW(stack_item_type, length);
	stack_item_type *new_ptr;
	for (    ptr = stack, new_ptr = new_stack; 
	         ptr != NULL;
	         ptr = ptr->next, new_ptr++) {
	    new_ptr->item = ptr->item;
	    new_ptr->next = new_ptr + 1;
	}
	new_stack[length - 1].next = NULL;
	return new_stack;
    } else {
	return NULL;
    }
}

static bitvector 
copy_register_mask(context_type *context, bitvector register_mask)
{
    int bits = context->mb->nlocals;
    int words = (bits + BITS_PER_INT - 1) / BITS_PER_INT;
    int bytes = words * sizeof(int);
    bitvector result = NEW(unsigned int, words);
    if (register_mask) 
	memcpy(result, register_mask, bytes);
    else 
	memset(result, 0, bytes);
    return result;
}





/* We create our own storage manager, since we malloc lots of little items, 
 * and I don't want to keep trace of when they become free.  I sure wish that
 * we had heaps, and I could just free the heap when done. 
 */

#define CCSegSize 2000

struct CCpool {			/* a segment of allocated memory in the pool */
    struct CCpool *next;
    char space[CCSegSize];
};

/* Initialize the context's heap. */
static void CCinit(context_type *context)
{
    context->CCroot = 0;
    context->CCfree_size = 0;	/* ensure we fail on first allocation */
    context->CCfree_ptr = 0;
}


/* Destroy the context's heap. */
static void CCdestroy(context_type *context)
{
    struct CCpool *this = context->CCroot;
    while (this) {
	struct CCpool *next = this->next;
	free(this);
	this = next;
    }
    /* These two aren't necessary.  But can't hurt either */
    context->CCroot = 0;
    context->CCfree_ptr = 0;
}

/* Allocate an object of the given size from the context's heap. */
static void *
CCalloc(context_type *context, int size, bool_t zero)
{
    register char *p;
    /* Round CC to the size of a pointer */
    size = (size + (sizeof(void *) - 1)) & ~(sizeof(void *) - 1);
    if (context->CCfree_size <  size) {
	struct CCpool *new = (struct CCpool *) malloc(sizeof(struct CCpool));
	new->next = context->CCroot;
	context->CCroot = new;
	context->CCfree_ptr = &new->space[0];
    }
    p = context->CCfree_ptr;
    context->CCfree_ptr += size;
    context->CCfree_size -= size;
    if (zero) 
	memcpy(p, 0, size);
    return p;
}


/* Get the signature associated with a particular field or method in 
 * the constant pool. 
 */
static char *
cp_index_to_signature(context_type *context, int cp_index)
{
    union cp_item_type *cp = context->class->constantpool;
    int index = cp[cp_index].i; /* value of Fieldref field */
    int key2 = index & 0xFFFF;	/* index to NameAndType  */
    int signature_index = cp[key2].i & 0xFFFF; 
    char *signature = cp[signature_index].cp;
    return signature;
}

/* Get the class associated with a particular field or method or class in the
 * constant pool.  If is_field is true, we've got a field or method.  If 
 * false, we've got a class.
 */
static fullinfo_type
cp_index_to_class_fullinfo(context_type *context, int cp_index, bool_t is_field)
{
    union cp_item_type *cp = context->class->constantpool;
    unsigned classkey = is_field ? (cp[cp_index].i >> 16) : cp_index;
    char *classname = GetClassConstantClassName(cp, classkey);
    if (classname[0] == SIGNATURE_ARRAY) {
	fullinfo_type result;
	/* This make recursively call us, in case of a class array */
	signature_to_fieldtype(context, &classname, &result);
	return result;
    } else {
	return MAKE_CLASSNAME_INFO(classname, 0);
    }
}


static void 
CCerror (context_type *context, char *format, ...)
{
    va_list args;
    struct methodblock *mb = context->mb;
    fprintf(stderr, "VERIFIER ERROR %s.%s%s:\n", 
	    fieldclass(&mb->fb)->name, mb->fb.name, mb->fb.signature, mb);
    va_start(args, format);
        vfprintf(stderr, format, args);        
    va_end(args);
    fprintf(stderr, "\n");
}



static char 
signature_to_fieldtype2(context_type *context, 
			char **signature_p, fullinfo_type *full_info_p,
			bool_t inside_array);
static char 
signature_to_fieldtype(context_type *context, 
		       char **signature_p, fullinfo_type *full_info_p)
{
    return signature_to_fieldtype2(context, signature_p, full_info_p, FALSE);
}


static char 
signature_to_fieldtype2(context_type *context, 
			char **signature_p, fullinfo_type *full_info_p,
			bool_t inside_array)
{
    char *p = *signature_p;
    char result;
    fullinfo_type full_info = MAKE_FULLINFO(0, 0, 0);
    switch(*p++) {
        case SIGNATURE_BOOLEAN: case SIGNATURE_BYTE: 
	    full_info = inside_array 
		           ? MAKE_FULLINFO(ITEM_Byte, 0, 0)
			   : MAKE_FULLINFO(ITEM_Integer, 0, 0);
	    result = 'I'; 
	    break;
        case SIGNATURE_CHAR:
	    full_info = inside_array 
		           ? MAKE_FULLINFO(ITEM_Char, 0, 0)
			   : MAKE_FULLINFO(ITEM_Integer, 0, 0);
	    result = 'I'; 
	    break;
        case SIGNATURE_SHORT: 
	    full_info = inside_array 
		           ? MAKE_FULLINFO(ITEM_Short, 0, 0)
			   : MAKE_FULLINFO(ITEM_Integer, 0, 0);
	    result = 'I'; 
	    break;
        case SIGNATURE_INT:
	    full_info = MAKE_FULLINFO(ITEM_Integer, 0, 0);
	    result = 'I'; 
	    break;
        case SIGNATURE_FLOAT:
	    full_info = MAKE_FULLINFO(ITEM_Float, 0, 0);
	    result = 'F'; 
	    break;
        case SIGNATURE_DOUBLE:
	    full_info = MAKE_FULLINFO(ITEM_Double, 0, 0);
	    result = 'D'; 
	    break;
        case SIGNATURE_LONG:
	    full_info = MAKE_FULLINFO(ITEM_Long, 0, 0);
	    result = 'L'; 
	    break;
        case SIGNATURE_ARRAY:
	    signature_to_fieldtype2(context, &p, &full_info, TRUE);
	    full_info = MAKE_FULLINFO(GET_ITEM_TYPE(full_info),
				       GET_INDIRECTION(full_info) + 1,
				       GET_EXTRA_INFO(full_info));
	    result = 'A';
	    break;

        case SIGNATURE_CLASS: {
	    char *finish = strchr(p, SIGNATURE_ENDCLASS);
	    char buffer[128];
	    memcpy(buffer, p, finish - p);
	    buffer[finish - p] = '\0';
	    full_info = MAKE_CLASSNAME_INFO_WITH_COPY(buffer);
	    result = 'A';
	    p = finish + 1;
	    break;
	}
        default:
	    result = 0; 
	    full_info = MAKE_FULLINFO(ITEM_Bogus, 0, 0);
	    break;
    }
    *signature_p = p;
    *full_info_p = full_info;
    return result;
}


/* Given an array type, create the type that has one less level of 
 * indirection.
 */

static fullinfo_type
decrement_indirection(fullinfo_type array_info)
{
    int type = GET_ITEM_TYPE(array_info);
    int indirection = GET_INDIRECTION(array_info) - 1;
    int extra_info = GET_EXTRA_INFO(array_info);
    if (   (indirection == 0) 
	&& ((type == ITEM_Short || type == ITEM_Byte || type == ITEM_Char)))
	type = ITEM_Integer;
    return MAKE_FULLINFO(type, indirection, extra_info);
}


/* See if we can assign an object of the "from" type to an object
 * of the "to" type.
 */

static bool_t is_assignable_to(context_type *context, 
			       fullinfo_type from, fullinfo_type to)
{
    return (merge_fullinfo_types(context, from, to, TRUE) == to);
}

/* Given two fullinfo_type's, find their lowest common denominator.  If
 * the assignable_p argument is non-null, we're really just calling to find
 * out if "<target> := <value>" is a legitimate assignment.  
 *
 * We treat all interfaces as if they were of type java/lang/Object, since the
 * runtime will do the full checking.
 */
static fullinfo_type 
merge_fullinfo_types(context_type *context, 
		     fullinfo_type value, fullinfo_type target,
		     bool_t for_assignment)
{
    if (value == target) {
	/* If they're identical, clearly just return what we've got */
	return value;
    }

    /* Both must be either arrays or objects to go further */
    if (GET_INDIRECTION(value) == 0 && GET_ITEM_TYPE(value) != ITEM_Object)
	return MAKE_FULLINFO(ITEM_Bogus, 0, 0);
    if (GET_INDIRECTION(target) == 0 && GET_ITEM_TYPE(target) != ITEM_Object)
	return MAKE_FULLINFO(ITEM_Bogus, 0, 0);
    
    /* If either is NULL, return the other. */
    if (value == NULL_FULLINFO) 
	return target;
    else if (target == NULL_FULLINFO)
	return value;

    /* If either is java/lang/Object, that's the result. */
    if (target == context->object_info)
	return target;
    else if (value == context->object_info) {
	/* Minor hack.  For assignments, Interface := Object, return Interface
	 * rather than Object, so that is_assignable_to() will get the right
	 * result.      */
	if (for_assignment && (WITH_ZERO_EXTRA_INFO(target) == 
			          MAKE_FULLINFO(ITEM_Object, 0, 0))) {
	    ClassClass *cb = object_fullinfo_to_classclass(context, target);
	    if (cb && cbIsInterface(cb)) 
		return target;
	}
	return value;
    }

    if (GET_INDIRECTION(value) > 0 || GET_INDIRECTION(target) > 0) {
	/* At least one is an array.  Neither is java/lang/Object or NULL */
	if (   (GET_INDIRECTION(value) == GET_INDIRECTION(target))
	    && (GET_ITEM_TYPE(value) == ITEM_Object) 
	    && (GET_ITEM_TYPE(target) == ITEM_Object)) {
	    /* I don't really like this, but I'm told its the right thing
	     * to do.  Fortunately, the Runtime system will catch them if they
	     * really are errors.
	     */
	    fullinfo_type value_base = WITH_ZERO_INDIRECTION(value);
	    fullinfo_type target_base = WITH_ZERO_INDIRECTION(target);
	    fullinfo_type  result_base = 
		merge_fullinfo_types(context, value_base, target_base,
					    for_assignment);
	    if (result_base == MAKE_FULLINFO(ITEM_Bogus, 0, 0))
		/* bogus in, bogus out */
		return result_base;
	    else 
		return MAKE_FULLINFO(ITEM_Object,
				     GET_INDIRECTION(value),
				     GET_EXTRA_INFO(result_base));
	} else if (   (GET_INDIRECTION(value) >  GET_INDIRECTION(target))
		   && (WITH_ZERO_INDIRECTION(target) == context->object_info)){
	    /* an n-th array of Object can hold an n-th array of any object, 
	     * including arrays of dimension > n */
	    return target;
	} else if (   (GET_INDIRECTION(value) < GET_INDIRECTION(target))
		   && (WITH_ZERO_INDIRECTION(value) == context->object_info)) {
	    /* an n-th array of Object can hold an n-th array of any object, 
	     * including arrays of dimension > n */
	    return value;
	} else {
	    /* Otherwise, we punt */
	    return MAKE_FULLINFO(ITEM_Bogus, 0, 0);
	}
    } else {
	/* Both are non-array objects. Neither is java/lang/Object or NULL */
	ClassClass *cb_value, *cb_target, *cb_super_value, *cb_super_target;
	void **addr;
	int value_info;

	/* Let's get the classes corresponding to each of these.  Treat 
	 * interfaces as if they were java/lang/Object.  See hack note above. */
	cb_target = object_fullinfo_to_classclass(context, target);
	if (cb_target == 0) 
	    return MAKE_FULLINFO(ITEM_Bogus, 0, 0);
	if (cbIsInterface(cb_target)) 
	    return for_assignment ? target : context->object_info;
	cb_value = object_fullinfo_to_classclass(context, value);
	if (cb_value == 0) 
	    return MAKE_FULLINFO(ITEM_Bogus, 0, 0);
	if (cbIsInterface(cb_value))
	    return context->object_info;
	
	/* If this is for assignment of target := value, we just need to see if
	 * cb_target is a superclass of cb_value.  Save ourselves a lot of 
	 * work.
	 */
	if (for_assignment) {
	    for (cb_super_value = cb_value; 
		 cbSuperclass(cb_super_value) != NULL; 
		 cb_super_value = unhand(cbSuperclass(cb_super_value))) {
		if (cb_super_value == cb_target) {
		    return target;
		}
	    }
	    return context->object_info;
	}

	/* Find out whether cb_value or cb_target is deeper in the class
	 * tree by moving both toward the root, and seeing who gets there
	 * first.                                                          */
	for (cb_super_value = cb_value, cb_super_target = cb_target;
	     cbSuperclass(cb_super_value) && cbSuperclass(cb_super_target); ) {
	    /* Optimization.  If either hits the other when going up looking
	     * for a parent, then might as well return the parent immediately */
	    if (cb_super_value == cb_target) 
		return target;
	    if (cb_super_target == cb_value) 
		return value;
	    cb_super_value= unhand(cbSuperclass(cb_super_value));
	    cb_super_target = unhand(cbSuperclass(cb_super_target));
	} 
	/* At most one of the following two while clauses will be executed. 
	 * Bring the deeper of cb_target and cb_value to the depth of the 
	 * shallower one. 
	 */
	while (cbSuperclass(cb_super_value)) { /* cb_value is deeper */
	    cb_super_value= unhand(cbSuperclass(cb_super_value)); 
	    cb_value= unhand(cbSuperclass(cb_value)); 
	}
	while (cbSuperclass(cb_super_target)) { /* cb_target is deeper */
	    cb_super_target= unhand(cbSuperclass(cb_super_target)); 
	    cb_target = unhand(cbSuperclass(cb_target)); 
	}
	
	/* Walk both up, maintaining equal depth, until a join is found.  We
	 * know that we will find one.  */
	while (cb_value != cb_target) { 
	    cb_value =  unhand(cbSuperclass(cb_value));
	    cb_target =  unhand(cbSuperclass(cb_target));
	}
	/* Get the info for this guy.  We know its cb_value, so we should
	 * fill that in, while we're at it.                      */
	value_info = Str2ID(&context->classHash, cb_value->name, &addr, FALSE);
	*addr = cb_value;
	return MAKE_FULLINFO(ITEM_Object, 0, value_info);
    } /* both items are classes */
}



/* Given a fullinfo_type corresponding to an Object, return the ClassClass *
 * structure of that type.
 *
 * This will return 0 on an illegal class.
 */

static ClassClass *
object_fullinfo_to_classclass(context_type *context, fullinfo_type classinfo)
{
    void **addr;
    ClassClass *cb;
	    
    int info = GET_EXTRA_INFO(classinfo);
    char *classname = ID2Str(context->classHash, info, &addr);
    if ((cb = *addr) != 0) {
	return cb;
    } else {
	*addr = cb = FindClassFromClass(0, classname, FALSE, context->class);
	if (cb == 0)
	    CCerror(context, "Cannot find class %s", classname);
	return cb;
    }
}


/* Below are for debugging. */


static void print_one_item_type(context_type *, fullinfo_type, bool_t);

static void 
print_stack(context_type *context, stack_item_type *stack)
{
    printf("(");
    for ( ; stack != 0; stack = stack->next) 
	print_one_item_type(context, stack->item, verify_verbose > 1);
    printf(")");
}
	

static void
print_registers(context_type *context, 
		int registers_used, fullinfo_type *registers, 
		bitvector register_mask)
{
    int i;
    printf("{");
    for (i = 0; i < registers_used; i++) 
	print_one_item_type(context, registers[i], verify_verbose > 1);
    printf("}");
    if (register_mask && registers_used != UNKNOWN_REGISTERS_USED) {
	int bits = context->mb->nlocals;
	char *separator = "";
	int i;
	printf("<");
	for (i = 0; i < bits; i++) {
	    if (IS_BIT_SET(register_mask, i)) {
		printf("%s%d%s", separator, i);
		separator = ",";
	    }
	}
	printf(">");
    }
}


static void 
print_one_item_type(context_type *context, fullinfo_type type, bool_t verbose) 
{
    int i;
    int indirection = GET_INDIRECTION(type);
    for (i = indirection; i-- > 0; )
	printf("[");
    switch (GET_ITEM_TYPE(type)) {
        case ITEM_Integer:       
	    printf("I"); break;
	case ITEM_Float:         
	    printf("F"); break;
	case ITEM_Double:        
	    printf("D"); break;
	case ITEM_Double_2:      
	    printf("d"); break;
	case ITEM_Long:          
	    printf("L"); break;
	case ITEM_Long_2:        
	    printf("l"); break;
	case ITEM_ReturnAddress: 
	    printf("a"); break;
	case ITEM_Object:        
	    if (!verbose) {
		printf("A");
	    } else {
		int extra = GET_EXTRA_INFO(type);
		if (extra == 0) {
		    printf("/Null/");
		} else {
		    char *name = ID2Str(context->classHash, extra, 0);
		    char *name2 = strrchr(name, '/');
		    printf("/%s/", name2 ? name2 + 1 : name);
		}
	    }
	    break;
	case ITEM_Char:
	    printf("C"); break;
	case ITEM_Short:
	    printf("S"); break;
	case ITEM_Byte:
	    printf("B"); break;
	default: 
	    printf("?"); break;
    }
    for (i = indirection; i-- > 0; )
	printf("]");
}


static void 
print_formatted_fieldname(context_type *context, int index)
{
    union cp_item_type *constant_pool = context->class->constantpool;
    unsigned char *type_table = 
	    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;
    unsigned type = CONSTANT_POOL_TYPE_TABLE_GET_TYPE(type_table, index);

    unsigned key = constant_pool[index].i; 
    unsigned classkey = key >> 16; 
    unsigned nametypekey = key & 0xFFFF;
    unsigned nametypeindex = constant_pool[nametypekey].i;
    unsigned fieldnameindex = nametypeindex >> 16;
    unsigned fieldtypeindex = nametypeindex & 0xFFFF;
    printf("  <%s.%s%s%s>",
	    GetClassConstantClassName(constant_pool, classkey),
	    constant_pool[fieldnameindex].cp, 
	    type == CONSTANT_Fieldref ? " " : "",
	    constant_pool[fieldtypeindex].cp);
}

